/*
   Module: sinsrv.c  - example TINE server, which illustrates
           the basic server API and alarm API
*/
#include "prolog.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "errors.h"
#include "rpctypes.h"
#include "swaplib.h"  /* for CF_STRUCT example */
#include "almsrv.h"
/* 
   toolkit.c is an optional module, which you might want to
   compile and link into your project.  It contains many
   useful routines for querying or manipulating data.  This
   example makes use of calls to getValuesAsFloat() and
   putValuesFromFloat(), as well as PropertyDataExchange(),
   all of which are modules from toolkit.c
   
   So here we include the header with the protypes:
*/
#include "toolkit.h" 

#ifdef MSDOS
# include "doscmd.h"
#elif defined(UNIX)
# include "unixcmd.h"
# define random(r) (rand()%(r))
#elif defined(WINNT)
# include "ntcmd.h"
# define random(r) (rand()%(r))
#else
# define random(r) ((i*(69069))%(r))
#endif

#define PRP_SINE      1
#define PRP_TEST      2
#define PRP_ECHO      3
#define PRP_ERRTEST   4
#define PRP_MODE      5
#define PRP_STRUCTTEST 6
#define PRP_SENDRPC   7
#define PRP_VALUE   8
#define PRP_ALARM   9

/*
   The code section below illustrates how one registers a structure
   to be exported.  

   Note the use of the short "reserved" in the structure below.  
   Although not necessary, it is strongly recommended that you 
   assume structures to be alligned on 4-byte boundaries and fill
   in any gaps with reserved or dummy tags.  This will make life
   much easier in Visual Basic 5.0 (32-bit) where structures are
   by definition alligned on 4-byte boundaries.  If you don't follow
   this suggestion, then you will have to (in Visual basic 5.0) read 
   your structure into a byte array and parse it out into your 
   desired VB structure.

   Please pay attention to all comments
*/

typedef struct
{
  float a[3];
  long  b[2];
  short c[1];
  short reserved;
  char  d[32];
} Test1Struct;
#define Test1StructSize ((sizeof(float)*3) +\
                         (sizeof(long)*2)  +\
                         (sizeof(short)*1) +\
                         (sizeof(short)*1) +\
                         32)
/* maximum structure array length you're willing to manage: */
#define MAX_TEST1 10

#define quit(i) { printf("Register struct: out of memory\n"); exit(i); }
void registerStructs(void)
{
  /* this must follow the order of the structure explicitly! */
  if (fillstruct("TEST1",OFFSETIN(Test1Struct,a),3,CF_FLOAT)) quit(1);
  if (fillstruct("TEST1",OFFSETIN(Test1Struct,b),2,CF_LONG)) quit(1);
  if (fillstruct("TEST1",OFFSETIN(Test1Struct,c),1,CF_SHORT)) quit(1);
  if (fillstruct("TEST1",OFFSETIN(Test1Struct,reserved),1,CF_SHORT)) quit(1);
  if (fillstruct("TEST1",OFFSETIN(Test1Struct,d),32,CF_TEXT)) quit(1);
  /* terminate the structure definition like this! */
  if (fillstruct("TEST1",sizeof(Test1Struct),MAX_TEST1,CF_NULL)) quit(1);
}

Test1Struct MyTest;

void fillStructs(void)
{
  int i;

  for (i=0; i<3; i++)
  {
    MyTest.a[i] = i+1;
    if (i<2) MyTest.b[i] = i+1;
  }
  MyTest.c[0] = 1;
  strncpy(MyTest.d,"This is a test",32);
}

/*
   End of structure sample routines.  Compare this to
   property STRUCTTEST below.
*/


/*
   Fake data by filling in a 1024 entry ring buffer with
   a sine curve.  Compare this to property SINE below.
*/
#define TRAVELLING 0
#define SIGNAL 1
#define NUMVALUES 4096 //1024

short noise[NUMVALUES];
long dataMode = TRAVELLING;
void fakeSine(float fbuff[NUMVALUES])
{
  int i;

  for (i=0; i<NUMVALUES; i++)
  {
    fbuff[i] = (float)((1.0 + sin(i*6.28/NUMVALUES))*255);
    noise[i] = random(30);
  }
  return;
}
float dbuf[NUMVALUES],sinbuf[NUMVALUES];
short tstbuf[NUMVALUES];

#define NUM_DEV 10
char *sinname[NUM_DEV] = {
  "SINEDEV_0","SINEDEV_1","SINEDEV_2","SINEDEV_3","SINEDEV_4",
  "SINEDEV_5","SINEDEV_6","SINEDEV_7","SINEDEV_8","SINEDEV_9" };

/*
   A server can also be a client, if it needs data from other
   servers.  If you want to test client connectivity inside this
   server program uncomment the following line
*/
/* #define CLIENT */
#define CLIENT
#ifdef CLIENT
#define DEVNAME "SINE/SINEDEV_0"
#define NUM 100
#define PATTERN 212

short echodat[NUM];
int NGenergyID,echoID;
float hpenergy = -1;
void showlink(int id, int cc)
{
  int i;
  if (cc)
  {
    printf("link %d : comp code %d %s\n>",id,cc,RPCLastErrorString);
    return;
  }
  if (id == echoID)
  {
    for (i=0; i<NUM; i++) if (echodat[i] != PATTERN+i)
      printf("ECHO failed: %d not equal to %d\n>",echodat[i],PATTERN+i);
  }
}
void showenergy(int id,int cc)
{
  static cnt = 0;
  if (cc)
  {
    printf("global %d : comp code %d %s\n>",id,cc,RPCLastErrorString);
    /* signal this process that the energy is not coming in */
    if (cnt++ > 3) hpenergy = -1;
    return;
  }
  cnt = 0;
  if (NGdebug > 1) printf("ENERGY is %f\n",hpenergy);
}

#endif

PropertyQueryStruct sinpqs[] = {
  { "SINE","[0:1000 bozos] ",1024,CF_FLOAT-CF_DOUBLE,CA_READ },
  { "TEST","Test Array",1024,CF_FLOAT-CF_DOUBLE,CA_READ },
  { "ECHO","Echo Test",10000,CF_FLOAT-CF_DOUBLE,CA_READ },
  { "STRUCTTEST","Struct test",1024,CF_STRUCT-CF_DOUBLE,CA_READ },
  { "ERRTEST","Error test",1024,CF_FLOAT-CF_DOUBLE,CA_READ },
  { "HURRICANE","Do-da",1024,CF_FLOAT-CF_DOUBLE,CA_READ },
};

int sinePropQueryFcn(char *devName,PropertyQueryStruct **pqs)
{
  *pqs = sinpqs;
  printf("prop query fcn called: %s\n",devName);
  return sizeof(sinpqs)/sizeof(PropertyQueryStruct);
}
void sininit(void)
{
  int cc;
  static short i,n;
# ifdef CLIENT
  static short dat[NUM];
  char devname[32];
  DTYPE dout,din;
# endif

  static int sine_init = 0;

  n = sizeof(sinpqs)/sizeof(PropertyQueryStruct);
//  if ((i=RegisterPropertyQueryFunction("SINSIM",sinePropQueryFcn,n)) != 0)
//    printf("RegisterPropertyQueryFunction : %s\n",erlst[i]);

  if (!sine_init)
  {
     sine_init = 1;
     fakeSine(sinbuf);
     registerStructs();
     fillStructs();
  }
  /* register device names, bind to device number: */
  for (i=0; i<NUM_DEV; i++) RegisterModuleName("SINSIM",sinname[i],i);
  /*
     you can register properties through the 'exports.csv' file or
     by calling RegisterProperty() (or RegisterPropertyEx()) or a
     combination of both.  'exports.csv' is first scanned (if it
     exists) and the cross-reference between the 'export' name and
     the 'equipment module' name is determined there, in which case,
     you can (and should) pass NULL as the first argument in any
     subsequent call to RegisterProperty().  If there is no 'exports.csv'
     file, then you must pass the export name at compile time as the
     first argument in the call to RegisterProperty().  If you use
     RegisterPropetyEx(), you can pass your own property identifier to
     be associated with the property.  A subsequent call to GetPropertyId()
     will then return this identifier.  Alternatively, use the PROPERTY_ID
     column in the 'exports.csv' file.
  */
  if ((cc=RegisterPropertyEx("SINSIM","MODE",1,CF_SHORT,CA_READ|CA_WRITE,"travelling -0 or noise -1",PRP_MODE)) < 0)
  {
    printf("Can't register property MODE : %s\n>",-cc);
  }
  if ((cc=RegisterPropertyEx("SINSIM","SENDRPC",1,CF_SHORT,CA_READ,"Send an RPC call",PRP_SENDRPC)) < 0)
  {
    printf("Can't register property SENDRPC : %s\n>",-cc);
  }
  if ((cc=RegisterPropertyEx("SINSIM","VALUE",1,CF_FLOAT,CA_READ|CA_WRITE,"read/set Value",PRP_VALUE)) < 0)
  {
    printf("Can't register property VALUE : %s\n>",-cc);
  }
  if ((cc=RegisterPropertyEx("SINSIM","GETALARM",1,CF_SHORT,CA_READ,"GetAlarm call",PRP_ALARM)) < 0)
  {
    printf("Can't register property GETALARM : %s\n>",-cc);
  }

# ifdef CLIENT
  dout.dFormat = CF_FLOAT;
  dout.dArrayLength = 1;
  dout.data.fptr = &hpenergy;
//  NGenergyID = recvNetGlobal("HPMAGEN",&dout,showenergy);
  if (NGenergyID < 0) printf("NETGLOBAL : %s\n>",erlst[-NGenergyID]);

  sprintf(devname,"%s",DEVNAME);
  dout.dFormat = CF_SHORT;
  dout.dArrayLength = NUM;
  dout.data.sptr = (UINT16 *)echodat; /* output from remote server data coming back */
  for (i=0; i<NUM; i++) dat[i] = PATTERN+i;
  din.dFormat = CF_SHORT;
  din.dArrayLength = NUM;
  din.data.sptr = (UINT16 *)dat; /* input data to remote server going out */
  echoID = AttachLink(devname,"ECHO",&dout,&din,CA_READ,1000,showlink,CM_POLL);
  if (echoID < 0) printf("ATTACH : %s\n>",erlst[-echoID]);
# endif
}

/*
   the section below illustrates how one uses the Alarm API
   (to lowest order) inside a hardware IO-loop.  If you want
   to set and clear alarms in response to events, you will
   probably want to use RemoveAlarm (or RemoveDeviceAlarm())
   instead of ClearAlarm().

   Note: the simple calls ClearAlarm() and SetAlarm() both work
   with device numbers (which is what the local alarm server uses
   internally anyway).  If you prefer to use device names, use the
   calls ClearDeviceAlarm() and SetDeviceAlarm().

   Uncomment the following line in order to activate it.
*/
#define FAKEALARMS
#define sine_error_1 512
#define sine_error_2 513
#define sine_error_3 514
#define sine_error_4 515
#define sine_error_5 516
#define sine_error_6 517
#define sine_error_7 518
#define sine_error_8 519
#define sine_error_9 520
void fakeAlarms(void)
{
  int i,sc,md;
  static time_t lasttime = 0;
  static step = 0;
  time_t t = time(NULL);
  short almdata[3];

  if (t == lasttime) return;  /* check at 1 Hz */

  switch (step)
  {
    case 0:     /* add a little bit of jitter  */
      step++;
      break;
    case 1:
      /* clear alarms at start of routine and see if they come back: */
      /* ClearDeviceAlarm() or ClearAlarm() increases the clear counter */
      /* No hardware IO in this example, so simulate some alarms:    */
      /* run through set alarms every other minute */
      if ((t/60)%2)
      {
        sc = sine_error_1 + random(8);   /* random alarm code          */
        md = random(NUM_DEV);            /* pick a random module       */
        almdata[0] = 1; almdata[1] = 2;
        SetDeviceAlarm("SINSIM",sinname[md],sc,(BYTE *)almdata,almNEWALARM);

        /* add some more alarms for good measure:                      */
        if (sc < sine_error_3) SetAlarm("SINSIM",md,sedac_error,NULL);
        else if (sc < sine_error_6) SetAlarm("SINSIM",md,gpib_error,NULL);
        else SetDeviceAlarm("SINSIM",sinname[md],hardware_error,NULL,almNEWALARM);

        /* note: a module will have to be cleared 3 times for the alarm to */
        /* disappear                                                       */
      }
      step++;
      break;
    case 2:
      lasttime = t;
      step = 0;
      break;
  }

  return;
}
void sinbkg(void)
{
  static int ptr = 0;
  static int cnt = 0;
  int i;
  /* do IO or something useful here */
  /* we'll simply update the data buffer dbuf for now */
  for (i=0; i<NUMVALUES; i++)
  {
    dbuf[i] = dataMode == TRAVELLING ? sinbuf[(ptr+i)%NUMVALUES] +  noise[i]:
                                       sinbuf[i] + noise[(ptr+i)%NUMVALUES];
  }
  ptr = (ptr+16)%NUMVALUES;
# ifdef FAKEALARMS
  fakeAlarms();
# endif
  if (NGdebug > 1) printf("sinsim background task active (count %d)\n>",cnt);
  cnt++;
}
int sineqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr,prpid,i,cc;
  static int cnt = 0;
  static float fvalue = 0;
  DTYPE doutEx;
  short sdat[100];
  AMS ams;

  cnt++;
  /*
     It is often easier to work internally with a device number.
     You can get the device number corresponding to the device
     name (devName) requested by the following call:
  */
  devnr = GetDeviceNumber("SINSIM",devName);
  if (devnr < 0) return illegal_equipment_number;
  /*
     You can likewise get the property ID associated with devProperty
     via the following call.  Whether it is easier to work with
     e.g. 
        if (!strcmp(devProperty,"SINE")) 
        {
        }
     or 
        switch(propid)
        {
          case PRP_SINE:
        }
     is up to you.
  */
  prpid = GetPropertyId("SINSIM",devProperty);

  switch (prpid)
  {
    case PRP_VALUE:
        if (din->dArrayLength > 0)
        {
          if ((cc=getValuesAsFloat(din,&fvalue,1)) != 0) return cc;
        }
        if (dout->dArrayLength > 0)
        {
          if ((cc=putValuesFromFloat(dout,&fvalue,1)) != 0) return cc;
        }
        return 0;
    case PRP_SINE:
        if (dout->dArrayLength == 0 || dout->dArrayLength > NUMVALUES) return out_of_range;
        if (access&CA_WRITE && din == NULL) return illegal_read_write;
        if ((cc=putValuesFromFloat(dout,dbuf,NUMVALUES)) != 0) return cc;
        return 0;
    case PRP_TEST:
        if (dout->dArrayLength == 0) return illegal_read_write;
        if (access & CA_WRITE && din->dArrayLength == 0) return illegal_read_write;
        memset(tstbuf,0,NUMVALUES*sizeof(short));
        if (din->dArrayLength > 0)
        {
          if ((cc=getValuesAsShort(din,tstbuf,NUMVALUES)) != 0) return cc;
        }
        for (i=0; i<dout->dArrayLength; i++) tstbuf[i] += i;
        if ((cc=putValuesFromShort(dout,tstbuf,NUMVALUES)) != 0) return cc;
        dout->dArrayLength = random(20);
        return 0;
    case PRP_ECHO:
        if (dout->dArrayLength == 0) return illegal_read_write;
        if (din->dArrayLength == 0) return illegal_read_write;
        memset(tstbuf,0,NUMVALUES*sizeof(short));
        if ((cc=getValuesAsShort(din,tstbuf,NUMVALUES)) != 0) return cc;
        if ((cc=putValuesFromShort(dout,tstbuf,NUMVALUES)) != 0) return cc;
        return 0;
    case PRP_SENDRPC:
        doutEx.dFormat = CF_SHORT;
        doutEx.dArrayLength = 100;
        doutEx.data.sptr = (UINT16 *)sdat; /* output from remote server data coming back */
        return ExecLink("SINE/#1","SINE",&doutEx,NULL,CA_READ);
    case PRP_MODE:
        return PropertyDataExchange(&dataMode,CF_LONG,1,dout,din,access);
    case PRP_ERRTEST:
        SetRPCCompletion("This is a test error");
        return 512;
    case PRP_ALARM:
        if ((cc=GetAlarm("SINSIM",devName,&ams)) != 0) return cc;
        printf("alarm for %s : %s <%d>\n",devName,ams.alarmTag,ams.alarmCode);
        return 0;
    case PRP_STRUCTTEST:
        if (din->dArrayLength)
        {
          if (din->dFormat == CF_STRUCT)
          {
            /* we'll insist that the caller tag his structure request: */
            if (!strncmp(din->dTag,"TEST1",8))
            {
              printf("structure TEST1 received (%d elements)\n",din->dArrayLength);
              for (i=0; i<din->dArrayLength; i++)
              {
                printf("\ta = %g %g %g\n\tb = %d %d\n\tc = %d\n\td = %s\n\n",
                     ((Test1Struct *)(din->data.vptr))[i].a[0],
                     ((Test1Struct *)(din->data.vptr))[i].a[1],
                     ((Test1Struct *)(din->data.vptr))[i].a[2],
                     ((Test1Struct *)(din->data.vptr))[i].b[0],
                     ((Test1Struct *)(din->data.vptr))[i].b[1],
                     ((Test1Struct *)(din->data.vptr))[i].c[0],
                     ((Test1Struct *)(din->data.vptr))[i].d );
              }
            }
            else
            {
              printf("Untagged structure received !!\n");
              return illegal_format;
            }
          }
          printf("received %d elements of format %d\n>",din->dArrayLength,din->dFormat);
        }
        /*
            Do somthing with the in-coming data,
            and send some interesting data back.
        */
        if (dout->dArrayLength)
        {
          if (dout->dFormat == CF_STRUCT)
          {
            /* we'll insist that the caller tag his structure request: */
            if (!strncmp(dout->dTag,"TEST1",8))
            {
              if (access & CA_WRITE)
              {
                if (strncmp(din->dTag,dout->dTag,8)) return illegal_read_write;
                memcpy(&MyTest,din->data.vptr,sizeof(Test1Struct));
              }
              for (i=0; i<dout->dArrayLength; i++)
              {
                ((Test1Struct *)(dout->data.vptr))[i].a[0] = MyTest.a[0] + cnt + i;
                ((Test1Struct *)(dout->data.vptr))[i].a[1] = MyTest.a[1] + cnt + i;
                ((Test1Struct *)(dout->data.vptr))[i].a[2] = MyTest.a[2] + cnt + i;
                ((Test1Struct *)(dout->data.vptr))[i].b[0] = MyTest.b[0] + cnt + i;
                ((Test1Struct *)(dout->data.vptr))[i].b[1] = MyTest.b[1] + cnt + i;
                ((Test1Struct *)(dout->data.vptr))[i].c[0] = MyTest.c[0] + cnt + i;
                strcpy(((Test1Struct *)(dout->data.vptr))[i].d,MyTest.d);
              }
            }
            else
            {
              printf("Untagged structure received !!\n");
              return illegal_format;
            }
            return 0;
          }
          return illegal_format;
        }
        return 0;
    default: 
        return illegal_property;
  }
}
/*
   PostSystemInit() is called in rpcserv.c, unless you have a good reason to do
   otherwise, you should leave rpcserv.c alone and make sure you have a
   function PostSystemInit of the following prototype.  This contains your own
   user specific initialization, such as the registration of equipment
   functions.
*/
void PostSystemInit(void)
{
  /*
     the first parameter is the exported equipment function name, if
     NULL, the registration function tries to find the export name
     via the local exports.csv file.  You should do it this way.  If
     you hard code a name in the call, then it makes it harder to
     use the server code on distributed front ends, where the indentical
     code runs on several front ends, each requiring a unique exported
     equipment function name.

     The second parameter is the local equipment function tag (6 characters
     maximum) and identifies the process on the local machine.  This tag
     is used in calls to the local alarm system and along with the server
     name is bound to the export equipment function name.  It is also the
     NODAL function name if you are compiling a NODAL front end.  As this
     tag name is never visible on the client side, it could just as well
     be a number.  However using a 6-character tag allows for more
     readable code without added header files.

     The third argument is the number of devices.  If you pass 0 here,
     then the value obtained from the exports.csv file will be used.

     The fourth argument is the equipment function itself.

     The fifth argument is an equipment function initialization routine
     or NULL.

     The sixth argument is an equipment function background routine
     or NULL.  If specificed, this routine will run in-process unless
     you are using VxWorks.

     The seventh argument is the polling rate of the background task.

     The eight argument is an equipment function exit routine
     or NULL.
  */

  RegisterEquipmentModule(NULL,"SINSIM",0,sineqm,sininit,sinbkg,1000,NULL);

  /*
     The default allowable minumum polling rate is 1000 milliseconds.
     To allow clients to poll for data at a faster rate, you can set
     this value as follows:
  */
  MinPollingRate = 200;
  /*
     On DOS and UNIX machines you can register your own command line
     functions and/or variables ...
  */
#if defined(MSDOS) || defined(UNIX) || defined(WINNT)
  RegisterUserCommand("sinemode",NULL,&dataMode,NULL,CA_READ);
# ifdef CLIENT
  RegisterUserCommand("hpenergy",NULL,NULL,&hpenergy,CA_READ);
# endif
#endif
}

