#ifndef BRDCAST_H_
#define BRDCAST_H_

#include "project.def"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifndef UDP_BROADCAST
# define UDP_BROADCAST FALSE /* default behavior for opening the broadcast socket */
#endif
#ifndef GLOBALGRACE
# define GLOBALGRACE  3      /* retries for missing heartbeat */
#endif

typedef struct
{
  char keyword[16];          /* global keyword */
  short size;                /* number of elements registered in global array */
  short format;              /* broadcast data format */
  time_t time;               /* timestamp of broadcast data */
  struct timeval tv;         /* timestamp of broadcast data */
  BYTE *data;                /* data pointer */
  BYTE *refdata;             /* for comparison */
  int  minPeriod,            /* minimum period between broadcasts */
       maxPeriod;            /* maximum period between broadcasts */
  double tolerance;          /* allowed deviation of data (data units) */
  UINT32 callbackID;         /* private callback ID */
  void (*fcn)(int,int);      /* call back function */
  short isPending;           /* pending flag */
  short isGrouped;           /* grouped flag */
  short hasUserCallbackId;   /* flag : TRUE if set by caller, otherwise false */
  short lastsize;            /* number of elements in last incoming/outgoing packet */
} GlobalListStruct;

#ifndef GLOBALSRV
# define GLOBALSRV "GLOBALS"
#endif
#ifndef MAXGLOBALS
# define MAXGLOBALS 25
#endif
#ifndef GLOBALHEARTBEAT
# define GLOBALHEARTBEAT 60
#endif

extern int nglobals, ngcasts;
extern int GlobalHeartbeat;
extern int MaxNumGlobals;

extern GlobalListStruct **GCastList;
extern GlobalListStruct **NGlobalList;

/* sendNetGlobal() */
/**
 * \brief registers and sends the accompanying keyword and data as a network global.
 *
 * The keyword supplied is appended to the globals table if not already registered and sent
 * immediately as a netork global per multicast if multicasting is enabled and/or per broadcast
 * if a brloadcast nets table has been registered (via the inclusion of a ipbcast.csv or ipxbcast.csv
 * file in the startup database set).  Broadcasts will then be scheduled according the the minPeriod,
 * maxPeriod, and tolerance specified, in which case the din data object should contain a reference 
 * to a data object which will be regularly updated by the server code.  Alternatively, the 
 * call can specify '-1' as minPeriod and maxPeriod, in which case the network global will only be sent
 * upon a call to sendNetGlobal().
 * If a callback function is supplied, is it called following the successful transmission of the
 * network global.
 * 
 * Returns 0 upon successful globals transmission, otherwise a TINE error code.
 *
 * @param keyword is the globals KEYWORD which identifies the data set being sent.
 * @param din is a pointer to the input data set, that is, the data set to be sent from the server 
 * If din is a NULL pointer or contains a NULL data set, sendNetGlobal() returns an error.
 * @param callback is the address of the user-supplied callback routine to be called when data are sent 
 * from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link 
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data 
 * transmission, cc will always be 0, otherwise it can be interpreted with GetRPCLastError().
 * @param minPeriod is the minimum period in milliseconds between scheduled broadcasts
 * @param maxPeriod is the maximum period in milliseconds between scheduled broadcasts.  If the data are
 * still within tolerance at the end of this period, the data set will nonetheless be sent as a network
 * global.
 * @param tolerance is the allowed range between the current and previous data sets.  If exceeded, the
 * current data set is then sent as a network global.
 *  
 */
EXPORT int sendNetGlobal(char *keyword,DTYPE *din,void (*callback)(int,int),
                  int minPeriod,int maxPeriod,double tolerance);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif
