/** \file netmex.h
 * \brief TINE NETMEX header
 *
 */
#ifndef NETMEX_H_
#define NETMEX_H_
#include "cprolog.h"

/*----nextmex.h--Mod 21.02.95:  Address for Ts,Stat,Vers --*/
/*----defines NETKEY structure for encapsulated calls------*/
typedef struct {
   char Eqpname[16];             /* points to server in db */
   char Property[32]; /* property (i.e. keyword) at server */
   short format;                  /* CF_CODE for data type */
   short no_el;              /* # of elements (init=total) */
   short EqNo;                    /* first element to send */
   short Ret_Code;                   /* 0=ok, else error # */
   int TimeStamp;              /* 4 Byte Int for timestamp */
   short Status;                       /* data status word */
   short Version;             /* version # for data header */
   void * data;                              /* data array */
   void (*fcn)(char *,int);      /*callback for async calls*/
   char DevName[16];          /* device name (re: keyword) */
} NETKEY;

/* MEX data structure used in MEX servers: */
typedef struct mexData
{
  char   keyword[32];
  char   description[96];
  NAME16 *devicenames;
  time_t timestamp;
  short  format;
  unsigned short size;
  short (*fcn)(float *data,unsigned short size,short index); /* calling function */
  struct mexData *next;
} MEXDATA;

int LinkNextMex(NETKEY *key,int PollRate,void (*function)(char *,int),short mode);
int LinkNetmex(NETKEY *key,int PollRate,void (*function)(char *,int),short mode);
int LinkNetmexEx(NETKEY *key,int PollRate,void (*function)(char *,int),
               short mode,void *inbuf,int inbuffmt,int inbuflen);
int NextMex(NETKEY* key);
int NetmexGetEx(NETKEY* key,void *inbuf,int inbuffmt,int inbuflen);
int NetmexGet(NETKEY* key);
int InitPack(char *Name, NETKEY* key);
int NextPack(char *Name);
int RegisterKeyword(char *keyword,short format,unsigned short size,
                    short (*fcn)(float *,unsigned short,short));
int RegisterKeywordEx(char *keyword,char *description,short format,unsigned short size,
                    short (*fcn)(float *,unsigned short,short));
int RegisterKeywordDeviceName(char *keyword,char *devicename,int index);
int RegisterNetmexKeywords(void);
void setKeywordRefreshRoutine(void (*fcn)(void),int rate); /* server background routine */
int initNetmexServer(void);

#define T_INIT  0              /* for header inits */
#define ST_INI -1
#define RC_INI -1
                                  /* first try for */
#define data_ok         0    /* defines for Status */
#define stale_data      1            /* start with */
#define header_mismatch 16384       /* maskable #s */

#define FlushNextMex(i) FlushLinks(i)

#define getSimpleFormat(f) (\
  (f) == CF_SHTINT ? CF_SHORT :\
  (f) == CF_LNGINT ? CF_LONG :\
  (f) == CF_FLTINT ? CF_FLOAT :\
  (f) == CF_DBLINT  ? CF_DOUBLE :\
  (f) == CF_TDS  ? CF_FLOAT :\
  (f) == CF_NAME8I  ? CF_NAME8 :\
  (f) == CF_NAME16I ? CF_NAME16 :\
  (f) == CF_NAME32I  ? CF_NAME32 :\
  (f) == CF_NAME48I  ? CF_NAME48 :\
  (f) == CF_NAME64I  ? CF_NAME64 : -1\
)

/**
 * \brief Retrieves the number of archived data records within the
 * time interval specified.
 *
 * \param keyword is the desired netmex keyword
 * \param start is the start time at which to begin scanning for data
 * \param stop is the stop time at which scanning for data should stop
 *
 * \return the number of stored points in the time interval specified.
 *
 */
int getNumberArchiveRecordsInInterval(char *keyword,time_t start,time_t stop);
/**
 * \brief Retrieves an array of archive records for the input time span.
 *
 * If the archive record is an array, the returned data contain an array
 * versus time at the given index. e.g. The keyword is an array of
 * temperatures for 100 different sensors.  This call should return
 * an array of one of the sensors (given by index) versus time.
 *
 * \param keyword is the desired netmex keyword
 * \param index is an optional index corresponding to the input device name
 * \param start is the start time at which to begin scanning for data
 * \param stop is the stop time at which scanning for data should stop
 * \param data returns a reference to an array of data (versus time)
 * covering the start-stop range specified and corresponding to the index
 * given if the keyword is an array.
 * \param timestamp returns a reference to an array of timestamps
 * associated with the data array returned.
 *
 *  \return the number of points in the given interval or -1 on error.
 */
int getArchiveRecordsInInterval(char *keyword,int index,time_t start,time_t stop,void **data,time_t **timestamps);
/**
 * \brief Retrieves an archive record at (or near) the given target
 * timestamp.
 *
 * If the record is an array, this call gives a snapshot of the array at
 * a given time, in contrast to retrieving an element of an array over
 * a time span.
 *
 * \param keyword is the desired netmex keyword
 * \param index is an optional index corresponding to the input device name
 * \param target is the target time for the data record.
 * \param data returns the data for the archive record at the target time
 * \param timestamp returns the actual timestamp of the stored data
 * (closest to but smaller than the target time).
 *
 * \return the number of points archived for the given record at the target time * requested or -1 on error.
 */
int getArchiveRecord(char *keyword,int index,time_t target,void **data,time_t **timestamp);

#endif /* NETMEX_H_ */
