#ifndef PROLOG_H_
#define PROLOG_H_

#include "project.def"
#include "rpcdefs.h"
#include "swaplib.h"
#include "rpctypes.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

/**
 * \brief Allocate memory for an array. Synonym for calloc() under normal circumstances.
 *
 * When some other form of memory managment is in place (e.g. in VxWorks in an EPICS context)
 * SystemCalloc can be re#defined in project.def to use it.
 *
 * \param nr Number of elements.
 * \param sz Size of an element in bytes.
 */
#ifndef SystemCalloc
# define SystemCalloc(nr,sz) calloc(nr,sz)
#endif

/**
 * \brief Free allocated memory. Synonym for free() under normal circumstances.
 *
 * When some other form of memory managment is in place (e.g. in VxWorks in an EPICS context)
 * SystemCalloc can be re#defined in project.def to use it.
 *
 * \param ptr pointer to memory area to be free()ed.
 */
#ifndef SystemFree
# define SystemFree(ptr) free(ptr)
#endif

#ifndef FEC_TIMEOUT
#define FEC_TIMEOUT      1   /* seconds */
#endif
#ifndef SERVER_TIMEOUT
#define SERVER_TIMEOUT   1   /* seconds */
#endif
#ifndef MAX_FDSETS
#define MAX_FDSETS      64
#endif
#ifndef PORTOFFSET
# define PORTOFFSET 0
#endif
#ifndef WATCHDOG
# define WATCHDOG 0
#endif

/* OS-specific definitions ... */
#if defined(MACOS)
#  define UNIX
#  define FREEBSD
#  define OS_TAG "MACOS"
#endif
#if defined(MSDOS)              /* MSDOS */
#  define OS_TAG "DOS"
#  define FS_DELIMITER '\\'
#  ifndef NumTCPChan
#    define NumTCPChan	  5
#  endif
#  define SERVER_POLLING_MODE
#  define millisleep(ms) delay(ms)
#  define ttyoutput(t) printf(t);
   struct timezone { int dummy; };
#elif defined(UNIX)              /* UNIX */
#  ifndef OS_TAG
#    define OS_TAG "UNIX"
#  endif
#  define FS_DELIMITER '/'
#  ifndef NumTCPChan
#    define NumTCPChan	  16
#  endif
#  ifndef _REENTRANT
#    define _REENTRANT
#  endif
#  define SERVER_SELECT_MODE
#  define soioctl(fd,REQ,buf) ioctl(fd,REQ,(char *)buf)
#  define soclose(i) close(i)
   int millisleep(int msecs);
   int ttyoutput(char *str);
#elif defined(VXWORKS)         /* VxWorks */
#  define OS_TAG "VXWORKS"
#  define FS_DELIMITER '/'
#  define SCAN_FOR_NAME_LISTS FALSE
#  ifndef NumTCPChan
#    define NumTCPChan	  8
#  endif
#  define SERVER_SELECT_MODE
#  define soioctl(fd,REQ,buf) ioctl(fd,REQ,(int)buf)
#  define soclose(i) close(i)
   int millisleep(int msecs);
#  define ttyoutput(t) printf(t);
#elif defined(VAX_VMS)        /* VMS */
#  define OS_TAG "VMS"
#  ifndef NumTCPChan
#    define NumTCPChan	  8
#  endif
#  define SERVER_SELECT_MODE
#  define soioctl(fd,REQ,buf) 0
#  define unlink(nm) remove(nm)
#  define millisleep(ms) ;
#  define soclose(i) close(i)
#  define shutdown(s,t) close(s)
#  define ttyoutput(t) printf(t);
   struct timezone { int dummy; };
#elif defined(WINNT)         /* Win32 */
#  define OS_TAG "WIN32"
#  define FS_DELIMITER '\\'
#  ifndef NumTCPChan
#    define NumTCPChan	  16
#  endif
#  define soerrno GetErrno()
#  define SERVER_SELECT_MODE
#  define soioctl(fd,REQ,buf) ioctlsocket(fd,REQ,(u_long *)buf)
#  define millisleep(ms) Sleep(ms)
#  define soclose(i) closesocket(i)
   int ttyoutput(char *str);
   struct timezone { int dummy; };
#elif defined(NIOS)          /* Altera NIOS */
#  define OS_TAG "NIOS"
#  define FS_DELIMITER '/'
#  define FS_RDONLY
#  ifndef NumTCPChan
#    define NumTCPChan	  5
#  endif
#  define ForceStructAlignment -1
#  define SERVER_POLLING_MODE
#  define millisleep(ms) sleep(ms)
#  define soioctl(fd,REQ,buf) ioctl(fd,REQ,(u_long *)buf)
#  define soread(fd,buf,nbyte)  recv(fd,(char *)buf,nbyte,0)
#  define sowrite(fd,buf,nbyte) send(fd,(char *)buf,nbyte,0)
#  define ttyoutput(t) printf(t);
   struct timezone { int dummy; };
#else                       /* ????? */ 
#  ifndef OS_TAG
#    define OS_TAG "UNKNOWN"
#  endif
#  define ttyoutput(t) printf(t);
   struct timezone { int dummy; };
#endif

#ifndef SCAN_FOR_DEVICE_LISTS
# define SCAN_FOR_DEVICE_LISTS TRUE
#endif

#ifndef SCAN_FOR_NAME_LISTS
# define SCAN_FOR_NAME_LISTS TRUE
#endif

#ifndef ForceStructAlignment
# define ForceStructAlignment 0
#endif

#ifndef soerrno
#  define soerrno errno
#endif

#ifndef SERVER_POLLING_MODE
#  define SERVER_SELECT_MODE
#endif

#ifdef MULTITHREADED
# ifdef INCLUDE_LEGACY_CODE
#   undef INCLUDE_LEGACY_CODE
# endif
# if defined(WINNT)
#   define WINTHREAD
# elif defined (UNIX)
#   define PTHREAD
# endif
# include "threader.h"
# define nDataOutBuffers 2
#else
# define nDataOutBuffers 1
#endif

/* 'us-or-them' definitions : */
#ifdef MSDOS   
#  include "dos.h"
#  define malloc_t void *
#  define SetTimeOut(TimeOut) { TimeOut = SERVER_TIMEOUT; }
#  define RDATE
#  define srvWorkArea     WorkArea
#  define srvWorkAreaSize WorkAreaSize
#else
#  define SetTimeOut(TimeOut) { TimeOut = SERVER_TIMEOUT+size/5000; }
#  define soperror(st) perror(st)
#  ifndef soread
#    define soread(fd,buf,nbyte)  read(fd,(char *)buf,nbyte)
#  endif
#  ifndef sowrite
#    define sowrite(fd,buf,nbyte) write(fd,(char *)buf,nbyte)
#  endif
#  if !defined(SUNOS)
#    define malloc_t size_t
#  endif
#  if !defined(WINNT)
   char *strrev(char *s);
#  endif
#  if !defined(WINNT) && !defined(NIOS)
   int stricmp(char *s1, char *s2); /* from the DOS world */
   int strnicmp(char *s1, char *s2,int len); /* from the DOS world */
   char *strupr(char *s);
   char *strlwr(char *s);
#  endif
   extern BYTE *srvWorkArea;
   extern UINT32 srvWorkAreaSize;
#endif                             /* MSDOS: definitions */

#ifndef VXWORKS
#  define inet_ntoa_b(n,b) strcpy(b,inet_ntoa(n))
#else
  struct hostent
  {
    char *h_name;
    char **h_aliases;
    int  h_addrtype;
    int  h_length;
    char *h_addr;
  };
  extern struct hostent globalHostent;
  extern char globalHostAddr[5];
#endif

#ifndef UNIX
  EXPORT struct timeval *gettimeofday(struct timeval *t,struct timezone *tz);
#endif

#ifndef IPXNW
typedef struct  /* an important entry from nwipxspx.h */
{
   BYTE network[4];
   BYTE node[6];
   BYTE socket[2];
} IPXAddress;
#endif

#define IsMemberIPControlNets(addr) IsMemberIPNets(NrOfIPNets,IPNetList,addr)
#define IsMemberIPBcastNets(addr) IsMemberIPNets(NrOfIPBcastNets,IPBcastList,addr)

#if defined(IPXNW) || defined(IPXBSD)
#  define IsMemberIPXControlNets(addr) IsMemberIPXNets(NrOfIPXNets,IPXNetList,addr)
#  define IsMemberIPXBcastNets(addr) IsMemberIPXNets(NrOfIPXBcastNets,IPXBcastList,addr)
#  define SendGCast(buf,size)  SendIPBroadcast(buf,size); SendIPXBroadcast(buf,size)
#  define AcceptRequest(tv) { AcceptIP(tv); AcceptIPX(tv); }
#  define Cleanup(tv)       { CleanupIPX(tv); CleanupIP(tv); CleanupStrandedDatagrams();}
#  define closeServers(i)      { closeIPXServer(); closeIPServer(); }
#  define SendToPeerM(c,d)       \
   (*(UINT16 *)(c)->IPXaddress.socket ? SendIPXData(c,d) : SendIPData(c,d))
#  define ntoa(c,b)     strcpy(b,\
   (*(UINT16 *)(c)->IPXaddress.socket ? ipxnet_ntoa((c)->IPXaddress.node) :\
				     inet_ntoa((c)->IPaddress.sin_addr)   )\
   )
#  define tCall(i,buf,to) (TCPloaded ? tUDPCall(i,buf,to) : tIPXCall(i,buf,to))
#  define IsMemberControlNets(c) ( *(DWORD *)(c)->IPXaddress.network ?\
				   IsMemberIPXControlNets(&(c)->IPXaddress) :\
				   IsMemberIPControlNets(&(c)->IPaddress)    )
#  define IsMemberBcastNets(c)   ( *(DWORD *)(c)->IPXaddress.network ?\
				   IsMemberIPXBcastNets(&(c)->IPXaddress) :\
				   IsMemberIPBcastNets(&(c)->IPaddress)    )
#else
#  define SendGCast(buf,size) SendIPBroadcast(buf,size)
#  define AcceptRequest(tv) AcceptIP(tv)
#  define Cleanup(tv)       { CleanupIP(tv); CleanupStrandedDatagrams(); }
#  define closeServers(i)      closeIPServer()
#  define SendToPeerM(c,d)   SendIPData(c,d)
#  define ntoa(c,b)         inet_ntoa_b((c)->IPaddress.sin_addr,b)
#  define tCall(i,buf,to)   tUDPCall(i,buf,to)
#  define IsMemberControlNets(c) IsMemberIPControlNets(&(c)->IPaddress)
#  define IsMemberBcastNets(c) IsMemberIPBcastNets(&(c)->IPaddress)
#endif         /* ! IPXNW || IPXBSD */

#ifdef NODAL
#  include "nodal.h"
#  define EQPCALL int (*)(double *,int,int,char *)
#  define PrepFormattedData(data,num,dstFmt,srcFmt)\
    reformat((BYTE *)data,dstFmt,(BYTE *)data,srcFmt,num)
#else
#  define EQPCALL int (*)(double *,short,short,short,char *)
   extern EXPORT int NGdebug;
   int dorpc(void);
   int initRPC(void);
typedef  struct /* an important entry from nodal.h */
{
   int  F_type;
   char F_name[7];
   char F_args[5];
   int (*F_entry)(double *,short,short,short,char *);
} FunHead;
#  define data_module        7
   int search_function(char name[7], FunHead **header, int *type);
#  define PrepFormattedData(data,num,dstFmt,srcFmt) 0
#endif

#ifdef MSDOS   /* initialization MACROS */
#  define initServers(i) {\
     if (initControlStructs()) printf("Cannot map header structures\n>");\
     if (initIPX()) printf("IPX/SPX server failed to initialize\n>");\
     if (initTCP()) printf("UDP/TCP server failed to initialize\n>");\
   }
#else
#  ifdef IPXBSD
#    define initServers(i) {\
       if (initControlStructs()) printf("Cannot map header structures\n>");\
       if (initTCP()) printf("UDP/TCP server failed to initialize\n>");\
       if (initIPX()) printf("IPX/SPX server failed to initialize\n>");\
     }
#  else
#    define initServers(i) {\
       if (initControlStructs()) printf("Cannot map header structures\n>");\
       if (initTCP()) printf("UDP/TCP server failed to initialize\n>");\
     }
#  endif
#endif        /* MSDOS: initialization MACROS */

#if defined(INTEL) || defined(VAX_VMS) || defined(NONALIGNED)
#  define PeekWord(w)    (w)
#  define PrepWord(d,s)  { d = s; }
#  define PrepDWord(d,s) { d = s; }
#else
#  define PeekWord(w)    (short)ISWAP(w)
#  define PrepWord(d,s)  { d = (short)ISWAP(s); }
#  define PrepDWord(d,s) { d = LSWAP(s); }
#endif

#ifdef MSDOS  /* Personal structure swapping */

# define AssignStructDataPtr(dptr,sptr,dir) { dptr = sptr;\
  switch (dir) {\
    case RTOL: i = (int)con->EqpSizeIn;\
	       if ((s=getStructFromTag(con->strTagIn)) == NULL) break;\
	       if ((i=i/s->rawlen) > s->num) return out_of_range;\
	       if (con->EqpSizeIn%s->rawlen) return invalid_structure_tag;\
	       break;\
    case LTOR: i = (int)con->EqpSizeOut;\
	       if ((s=getStructFromTag(con->strTagOut)) == NULL) break;\
	       if ((i=i/s->rawlen) > s->num) return out_of_range;\
	       if (con->EqpSizeOut%s->rawlen) return invalid_structure_tag;\
         i = dout.dArrayLength; dout.dArrayLength *= s->rawlen;\
	       break;\
    case NONE: i = (int)con->EqpSizeOut;\
	       if ((s=getStructFromTag(con->strTagOut)) == NULL) break;\
	       if ((i=i/s->rawlen) > s->num) return out_of_range;\
	       if (con->EqpSizeOut%s->rawlen) return invalid_structure_tag;\
	       break;\
  }\
  }

# define PrepOutgoingCLNQS(buf,c) memcpy(buf,c,sizeof(CLNQS));

#else

# define AssignStructDataPtr(dptr,sptr,dir) { dptr = sptr;\
  switch (dir) {\
    case RTOL: i = con->EqpSizeIn;\
	       if (i == 0 || (s=getStructFromTag(con->strTagIn)) == NULL) break;\
	       if ((i=i/s->rawlen) > s->num) return out_of_range;\
	       if (con->EqpSizeIn%s->rawlen) return invalid_structure_tag;\
	       StructSwapEx(con->strTagIn,(BYTE *)(sptr),s->sdata,RTOL,i,s->size);\
	       dptr = s->sdata;\
	       break;\
    case LTOR: i = con->EqpSizeOut;\
	       if (i == 0 || (s=getStructFromTag(con->strTagOut)) == NULL) break;\
	       if ((i=i/s->rawlen) > s->num) return out_of_range;\
	       if (con->EqpSizeOut%s->rawlen) return invalid_structure_tag;\
         i = dout.dArrayLength; dout.dArrayLength *= s->rawlen;\
	       StructSwapEx(con->strTagOut,(BYTE *)(dptr),s->sdata,LTOR,i,s->size);\
	       break;\
    case NONE: i = con->EqpSizeOut;\
	       if (i == 0 || (s=getStructFromTag(con->strTagOut)) == NULL) break;\
	       if ((i=i/s->rawlen) > s->num) return out_of_range;\
	       if (con->EqpSizeOut%s->rawlen) return invalid_structure_tag;\
         memset(s->sdata,0,i*s->size);\
	       dptr = s->sdata;\
	       break;\
  }\
  }

# define PrepOutgoingCLNQS(buf,c) {\
   _StructSwap(ClientQueryStructStruct,(unsigned char *)buf,(unsigned char *)c,LTOR);\
  }

#endif /* MSDOS */

#ifdef INTEL  /* Data Preparation MACROS: */

#define PrepPCHeader(pD,cns,con,ts,tsms) {\
  if ((cns)->client->protocol == 4) {\
    Pheader = (ProducerHeaderStruct4 *)pD;\
    Pheader->msgsize        = msgsize;\
    Pheader->CompletionCode = (cns)->status.statusCode;\
    Pheader->timestamp      = (ts);\
    Pheader->starttime      = (cns)->status.starttime;\
    Pheader->SubscriptionID = (cns)->status.id;\
    Pheader->counter        = (cns)->status.counter;\
  }\
  else {\
    PheaderEx = ForceStructAlignment ? &tmpPEx : (ProducerHeaderStruct *)pD;\
    PheaderEx->msgsize        = msgsize;\
    PheaderEx->CompletionCode = (cns)->status.statusCode;\
    PheaderEx->timestampMSEC  = (tsms);\
    PheaderEx->EqpFormat      = (con).EqpFormatOut;\
    PheaderEx->timestamp      = (ts);\
    PheaderEx->ClnStarttime   = (cns)->status.starttime;\
    PheaderEx->SubscriptionID = (cns)->status.id;\
    PheaderEx->counter        = (BYTE)(cns)->status.counter;\
    PheaderEx->blknum         = (cns)->status.blknum;\
    PheaderEx->numblks        = (cns)->status.numblks;\
    PheaderEx->mtu            = (cns)->status.mtu;\
    if (ForceStructAlignment) memcpy(pD,&tmpPEx,sizeof(ProducerHeaderStruct));\
  }\
}

#define PrepPCData(pD,con,siz) {\
  if (siz >= 0) memcpy(pD,(con)->Data,(size_t)siz);\
  else memcpy((con)->Data,pD,(size_t)abs((siz)));\
}

#define CopySubData(pD,con,siz,dir) {\
  if (dir == LTOR) memcpy(pD,(con)->DataOut,(size_t)siz);\
  else memcpy((con)->DataIn,pD,(size_t)siz);\
}

#define PrepSubData(pD,sz,f,dir) ;

#define PrepSubHeader(p,dat) {\
  if (p == 4) {\
    reg = (struct subscription4 *)dat;\
    memset(regEx,0,sizeof(struct subscription));\
    regEx->msgsize = reg->msgsize;\
    regEx->PollingRate = reg->PollingRate;\
    regEx->id = reg->id;\
    regEx->mode = reg->mode;\
    regEx->starttime = reg->starttime;\
    regEx->numblks = 1;\
    regEx->blknum = 1;\
    regEx->mtu = reg->msgsize;\
    strncpy(regEx->contract.EqpName,reg->contract.EqpName,8);\
    strncpy(regEx->contract.EqpProperty,reg->contract.Property,12);\
    regEx->contract.EqpSizeOut = abs(reg->contract.EqpSize);\
    regEx->contract.EqpSizeIn = reg->contract.EqpSize < 0 ? abs(reg->contract.EqpSize): 0;\
    regEx->contract.EqpAccess = reg->contract.EqpSize < 0 ? CA_WRITE|CA_READ : CA_READ;\
    if (reg->contract.EqpFormat == 1) reg->contract.EqpFormat = CF_TEXT;\
    regEx->contract.EqpFormatOut = BFMT(reg->contract.EqpFormat);\
    regEx->contract.EqpFormatIn = reg->contract.EqpSize < 0 ? regEx->contract.EqpFormatOut : BFMT(CF_NULL);\
    sprintf(regEx->contract.EqpDeviceName,"#%d",reg->contract.EqpNumber);\
  } else {\
    if (ForceStructAlignment) {\
      memcpy(&tmpRegEx,dat,sizeof(struct subscription));\
      regEx = &tmpRegEx;\
    } else {\
      regEx = (struct subscription *)dat;\
    }\
  }\
}
#define PrepIncomingHeader(h,dat) { h = (HeaderStruct *)dat; }
#define PrepIncomingInfo(e,dat) { e = (EqpInfoStruct *)dat; }
#define PrepStreamData(dst,src,num,fmt) memcpy(dst,src,abs(num)*getFormatSize(fmt))
#define PrepOutgoingFecaddr(fecstruct,buf) memcpy(buf,fecstruct,RPCFECSTRUCT_SIZE)
#define PrepOutgoingEqpaddr(datastruct,buf) memcpy(buf,datastruct,RPCDATASTRUCT_SIZE)
#define PrepOutgoingAlarmData(buf,n,fmt) ;
#define PrepOutgoingAMS(amsstruct,buf,n) {if ((void *)buf != (void *)amsstruct) memcpy(buf,amsstruct,n*AMS_SIZE); }
#define PrepOutgoingADS(adsstruct,buf,n) {if ((void *)buf != (void *)adsstruct) memcpy(buf,adsstruct,n*ADS_SIZE); }
#define PrepIncomingAMRB(rb,buf) memcpy(rb,buf,AMRB_SIZE);
#define PrepOutgoingAQS(a,buf) {if ((void *)buf != (void *)a) memcpy(buf,a,ACTIVITYQUERYSTRUCT_SIZE); }
#define PrepOutgoingPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*PROPERTYQUERYSTRUCT_SIZE); }
#define PrepOutgoingPQSX(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*PROPERTYQUERYSTRUCTEX_SIZE); }
#define PrepOutgoingXPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*XPROPERTYQUERYSTRUCT_SIZE); }
#define PrepOutgoingCONQS(c) ;
#define PrepIncomingPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*PROPERTYQUERYSTRUCT_SIZE); }
#define PrepIncomingPQSX(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*PROPERTYQUERYSTRUCTEX_SIZE); }
#define PrepIncomingXPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*XPROPERTYQUERYSTRUCT_SIZE); }

#else     /* non-INTEL world must SWAP ! */

#define PrepPCHeader(pD,cns,con,ts,tsms) {\
  if ((cns)->client->protocol == 4) {\
    Pheader->msgsize        = msgsize;\
    Pheader->CompletionCode = (cns)->status.statusCode;\
    Pheader->timestamp      = (ts);\
    Pheader->starttime      = (cns)->status.starttime;\
    Pheader->SubscriptionID = (cns)->status.id;\
    Pheader->counter        = (cns)->status.counter;\
   _StructSwap(ProducerHeaderStruct4Struct,(unsigned char *)pD,(unsigned char *)Pheader,LTOR);\
  } else {\
    PheaderEx->msgsize        = msgsize;\
    PheaderEx->CompletionCode = (cns)->status.statusCode;\
    PheaderEx->timestampMSEC  = (tsms);\
    PheaderEx->EqpFormat      = (con).EqpFormatOut;\
    PheaderEx->timestamp      = (ts);\
    PheaderEx->ClnStarttime   = (cns)->status.starttime;\
    PheaderEx->SubscriptionID = (cns)->status.id;\
    PheaderEx->counter        = (BYTE)(cns)->status.counter;\
    PheaderEx->mtu            = (cns)->status.mtu;\
    PheaderEx->blknum         = (cns)->status.blknum;\
    PheaderEx->numblks        = (cns)->status.numblks;\
   _StructSwap(ProducerHeaderStructStruct,(unsigned char *)pD,(unsigned char *)PheaderEx,LTOR);\
  }\
}

#define PrepPCData(pD,con,siz) {\
  if (siz >=0) {\
    memcpy(pD,(con)->Data,(size_t)siz);\
    SwapBytes((char *)pD,abs((con)->contract.EqpSize),(con)->contract.EqpFormat,LTOR);\
  } else {\
    SwapBytes((char *)pD,abs((con)->contract.EqpSize),(con)->contract.EqpFormat,RTOL);\
    memcpy((con)->Data,pD,(size_t)abs(siz));\
  }\
}
#define PrepSubData(pD,sz,f,dir) {\
    SwapBytes((char *)pD,sz,LFMT(f),dir);\
}
#define PrepSubHeader(p,dat) {\
  if (p == 4) {\
    memset(reg,0,sizeof(struct subscription4));\
    memset(regEx,0,sizeof(struct subscription));\
    _StructSwap(SUBSCRIPTION4Struct,(unsigned char *)dat,(unsigned char *)reg,RTOL);\
    regEx->msgsize = reg->msgsize;\
    regEx->PollingRate = reg->PollingRate;\
    regEx->id = reg->id;\
    regEx->mode = reg->mode;\
    regEx->starttime = reg->starttime;\
    regEx->numblks = 1;\
    regEx->blknum = 1;\
    regEx->mtu = reg->msgsize;\
    strncpy(regEx->contract.EqpName,reg->contract.EqpName,8);\
    strncpy(regEx->contract.EqpProperty,reg->contract.Property,12);\
    regEx->contract.EqpSizeOut = abs(reg->contract.EqpSize);\
    regEx->contract.EqpSizeIn = reg->contract.EqpSize < 0 ? abs(reg->contract.EqpSize): 0;\
    regEx->contract.EqpAccess = reg->contract.EqpSize < 0 ? CA_WRITE|CA_READ : CA_READ;\
    if (reg->contract.EqpFormat == 1) reg->contract.EqpFormat = CF_TEXT;\
    regEx->contract.EqpFormatOut = BFMT(reg->contract.EqpFormat);\
    regEx->contract.EqpFormatIn = reg->contract.EqpSize < 0 ? regEx->contract.EqpFormatOut : BFMT(CF_NULL);\
    sprintf(regEx->contract.EqpDeviceName,"#%d",reg->contract.EqpNumber);\
  } else {\
    memset(regEx,0,sizeof(struct subscription));\
    _StructSwap(SUBSCRIPTIONStruct,(unsigned char *)dat,(unsigned char *)regEx,RTOL);\
  }\
}

#define PrepIncomingHeader(h,dat) {\
  _StructSwap(HeaderStructStruct,(unsigned char *)dat,(unsigned char *)h,RTOL);\
}
#define PrepIncomingInfo(e,dat) {\
  e = (EqpInfoStruct *)dat;\
  _StructSwap(EqpInfoStructStruct,(unsigned char *)dat,(unsigned char *)e,RTOL);\
}
#define PrepStreamData(dst,src,num,fmt) {\
  if (num >= 0) {\
    SwapBytes((char *)src,num,fmt,LTOR);\
    memcpy(dst,src,num*getFormatSize(fmt));\
  } else {\
    memcpy(dst,src,-num*getFormatSize(fmt));\
    SwapBytes((char *)dst,-num,fmt,RTOL);\
  }\
}
#define PrepOutgoingFecaddr(fecstruct,buf)  {\
   _StructSwap(RPCFecStructStruct,(unsigned char *)buf,(unsigned char *)fecstruct,LTOR);\
}
#define PrepOutgoingEqpaddr(datastruct,buf)  {\
   _StructSwap(RPCDataStructStruct,(unsigned char *)buf,(unsigned char *)datastruct,LTOR);\
}
#define PrepOutgoingAlarmData(buf,n,fmt)  {\
   SwapBytes((char *)buf,n,fmt,LTOR);\
}
#define PrepOutgoingAMS(amsstruct,buf,n)  {\
   _StructSwapEx(AMSStructStruct,(unsigned char *)buf,(unsigned char *)amsstruct,LTOR,n,sizeof(AMS));\
}
#define PrepOutgoingADS(adsstruct,buf,n)  {\
   _StructSwapEx(ADSStructStruct,(unsigned char *)buf,(unsigned char *)adsstruct,LTOR,n,sizeof(ADS));\
}
#define PrepIncomingAMRB(rb,buf) {\
   memcpy(rb,buf,AMRB_SIZE);\
   SwapBytes((char *)rb,4,CF_LONG,RTOL);\
}
#define PrepOutgoingAQS(a,buf) {\
   _StructSwap(ActivityQueryStructStruct,(unsigned char *)buf,(unsigned char *)a,LTOR);\
}
#define PrepOutgoingPQS(p,buf,n) {\
   _StructSwapEx(PropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(PropertyQueryStruct));\
}
#define PrepOutgoingPQSX(p,buf,n) {\
   _StructSwapEx(PropertyQueryStructExStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(PropertyQueryStructEx));\
}
#define PrepOutgoingXPQS(p,buf,n) {\
   _StructSwapEx(XPropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(XPropertyQueryStruct));\
}
#define PrepOutgoingCONQS(c) {\
   _LSWAP(c.contract.EqpSizeIn);\
   _LSWAP(c.contract.EqpSizeOut);\
   _LSWAP(c.pollrate);\
   _LSWAP(c.nconsumers);\
}
#define PrepIncomingPQS(p,buf,n) {\
   _StructSwapEx(PropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(PropertyQueryStruct));\
}
#define PrepIncomingPQSX(p,buf,n) {\
   _StructSwapEx(PropertyQueryStructExStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(PropertyQueryStructEx));\
}
#define PrepIncomingXPQS(p,buf,n) {\
   _StructSwapEx(XPropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(XPropertyQueryStruct));\
}

#endif  /* INTEL (data prepartion MACROS) */

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* PROLOG_H_ */
