
#ifndef RPCTYPES_H_
#define RPCTYPES_H_

#include "project.def"

#include <time.h>
#include <stdarg.h>
#include "rpcdefs.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

/**
 * \struct NAME8
 * \brief Defines a TINE 8-character fixed-length string data object
 *
 * - name is the string buffer containing the string information
 */
typedef struct
{
  char name[8];
} NAME8;

/**
 * \struct NAME16
 * \brief Defines a TINE 16-character fixed-length string data object
 *
 * - name is the string buffer containing the string information
 */
typedef struct
{
  char name[16];
} NAME16;

/**
 * \struct NAME32
 * \brief Defines a TINE 32-character fixed-length string data object
 *
 * - name is the string buffer containing the string information
 */
typedef struct
{
  char name[32];
} NAME32;

/**
 * \struct NAME48
 * \brief Defines a TINE 48-character fixed-length string data object
 *
 * - name is the string buffer containing the string information
 */
typedef struct
{
  char name[48];
} NAME48;

/**
 * \struct NAME64
 * \brief Defines a TINE 64-character fixed-length string data object
 *
 * - name is the string buffer containing the string information
 */
typedef struct
{
  char name[64];
} NAME64;

/**
 * \struct NAME8I
 * \brief Defines a TINE pairwise data object containing an 8-character fixed-length string 
 * followed by a 4-byte integer
 *
 * - name is the string buffer containing the string information
 * - ival is the long integer value
 */
typedef struct
{
  char name[8];
  UINT32 ival;
} NAME8I;

/**
 * \struct NAME16I
 * \brief Defines a TINE pairwise data object containing a 16-character fixed-length string 
 * followed by a 4-byte integer
 *
 * - name is the string buffer containing the string information
 * - ival is the long integer value
 */
typedef struct
{
  char name[16];
  UINT32 ival;
} NAME16I;

/**
 * \struct NAME32I
 * \brief Defines a TINE pairwise data object containing a 32-character fixed-length string 
 * followed by a 4-byte integer
 *
 * - name is the string buffer containing the string information
 * - ival is the long integer value
 */
typedef struct
{
  char name[32];
  UINT32 ival;
} NAME32I;

/**
 * \struct NAME48I
 * \brief Defines a TINE pairwise data object containing a 48-character fixed-length string 
 * followed by a 4-byte integer
 *
 * - name is the string buffer containing the string information
 * - ival is the long integer value
 */
typedef struct
{
  char name[48];
  UINT32 ival;
} NAME48I;

/**
 * \struct NAME64I
 * \brief Defines a TINE pairwise data object containing a 64-character fixed-length string 
 * followed by a 4-byte integer
 *
 * - name is the string buffer containing the string information
 * - ival is the long integer value
 */
typedef struct
{
  char name[64];
  UINT32 ival;
} NAME64I;

typedef struct
{
  UINT16 sval;
  UINT32 ival;
} SHTINT;

typedef struct
{
  UINT32 lval;
  UINT32 ival;
} LNGINT;

/**
 * \struct INTINT
 * \brief Defines a TINE pairwise data object containing a 4-byte integer doublet
 *
 * - i1val is the first long integer value
 * - i2val is the second long integer value
 *
 */
typedef struct
{
  UINT32 i1val;
  UINT32 i2val;
} INTINT;

/**
 * \struct FLTINT
 * \brief Defines a TINE pairwise data object containing a float value followed by
 * a 4-byte integer value
 *
 * - fval is the float value
 * - ival is the long integer value
 *
 */
typedef struct
{
  float  fval;
  UINT32 ival;
} FLTINT;

/**
 * \struct DBLINT
 * \brief Defines a TINE pairwise data object containing a double value followed by
 * a 4-byte integer value
 *
 * - dval is the double value
 * - ival is the long integer value
 *
 * \note This structure has limited applicability and should only be used where structures 
 * on compiled with 4-byte boundaries.
 * 
 */
typedef struct
{
  double dval;
  UINT32 ival;
} DBLINT;

/**
 * \struct DBLDBL
 * \brief Defines a TINE pairwise data object containing a pairwise double doublet.
 *
 * - d1val is the first double value
 * - d2val is the second double value
 *
 * \note This structure is required for archive data-retrieval when millisecond-resolution
 * double timestamps are requested.  In this case the data are returned as pair-wise value
 * plus timestamp entries.
 */
typedef struct
{
  double d1val;
  double d2val;
} DBLDBL;

/**
 * \struct FLTFLT
 * \brief Defines a TINE pairwise data object containing a pairwise float doublet.
 *
 * - f1val is the first float value
 * - f2val is the second float value
 *
 * \note This is formally synonymous with typedef DPOINT
 */
typedef struct
{
  float f1val;
  float f2val;
} FLTFLT;

/**
 * \struct DPOINT
 * \brief Defines a TINE pairwise data object containing a pairwise float doublet.
 *
 * - x is the first float value
 * - y is the second float value
 *
 */
typedef struct
{
  float x;
  float y;
} DPOINT;

/**
 * \struct NAME16FI
 * \brief Defines a TINE triplet data object containing a 16-character fixed-length string 
 * value followed by a float value followed by 4-byte integer value
 *
 * - name is the string buffer containing the string information
 * - fval is the float value
 * - ival is the long integer value
 *
 * \note This structure has limited applicability and should only be used where structures 
 * on compiled with 4-byte boundaries.
 * 
 */
typedef struct
{
  char   name[16];
  float  fval;
  UINT32 ival;
} NAME16FI;

/**
 * \struct NAME16II
 * \brief Defines a TINE triplet data object containing a 16-character fixed-length string 
 * value followed by two 4-byte integer values
 *
 * - name is the string buffer containing the string information
 * - ival1 is the first long integer value
 * - ival2 is the second long integer value
 *
 * \note This structure has limited applicability and should only be used where structures 
 * on compiled with 4-byte boundaries.
 * 
 */
typedef struct
{
  char   name[16];
  UINT32 i1val;
  UINT32 i2val;
} NAME16II;

/**
 * \struct NAME16DBLDBL
 * \brief Defines a TINE triplet data object containing a 16-character fixed-length string 
 * value followed by a double doublet
 *
 * - name is the string buffer containing the string information
 * - d1val is the first double value
 * - d2val is the second double value
 *
 */
typedef struct
{
  char   name[16];
  double d1val;
  double d2val;
} NAME16DBLDBL;

/**
 * \struct NAME32DBLDBL
 * \brief Defines a TINE triplet data object containing a 32-character fixed-length string 
 * value followed by a double doublet
 *
 * - name is the string buffer containing the string information
 * - d1val is the first double value
 * - d2val is the second double value
 *
 */
typedef struct
{
  char   name[32];
  double d1val;
  double d2val;
} NAME32DBLDBL;

/**
 * \struct NAME64DBLDBL
 * \brief Defines a TINE triplet data object containing a 64-character fixed-length string 
 * value followed by a double doublet
 *
 * - name is the string buffer containing the string information
 * - d1val is the first double value
 * - d2val is the second double value
 *
 */
typedef struct
{
  char   name[64];
  double d1val;
  double d2val;
} NAME64DBLDBL;

/**
 * \struct FLTINTINT
 * \brief Defines a TINE triplet data object containing a float
 * value followed by a long integer doublet
 *
 * - fval is the float value
 * - i1val is the first long integer value
 * - i2val is the second long integer value
 *
 */
typedef struct
{
  float  fval;
  UINT32 i1val;
  UINT32 i2val;
} FLTINTINT;

/**
 * \struct INTFLTINT
 * \brief Defines a TINE triplet data object containing a long integer value
 * followed by a float followed by a second long integer value.
 *
 * - i1val is the first long integer value
 * - fval is the float value
 * - i2val is the second long integer value
 *
 * \note This structure is synonymous with a 'TDS',i.e. Time-stamp - Data - Status,
 * object used in certain calls from DOOCS.
 */
typedef struct
{
  UINT32 i1val;
  float  fval;
  UINT32 i2val;
} INTFLTINT;

typedef struct
{
  float  xval;
  float  yval;
  UINT32 ival;
} XYS;

/**
 * \struct FLTFLTINT
 * \brief Defines a TINE triplet data object containing a float doublet followed by a 
 * long integer value.
 *
 * - f1val is the first float value
 * - f2val is the second float value
 * - ival is the long integer value
 *
 */
typedef struct
{
  float f1val;
  float f2val;
  UINT32 ival;
} FLTFLTINT;

/**
 * \struct INTINTINT
 * \brief Defines a TINE triplet data object containing a set of three
 * long integer values.
 *
 * - i1val is the first long integer value
 * - i2val is the second long integer value
 * - i3val is the third long integer value
 *
 */
typedef struct
{
  UINT32 i1val;
  UINT32 i2val;
  UINT32 i3val;
} INTINTINT;

/**
 * \struct DADDRESS
 * \brief Defines a TINE quadruplet data object containing four long integer values
 *
 * - i1val is the first long integer value
 * - i2val is the second long integer value
 * - i3val is the third long integer value
 * - i4val is the fourth long integer value
 *
 */
typedef struct
{
  UINT32 i1val;
  UINT32 i2val;
  UINT32 i3val;
  UINT32 i4val;
} DADDRESS;

/**
 * \struct DWINDOW
 * \brief Defines a TINE quadruplet data object containing a long integer doublet
 * giving a time range followed by another long integer doublet giving a value range
 *
 * - t1val is the first long integer time_t value
 * - t2val is the second long integer time_t value
 * - i1val is the first long integer value
 * - i2val is the second long integer value
 *
 */
typedef struct
{
  UINT32 t1;
  UINT32 t2;
  UINT32 i1val;
  UINT32 i2val;
} DWINDOW;

/**
 * \struct FWINDOW
 * \brief Defines a TINE quadruplet data object containing a long integer doublet
 * followed by a float doublet
 *
 * - i1val is the first long integer value
 * - i2val is the second long integer value
 * - f1val is the first float value
 * - f2val is the second float value
 *
 */
typedef struct
{
  UINT32 i1val;
  UINT32 i2val;
  float  f1val;
  float  f2val;
} FWINDOW;

/**
 * \struct FLTINTFLTINT
 * \brief Defines a TINE quadruplet data object containing a float value followed by a
 * long integer followed by a float value followed by a long integer
 *
 * - f1val is the first float value
 * - i1val is the first long integer value
 * - f2val is the second float value
 * - i2val is the second long integer value
 *
 */
typedef struct
{
  float  f1val;
  UINT32 i1val;
  float  f2val;
  UINT32 i2val;
} FLTINTFLTINT;

typedef struct
{
  UINT32 t;
  float  d;
  UINT16 s;
  UINT16 i;
} TDSI;

/**
 * \struct DFILTER
 * \brief Defines a TINE quadruplet data object containing a long integer value followed by a
 * float triplet
 *
 * - ival is the long integer value
 * - f1val is the first float value
 * - f2val is the second float value
 * - f3val is the third float value
 *
 */
typedef struct
{
  UINT32 ival;
  float  f1val;
  float  f2val;
  float  f3val;
} DFILTER;

/**
 * \struct DUSTRING
 * \brief Defines a TINE quintet data object containing a long integer value
 * followed by a float doublet followed by a second long integer value,
 * followed by an 80-character fixed-length string
 *
 * - ival is the first long integer value
 * - f1val is the first float value
 * - f2val is the second float value
 * - tm is the seond long integer value (intpreted internally as a time_t)
 * - str is the string buffer
 *
 */
typedef struct
{
  UINT32 ival;
  float  f1val;
  float  f2val;
  UINT32 tm;
  char   str[80];
} DUSTRING;

/**
 * \struct DUNAME
 * \brief Defines a TINE quintet data object containing a long integer value
 * followed by a float triplet followed
 * followed by a 16-character fixed-length name
 *
 * - ival is the first long integer value
 * - f1val is the first float value
 * - f2val is the second float value
 * - f3val is the third float value
 * - name is the name buffer
 *
 */
typedef struct
{
  UINT32 ival;
  float  f1val;
  float  f2val;
  float  f3val;
  char   name[16];
} DUNAME;

/**
 * \struct hdrSPECTRUM
 * \brief Defines a TINE quintuplet data object containing an 80-character fixed-length string
 * followed by a long integer timestamp followed by a float doublet followed by a long integer
 * value 
 *
 * - comment is the string buffer
 * - tm is the first long integer value interpreted as a time_t
 * - s_start is the first float value
 * - s_inc is the second float value
 * - status is the second long integer value
 *
 * \note This structure defines the header structure for data type CF_SPECTRUM.
 */
typedef struct
{
  char   comment[80];
  UINT32 tm;
  float  s_start;
  float  s_inc;
  UINT32 status;
} hdrSPECTRUM;

/**
 * \struct DSPECTRUM
 * \brief Defines a TINE sextuple data object containing an 80-character fixed-length string
 * followed by a long integer timestamp followed by a float doublet followed by a long integer
 * value followed by 1024-value float spectrum array
 *
 * - comment is the string buffer
 * - tm is the first long integer value interpreted as a time_t
 * - s_start is the first float value
 * - s_inc is the second float value
 * - status is the second long integer value
 * - d_spect_array is the spectrum array
 *
 */
typedef struct
{
  char   comment[80];
  UINT32 tm;
  float  s_start;
  float  s_inc;
  UINT32 status;
  float  d_spect_array[1024];
} DSPECTRUM;

/**
 * \struct DUNION
 * \brief Dereferneces the data set in a DTYPE object
 *
 */
typedef union
{
  void   *vptr;
  char   *cptr;
  UINT8  *bptr;
  UINT16 *sptr;
  UINT32 *lptr;
  float  *fptr;
  double *dptr;
} DUNION;

/**
 * \struct DTYPE
 * \brief Defines a TINE data object
 *
 * - dArrayLength is the size of the data set in terms of the format size specified.
 * - dFormat is the TINE format of the data set
 * - dTimestamp is the associated (threadsafe) timestamp of the returned data set, valid only for
 * synchronous calls, as the data object is usually no longer in scope for asynchronous calls.
 * - dTag is an optional data tag used typically to identify user-defined structures
 * - data is a reference to the data set
 */
typedef struct
{
  UINT32 dArrayLength;
  short  dFormat;
  short  reserved;
  double dTimeStamp;
  char   dTag[16];
  DUNION data;
} DTYPE;

#define RPCERR_SIZE 32
extern EXPORT char RPCLastErrorString[RPCERR_SIZE];

#define RPC_ERROR_LCL 0
#define RPC_ERROR_RMT 1
/* RPCLastErrorSource */
/**
 * \brief Gives the signature of the last Link Error 
 * 
 * If a call to ExecLink() or ExecLinkEx() fails it will return a non-zero
 * TINE error code, which can be interpretted by GetRPCLastError().  It is
 * sometime useful to know whether the return code was locally generated
 * (for instance a time out) or remotely generated (for instance 'illegal property').
 * To this end you can examine the value of RPCLastErrorSource and check it
 * against RPC_ERROR_LCL or RPC_ERROR_RMT
 *
 */
extern short RPCLastErrorSource;

/* ======================== servers.h ==remove far's 26.01.95===========*/

int Subscribe(char **Values,char *eqpname,int EqNo,char *property,int size,
		   int format,int pollingRate,void (*function)(int,int),int mode);

int hotRpc(char **Values, char *eqpname, int EqNo, char *property, int rwFlg,
		 int format, int pollingRate, void (*function)(int,int));

/* CloseLink() */
/**
 * \brief Cancels an active data link
 * 
 * Use this call to cancel an active data link (alias: CloseLink()).  Note that
 * the parameter passed must be the returned link id when the original call to
 * AttachLink() or AttachLinkEx() was made.  
 * 
 * \param i the link index to be canceled
 *
 * \see AttachLink(), AttachLinkEx()
 * 
 * \b Example:

   CloseLink(idEnergyLink);
 
 */
EXPORT void CloseLink(int i);

/**
 * \def Cancel(i)
 * 
 * Identical to CloseLink(i)
 */
#define Cancel(i) CloseLink(i) 
extern int connectionTableID;

int RPC(char *Values, char *eqpname, int EqNo, char *property, int rwFlg, int format, int datasize);
int aRPC(char *Values, char *eqpname, int EqNo, char *property, int rwFlg, int format, int datasize, void (*function)(int,int));

int PACK(char *Name,char *Values,char *eqpname, int EqNo,
         char *property,int rwFlg,int format,int datasize,char **info);

extern int SystemLocalDataTimeSynchronization;

/* SystemRunning */
/**
 * \brief Determines whether a server has been successfully initialized or not.
 * 
 * A server can query this flag at any stage to determine if all initialization step have been
 * successfully completed or not.  You should never set this flag yourself.
 *
 * \b Default: FALSE
 *  
 */
extern int SystemRunning;
/* SystemSupressHeartbeatNofification */
/**
 * \brief Determines whether CM_REFRESH data links signal incoming data by calling the 
 * corresonding callback routine even when the incoming data is a HEARTBEAT notification.
 * 
 * A client may wish to receive data notification events only when the incoming data have
 * changed or the data link has an error, and not receive and respond to heartbeat notifications.
 * Setting this global variable to TRUE will archieve this aim.
 *
 * \b Default: FALSE
 *  
 * \see SetNotificationTolerance()
 */
extern int SystemSupressHeartbeatNotification;
/* SystemPresetMemory */
/**
 * \brief Determines whether data requests preset the calling buffer to zero before issuing the call
 * 
 * If a server always returns the amount of data requested, then presetting the calling buffer
 * contents can be entirely superfluous, and in the case of large data call, can result in a
 * non-negligible increase in the server CPU load. On the other hand, if a server returns less than the
 * amount of data asked for, or the caller asks for an array of fixed string length data, there
 * could be unforseen side-effects if the calling buffer has not been zeroed out.
 *
 * \b Default: TRUE
 *  
 */
extern EXPORT int SystemPresetMemory;

/* SystemReset() */
/**
 * \brief Flushes all entry tables and returns all counters and semaphores to their initial state.
 * 
 *  Useful under some circumstances, for instance following program suspension or interrupt 
 * due to CNTL-C, etc.
 *
 * \param level is reserved for future use.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see _SystemInit(), _SystemCycle(),
 *
 * \b Alias: SystemReset
 *
 * \b Example:
 *
    if ((cc=SystemReset(0)) != 0) printf("Could not reset system : %s\n", erlst[cc]);

 */
EXPORT int _SystemReset(int level);
#define SystemReset(i) _SystemReset(i)

/* SystemCycle() */
/**
 * \brief Required call for the TINE engine to function property.
 * 
 * Should be called within an infinite loop.  Prior to the start of this loop, all other 
 * device server registration should occur.  
 * Clients with asynchronous links must see to it that _SystemCycle() is called regularly 
 * (unless SystemDelay(-1) is used to block the main process).  
 *
 * \param chkcmd is used to signal the presence of a local 
 * command interpreter.  Developers wishing to bypass the built-in command interpreter 
 * in favor of another should pass FALSE for this argument
 *  
 * \note SystemCycle does will in general NOT monopolize the CPU as for most Operating
 * systems, blocking sockets are used and the TINE Engine will
 * itself block on select() for a duration not greater than the SystemPollingRate. A
 * noteable exception is MSDOS, where non-blocking sockets are used and the TINE engine 
 * becomes a defacto operating system.
 *
 * \return 0 under normal circumstances. A non-zero return values indicates an immediate
 * need to re-enter (i.e. following a complete cycle, there are still incomplete data 
 * transfer pending).
 *
 * \see _SystemInit(), _SystemDelay(),
 *
 * \b Alias: SystemCycle
 *
 * \b Example:
 *
 * \include eg_main.c
 */
EXPORT int _SystemCycle(int chkcmd);
#define SystemCycle(i) _SystemCycle(i)

/* SystemDelay() */
/**
 * \brief Blocks any activity inside a TINE task or equipment modules for the 
 * specified amount of time.
 * 
 * The TINE engine will continue to run and process data which is not related to
 * 1) the equipment module being blocked if _SystemDelay() is called inside an
 * equipment module or 2) the background task if _SystemDelay() is called inside a
 * backgroung task.
 *
 * \param msec is the amount of time in milliseonds to suspend the calling routine.
 *  
 * \note The TINE engine is designed to run single-threaded.  You are free to use
 * threads and you can specify multi-threaded backgound tasks if you want the TINE
 * engine itself to explicitly use threads.  In such cases, calls to 'sleep()' would
 * have their intended effect.  Under single-threaded conditions (i.e. MSDOS, VMS, or
 * you're avoided threads because they complicate things) you should not use 'sleep()' 
 * since you will block then entire TINE engine.  Use _SystemDelay() instead if you want
 * to wait at a specific spot in your code until some action has finished.
 *
 * \return none
 *
 * \see _SystemCycle(), _SystemInit(),
 *
 * \b Alias: SystemDelay
 *
 * \b Example
 *
 * \include eg_SystemDelay.c
 *
 */
EXPORT void _SystemDelay(int msec);
#define SystemDelay(i) _SystemDelay(i)

/* SystemInit() */
/**
 * \brief Required call for the TINE server engine to function property.
 * 
 * Should be called prior to the start of the main TINE server engine.
 * Any deviations from default system parameters should be made before any call to 
 * _SystemInit().  This might include server settings which fix the size of the local 
 * connection table entries, the maximum date transport size, etc.
 * Note that _SystemInit() will begin be allocating space for its connection tables, so
 * modifying these values away from their defaults will have no effect after _SystemInit()
 * has been called.  Also, if the FEC Name is given by calling RegisterFecInformation() or 
 * RegisterFecNameEx(), then this should also preceed a call to _SystemInit().
 *
 *
 * \param netchk is used only for MS-DOS servers and if TRUE will initialize a network 
 * SAP watchdog to check connectivity to a NetWare file server.  For non-DOS servers it 
 * plays no role.
 *  
 * \note _SystemInit() does not and should not be called if you are writing a 'client-only'
 * application.  _SystemInit() will attempt to register various server settings which 
 * play no role, and will attempt to initiaze server sockets which could lead to confusing
 * situations on the local computer.  _SystemInit() does not initialize any client-side 
 * functioniality.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted 
 * by a call to GetRPCLastError().
 *
 * \see _SystemCycle(), _SystemDelay(),
 *
 * \b Alias: SystemInit
 *
 * \b Example:
 *
 * \include eg_main.c
 */
EXPORT int _SystemInit(int netchk);
#define SystemInit(i)  _SystemInit(i)
/* SystemScheduleProperty() */
/**
 * \brief Schedules the given property for immediate delivery to all attached clients.
 * 
 * When clients are attached to a particular property, they have specified a polling rate, which
 * defines the maximum latency for receiving updates at the client side.  This is usually fine.
 * However, if a server knows that important data have changed and which properties depend on these
 * data, it can signal the scheduler to call the given properties immediately and send the data to
 * all attached callers, regardless of their registered polling rates.  In this way, a server can
 * signal an event to its attached clients.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param prp is (comma separated) list of properties which are to be scheduled.
 *  
 * \return The error string associated with the input error number
 *
 * \b Alias: SystemScheduleProperty()
 *
 * \b Example:
 * 
 * \include eg_SystemScheduleProperty.c
 */
EXPORT int _SystemScheduleProperty(char *eqm, char *prp);
#define SystemScheduleProperty(a,b) _SystemScheduleProperty(a,b)
EXPORT int _SystemAssignBufferSpace(UINT32 rcvBufferSpace,UINT32 sndBufferSpace);
#define SystemAssignBufferSpace(a,b) _SystemAssignBufferSpace(a,b)

EXPORT void _SystemFreeResources(int exitOnFree);
#define SystemFreeResources(e) _SystemFreeResources(e)

/* SetBurstLimit() */
/**
 * \brief Sets the burst limit in number of packets which are allowed to be set consecutively
 * 
 * For UDP transport (the default) TINE does not make use of flow control in any systematic way,
 * other than adjusting this parameter setting at the server side.  This parameter works together
 * with the cycle delay and determines how many UDP packets (datagrams) are allowed to be sent
 * consecutively within the transport loop before a brake is applies (cycle delay).  If all clients
 * and servers are on the same speed ethernet, this value can be adjusted to a much larger value.
 * However, a fast server sending to a slow client (a remote client via a DSL link for instance) might
 * send more datagrams in a burst than can be received.  The default value should compensate for this
 * effect.
 *
 * \param npackets is the number of packet allowed to be set consecutively for a given call
 *  (default: 20).
 *  
 * \return The current setting for the burst limit.
 *
 * \see SetCycleDelay().
 */
EXPORT int SetBurstLimit(int npackets);

/* SetCycleDelay() */
/**
 * \brief Sets the delay time in milliseconds following a 'burst' of ethernet packets (as defined by
 * SetBurstLimit().
 * 
 * For UDP transport (the default) TINE does not make use of flow control in any systematic way,
 * other than adjusting this parameter setting at the server side.  This parameter works together
 * with the burst limit and determines how long to delay further datagram transmission when the burst
 * limit has been reached. If all clients
 * and servers are on the same speed ethernet, this value can be adjusted to 0.
 * However, a fast server sending to a slow client (a remote client via a DSL link for instance) might
 * send more datagrams in a burst than can be received.  The default value should compensate for this
 * effect.
 *
 * \param msec is the time in milliseconds to delay before sending further ethernet packets.
 *  (default: 20 msec).
 *  
 * \return The current setting for the cycle delay.
 *
 * \see SetBurstLimit().
 */
EXPORT int SetCycleDelay(int msecs);

/* setWorkAreaSize() */
/**
 * \brief Sets the server-side work area size which is coupled to the maximum transport size for
 * supported by a single-threaded server.
 * 
 * Server-side equipment function calls are double buffered, one buffer being a dynamic allocation
 * of the caller's requested data size, the other buffer being in the case of single threaded servers,
 * a temporary use of the server's work area.  In order to allow transfer of 'large' data sets, this
 * size needs to be adjustsed accordingly prior to server initialization.  Note that multi-threaded
 * servers will dynamically allocate both buffers on a call by call basis.
 *
 * \param size is the size in bytes to be reserved for the server's work area and hence the maximum
 *  allowed transport size for any property supported by the server.
 *  (default: 64 KBytes on most operating systems).
 *  
 * \return The current setting for the server's work area.
 *
 */
EXPORT UINT32 setWorkAreaSize(UINT32 size);

/* GetFeclogPath() */
/**
 * \brief Gets the current setting for the fec log repository.
 * 
 * \return A string pointing to the current setting for the fec log repository
 *
 */
EXPORT char *GetFeclogPath(void);

/* GetFecHome() */
/**
 * \brief Gets the current setting for the fec home database repository.
 * 
 * \return A string pointing to the current setting for the fec home database repository
 *
 */
EXPORT char *GetFecHome(void);

/* GetRPCLastError() */
/**
 * \brief The error string associated with the input error number
 * 
 * \param cc is the completion code for which the error string is desired.  If cc is in fact
 * the last Link Error then the returned error string is used, otherwise the system error
 * string for the completion code in question is returned
 *  
 * \return The error string associated with the input error number.
 *
 * \see GetCompletionStatus()
 *
 * \b Example:
 *
 * \include eg_GetRPCLastError.c
 */
EXPORT char *GetRPCLastError(short cc);
EXPORT short GetLastLinkError(short cc, char  *errstr);

/* SetRPCCompletion() */
/**
 * \brief Sets the error string to accompany the current server call
 * 
 * For user-defined errors codes (equal to or above 512) the accompanying error string
 * should be defined by calling this routine prior to returning from the equipment module
 * call.  When system error codes such as 'illegal_format' are used, a call to SetRPCCompletion() 
 * is not necessary.
 *
 * \param errstr is the 32-character string which is to accompany the outgoing error code.
 *  
 * \b Example:
 *
 * \include eg_SetRPCCompletion.c
 */
EXPORT void SetRPCCompletion(char *errstr);

/* SetRPCRedirection() */
/**
 * \brief Sets the redirection string to accompany the current server call
 * 
 * This routine can be used when Property Query Functions are in play.  In such cases the
 * equipment module is responsible for deciding if a call should be redirected to another
 * server or not.  If so, the server can call this routine to set the redirection information.
 *
 * \param rdrSrv is the device server to which the call should be redirected
 * \param rdrPrp is the device property to which the call should be redirected. (Note that
 * property lenght is restricted to 16 characters for redirected calls.
 *  
 */
EXPORT void SetRPCRedirection(char *rdrSrv,char *rdrPrp);

extern int (*lookupRedirectionNameStub)(char *eqm,char *prpName,char *devName);

/* SetlookupRedirectionNameStub() */
/**
 * \brief Sets a stub to an existing lookupRedirectionNameStub function for cases
 * where a Property Query Function is in force and the server needs to redirect 
 * calls.
 *
 * A lookupRedirectionNameStub function should return server_redirection if the 
 * property and device combination is to be redirected.  It should also 
 * call the SetRPCRedirection() routine to establish the redirection string.
 *
 * \param fcn is the address of the lookupRedirectionNameStub function.
 */
EXPORT void SetlookupRedirectionNameStub(int (*fcn)(char *eqm,char *prpName,char *devName));

/* ExecLink() */
/**
 * \brief Executes a synchronous link
 * 
 * Synchronous data exchange. ExecLink() does not complete until the data transfer has completed or 
 * a timeout has ensued. ExecLink() differs from ExecLinkEx() in that the default timeout 
 * of 1000 msec is assumed
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to 
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only 
 * be successful when the client and server are running together in the same process). 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any set of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.)
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see ExecLocalLink(), ExecLinkEx(), DTYPE
 *  
 *  
 * \b Example:
 *
 * \include eg_ExecLink.c
 */
EXPORT int ExecLink(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);

/* ExecLinkEx() */
/**
 * \brief Executes a synchronous link (Extended call).
 * 
 * Synchronous data exchange. ExecLinkEx() does not complete until the data transfer has completed or 
 * a timeout has ensued. ExecLinkEx() differs from ExecLink() in that the user may specify a timeout 
 * period for the call to complete.  
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to 
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only 
 * be successful when the client and server are running together in the same process). 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \see ExecLocalLink(), ExecLink(), DTYPE
 *  
 * \b Example:
 *
 * \include eg_ExecLinkEx.c
 */
EXPORT int ExecLinkEx(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access,UINT16 timeout);

/* AttachLink() */
/**
 * \brief Initiates an asynchronous link
 * 
 * Asynchronous data exchange. AttachLink() returns immediately with a positive link index if the 
 * device name can be resolved and there are sufficient resources on the client side.  Otherwise,
 * a call to AttachLink() returns a negative completion code which can be interpreted with GetRPCLastError().
 *
 * AttachLink() differs from AttachLinkEx() in that the callback will always return the link identifier 
 * as the callback id parameter in the callback routine.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to 
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only 
 * be successful when the client and server are running together in the same process). 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any of the TINE access codes ORed together (CA_READ,
 * CA_WRITE, etc.)
 * \param pollingRate is the polling rate in milliseconds as seen at the server.  For SINGLE transfers, 
 * this value also serves as a timeout parameter.
 * \param callback is the address of the user-supplied callback routine to be called when data arrive from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link 
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data 
 * transmission, cc will always be 0, otherwise it can be interpreted with GetRPCLastError().
 * \param mode is the transfer mode (see discussion on Data Transfer Modes).
 *
 * \return A positive link handle if successful, otherwise the negative of a TINE completion code which can be 
 * interpreted by a call to GetRPCLastError().
 *
 * \see AttachLinkEx(), DTYPE
 * 
 * \b Example:
 *
 * \include eg_AttachLink.c
 * 
 */
EXPORT int AttachLink(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access,
               int pollingRate,void (*callback)(int,int),int mode);

/* AttachLinkEx() */
/**
 * \brief Initiates an asynchronous link
 * 
 * Asynchronous data exchange. AttachLinkEx() returns immediately with a positive link index if the 
 * device name can be resolved and there are sufficient resources on the client side.  Otherwise,
 * a call to AttachLinkEx() returns a negative completion code which can be interpreted with GetRPCLastError().
 *
 * AttachLinkEx() differs from AttachLink() in that the caller can supply a callback identify to be
 * returned in the callback function.  In lieu of this, the callback will return the link identifier.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". Note that, if devName references a locally registered equipment module,
 * an internal call to ExecLocalLink() is made.  For instance: "/LOCAL/BPMEQM/WL167" sets the context to 
 * "LOCAL", and "BPMEQM" must be a locally registered equipment module. In such a case, the call can only 
 * be successful when the client and server are running together in the same process). 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any of the TINE access codes ORed together
 * (CA_READ, CA_WRITE, etc.).
 * \param pollingRate is the polling rate in milliseconds as seen at the server.  For SINGLE transfers, 
 * this value also serves as a timeout parameter.
 * \param callback is the address of the user-supplied callback routine to be called when data arrive from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link 
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data 
 * transmission, cc will always be 0, otherwise it can be interpreted with GetRPCLastError().
 * \param mode is the transfer mode (see discussion on Data Transfer Modes).
 * \param callbackID is supplied by the caller as an identifier to be returned in the callback funtion
 *
 * \return A positive link handle if successful, otherwise the negative of a TINE completion code which can be 
 * interpreted by a call to GetRPCLastError().
 *
 * \see AttachLink(), DTYPE
 *  
 * \b Example:
 *
 * \include eg_AttachLinkEx.c
 */
EXPORT int AttachLinkEx(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access,
               int pollingRate,void (*callback)(int,int),int mode,UINT32 callbackID);

/* ModifyLinkAttributes() */
/**
 * \brief Allows the caller to assign new link attributes to an active link
 * 
 * A TINE client by making this call can re-assign a callback function, access mode or polling rate currently in place for
 * an active link.
 *
 * \param i is the link identifier for the link in question (the returned identifier when the call to
 * AttachLink() or AttachLinkEx() was originally made
 * \param access is the data access mode.  This can be any combination of TINE access code ORed
 * together (e.g. CA_READ|CA_WRITE).
 * \param pollingRate is the polling rate in milliseconds as seen at the server.  For SINGLE transfers, 
 * this value also serves as a timeout parameter.
 * \param callback is the address of the user-supplied callback routine to be called when data arrive from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link 
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data 
 * transmission, cc will always be 0, otherwise it can be interpreted with GetRPCLastError().
 * \param mode is the transfer mode (see discussion on Data Transfer Modes).
 * \param callbackID is supplied by the caller as an identifier to be returned in the callback funtion
 *
 * \return 0 if successful, otherwise a TINE completion code which can be 
 * interpreted by a call to GetRPCLastError().
 *
 * \see AttachLink(), AttachLinkEx(), DTYPE
 *  
 */
EXPORT int ModifyLinkAttributes(int i,short access,int pollingRate,
               void (*callback)(int,int),int mode,UINT32 callbackID);

/* SetNotificationTolerance() */
/**
 * \brief Allows the caller to apply a tolerance pertaining to link notification.
 * 
 * When a TINE client subscribes to data using CM_REFRESH mode, the data is sent from server to client
 * when the data change without regard to any applied tolerance (zero tolerance).  The client application
 * may nonetheless wish to be notified only when the data have changed outside a given tolerance.
 * This can be achieved by calling this routine.  The same applies to CM_POLL mode, where the data are
 * sent server to client according to the schedule.  In the latter case, all data comparisons are made 
 * entirely at the client side.  Note that if this feature is desired in CM_POLL mode, the CM_NODUPLICATES 
 * switch must be applied to the mode parameter in the call to AttachLink() or AttachLinkEx().  In
 * CM_POLL mode, it is sufficient that the SystemSupressHeartbeatNofification global switch is set to
 * TRUE to use this feature (otherwise append CM_NODUPLICATES).
 *
 * \param linkId is the link identifier for the link in question (the returned identifier when the call to
 * AttachLink() or AttachLinkEx() was originally made
 * \param toleranceAbsolute if non-zero is a tolerance in the data units of the in-coming data.  It is
 * checked against irrespective of the current data values.
 * \param tolerancePercent if non-zero is a tolerance expressed as a percent, and refers to the previous
 * data set.
 *
 * If both an absolute tolerance and a percent tolerance are used, the absolute value of both contributions
 * are added together to define the total tolerance which should be checked against.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be 
 * interpreted by a call to GetRPCLastError().
 *
 * \see AttachLink(), AttachLinkEx()
 *  
 */
EXPORT int SetNotificationTolerance(int linkId, float toleranceAbsolute, float tolerancePercent);

/* recvNetGlobal() */
/**
 * \brief Initiates a net globals data link
 * 
 * A TINE client requests to receive network global parameters via this routine.  Note that the 
 * client must reside on a network which itself is receiving broadcast data or onto which multicasts
 * are being routed in order for this call to take effect.  
 *
 * recvNetGlobal() differs from recvNetGlobalEx() in that the callback will return the link identifier
 * in the callback function.
 *
 * \param keyword is the name of the network global which is desired.
 * For example: "HEDCCur". 
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned.
 * \param callback is the address of the user-supplied callback routine to be called when data arrive from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link 
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data 
 * transmission, cc will always be 0, otherwise it can be interpreted with GetRPCLastError().
 *
 * \return A positive link handle if successful, otherwise the negative of a TINE completion code which can be 
 * interpreted by a call to GetRPCLastError().
 *
 * \see recvNetGlobalEx(), sendNetGlobal(), DTYPE
 *  
 * \b Example:
 *
 * \include eg_recvNetGlobal.c
 */
EXPORT int recvNetGlobal(char *keyword,DTYPE *dout,void (*callback)(int,int));

/* recvNetGlobalEx() */
/**
 * \brief Initiates a net globals data link (Extended call).
 * 
 * A TINE client requests to receive network global parameters via this routine.  Note that the 
 * client must reside on a network which itself is receiving broadcast data or onto which multicasts
 * are being routed in order for this call to take effect.  
 *
 * recvNetGlobalEx() differs from recvNetGlobal() in that the caller can supply a callback identify to be
 * returned in the callback function.  In lieu of this, the callback will return the link identifier.
 *
 * \param keyword is the name of the network global which is desired.
 * For example: "HEDCCur". 
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned.
 * \param callback is the address of the user-supplied callback routine to be called when data arrive from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link 
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data 
 * transmission, cc will always be 0, otherwise it can be interpreted with GetRPCLastError().
 * \param callbackID is supplied by the caller as an identifier to be returned in the callback funtion
 *
 * \return A positive link handle if successful, otherwise the negative of a TINE completion code which can be 
 * interpreted by a call to GetRPCLastError().
 *
 * \see recvNetGlobal(), sendNetGlobal(), DTYPE
 *  
 * \b Example:
 *
 * \include eg_recvNetGlobalEx.c
 */
EXPORT int recvNetGlobalEx(char *keyword,DTYPE *dout,void (*callback)(int,int),UINT32 callbackID);

int PACKRPC(char *Name);

/* RegisterFecInformation() */
/**
 * \brief Assigns a FEC Name and descriptive information to the server process.
 * 
 * Servers must be assigned to a Front End Computer (FEC), which for many operating systems
 * corresponds to a process running on the computer rather than the computer itself (meaning
 * there can in some cases be more than one servers with different 'FEC' names running on 
 * the same computer.  There can be several device servers attached to the same FEC (i.e.
 * sharing the same address space), so the FEC name is a distinct quantity from the
 * device server name (see RegisterExport()), and this name must also be system-wide 
 * unique.  The call RegisterFecInformation() supercedes RegisterFecNameEx() and RegisterFecName()
 * and differs from them in that it accepts 'Context' and 'SubSystem' as parameters, both of
 * which will be applied to all device servers attaching to the same FEC name.
 *
 * The FEC name can also be (and frequently is) registered via the presence of an
 * 'fecid.csv' startup database file.
 *
 * \param name is the FEC name indentifying the Front End Computer and to which all registered 
 *  equipment modules are bound.
 * \param sub is the subsystem to which attached device servers belong.  Note that the subsystem
 * (unlike context) is itself not part of the name space, although the name space can be 
 * queried for device servers belonging to a subsystem.
 *  System.  This information is now deduced automatically from the library build.
 * \param cntxt is a 16-character string giving the context of all registered equipment modules.
 * \param dsc is a 32-character desciption of the FECs server duties.  A 'subsystem' can be 
 *  associated with ALL equipment modules found on the FEC by prefixing the description with
 *  a subsystem tag of the form "[<tag>]".  This information will be parsed by the Equipment
 *  name server and used for sub-system specific queries.  For example, the associate the FEC
 *  "HEPHF" with the subsystem "RF", the description field might be "[RF]Hera Proton HF Control".
 * \param loc is a 32-character string giving the phyical location of the FEC.
 * \param hwd is a 32-character brief description of the IO hardware found on the FEC.
 * \param resp is a 32-character string listing the developer(s) responsible for the FEC. Note:
 *  The Equipment Name Server will allow the removal of FECs and associated equipment modules
 *  to the user(s) specified here.
 * \param poff is the 'Port Offset' to be applied to the FEC.  This parameter plays an important
 * role where more that one FEC is to run on a machine running an operating system using virtual 
 * memory.  In such cases 'Front End Computer' is a misnomer, since "FEC" really refers to a 
 * process running on the computer to which one or more equipment modules are bound.  For 
 * operating systems where all processes run in the same address space (such as VxWorks, MS-DOS,
 * Win16, NIOS) there is in fact only one such processes managing all requistered equipment
 * modules.  For systems using virtual memory (such as Linux, Solaris, HP-UX, Win32, etc.),
 * you can have many such processes running independently of one another.  Since each such
 * process must listen on a unique set of server ports, you must see to it that all such 
 * FEC processes are registered with a unique 'Port Offset'.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \note FEC Names can also be registered by including the startup file 'fecid.csv'.  In the 
 * file, most of the input parameters appear as (optional) column entries.  It is frequently 
 * preferable to work with startup configuation files, since it is frequently desireable to 
 * avoid hard-coding names and description parameters.
 *
 * \see RegisterFecName(), RegisterFecNameEx().
 *
 */
EXPORT int RegisterFecInformation(char *name,char *sub,char *cntxt,char *dsc,char *loc,char *hdw,char *resp,UINT16 poff);

/* RegisterFecNameEx() */
/**
 * \brief Assigns a FEC Name to the server process.  Extended call.
 * 
 * Servers must be assigned to a Front End Computer (FEC), which for many operating systems
 * corresponds to a process running on the computer rather than the computer itself (meaning
 * there can in some cases be more than one servers with different 'FEC' names running on 
 * the same computer.  There can be several device servers attached to the same FEC (i.e.
 * sharing the same address space), so the FEC name is a distinct quantity from the
 * device server name (see RegisterExport()), and this name must also be system-wide 
 * unique.  The extended call RegisterFecNameEx() differs from RegisterFecName() in that
 * it accepts the server 'Context' as an additional parameter.  The context will be applied
 * to all device servers attaching to the same FEC name.
 *
 * The FEC name can also be (and frequently is) registered via the presence of an
 * 'fecid.csv' startup database file.
 *
 * \param name is the FEC name indentifying the Front End Computer and to which all registered 
 *  equipment modules are bound.
 * \param dsc is a 32-character desciption of the FECs server duties.  A 'subsystem' can be 
 *  associated with ALL equipment modules found on the FEC by prefixing the description with
 *  a subsystem tag of the form "[<tag>]".  This information will be parsed by the Equipment
 *  name server and used for sub-system specific queries.  For example, the associate the FEC
 *  "HEPHF" with the subsystem "RF", the description field might be "[RF]Hera Proton HF Control".
 * \param os is included for backward compatibility.  Historicall it defined the FEC's Operating 
 *  System.  This information is now deduced automatically from the library build.
 * \param loc is a 32-character string giving the phyical location of the FEC.
 * \param hwd is a 32-character brief description of the IO hardware found on the FEC.
 * \param resp is a 32-character string listing the developer(s) responsible for the FEC. Note:
 *  The Equipment Name Server will allow the removal of FECs and associated equipment modules
 *  to the user(s) specified here.
 * \param poff is the 'Port Offset' to be applied to the FEC.  This parameter plays an important
 * role where more that one FEC is to run on a machine running an operating system using virtual 
 * memory.  In such cases 'Front End Computer' is a misnomer, since "FEC" really refers to a 
 * process running on the computer to which one or more equipment modules are bound.  For 
 * operating systems where all processes run in the same address space (such as VxWorks, MS-DOS,
 * Win16, NIOS) there is in fact only one such processes managing all requistered equipment
 * modules.  For systems using virtual memory (such as Linux, Solaris, HP-UX, Win32, etc.),
 * you can have many such processes running independently of one another.  Since each such
 * process must listen on a unique set of server ports, you must see to it that all such 
 * FEC processes are registered with a unique 'Port Offset'.
 * \param context is a 16-character string giving the context of all registered equipment modules.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \note FEC Names can also be registered by including the startup file 'fecid.csv'.  In the 
 * file, most of the input parameters appear as (optional) column entries.  It is frequently 
 * preferable to work with startup configuation files, since it is frequently desireable to 
 * avoid hard-coding names and description parameters.
 *
 * \see RegisterFecName(), RegisterFecInformation().
 *
 */
EXPORT int RegisterFecNameEx(char *name,char *dsc,char *os,char *loc,char *hdw,char *resp,UINT16 poff,char *context);

/* RegisterFecName() */
/**
 * \brief Assigns a FEC Name to the server process
 * 
 * Servers must be assigned to a Front End Computer (FEC), which for many operating systems
 * corresponds to a process running on the computer rather than the computer itself (meaning
 * there can in some cases be more than one servers with different 'FEC' names running on 
 * the same computer.  There can be several device servers attached to the same FEC (i.e.
 * sharing the same address space), so the FEC name is a distinct quantity from the
 * device server name (see RegisterExport()), and this name must also be system-wide 
 * unique.
 *
 * The FEC name can also be (and frequently is) registered via the presence of an
 * 'fecid.csv' startup database file.
 *
 * \param name is the FEC name indentifying the Front End Computer and to which all registered 
 *  equipment modules are bound.
 * \param dsc is a 32-character desciption of the FECs server duties.  A 'subsystem' can be 
 *  associated with ALL equipment modules found on the FEC by prefixing the description with
 *  a subsystem tag of the form "[<tag>]".  This information will be parsed by the Equipment
 *  name server and used for sub-system specific queries.  For example, the associate the FEC
 *  "HEPHF" with the subsystem "RF", the description field might be "[RF]Hera Proton HF Control".
 * \param os is included for backward compatibility.  Historicall it defined the FEC's Operating 
 *  System.  This information is now deduced automatically from the library build.
 * \param loc is a 32-character string giving the phyical location of the FEC.
 * \param hwd is a 32-character brief description of the IO hardware found on the FEC.
 * \param resp is a 32-character string listing the developer(s) responsible for the FEC. Note:
 *  The Equipment Name Server will allow the removal of FECs and associated equipment modules
 *  to the user(s) specified here.
 * \param poff is the 'Port Offset' to be applied to the FEC.  This parameter plays an important
 * role where more that one FEC is to run on a machine running an operating system using virtual 
 * memory.  In such cases 'Front End Computer' is a misnomer, since "FEC" really refers to a 
 * process running on the computer to which one or more equipment modules are bound.  For 
 * operating systems where all processes run in the same address space (such as VxWorks, MS-DOS,
 * Win16, NIOS) there is in fact only one such processes managing all requistered equipment
 * modules.  For systems using virtual memory (such as Linux, Solaris, HP-UX, Win32, etc.),
 * you can have many such processes running independently of one another.  Since each such
 * process must listen on a unique set of server ports, you must see to it that all such 
 * FEC processes are registered with a unique 'Port Offset'.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \note FEC Names can also be registered by including the startup file 'fecid.csv'.  In the 
 * file, most of the input parameters appear as (optional) column entries.  It is frequently 
 * preferable to work with startup configuation files, since it is frequently desireable to 
 * avoid hard-coding names and description parameters.
 *
 * \see RegisterFecNameEx(), RegisterFecInformation().
 *  
 */
EXPORT int RegisterFecName(char *name,char *desc,char *os,char *loc,char *hdw,char *resp,UINT16 poff);

/* RegisterPropertyEx() */
/**
 * \brief Assigns pertinent information for the specified property.
 * 
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * The general behavior of the property (including descriptions) can be specified with this
 * call.  Properties can also be (and frequently are) registered via the presence of an
 * 'exports.csv' startup database file.
 *
 * \note Properties which are overloaded can and should be registered more than once
 * with the appropriate characteristics supplied in the registration call.  However,
 * if input data sets require different data set specifications than the output data
 * sets then the call to RegisterPropertyInformation() should be used.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param prop is property name in question (up to 32 characters, preferably 16 or less).
 * \param siz is expected data size (i.e. maximum allowed data size) for this property
 * \param fmt is exected data format for this property
 * \param acc is the data 'access' (CA_READ or CA_WRITE) accepted by the server.
 * \param dsc is the 64-character description of what the property reads or writes.
 * \param propId is the property identifier to be associated with the property name 
 * returned in calls to GetPropertyId().
 *
 * \return The (positive) associated property identifier if successful, otherwise the
 * negative of a TINE return code.
 *
 * \see RegisterProperty(), RegisterPropertyInformation(), GetPropertyId()
 *  
 */
EXPORT int RegisterPropertyEx(char *eqm,char *prop,int siz,short fmt,short acc,char *dsc,int prpId);

/* RegisterProperty() */
/**
 * \brief Assigns pertinent information for the specified property.
 * 
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * The general behavior of the property (including descriptions) can be specified with this
 * call.  Properties can also be (and frequently are) registered via the presence of an
 * 'exports.csv' startup database file.
 *
 * \note Properties which are overloaded can and should be registered more than once
 * with the appropriate characteristics supplied in the registration call.  However,
 * if input data sets require different data set specifications than the output data
 * sets then the call to RegisterPropertyInformation() should be used.  If the property
 * identifier is important, then either RegisterPropertyInformation() or 
 * RegisterPropertyEx should be used, otherwise, the associated property identifier will
 * be supplied by the server engine and be given via the return code.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param prop is property name in question (up to 32 characters, preferably 16 or less).
 * \param siz is expected data size (i.e. maximum allowed data size) for this property
 * \param fmt is exected data format for this property
 * \param acc is the data 'access' (CA_READ or CA_WRITE) accepted by the server.
 * \param dsc is the 64-character description of what the property reads or writes.
 * returned in calls to GetPropertyId().
 *
 * \return The (positive) associated property identifier if successful, otherwise the
 * negative of a TINE return code.
 *
 * \see RegisterPropertyEx(), RegisterPropertyInformation(), GetPropertyId()
 *  
 */
EXPORT int RegisterProperty(char *eqm,char *prop,int siz,short fmt,short acc,char *dsc);

/* RegisterPropertyInformation() */
/**
 * \brief Assigns pertinent information for the specified property.
 * 
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * The general behavior of the property (including descriptions) can be specified with this
 * call.  Properties can also be (and frequently are) registered via the presence of an
 * 'exports.csv' startup database file.
 *
 * \note Properties which are overloaded can and should be registered more than once
 * with the appropriate characteristics supplied in the registration call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param prop is property name in question (up to 32 characters, preferably 16 or less).
 * \param dout is a template data object (minus a data pointer) giving the default format
 * size, and structure tag (if any) of the output data set to be returned by the server 
 * to the caller.
 * \param din is a template data object (minus a data pointer) giving the default format
 * size, and structure tag (if any) of the input data set to be supplied by the caller.
 * \param acc is the data 'access' (CA_READ or CA_WRITE) accepted by the server.
 * \param atype is the data 'array type' (AT_SCALAR, AT_SINGLE, etc.) associated with the
 * returned data.
 * \param rowlength if > 0 gives the row length of the returned array data.  This is 
 * most relevant for double arrays, where the number of rows is then given by the output
 * data size divided by the row length.
 * \param dsc is the 64-character description of what the property reads or writes.
 * \param propId is the property identifier to be associated with the property name 
 * returned in calls to GetPropertyId().
 * \param rdr is a redirection string, if the property is to be redirected
 * to another server.  Most properties are not redirected, and this parameter is usually
 * NULL.  If used, if will be parsed according to 
 * \<device server\>/\<device name\>[device property], where the 'device name' and 
 * 'device property' are optional.  If 'device name' is not present, all devices will be
 * redirected to the specified device server.  If 'device property' is not present
 * then the identical property name as registered will be redirected to the device server
 * specified.  If on the other hand, the device property is present in the string to be
 * parsed, then the property registered will be redirected to the device server specified
 * as well as to the device property specified.
 *
 * \return The (positive) associated property identifier if successful, otherwise the
 * negative of a TINE return code.
 *
 * \see RegisterProperty(), RegisterPropertyEx(), GetPropertyId()
 *  
 */
EXPORT int RegisterPropertyInformation(char *eqm,char *prop,DTYPE *dout,DTYPE *din,
                                       short acc,short atype,UINT16 rowlength,char *dsc,int propId,char *rdr);

/* RedirectProperty() */
/**
 * \brief Applies the redirection string to the given property
 * 
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * A server can export properties or devices which live on other servers.  To do so a
 * redirection string needs to be applied to the property in question.  As different registered
 * devices can also be redirected to different servers.  This routine can be called several
 * times for the same property, once for each remote device.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param prop is property name in question (up to 32 characters, preferably 16 or less).
 * \param rdr is the redirection string of the form "/<context>/<server>/<device>/[property]"
 * where <server> is the only non-optional entry in the redirection string.  If <device> is
 * omitted, all calls for the given property will be redirected to the given <server>.  If
 * device is included, only those calls to the specified device and property will be redirected.
 * If <property> is specified then the given property name will be redirected to the remote 
 * <server> and the remote <property>.
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \see RegisterProperty(), RegisterPropertyEx(), RegisterPropertyInformation(), RedirectDeviceName()
 *  
 */
EXPORT int RedirectProperty(char *eqm,char *prop,char *rdr);

/* RegisterDeviceName() */
/**
 * \brief Assigns a device name to the specified device number
 * 
 * Many servers handle instances of a given device type, such as a set of
 * vacuum pumps, beam position monitors, magnet power supply controllers, etc. It is 
 * frequently very efficient to code in terms of device numbers which
 * might be an index into an array, but
 * to specify modules in calls by their human-readable names.  For such purposes a device 
 * name can be assigned to a number at initialization via a call to RegisterDeviceName()
 * (alias: RegisterModuleName()).  An alternative is to provide a startup database file '\<EQPNAME>\>-devices.csv' containing
 * a cross-reference for numbers and names.  Internally, a hash table is maintained for 
 * fast lookups inside equipment module routines.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param devname is device name to be assigned
 * \param devnr is the device number associated with the module name specified.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see GetDeviceName(), GetDeviceNumber()
 *  
 */
EXPORT int RegisterDeviceName(char *eqm,char *devname,int devnr);
#define RegisterModuleName(eqm,dev,devnr) RegisterDeviceName(eqm,dev,devnr)

/* AssignPropertyList() */
/**
 * \brief Assigns the given property list to the given device to be used in device-oriented queries
 * 
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * Most device servers provide a list of all instances, a device list, and a list of all properties,
 * where every instance (device) supports all exported properties.  Under some circumstances,
 * it is desireable to support only a (device-specific) subset of properties per device.  This
 * can be achieved by supplying a property query function, or by configuration file, or by using 
 * this call to assign a property list to the specified device.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param devname is device name in question (up to 32 characters, preferably 16 or less).
 * \param listname is a 'tag' name to be assigned to the list. Where many devices share the same
 * property list, this tag is used to avoid unnecessarily allocating new memory for the same list.
 * \param listsize is the length of the property list.
 * \param list is the property list as an array of NAME32 entities.  If the list provided is not
 * at least as long as indicated by parameter listsize, the server's behavior is undefined.
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \see RegisterPropertyQueryFunction(), RegisterXPropertyQueryFunction().
 *  
 */
EXPORT int AssignPropertyList(char *eqm,char *devname,char *listname,int listsize,NAME32 *prplist);

/* RedirectDeviceName() */
/**
 * \brief Applies the redirection string to the given device for all properties
 * 
 * Servers export their behavior through Properties, which can be thought of as 'properties'
 * in the Object Oriented sense with Get and Set methods.  That is, any read or write access
 * for a given property will call into the associated equipment module and be handled there.
 * A server can export devices which live on other servers.  To do so a
 * redirection string needs to be applied to the device in question.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param devname is device name in question (up to 32 characters, preferably 16 or less).
 * \param rdr is the redirection string whose general form is "/<context>/<server>/<device>/[property]".
 * The only relevant tag in the redirection string is the <server> tag as a call to RedirectDeviceName
 * will apply to all properties for the specified device.  The <context> tag is as yet always
 * assumed to be the same context as the redirecting server.  Thus the redirection string can
 * be specified as simply "<server>"
 *
 * \return 0 upon success, otherwise a TINE error code.
 *
 * \see RedirectProperty(), RegisterProperty(), RegisterPropertyEx(), RegisterPropertyInformation()
 *  
 */
EXPORT int RedirectDeviceName(char *eqm,char *devname,char *rdr);

/* GetDeviceName() */
/**
 * \brief Gives the registered device name for the specified equipment module and device number
 * 
 * A server can determine the name which has be associated with a device number by making this
 * call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param devnr is the device number to be looked up.
 *
 * \return The device name corresponding to the input device number if successful, otherwise a
 * NULL pointer (indicating that the name input is not registered as a device on the server.)
 *
 * \see RegisterDeviceName(), GetDeviceNumber()
 *
 * \b Alias: GetModuleName
 *
 * \b Example:
 *
   printf("device %d is  %s\n",i,GetDeviceName("EQP1",i));

 */
EXPORT char *GetDeviceName(char *eqm,int devnr);
#define GetModuleName(eqm,devnr) GetDeviceName(eqm,devnr)

/* GetExportedDeviceList() */
/**
 * \brief Returns the list of exported device names explicitly associated with the
 * equipment module passed as the input parameter.
 * 
 * A server can retieve the list of explicitly exported device names by making this 
 * call.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 *
 * \return A array of NAME32 objects containing the exported device names or a NULL
 * pointer if there is an error, such as the equipment module given is not registered or
 * there are no exported device names associated with the equipment module.
 *
 * \see RegisterDeviceName()
 *
 */
EXPORT NAME32 *GetExportedDeviceList(char *eqm);

/* GetRegisteredPropertyList() */
/**
 * \brief Returns the list of registered properties explicitly associated with the
 * equipment module passed as the first argument input parameter.
 * 
 * A server can retieve the list of registered properties by making this 
 * call.  This is sometimes useful when the property list is maintained in an
 * 'exports.csv' file and the server application needs to retrieve this list.
 *
 * \param eqm (input) is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM".
 * \param prpNames (output) is the address location which is to hold the property list
 * upon completion
 * \param nprps (intput/output) is a pointer to an integer which as input specifies the size of the
 * buffer pointed to by the second parameter. Upon return it contains the actual number of 
 * registered properties.
 *
 * \return '0' upon success, othereise a tine return code.
 *
 */
EXPORT int GetRegisteredPropertyList(char *eqm,NAME32 *prpNames,int *nprps);

/* getLocalName() */
/**
 * \brief Returns the local equipment module name associated with the export name given.
 * 
 * \param exportName is the exported device server name for which the local equipment module 
 * name is desired.
 *
 * \return a string pointing to the local equipment module name.
 *
 */
EXPORT char *getLocalName(char *exportName);

EXPORT int GetNumberEquipmentModules(void);

/* GetNumberRegisteredDevices() */
/**
 * \brief Gives the number of registered devices explicitly associated with the equiment module name
 * given.
 * 
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 *
 * \return The number of registered devices if successful, or the 
 * negative of a TINE return code (indicating that the name input is not registered as a device on the server.)
 *
 * \see RegisterDeviceName(), GetDeviceNumber()
 *  
 */
EXPORT int GetNumberRegisteredDevices(char *eqm);

/* GetDeviceNumberEx() */
/**
 * \brief Gives the registered device number for the specified device name and property name
 * 
 * Callers frequently address specific devices by their human-readable names rather than
 * a device number (which might be an entry point into an array, for example).  If the
 * caller registers device names via calls to RegisterDeviceName() or has registered a "<property>.NAM"
 * meta property (corresponding to a property-specific set of device names for <property>), then the 
 * cross reference exists in the form of a hash table.  A server's equipment module can thus rapidly 
 * determine the corresponding device number from an input device name by calling GetDeviceNumberEx().
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param devname is the device name to be looked up
 * \param prpname if non-NULL should be the targetted property name for which the device names apply
 *
 * \return The device number corresponding to the input device name if successful, otherwise the
 * negative of a TINE return code (indicating that the name input is not registered as a device on the server.)
 *
 * \see RegisterDeviceName(), GetDeviceNumber()
 *  
 */
EXPORT int GetDeviceNumberEx(char *eqm,char *devname,char *prpname);
#define GetModuleNumberEx(eqm,dev,prp) GetDeviceNumberEx(eqm,dev,prp)

/* GetDeviceNumber() */
/**
 * \brief Gives the registered device number for the specified device name
 * 
 * Callers frequently address specific devices by their human-readable names rather than
 * a device number (which might be an entry point into an array, for example).  If the
 * caller registers device names via calls to RegisterDeviceName() then the cross reference
 * exists in the form of a hash table.  A server's equipment module can thus rapidly determine
 * the corresponding device number from an input device name by calling GetDeviceNumber().
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param devname is the device name to be looked up
 *
 * \return The device number corresponding to the input device name if successful, otherwise the
 * negative of a TINE return code (indicating that the name input is not registered as a device on the server).
 *
 * \see RegisterDeviceName(), GetDeviceNumberEx()
 *
 * \b Alias: GetModuleNumber
 *  
 * \b Example:
 *
 * \include eg_GetDeviceNumber.c
 */
EXPORT int GetDeviceNumber(char *eqm,char *devname);
#define GetModuleNumber(eqm,dev) GetDeviceNumber(eqm,dev)

/* GetPropertyId() */
/**
 * \brief Gives the associated property identifier for the given property name
 * 
 * Callers will alway address properties by their human-readable names. 
 * Equipment modules can either initiate a series of strcmp()s in order to determine
 * which property is being called or make direct use of the associated property identifier
 * in a case switch or address table.  Note that properties registered with RegisterProperty()
 * simply return an index in the server's ordered list.  You can associate any indentifier you
 * like by calling either RegisterPropertyEx() or RegisterPropertyInformation() to register 
 * the property.  The cross reference from property name to property identifier
 * exists in the form of a hash table.  A server's equipment module can thus rapidly determine
 * the corresponding identifier from an input property by calling GetPropertyId().
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param prpName is the property name to be looked up
 *
 * \return The property identifier corresponding to the input property name if successful, 
 * otherwise -1 (indicating that the name input is not registered as a property on the server).
 *
 * \see RegisterProperty(), RegisterPropertyEx(), RegisterPropertyInformation()
 *  
 * \b Example:
 *
 * \include eg_GetPropertyId.c
 */
EXPORT int GetPropertyId(char *eqm,char *prpName);

int RegisterExport(char *tagname,char *eqpname,int numModules);

/* JoinEquipmentGroup() */
/**
 * \brief Instructs the equipment module to join the specified equipment group.
 * 
 * It is sometimes desireable to access a collection of like devices via a 
 * device group, rather than a device server.  This is of course true when the
 * devices in the group span more than one physical device server.
 * You can use this API call to instruct the equipment module to join the specified 
 * group.  You should use this call prior to calling SystemCycle().
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param groupname is the name of the group the equipment module is to join.
 * \param groupindex is an ordered index, which can specify which location within the
 * group this equipment module's devices are to have.  If '0' is passed, the group
 * server will assign an order.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
EXPORT int JoinEquipmentGroup(char *eqm,char *groupname,int groupindex);

/* RegisterEquipmentModule() */
/**
 * \brief Registers an equipment module with the TINE server engine
 * 
 * The point of contract between the TINE server engine and server-specific code is the
 * equpment module, which can be regarded as an interupt service routine.
 *
 * \param expName is the exported device server name to be associated with this equipment module.
 * If this is NULL, then it is assumed that an 'exports.csv' local database file exists, which
 * contains the export name.  Otherwise it will contain a system-wide unique name (e.g. "HERA-P-BPM")
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param ndevices is this number of devices managed via this equipment module.  If this is 0, then
 * it is assumed that an 'exports.csv' local database file exists, which contains the number of devices.
 * \param fcn is a reference to the equipment module, which must have the prototype shown.  Namely, it
 * is an interrupt service routine which will receive the device name, device property, the output data
 * set, the input data set, and the data access flags as parameters.
 * \param ini is a reference to an initialization routine which should be called prior to enabling the 
 * equipment module.  It does not take arguments.
 * \param tsk is a reference to a background task which will be called at regular intervals, and is
 * frequently used for hardware IO, etc.  
 * Where multi-threaded servers are specified, or VxWorks tasks are available these tasks exists as 
 * separate threads, or tasks.  In other cases, they are managed by the TINE server engine scheduler.
 * \param rate is the rate in milliseconds used to schedule the background task supplied by the 'tsk'
 * parameter.  
 * \param exe is a reference to an exit routine which should be called prior to shutting down the 
 * equipment module.  It does not take arguments, and applies of course to graceful exits of the server.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
EXPORT int RegisterEquipmentModule(
    char *expName,
    char *eqm,
    int ndevices,
    int (*fcn)(char *,char *,DTYPE *,DTYPE *,short),
    void (*ini)(void),
    void (*tsk)(void),
    int rate,
    void (*exi)(void)
);
int RegisterNodalModule(
    char *expName,
    char *eqpname,
    int NumModules,
#   ifdef NODAL
    int (*fcn)(double *,int,int,char *),
#   else
    int (*fcn)(double *,short,short,short,char *),
#   endif
    void (*ini)(void),
    void (*tsk)(void),
    int rate,
    void (*exi)(void)
);
int PutRPCErrorValue(int i, void *errValue);

/**
 * \struct PropertyQueryStruct
 * \brief Defines a simple property query structure
 *
 * - prpName is the property name
 * - prpDescription is the property description
 * - prpSize is the maximum allowed size of an output data set associated with the property
 * - prpFormat is the default output data format accepted for this property
 * - prpAccess is the allowed data access for this property (either CA_READ, CA_WRITE or both)
 * GT_LIN, CF_LOG, etc. (see rpcdefs.h).
 *
 * \note As there is no way to distinguish input data set from output data sets with this
 * simple form of the query structure, the assumption that the input size and type are the
 * same as the output size and type (if input data are required) is usually made.  The 
 * extended property query structure 'PropertyQueryStructEx' is more complete in this regard.
 *
 */
typedef struct
{
  char  prpName[32];
  char  prpDescription[32];
  UINT16 prpSize;
  BYTE prpFormat;
  BYTE prpAccess;
} PropertyQueryStruct;
#define PROPERTYQUERYSTRUCT_SIZE (32+32+2+1+1)

/**
 * \struct PropertyQueryStructEx
 * \brief Defines an extended property query structure
 *
 * - prpName is the property name
 * - prpDescription is the property description
 * - prpRedirection is a redirection string if the property is redirected to another server
 * - prpTag is the data tag for output data sets, applicable if the output data format is
 * CF_STRUCT (i.e. a user defined structure)
 * - prpTagIn is the data tag for input data sets, applicable if the input data format is
 * CF_STRUCT (i.e. a user defined structure)
 * - prpUnits is a string containing the natural units of the data returned by the property
 * - prpMinValue is the lower limit of valid data associated with the property
 * - prpMaxValue is the upper limit of valid data associated with the property
 * - prpSize is the maximum allowed size of an output data set associated with the property
 * - prpSizeIn is the maximum allowed size of an input data set associated with the property
 * - prpNumOverloads is the number of overloads associated with this property name 
 * - prpHistoryDepthShort is the depth of the short term history kept for this property
 * - prpHistoryDepthLong is the depth of the long term history kept for this property
 * - prpFormat is the default output data format accepted for this property
 * - prpFormatIn is the default input data format accepted for this property
 * - prpAccess is the allowed data access for this property (either CA_READ, CA_WRITE or both)
 * - prpGraphType is the preferred graph type for output data associated with this property, e.g.
 * GT_LIN, CF_LOG, etc. (see rpcdefs.h).
 */
typedef struct
{
  char  prpName[32];
  char  prpDescription[64];
  char  prpRedirection[32];
  char  prpTag[8];
  char  prpTagIn[8];
  char  prpUnits[16];
  float prpMinValue;
  float prpMaxValue;
  UINT32 prpSize;
  UINT32 prpSizeIn;
  UINT32 prpNumOverloads;
  UINT16 prpHistoryDepthShort;
  UINT16 prpHistoryDepthLong;
  BYTE prpFormat;
  BYTE prpFormatIn;
  BYTE prpAccess;
  BYTE prpGraphType;
} PropertyQueryStructEx;
#define PROPERTYQUERYSTRUCTEX_SIZE (32+64+32+8*2+16+4*5+2*2+4)

/**
 * \struct XPropertyQueryStruct
 * \brief Defines an extended property query structure
 *
 * - prpName is the property name
 * - prpDescription is the property description
 * - prpRedirection is a redirection string if the property is redirected to another server
 * - prpTag is the data tag for output data sets, applicable if the output data format is
 * CF_STRUCT (i.e. a user defined structure)
 * - prpTagIn is the data tag for input data sets, applicable if the input data format is
 * CF_STRUCT (i.e. a user defined structure)
 * - prpUnits is a string containing the natural units of the data returned by the property
 * - prpMinValue is the lower limit of valid data associated with the property
 * - prpMaxValue is the upper limit of valid data associated with the property
 * - prpSize is the maximum allowed size of an output data set associated with the property
 * - prpSizeIn is the maximum allowed size of an input data set associated with the property
 * - prpNumOverloads is the number of overloads associated with this property name 
 * - prpHistoryDepthShort is the depth of the short term history kept for this property
 * - prpHistoryDepthLong is the depth of the long term history kept for this property
 * - prpFormat is the default output data format accepted for this property
 * - prpFormatIn is the default input data format accepted for this property
 * - prpAccess is the allowed data access for this property (either CA_READ, CA_WRITE or both)
 * - prpGraphType is the preferred graph type for output data associated with this property, e.g.
 * GT_LIN, CF_LOG, etc. (see rpcdefs.h).
 * - rngUnits is a string containing the natural units of the data range axis if the 
 * property returns an AT_SPECTRUM array type.
 * - rngMinValue is the lower limit for the horizontal display axis, relevant if the 
 * property returns an AT_SPECTRUM array type.
 * - rngMaxValue is the upper limit for the horizontal display axis, relevant if the 
 * property returns an AT_SPECTRUM array type.  The display increment can be determined
 * by (rngMaxValue-rngMinValue)/prpSize.
 * - numRows is the number of rows when the property returns an AT_DOUBLE array type.
 * - rowSize gives the row length when the property returns an AT_DOUBLE array type.
 * - prpArrayType is the property array type (see the array type definitions).
 * - reserved is reserved space for future use.
 */
typedef struct
{
  char  prpName[32];
  char  prpDescription[64];
  char  prpRedirection[32];
  char  prpTag[8];
  char  prpTagIn[8];
  char  prpUnits[16];
  float prpMinValue;
  float prpMaxValue;
  UINT32 prpSize;
  UINT32 prpSizeIn;
  UINT32 prpNumOverloads;
  UINT16 prpHistoryDepthShort;
  UINT16 prpHistoryDepthLong;
  BYTE prpFormat;
  BYTE prpFormatIn;
  BYTE prpAccess;
  BYTE prpGraphType;
  char  rngUnits[16];
  float rngMinValue;
  float rngMaxValue;
  UINT16 numRows;
  UINT16 rowSize;
  UINT16 prpArrayType;
  UINT16 reserved[3];
} XPropertyQueryStruct;
#define XPROPERTYQUERYSTRUCT_SIZE (PROPERTYQUERYSTRUCTEX_SIZE + 16+2*4+6*2)

/* RegisterPropertyQueryFunction() */
/**
 * \brief Registers an extended property query function
 * 
 * If a server's properties are all registered via a call to RegisterPropertyInformation(), RegisterPropertyEx(),
 * or RegisterProperty(), then there is no need to deal with property query functions.  Some servers may wish to
 * have a different property list for any given device, in which case a one-time query of the list of 
 * properties will not apply in general.  This is infrequently the case, but when it arises, a server will need
 * to field all property queries explicitly by registering its own property query structure.  
 * The query function must have the propotype shown in the parameter 'fcn'.  That is, it should be a routine
 * which accepts a device name as a parameter, and returns via a second paramter a reference to a list of 
 * extended property query structures (PropertyQueryStruct).
 *
 * \note This is a legacy call and deals with lists of 'simple' property query structures as opposed to
 * extended property query structures.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param fcn is the extended property query function which will handle property information queries
 * \param numprops is the number of properties which are registered.
 *  
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see RegisterXPropertyQueryFunction()
 *
 * \b Example:
 *
 * \include eg_RegisterPropertyQueryFunction.c
 */
EXPORT short RegisterPropertyQueryFunction(char *eqm,int (*fcn)(char *devName,PropertyQueryStruct **pqs),int numprops);

/* RegisterXPropertyQueryFunction() */
/**
 * \brief Registers an extended property query function
 * 
 * If a server's properties are all registered via a call to RegisterPropertyInformation(), RegisterPropertyEx(),
 * or RegisterProperty(), then there is no need to deal with property query functions.  Some servers may wish to
 * have a different property list for any given device, in which case a one-time query of the list of 
 * properties will not apply in general.  This is infrequently the case, but when it arises, a server will need
 * to field all property queries explicitly by registering its own property query structure.  
 * The query function must have the propotype shown in the parameter 'fcn'.  That is, it should be a routine
 * which accepts a device name as a parameter, and returns via a second paramter a reference to a list of 
 * extended property query structures (PropertyQueryStructEx).
 *
 * \note This extended call deals with lists of the more modern extended property query structure as opposed to
 * the legacy 'simple' property query structure.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param fcn is the extended property query function which will handle property information queries
 * \param numprops is the number of properties which are registered.
 *  
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see RegisterPropertyQueryFunction()
 *
 * \b Example:
 *
 * \include eg_RegisterXPropertyQueryFunction.c
 */
EXPORT short RegisterXPropertyQueryFunction(char *eqm,int (*fcn)(char *devName,XPropertyQueryStruct **xpqs),int numprops);
int RegisterLocalAlias(char *eqpname,char *alias);

/* AssignNumModules() */
/**
 * \brief Fixes the number of modules (devices) attached to the specified equipment module
 * 
 * Servers generally do not need to make this call, as the number of modules or devices is either specified
 * in the local database file 'exports.csv', or has been passed as an argument in a call to 
 * RegisterEquipmentModule().  However, in cases where the number so assigns represents an upper limit, and the
 * actual number of modules is only known after the server is running, then a call to AssignNumModules() can
 * be made to reduce the querieable number of modules to this number.  
 *
 * \note An upper limit of the number of modules must have already been specified either through the exports.csv
 * file or through a call to RegisterEquipmentModule().  The number specied in a call to AssignNumModules() 
 * must be smaller than this number.
 *
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 * \param num is the number of modules (devices) to be associated with the equipment module
 *  
 * \return 0 is successful or a TINE completion code
 *
 */
EXPORT int AssignNumModules(char *eqm,int num);

/* FindNameServerOnNetwork() */
/**
 * \brief Issues a multicast (or broadcast) to which the TINE equipment name server repsonds
 * 
 * In lieu of accessing the file 'cshosts.csv' (whose location is determined by the
 * environment variable TINE_HOME) a server can make this call to ascertain the
 * address(es) of the Equipment Name Server (ENS).  This is convenient for platforms which do
 * not have a hard disk and do not know a-priori the address of the equipment name server.
 *  
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see SetNameServerAddress()
 */
EXPORT int FindNameServerOnNetwork(void);

/* SetNameServerAddress() */
/**
 * \brief Sets the address of the Equipment Name Server via API call
 * 
 * In lieu of accessing the file 'cshosts.csv' (whose location is determined by the
 * environment variable TINE_HOME) a server can make this call to establish the
 * address(es) of the Equipment Name Server (ENS).  This is convenient for platforms which do
 * not have a hard disk.
 *
 * \param ens a string containing a list of the IP addresses of the ENSes to be used, 
 * separated by commas (','), and optionally providing a port offset (separated by a colon ':').
 * 
 * \b Example (sets two ENS addresses, the second with port offset = 2):
 * 
 * SetNameServerAddress("131.169.120.141,131.169.120.146:2");
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
EXPORT int SetNameServerAddress(char *ens);

/* GetNumContracts() */
/**
 * \brief Returns the current number of contracts associated with the equipment module given
 * 
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 *  
 * \return The total number of contracts associated with the given equipment module.  If a
 * NULL parameter is passed, the function returns the total number of contracts for all 
 * registered equipment modules.
 *
 * \b Example:
 *
   printf("EQP1 has %d active contracts\n",GetNumContracts("EQP1");

 */
EXPORT int GetNumContracts(char *eqm);

/* GetNumConsumers() */
/**
 * \brief Returns the current number of consumers associated with the equipment module given
 * 
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 *  
 * \return The total number of consumers (i.e. clients) associated with the given equipment module.  If a
 * NULL parameter is passed, the function returns the total number of consumers for all 
 * registered equipment modules.  This call is synonymous with GetNumCallers().
 *
 * \b Example:
 *
   printf("EQP1 has %d active clients\n",GetNumConsumers("EQP1");

 */
EXPORT int GetNumConsumers(char *eqm);

/* GetNumCallers() */
/**
 * \brief Returns the current number of callers associated with the equipment module given
 * 
 * \param eqm is the local equipment module name (maximum 6 characters in length)
 * For example: "BPMEQM". 
 *  
 * \return The total number of callers (i.e. clients) associated with the given equipment module.  If a
 * NULL parameter is passed, the function returns the total number of consumers for all 
 * registered equipment modules.  This call is synonymous with GetNumConsumers().
 *
 */
EXPORT int GetNumCallers(char *eqm);

/* GetCurrentLinkStatus() */
/**
 * \brief Returns the completion code associated with the input link index
 * 
 * \param i is the link index of the link for which the current completion code is desired.  
 *  
 * \return The completion code associated with the input link index
 *
 * \b Example:
 *
 * \include eg_GetCurrentLinkStatus.c
 */
EXPORT int GetCurrentLinkStatus(int i);

/* GetCurrentLinkStatusFromCallbackId() */
/**
 * \brief Returns the completion code associated with the 
 * callback identifier supplied
 * 
 * \param id is the callback id of the link for which the current completion code is desired.  
 *  
 * \return The completion code associated with the callback identifier supplied
 *
 * \b Example:
 *
 * \include eg_GetCurrentLinkStatusFromCallbackId.c
 */
EXPORT int GetCurrentLinkStatusFromCallbackId(int id);

/* GetCaller() */
/**
 * \brief Gives the user name of the current caller
 * 
 * If a call to an equipment module is made on behalf of more than one user, then GetCaller() only
 * refers to the first in the list.  Typically, GetCaller() is used to see who is making a WRITE access
 * call, which is usually synchronous and has only one user attached.  When more information is required
 * or the contract is liable to be monitored by many clients, then GetCallerInfo() should be used instead.
 *  
 * \return The user name of the current caller
 *
 * \see GetCallerInfo()
 *
 * \b Example:
 *
    if (!strcmp("SMITH",GetCaller()) return access_denied;// dont let SMITH issue this command

 */
EXPORT char *GetCaller(void);

/* GetCallerInfo() */
/**
 * \brief Returns the user name(s) and network address(es) of all callers interested in the current contract
 * 
 * \param un (out) if non-NULL is a reference to an array of NAME16 objects to be filled with the caller list
 * \param ipx (out) if non-NULL is a reference to a byte array to be filled with the IPX addresses of the caller(s).
 * \param ip (out) if non-NULL is a reference to an unsigned long array to be filled with the IP addresses of the caller(s).
 * \param prot (out) if non-NULL is a reference to a short array to be filled with the requested protocol level of the caller(s).
 * \param num (in/out) is a reference to an int value which initially contains the buffer size of the array
 * references used in the parameter list. Upon return of the call, it contains the actual number of callers.
 *
 * \note The non-NULL arguments supplied must point to buffer with enough room to hold at least the input 'num'
 * entries, otherwise a system crash is likely.
 *
 * \return 0 if successful, or a TINE completion code
 *
 * \see GetCaller()
 * \b Example:
 *
 * \include eg_GetCallerInfo.c
 */
EXPORT int GetCallerInfo(NAME16 *un,BYTE *ipx,UINT32 *ip,short *prot,int *num);

/* GetCompletionDataSize() */
/**
 * \brief Returns the most recent data size of the link index supplied
 * 
 * \param i is the link index for which the completed data size is desired
 *  
 * \return The most recent data size of the link index supplied
 *
 * \see GetCompletionDataSizeFromCallbackId()
 *
 * \b Example:
 *
 * \include eg_GetCompletionDataSize.c
 */
EXPORT int GetCompletionDataSize(int i);
#define getCompletionDataSize(i) GetCompletionDataSize(i)

/* GetCompletionDataSizeFromCallbackId() */
/**
 * \brief Returns the most recent data size of the link 
 * associated with the callback id supplied
 * 
 * \param id is the callback id of the link for which the completed data size is desired
 *  
 * \return The most recent data size of the link associated with the callback id supplied
 *
 * \see GetCompletionDataSize()
 *
 * \b Example:
 *
 * \include eg_GetCompletionDataSizeFromCallbackId.c
 */
EXPORT int GetCompletionDataSizeFromCallbackId(int id);
#define getCompletionDataSizeFromCallbackId(i) GetCompletionDataSizeFromCallbackId(i)

/* GetCompletionDataType() */
/**
 * \brief Returns the most recent data type of the link index
 * supplied
 * 
 * \param i is the link index for which the data type is desired
 *  
 * \return The most recent data type of the link index supplied
 *  
 * \see GetCompletionDataTypeFromCallbackId()
 *
 * \b Example:
 *
 * \include eg_GetCompletionDataType.c
 */
EXPORT int GetCompletionDataType(int i);
#define getCompletionDataType(i) GetCompletionDataType(i)

/* GetCompletionDataTypeFromCallbackId() */
/**
 * \brief Returns the most recent data type of the link 
 * associated with the callback id supplied
 * 
 * \param id is the callback id of the link for which the completed data type is desired
 *  
 * \return The most recent data type of the link associated with the callback id supplied
 *
 * \see GetCompletionDataType()
 *
 * \b Example:
 *
 * \include eg_GetCompletionDataTypeFromCallbackId.c
 */
EXPORT int GetCompletionDataTypeFromCallbackId(int id);
#define getCompletionDataTypeFromCallbackId(i) GetCompletionDataTypeFromCallbackId(i)

/* GetCompletionSource() */
/**
 * \brief Returns the error source associated with the input link index
 * 
 * \param i is the link index of the link for which the current status string is desired.  
 *
 * \return The error source associated with the input link index as one of RPC_SOURCE_LCL
 * or RPC_SOURCE_RMT, or -1 if the link index is invalid.
 *
 * \see GetCompletionSourceFromCallbackId()
 *
 */
EXPORT short GetCompletionSource(int i);
#define getCompletionSource(i) GetCompletionSource(i)

/* GetCompletionStatus() */
/**
 * \brief Returns the error string associated with the input link index
 * 
 * \param i is the link index of the link for which the current status string is desired.  
 *
 * \return The error string associated with the input link index, or NULL if the
 * the link index is invalid.
 *
 * \see GetCompletionStatusFromCallbackId()
 *
 * \b Example:
 *
 * \include eg_GetCompletionStatus.c
 */
EXPORT char *GetCompletionStatus(int i);
#define getCompletionStatus(i) GetCompletionStatus(i)

/* GetCompletionSourceFromCallbackId() */
/**
 * \brief Returns the error source associated with the callback identifier supplied.
 * 
 * \param id is the callback id of the link for which the current status string is desired.  
 *
 * \return The error source associated with the callback identifier supplied as one of 
 * RPC_SOURCE_LCL or RPC_SOURCE_RMT, or -1 if the callback identifier is invalid.
 *
 */
EXPORT short GetCompletionSourceFromCallbackId(int id);
#define getCompletionSourceFromCallbackId(i) GetCompletionSourceFromCallbackId(i)

/* GetCompletionStatusFromCallbackId() */
/**
 * \brief Returns the error string associated with the 
 * callback identifier supplied
 * 
 * \param id is the callback id of the link for which the current status string is desired.  
 *  
 * \return The error string associated with the callback identifier supplied, or NULL
 * if the callback id is invalid.
 *
 * \see GetCompletionStatus()
 *
 * \b Example:
 *
 * \include eg_GetCompletionStatusFromCallbackId.c
 */
EXPORT char *GetCompletionStatusFromCallbackId(int id);
#define getCompletionStatusFromCallbackId(i) GetCompletionStatusFromCallbackId(i)

/* GetLinkName() */
/**
 * \brief Returns the full link name associated with the 
 * input link index
 * 
 * \param idis the link index of the link for which the link name is desired.  
 *  
 * \return The full link name as "/<Context>/<Server>/<Device>[<Property>]", or 
 * "link not found" if the callback id is invalid.
 *
 */
EXPORT char *GetLinkName(int i);
#define getLinkName(i) GetLinkName(i)

/* GetLinkNameFromCallbackId() */
/**
 * \brief Returns the full link name associated with the 
 * callback identifier supplied
 * 
 * \param id is the callback id of the link for which the link name is desired.  
 *  
 * \return The full link name as "/<Context>/<Server>/<Device>[<Property>]", or 
 * "link not found" if the callback id is invalid.
 *
 */
EXPORT char *GetLinkNameFromCallbackId(int id);
#define getLinkNameFromCallbackId(i) GetLinkNameFromCallbackId(i)

/* GetExportedName() */
/**
 * \brief Returns the exported name associated with the equipment function name given.
 * 
 * A server as coded, frequently does not know its system wide identity (i.e. its 'export name'),
 * as this is information is contained in a registration file such as 'exports.csv'.  Nonetheless
 * it might be desireable to know what name was actually exported, for use in say conditional
 * trapping or logging.  This call returns the name finally exported after the server has been
 * initialized.
 *  
 * \param eqpname is the equipment function name (local name) for which the exported name
 * is desired.
 *
 * \return If successful, the export name associated with the equipment function name supplied.
 * Otherwise a NULL pointer is returned.
 *
 * \see GetCompletionStatus()
 */
EXPORT char *GetExportedName(char *eqpname);

/* GetExportedContext() */
/**
 * \brief Returns the exported context associated with the equipment function name given.
 * 
 * A server as coded, frequently does not know the context under which it is registered,
 * as this is information is often contained in a registration file such as 'fecid.csv'.  Nonetheless
 * it might be desireable to know what name was actually exported, for use in say conditional
 * trapping or logging.  This call returns the name finally exported after the server has been
 * initialized.
 *  
 * \param eqpname is the equipment function name (local name) for which the exported context
 * is desired.
 *
 * \return If successful, the export context associated with the equipment function name supplied.
 */
EXPORT char *GetExportedContext(char *eqpname);

/* GetExportedFecName() */
/**
 * \brief Returns the FEC name exported at the time of server initialization.
 * 
 * \return If successful, the fec name registered at the equipment name server.
 */
EXPORT char *GetExportedFecName(void);

/* GetLastGlobalDataSize() */
/**
 * \brief Returns the most recent data size of the global link index supplied
 * 
 * \param i is the global link index for which the data size is desired.
 * If i is -1 the last recorded incoming global data size (not a thread-safe thing to do).
 *  
 * \return The most recent data size of the global link index supplied
 *
 * \see GetLastGlobalDataSizeFromCallbackId()
 *
 */
EXPORT int GetLastGlobalDataSize(int id);

/* GetLastGlobalDataSizeFromCallbackId() */
/**
 * \brief Returns the most recent data size of the global link callback id supplied
 * 
 * \param i is the global link callback id for which the data size is desired.
 * 
 * \return The most recent data size of the global link callback id supplied
 *
 * \note This is a best effort to determine the propery entry table for the callback id
 * supplied as an argument.  It is the duty of the caller to manage a unique  set of
 * callback ids for his application.
 * 
 * \see GetLastGlobalDataSize()
 *
 */
EXPORT int GetLastGlobalDataSizeFromCallbackId(int id);

/* SystemVersion() */
/**
 * \brief Returns the system version as a 2-byte array containg the major and minor version numbers.
 * 
 * \return The tine version in use.
 *
 * \b Example:
 * 
 * \include eg_SystemVersion.c
 */
EXPORT BYTE *SystemVersion(void);

/* feclog() */
/**
 * \brief Puts entries into a server's log file.
 * 
 * Server be default maintain a log file (fec.log) which contains system information at startup
 * and shutdown time, as well as WRITE access command calls, as well as other notifications.  
 * The server can all place its own text in the server's log file by using this routine.  All
 * entries will be appended with a time stamp and a new-line character.
 *
 * \param text is the text string (up to 200 characters) to be logged.
 *
 * \return 0 if successful, otherwise a TINE error code.
 *
 * \see GetCompletionStatus()
 * \see gFeclogPath
 * \see vfeclog()
 * \see FelclogDepth
 */
EXPORT int feclog(char *text,...);

/* ExecLocalLink() */
/**
 * \brief Executes a synchronous link within the local process
 * 
 * Synchronous data exchange within a process. ExecLocalLink() does not complete until the data transfer 
 * has completed or a timeout has ensued. ExecLocalLink() has marginal applicability other than processing
 * and returning the result of a call within a server, which could be of use if multiple equipment functions
 * are bound together or a server's background task needs to know the outcome of a call into the server, etc.
 *
 * \param devName is the local device name (/LOCAL/\<Local Equipment Name\>/\<Device\>) of the device to contact.
 * For example: "/LOCAL/BPMEQM/WL167". 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  The structure element dout->data should be a pointer to a pre-allocated data buffer,
 * which is to receive the returned data.  If dout is a NULL pointer, it is assumed that no data are to 
 * be returned.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  If din is a NULL pointer, it is assumed that no data are to be sent to the server
 * \param access is the data access mode.  This can be any combination of TINE access code ORed
 * together (e.g. CA_READ|CA_WRITE).
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 * \see ExecLink(), ExecLinkEx(), DTYPE
 *  
 */
EXPORT int ExecLocalLink(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);

/* getDataTimeString() */
/**
 * \brief Returns the TINE data timestamp as a human-readable string
 * 
 * A call to getDataTimeStamp() returns an 8-byte double containing the current 
 * TINE system timestamp. This will be the UTC timestamp appended with 
 * milliseconds after the decimal (if available).  This timestamp can be 
 * converted to a human readable string with this call.  The format will depend
 * on the value of the second parameter.  If useLongStringFormat is non-zero,
 * the a date of the form:
 * 
 * "Mon Jan 10 12:08:43.355 CET 2005" 
 *
 * will be returned.  If useLongStringFormat is zero, then the date string will
 * be of the form (dd.mm.yy hh:mm:ss.msec TZ):
 *
 * "10.01.05 12:08:43.355 CET"
 *  
 * \param ts is the TINE Data Timestamp for which the string conversion is to be
 * applied.
 * \param useLongStringFormat is a flag which if non-zero returns a longer representation
 * of the timestamp.
 *
 * \note This call is not thread-safe.  See the discussion in getDataTimeStamp() for 
 * details.
 *
 * \return A pointer to a string containing the timestamp representation.
 *
 * \see getDataTimeStamp, getCurrentDataTimeStamp, getCurrentDataTimeStampFromCallbackId
 * \see ctime()
 * \see localtime()
 * \see findDaylight()
 */
EXPORT char *getDataTimeString(double ts,int useLongStringFormat);

/* getDataTimeStampAsTimeval() */
/**
 * \brief Returns the last established data timestamp as a pointer to a struct timeval.
 * 
 * A call to getDataTimeStampAsTimeval() immediately following a call to ExecLink() or within a
 * callback routine will yield the data timestamp of the call.  
 *  
 * \note This call is not thread-safe.  See the discussion in getDataTimeStamp() for 
 * details.
 *
 * \return A pointer to a struct timeval containing the data timestamp currently in place.
 *
 * \see getDataTimeStamp, getCurrentDataTimeStamp, getCurrentDataTimeStampFromCallbackId
 */
EXPORT struct timeval *getDataTimeStampAsTimeval(void);

/* getDataTimeStamp() */
/**
 * \brief Returns the last established data timestamp
 * 
 * A call to getDataTimeStamp() immediately following a call to ExecLink() or within a
 * callback routine will yield the data timestamp of the call.  
 *  
 * \note This call is not thread-safe.  Synchronous data acquisition on two different threads
 * could conceivably interfer with the value returned from this routine.  A thread-safe way of
 * obtaining the data timestamp associated with a call is to use the .dTimestamp field of the
 * DTYPE object for synchronous calls such as ExecLink() or ExecLinkEx() or to use
 * getCurrentDataTimeStampFromCallbackId() or getCurrentDataTimeStamp() (which take a link
 * identifier as input parameter) for asynchronous links.
 *
 * \return The data timestamp currently in place.
 *
 * \see getDataTimeStampAsTimeval, getCurrentDataTimeStamp, getCurrentDataTimeStampFromCallbackId
 */
EXPORT double getDataTimeStamp(void);

/* getCurrentDataTimeStamp() */
/**
 * \brief Returns the data timestamp associated with the input link index
 * 
 * \param i is the link index of the link for which the current data timestamp is desired.  
 *  
 * \return The data timestamp associated with the input link index
 *
 * \see getDataTimeStampAsTimeval, getDataTimeStamp, getCurrentDataTimeStampFromCallbackId
 */
EXPORT double getCurrentDataTimeStamp(int i);

/* getCurrentDataTimeStampFromCallbackId() */
/**
 * \brief Returns the data timestamp associated with the 
 * callback identifier supplied
 * 
 * \param i is the callback identifier of the link for which the current data timestamp is desired.  
 *  
 * \return The data timestamp associated with the input callback identifier
 *
 * \see getDataTimeStampAsTimeval, getDataTimeStamp, getCurrentDataTimeStamp
 */
EXPORT double getCurrentDataTimeStampFromCallbackId(int id);

/* setDataTimeStamp() */
/**
 * \brief Sets the intrinsic data timestamp to the value given
 * 
 * This call is typically used inside an equipment function to set the 
 * data timestamp associated with the request being made.  For instance,
 * when the data are taken from the hardware IO loop, a call to makeDataTimeStamp() 
 * is made to get the current system timestamp. Then when a request for data is made
 * the caller can know the precise time of data acquisition.
 * If this routine is not called, then the current system time will be used to stamp the
 * data.
 *
 * \param ts is the timestamp to be associated with the next data set.
 *  
 * \see putDataTimeStamp, makeDataTimeStamp
 */
EXPORT void setDataTimeStamp(double ts);

/* putDataTimeStamp() */
/**
 * \brief Returns a data timestamp according to the input parameters given
 * 
 * This call is has the same purpose as makeDataTimeStamp(), but can assign a TINE
 * data timestamp according to the input seconds, milliseconds and system time offset.
 * This call is used only under special circumstances.
 *
 * \param toffset is the systematic time offset to be applied.
 * \param tsec is the data timestamp in seconds
 * \param tmsec is the millisecond part of the data timestamp
 *  
 * \return A tine data timestamp.
 *
 * \see setDataTimeStamp, makeDataTimeStamp
 */
EXPORT double putDataTimeStamp(double toffset, time_t tsec, int tmsec);

/* makeDataTimeStamp() */
/**
 * \brief Returns a data timestamp according to the current system time.
 * 
 * \return A tine data timestamp.
 *
 * \see setDataTimeStamp, putDataTimeStamp
 */
EXPORT double makeDataTimeStamp(void);

extern double gDataTimeStamp;
extern double gServerDataTimeStamp;
extern double gDataTimeStampOffset;
#define RPCtimestamp ((SINT32)gDataTimeStamp)
#define RPCtimestampMSEC (int)((gDataTimeStamp - (SINT32)gDataTimeStamp) * 1000)
/* RequireAcknowledgments */
/**
 * \brief Determines whether acknowledgements are expected following data changes where contracts use CM_REFRESH mode
 * 
 * Under CM_POLL link conditions the server is not directly interested as to whether the client received its data 
 * packet or not, since the client is expecting data to arrive at the polling interval and will complain if it is
 * missing.  However,when a client establishes a data link and requests CM_REFRESH mode, data are are only returned to the client
 * when they have changed, or when the link heartbeat (1 minute) has expired.  A server will under these conditions
 * require an acknowledgement from the client if data do in fact change.  A missing acknowledgement will cause
 * the server to re-issue the data transfer.  This default behavior will of course increase the network traffic 
 * when data are frequently changing.  Under some circumstance it is undesireable to require acknowledgments and
 * this feature can be turned off by setting this value to FALSE.
 * 
 *
 * \b Default: TRUE
 *  
 */
extern int RequireAcknowledgments;
extern int AllowRemoteExit;
extern short SychronizeFecClock;
extern WORD watchdogBaseAddress;
extern UINT32 SysPoll;
/* MinPollingRate */
/**
 * \brief Determines the minimum polling rate in milliseconds a server will allow.
 * 
 * A server will not honor polling rates smaller than this value.  An attempt to poll a property a 1 msec
 * for instance will be serviced at the MinPollingRate instead.
 * If it is known that the server should support very fast polling rates, 
 * then this value should be set accordingly at initialization time.
 *
 * \b Default: 100 (Or #define MIN_POLLRATE or #define SYS_POLL in project.def)
 *  
 */
extern EXPORT unsigned short MinPollingRate;
extern EXPORT unsigned short MaxPollingRate;
extern int defaultClientTimeout;
#define defaultClientTimeoutMsec defaultClientTimeout
/* MaxNumContracts */
/**
 * \brief Determines the maximum number of contracts a server will service.
 * 
 * A server manages both a contract table and a client table, which are independent of one another 
 * except that when there is at least on contract there must be at least one client.  The size of this table
 * is pre-allocated at initialization time.  If it is known that the server will have to support a large number 
 * contracts. then this value should be set accordingly at initialization time.
 *
 * \b Default: 150 (Or #define NCONTRACTS in project.def)
 *  
 */
extern EXPORT int MaxNumContracts;
/* MaxNumClients */
/**
 * \brief Determines the maximum number of clients a server will service.
 * 
 * A server manages both a contract table and a client table, which are independent of one another 
 * except that when there is at least on contract there must be at least one client.  The size of this table
 * is pre-allocated at initialization time.  If it is known that the server will have to support a large number 
 * clients. then this value should be set accordingly at initialization time.
 *
 * \b Default: 50 (Or #define NCONSUMERS in project.def)
 *  
 */
extern EXPORT int MaxNumClients;
/* MaxNumConnections */
/**
 * \brief Determines the maximum number of entries in the connection table.
 * 
 * A client's connections are managed and maintained in a connection table.  The size of this table
 * is pre-allocated at initialization time.  This allows for fast lookups, since a connection ID is 
 * simply an entry into the table.  If it is known that the client will need a large number of simultaneous
 * links then this value should be set accordingly at initialization time.
 *
 * \b Default: 100 (Or #define MAX_ENTRIES in project.def)
 *  
 */
extern EXPORT int MaxNumConnections;
/* MaxRPCTransportSize */
/**
 * \brief Determines the maximum size in bytes of any call supported by the server
 * 
 * Any call which attempts to access more than the MaxRPCTransportSize will fail with 
 * 'invalid transport size'.  If it is known before hand that the server should support calls for
 * very large amounts of data, and there is available main memory on the server, this value should be
 * set accordingly prior to the call to SystemInit().
 *
 * \b Default: 64 Kb for most operating systems.  Exceptions: MSDOS (16 Kbyte), Win16 (32 Kbyte).
 *  
 */
extern EXPORT UINT32 MaxRPCTransportSize;
extern UINT16 PortOffset;
/* FeclogDepth */
/**
 * \brief Determines the approximate size of the server's log file in lines.
 * 
 * The server's log file, if located on the local disk, is a file called 'fec.log'.  When it's length
 * surpasses the FecLogDepth, this file is moved to 'fec.bak' and a new 'fec.log' file is started.  A
 * call for the server's log file will always be able to scan both the 'fec.bak' file and 'fec.log', and
 * will thus always be able to scan at least FeclogDepth entries.
 *
 * \b Default: 100
 *  
 */
extern EXPORT int FeclogDepth;
extern int leftFeclogTime;
/* StartupDebug */
/**
 * \brief Determines whether server initialization messages are displayed or not
 * 
 * A running server is quiet regarding information displayed on the standard output. An
 * initializing server is however by default not quiet, and will echo various steps of the
 * initialization process on the screen, including whether of not optional configuration files
 * are found or not, etc.  This information can be surpressed by setting StartupDebug to FALSE
 * prior to calling SystemInit().
 *
 * \note Critical errors will be displayed on the screen in all circumstances.
 *
 * \b Default: TRUE
 *  
 */
extern int StartupDebug;
extern int queryDeviceNamesFromProperties;

/* LastCompletionDataSize */
/**
 * \brief Supplies the data size of the most recent link
 * 
 * A global variable which supplies the data size of the most recent link.  
 *
 * \note This is not thread safe.  In multithreaded applications, use 
 * GetCompletionDataSize() or GetCompletionDataSizeFromCallbackId()
 *
 * \see GetCompletionDataSize(), GetCompletionDataSizeFromCallbackId()
 *  
 */
extern UINT32 LastCompletionDataSize;

/* LastCompletionDataType */
/**
 * \brief Supplies the data type of the most recent link
 * 
 * A global variable which supplies the data type of the most recent link.  
 *
 * \note This is not thread safe.  In multithreaded applications, use 
 * GetCompletionDataType() or GetCompletionDataTypeFromCallbackId()
 *
 * \see GetCompletionDataType(), GetCompletionDataTypeFromCallbackId()
 *  
 */
extern int LastCompletionDataType;
extern int allowLegacyConnections;
extern int longDeviceNames;
/* putCommandsInFeclog */
/**
 * \brief Determines whether all in-coming WRITE access calls are automatically included in the server's log file
 * 
 * \b Default: TRUE
 *  
 */
extern EXPORT int putCommandsInFeclog;
/* nofeclog */
/**
 * \brief Determines whether a server is to keep a log file on the local disk
 * 
 * The location of the log file is
 * determined by the FECDB environment variable or, if not set, the current working directory is used.
 * If set to TRUE, then no log file will be maintained on the local disk.  A log file will nonetheless be
 * maintained in a local ring-buffer, with the caveat that all entries will disappear upon server restart.
 * To eliminate the log file entirely, 'nofeclog' should be set to TRUE and 'FeclogDepth' should be set to 0.
 *
 * \b Default: FALSE for most operating systems.  Exceptions: VxWorks, NIOS.
 *  
 */
extern EXPORT int nofeclog;

/* useGlobalSynchronization */
/**
 * \brief Determines whether data timestamps are to be externally synchronized
 * 
 * If set to TRUE, data timestamps will be synchronized to an externally supplied timestamp.  This 
 * requires a TINE Time Server to be in operation and sending timestamps at regular intervals either as
 * multicast or as broadcast.
 *
 * \b Default: FALSE
 *  
 */
extern EXPORT int useGlobalSynchronization;

extern time_t appmktime;

/* SetAppDate */
/**
 * \brief Sets the compilation date of the current running server process
 * 
 * The server's application creation date is queriable from a TINE server.
 * However this information needs to be passed to the server engine at initialization
 * time via this call.
 *
 * \param appdate is the creation time as a UNIX time variable.
 *  
 * \see SetAppVersion()
 */
EXPORT void SetAppDate(time_t appdate);

extern char appmkversion[64];
/* SetAppVersion */
/**
 * \brief Sets the application version of the current running server process
 * 
 * The server's application version number is queriable from a TINE server.
 * However this information needs to be passed to the server engine at initialization
 * time via this call.
 *
 * \param major is the major version number of the application
 * \param minor is the minor version number of the application
 * \param revision is the revision version number of the application
 *  
 * \see SetAppVersion()
 */
EXPORT void SetAppVersion(int major,int minor,int revision);

EXPORT int setDefaultClientTimeout(int timeoutMilliseconds);

extern double gSyncTolerance;

/* setSynchronizationTolerance */
/**
 * \brief Sets the tolerance used in deciding whether to apply a timestamp offset or not.
 * 
 * If a TINE time server is running and supplying a system time stamp, and this server
 * is accepting this to synchronize its local clock when applying data time stamps,
 * the tolerance used in applying this offset can be set by this routine.
 *
 * \param toleranceInSeconds is the maximum time difference allowed before setting the
 * data timestamp offset. (Default: 0.5 seconds)
 * 
 *  
 * \see getSynchronizationTolerance()
 */
EXPORT void SetSynchronizationTolerance(double toleranceInSeconds);

/* getSynchronizationTolerance */
/**
 * \brief Gets the tolerance used in deciding whether to apply a timestamp offset or not.
 * 
 * If a TINE time server is running and supplying a system time stamp, and this server
 * is accepting this to synchronize its local clock when applying data time stamps,
 * the tolerance used in applying this offset can be read by this routine.
 *
 * \return toleranceInSeconds is the maximum time difference allowed before setting the
 * data timestamp offset.
 * 
 * \see setSynchronizationTolerance()
 */
EXPORT double GetSynchronizationTolerance(void);

EXPORT int hasDevicePropertyList(char *eqm);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* RPCTYPES_H_ */
