/** \file clnip.c
 *
 * some client-side IP routines ...
 *
 */
#include "cprolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "servdef.h"
#include "srvip.h"
#include "brdcast.h"
#include "database.h"
#ifdef UNIX
#  include "unixcmd.h"
#  ifdef NOTTY
int foregroundTTY = 0;
int foregroundCln = 0;
#  else
int foregroundTTY = 1;
int foregroundCln = 1;
#  endif
#endif

int ConsumerSocket = 0;
int MCastConsumerSocket = 0;
int ListenForMCasts = FALSE;
#ifndef UDP_GLOBALS
# define UDP_GLOBALS 0
#endif
int ListenForGlobals = UDP_GLOBALS;
extern fd_set rwset;
fd_set conset;
struct timeval conto;
struct sockaddr_in srv_name;
int GlobalSocket = 0;
int defaultClientTimeout = 20;
int clientTimeout = 0;
int clientTimeoutMsec = 0;
int ReturnToClientCycle = 0;
struct sockaddr_in IncomingIPfec;
extern IPXAddress IncomingIPXfec;
IDLE_CONNECTION *tcpIdleConnectionList = NULL;

EXPORT void setClientTimeout(int timeoutMilliseconds)
{
  if (defaultClientTimeout >= 0) 
  {
    clientTimeout = timeoutMilliseconds/1000;
    clientTimeoutMsec = timeoutMilliseconds%1000;
  }
}
/**
 * \internal
 * \brief Initialize file descriptor sets and bind sockets for an IP client.
 *
 *
 * Binds sockets for the consumer, MCast, and GCast,
 * and adds these sockets to conset.
 * 
 * Initialize the file descriptor sets conset, rwset, ext-set.
 *
 * Also adds stdin to conset if stdin is a tty.
 * 
 * \param port consumer port number.
 * \todo describe MCast and GCast.
 * \todo document global fd sets.
 */
int initIPClient(unsigned short port)
{
  int cc = 0;
  int optval = 1, optlen = sizeof(int);
  if (defaultClientTimeout < 10) defaultClientTimeout *= 1000;
# ifndef MSDOS
  TCPloaded = 1;
  if (ProducerSocket != 0) defaultClientTimeout = 0;
# ifdef WINNT
  if (!WinsockLoaded()) TCPloaded = 0;
  MaxPollingRate = 10; MinPollingRate = 10;
# endif
# else
  if (!TCPloaded) ccerr(tcp_not_supported);
  defaultClientTimeout = 0;
# endif
  clientTimeout = defaultClientTimeout/1000;
  clientTimeoutMsec = defaultClientTimeout%1000;
  if (MaxPollingRate > 1000) MaxPollingRate = 1000;
  if (SysPoll > 1000) SysPoll = 1000;
  /* set up UDP Consumer Sockets */
  conto.tv_sec = clientTimeout; conto.tv_usec = clientTimeoutMsec*1000;
  FD_ZERO(&conset); FD_ZERO(&rwset); FD_ZERO(&ext_set);
  if (listenUDP(&ConsumerSocket,port) != 0) ccerr(udp_socket_error);
# ifndef MSDOS
  if (setsockopt(ConsumerSocket,SOL_SOCKET,SO_BROADCAST,(char *)&optval,optlen))
  {
    feclog("cannot set broadcast option on consumer socket !");
  }
# endif /* MSDOS */
  FD_SET(ConsumerSocket,&conset);
  if (ListenForMCasts && MCastConsumerSocket == 0)
  {
    setupGlobalsSocket(MCastConsumerSocket,(UINT16)gtMCastPort);
    if (MCastConsumerSocket == 0) ListenForMCasts = 0;
  }
  /* does it make sense to listen for broadcasts ? */
  if (ListenForGlobals && GlobalSocket == 0) 
  {
    setupGlobalsSocket(GlobalSocket,(UINT16)gtGCastPort);
    if (GlobalSocket == 0) ListenForGlobals = 0;
  }
err:
# if defined(UNIX) && !defined(NOTTY)
  if (foregroundTTY == 0) foregroundCln = 0;
  else if ((foregroundCln=isatty(STDIN_FILENO)) != 0) FD_SET(STDIN_FILENO,&conset);
  foreground = foregroundCln;
# endif
  if (cc)
  {
    feclog("Consumer socket : %s",erlst[cc]);
  }
  return cc;
}

int ListenForIPMulticasts(void)
{
  int cc = 0;
  if (ListenForMCasts) ccerr(0);
  if (!TCPloaded) ccerr(tcp_not_supported);
  setupGlobalsSocket(MCastConsumerSocket,(UINT16)gtMCastPort);
  if (MCastConsumerSocket == 0) ccerr(tcp_socket_error);
# if defined(UNIX)
  gUseLoopback = TRUE;
# endif
  ListenForMCasts = TRUE;
err:
  if (cc) ListenForMCasts = FALSE;
  return cc;
}

int ListenForIPNetGlobals(void)
{
  int cc = 0;
  if (ListenForGlobals) ccerr(0)
  if (!TCPloaded) ccerr(tcp_not_supported);
  setupGlobalsSocket(GlobalSocket,(UINT16)gtGCastPort);
  if (GlobalSocket == 0) ccerr(tcp_socket_error);
  ListenForGlobals = TRUE;
err:
  if (cc) ListenForGlobals = FALSE;
  return cc;
}

void closeIPClients(void)
{
  int i,done = 0;
  char str[256];
  if (!TCPloaded) return;
  if (soclose(ConsumerSocket)) soperror("UDP soclose");
  if (ListenForMCasts)
  {
    if (MCastConsumerSocket != 0 && soclose(MCastConsumerSocket)) soperror("UDP soclose");
  }
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (RPCFec[connTable[i]->FECindex].Prot == TCP)
    {
      if (!done)
      {
        done = -1;
        shutdown(connTable[i]->TCPsocket,2);
        soclose(connTable[i]->TCPsocket);
      }
      connTable[i]->TCPsocket = 0;
      sprintf(str,"TCP link to %s %s terminated",connTable[i]->name, connTable[i]->prop);
      feclog(str);
    }
  }
  return;
}

void closeIPConnectionToFec(int index)
{
  int i,done = 0;
  if (!TCPloaded) return;
  if (RPCFec[index].Prot != TCP) return;
  RPCFec[index].Prot = UDP;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (index != connTable[i]->FECindex) continue;
    if (!done)
    {
      done = -1;
      shutdown(connTable[i]->TCPsocket,2);
      soclose(connTable[i]->TCPsocket);
    }
    connTable[i]->TCPsocket = 0;         /* signal for new connection */
    feclog("TCP link to %s %s terminated",connTable[i]->name, connTable[i]->prop);
  }
  return;
}
int connectTCP(int port,int i)  /* TCP client for connection table entry #i */
{
  ConnTableEntry *c;
  int fec,cc=0;
  int optval = 1, mode = 1;
  struct timeval tv0,tv;
  UINT16 poff;

  if ( i < 0 || i >= nConnectionTableEntries) return un_allocated;
  c = connTable[i]; 
  fec = c->FECindex;
  memset(&srv_name,0,sizeof(srv_name));
  memcpy(&srv_name.sin_addr,&RPCFec[fec].IPh_addr,sizeof(struct in_addr));
  srv_name.sin_family=AF_INET;
  poff = GetPortOffset(&RPCFec[fec]);
  srv_name.sin_port = htons((UINT16)(port + poff));
  if ((c->TCPsocket=(int)socket(AF_INET,SOCK_STREAM,0))<0)
  {
    soperror("socket");
    return tcp_socket_error;
  }
  if (setsockopt(c->TCPsocket,SOL_SOCKET,SO_KEEPALIVE,(char *)&optval,sizeof(optval)))
    ccerr(tcp_socket_error);
  if (soioctl(c->TCPsocket,FIONBIO,(char *)&mode)) ccerr(tcp_connect_error);

  gettimeofday(&tv0,(struct timezone *)NULL);  tv = tv0;

  while (connect(c->TCPsocket,(struct sockaddr *)&srv_name,sizeof(srv_name)) < 0)
  {
    if (soerrno == EISCONN) break;
    gettimeofday(&tv,(struct timezone *)NULL);
    if (MSECS(tv,tv0) > DEFAULT_TCP_CONNECTION_TIMEOUT) ccerr(connection_timeout);
  }
err:
  if (cc) 
  {
    soclose(c->TCPsocket);
    c->TCPsocket = 0;
  }
  return cc;
}

void doUDPConsumer(void)
{
  short i,k,n=0,nread,nleft;
  int retry=2,nready;
  unsigned int len = sizeof(IncomingIPfec);
  ConnTableEntry *c;
  TCPBCKT *bckt;
  BYTE *bufptr=clnWorkArea;
# ifdef UNIX
  char cmdbuf[CMDSIZE];
  int sck;
# endif
  if (!TCPloaded) return;
  conset = rwset;            /* fd sets from the server side */
  FD_SET(ConsumerSocket,&conset);
  if (ListenForMCasts) 
  {
    if (MCastConsumerSocket != 0) FD_SET(MCastConsumerSocket,&conset);
  }
  if (GlobalSocket) FD_SET(GlobalSocket,&conset);
  for (i=0; i<nConnectionTableEntries; i++)    /* possible TCP connections */
    if ((c=connTable[i])->TCPsocket != 0) FD_SET(c->TCPsocket,&conset);
# ifdef UNIX
  if (foregroundCln) FD_SET(STDIN_FILENO,&conset);
  if (externalFdSets) FD_MERGE(ext_set,&conset);  /* external fd sets ? */
# ifndef NO_TINE_PIPE
  for (i=0; i<RPCNumFecsRead; i++) if (RPCFec[i].Prot == PIPE)
  {
     memcpy(&sck,RPCFec[i].Network,4);
     if (sck > 0) FD_SET(sck,&conset);
  }
# endif
# endif
# ifdef IPXBSD
  FD_SET(IPXConsumerSocket,&conset);
  if (ListenForMCasts && IPXMCastConsumerSocket) FD_SET(IPXMCastConsumerSocket,&conset);
# endif
  if (ReturnToClientCycle || ReturnToServerCycle)
  {
    conto.tv_sec = conto.tv_usec = 0;
  }
  else
  {
    conto.tv_sec =  clientTimeout;
    conto.tv_usec = clientTimeoutMsec*1000;
  }

retry_select:
  for (k=0; (nready=select(maxFdSets,&conset,NULL,NULL,&conto)) > 0 && k<gBurstLimit; k++)
  {
    conto.tv_usec = 0; conto.tv_sec = 0;
    WaitForWorkArea(0);
    IncomingNetworkProtocol = TCP;
    if (NGdebug) sprintf(fecprot,"TCP");
    for (i=0; i<nConnectionTableEntries; i++)    /* possible TCP connections */
      if ((c=connTable[i])->TCPsocket != 0) recvIncomingIPData(c->TCPsocket);
    IncomingNetworkProtocol = UDP;
    if (NGdebug) sprintf(fecprot,"UDP");
    recvIncomingIPData(ConsumerSocket); n = 0;
    if (ListenForMCasts) 
    {
      if (MCastConsumerSocket != 0) recvIncomingIPData(MCastConsumerSocket);
    }
    if (GlobalSocket > 0 && FD_ISSET(GlobalSocket,&conset))
    {
      if ((nread=recvfrom(GlobalSocket,(char *)clnWorkArea,BRDCST_BUFFER_SIZE,0,(struct sockaddr *)&IncomingIPfec,(int *)&len)) > 0)
      {
        if (NGdebug) sprintf(fecprot,"GLB");
        if (nread == PeekWord(*(short *)clnWorkArea)) getGlobalUpdate(clnWorkArea);
      }
    }
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    for (i=0; i<RPCNumFecsRead; i++) if (RPCFec[i].Prot == PIPE)
    {
       IncomingNetworkProtocol = PIPE;
       if (NGdebug) sprintf(fecprot,"LCL");
       memcpy(&sck,RPCFec[i].Network,4);
       if (sck > 0) recvIncomingIPData(sck);
       if (sck == 0) { RPCFec[i].Prot = 0; memset(RPCFec[i].Network,0,4); }
    }
#   endif
    ReleaseWorkArea(0);
#   ifdef UNIX
    if (foregroundCln && FD_ISSET(STDIN_FILENO,&conset) &&
        IPCSocket == 0 && GetCommand(STDIN_FILENO,cmdbuf,CMDSIZE) > 0)
      InterpretCommand(cmdbuf);
#   ifndef NO_TINE_PIPE
    for (i=0; i<RPCNumFecsRead; i++) if (RPCFec[i].Prot == PIPE)
    {
       memcpy(&sck,RPCFec[i].Network,4);
       if (sck > 0) FD_SET(sck,&conset);
    }
#   else
    sck = sck; /* aviod a warning */
#   endif /* NO_TINE_PIPE */
#   endif /* UNIX */
    for (i=0; i<nConnectionTableEntries; i++)    /* possible TCP connections */
      if ((c=connTable[i])->TCPsocket != 0) FD_SET(c->TCPsocket,&conset);
    FD_SET(ConsumerSocket,&conset);
    if (ListenForMCasts) 
    {
      if (MCastConsumerSocket != 0) FD_SET(MCastConsumerSocket,&conset);
    }
    if (GlobalSocket) FD_SET(GlobalSocket,&conset);
#   ifdef IPXBSD
    if (IPXConsumerSocket && FD_ISSET(IPXConsumerSocket,&conset)) break;
    if (IPXMCastConsumerSocket && FD_ISSET(IPXMCastConsumerSocket,&conset)) break;
#   endif
  }
  if (nready == -1)
  {
    if (soerrno != EWOULDBLOCK)
    {
      if (NGdebug > 2) soperror("UDP select");
      if (retry-- > 0) goto retry_select;
    }
  }
  ReturnToClientCycle = (k == gBurstLimit) ? 1 : 0;
}
IDLE_CONNECTION *RemoveIdleConnection(IDLE_CONNECTION *ic)
{
  IDLE_CONNECTION *nxt=NULL;
  if (ic == NULL) return ic;
  nxt = ic->nxt;
  if (ic->prv != NULL) ic->prv->nxt = ic->nxt;
  else if (ic == tcpIdleConnectionList) tcpIdleConnectionList = ic->nxt;
  if (NGdebug) dbglog("TCP connection to %s (sck %d) no longer idle",RPCFec[ic->idx].Name,ic->sck);
  SystemFree(ic);
  return nxt;
}
IDLE_CONNECTION *AddIdleConnection(int sck,int idx)
{
  IDLE_CONNECTION *ic=NULL;

  if ((ic=getIdleConnection(idx)) != NULL) return ic;
  if ((ic=(IDLE_CONNECTION *)SystemCalloc(1,sizeof(IDLE_CONNECTION))) == NULL) return NULL;
  ic->sck = sck;
  ic->idx = idx;
  ic->ttl = 300000;
  ic->nxt = tcpIdleConnectionList;
  if (tcpIdleConnectionList != NULL) tcpIdleConnectionList->prv = ic;
  tcpIdleConnectionList = ic;
  if (NGdebug) dbglog("mark TCP connection to %s (sck %d) as idle",RPCFec[idx].Name,ic->sck);
  return ic;
}
IDLE_CONNECTION *getIdleConnection(int idx)
{
  IDLE_CONNECTION *ic=NULL;

  for (ic=tcpIdleConnectionList; ic != NULL; ic=ic->nxt)
  {
    if (ic->idx == idx) return ic;
  }
  return NULL;
}
IDLE_CONNECTION *getIdleConnectionBySocket(int sck)
{
  IDLE_CONNECTION *ic=NULL;

  for (ic=tcpIdleConnectionList; ic != NULL; ic=ic->nxt)
  {
    if (ic->sck == sck) return ic;
  }
  return NULL;
}
