#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/stat.h>
#include "errors.h"
#include "rpctypes.h"
#include "servdef.h"
#include "srvip.h"
#include "toolkit.h"
#include "database.h"

#define RPTR_DEFAULT_PORT 101
#define LSNEQM "LSNEQM"
#define STD_TRPTR_FEC "TINERPTRFEC"
#define STD_TRPTR_SRV "TINERPTRSRV"

double blob[8000];
PropertyQueryStruct prpinfo;
BYTE *bval = (BYTE *)blob;
short *sval = (short *)blob;
UINT32 *lval = (UINT32 *)blob;
float *fval = (float *)blob;
double *dval = blob;
char *cval = (char *)blob;
NAME16 *n16val = (NAME16 *)blob;
NAME32 *n32val = (NAME32 *)blob;

void tmonitorCallback(int id,int cc)
{
  int i;
  printf("\n");
  if (cc) 
  {
    printf("error: %s\n>",RPCLastErrorString);
    return;
  }
  switch (LFMT(prpinfo.prpFormat))
  {
    case CF_BYTE:
      for (i=0; i<prpinfo.prpSize; i++) printf("%d ",bval[i]); break;
    case CF_SHORT:
      for (i=0; i<prpinfo.prpSize; i++) printf("%d ",sval[i]); break;
    case CF_LONG:
      for (i=0; i<prpinfo.prpSize; i++) printf("%d ",lval[i]); break;
    case CF_FLOAT:
      for (i=0; i<prpinfo.prpSize; i++) printf("%g ",fval[i]); break;
    case CF_DOUBLE:
      for (i=0; i<prpinfo.prpSize; i++) printf("%g ",dval[i]); break;
    case CF_TEXT:
      printf("%s",cval); break;
    case CF_NAME16:
      for (i=0; i<prpinfo.prpSize; i++) printf("%16s ",n16val[i].name); break;
    case CF_NAME32:
      for (i=0; i<prpinfo.prpSize; i++) printf("%32s ",n32val[i].name); break;
    default: printf("output format type %d not implemented\n",LFMT(prpinfo.prpFormat));
  }
}
int _tmonitor(void *buf,char *dev,char *prp,int fmt,int len,int tmr)
{
  int id;
  DTYPE dout;

  dout.dFormat = fmt;
  dout.dArrayLength = len;
  dout.dTag[0] = 0;
  dout.data.vptr = buf;
  id = AttachLink(dev,prp,&dout,NULL,CA_READ,tmr,tmonitorCallback,CM_POLL);
  return id < 0 ? -id : 0;
}
int _tput(void *buf,char *dev,char *prp,int fmt,int len,int tmr)
{
  DTYPE din;

  din.dFormat = fmt;
  din.dArrayLength = len;
  din.dTag[0] = 0;
  din.data.vptr = buf;

  return ExecLinkEx(dev,prp,NULL,&din,CA_WRITE,(UINT16)tmr);
}
int _tget(void *buf,char *dev,char *prp,int fmt,int len,int tmr)
{
  DTYPE dout;

  dout.dFormat = fmt;
  dout.dArrayLength = len;
  dout.dTag[0] = 0;
  dout.data.vptr = buf;
  return ExecLinkEx(dev,prp,&dout,NULL,CA_READ,(UINT16)tmr);
}
int GetPropertyInformation(char *srv,char *prp,PropertyQueryStruct *srvProps)
{
  DTYPE dout,din;
  int cc = 0;
 
  din.dFormat = CF_NAME32;
  din.dArrayLength = 1;
  din.data.vptr = prp;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = PROPERTYQUERYSTRUCT_SIZE;
  dout.data.vptr = srvProps;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,&din,CA_READ,500)) != 0) return cc;
  srvProps->prpSize = ISWAP(srvProps->prpSize);
  return 0;
}
char *fmtstr(int fmt)
{
  static char s[80];
  switch (LFMT(fmt))
  {
    case CF_BYTE: sprintf(s,"BYTE"); break;
    case CF_SHORT: sprintf(s,"short"); break;
    case CF_LONG: sprintf(s,"UINT32"); break;
    case CF_FLOAT: sprintf(s,"float"); break;
    case CF_DOUBLE: sprintf(s,"double"); break;
    case CF_TEXT: sprintf(s,"char"); break;
    case CF_NAME16: sprintf(s,"NAME16"); break;
    case CF_NAME32: sprintf(s,"NAME32"); break;
    default: sprintf(s,"type <%d>",LFMT(fmt)); break;
  }
  return s;
}
int tinfo(char *dev,char *prp)
{
  int cc;
  float max,min;
  char egu[80];
 
  if ((cc=GetPropertyInformation(dev,prp,&prpinfo)) != 0) return cc;
  if ((cc=GetDevicePropertyEGU(dev,prp,&max,&min,egu)) != 0) return cc;
  printf("\nDevice : %s\nProperty : %s\nDescription : %s\nSize : %d\nFormat %s\nMaximum : %f\nMinimum : %f\nUnits : %s\n",
         dev,prpinfo.prpName,prpinfo.prpDescription,prpinfo.prpSize,
         fmtstr(prpinfo.prpFormat),max,min,egu);
  return 0;
}
int tserversEx(char *cnt,char *sub)
{
  int i,k,cc,n = 256;
  NAME16 srvs[256];
  char srv[64];
  DTYPE dout,din,*dinptr=NULL;

  strncpy(srv,"ENS",16);
  if (cnt != NULL && strlen(cnt) > 0)
  {
    strcat(srv,"/"); strncat(srv,cnt,16);
  }
  if (sub != NULL)
  {
    din.dFormat = CF_TEXT;
    din.dArrayLength = strlen(sub);
    din.data.cptr = sub;
    dinptr = &din;
  }
  dout.dFormat = CF_NAME16;
  dout.dArrayLength = n;
  dout.data.vptr = srvs;
  memset(srvs,0,n*16);
  if ((cc=ExecLinkEx(srv,"TAGS",&dout,dinptr,CA_READ,200)) != 0) return cc;
  printf("Servers in context %s",cnt);
  if (sub != NULL) printf(" subsystem %s",sub);
  printf(":\n\n");
  for (i=0,k=0; i<n; i++)
  {
    if (strlen(srvs[i].name) == 0) continue;
    printf("%s\n",srvs[i].name);
    k++;
  }
  printf("\n%d servers found\n",k);
  return 0;
}
int tservers(char *cnt)
{
  int i,cc,n = 256;
  NAME16 srvs[256];
  if ((cc=GetDeviceServersEx(cnt,srvs,&n)) != 0) return cc;
  printf("Servers in context %s",cnt);
  printf(":\n\n");
  for (i=0; i<n; i++) 
  {
    printf("%s\n",srvs[i].name);
  }
  return 0;
}
int tprops(char *dev)
{
  int i,cc,n = 1024,hasQryFcn=FALSE;
  NAME32 *prps = n32val;
  if ((cc=GetDeviceProperties(dev,prps,&n)) != 0)
  {
    if (cc & has_query_function) hasQryFcn = TRUE;
    if (cc & CE_SENDDATA) cc = 0;
  }
  if (cc) return cc;
  printf("%d properties found :\n",n);
  for (i=0; i<n; i++) 
  {
    if (strlen(prps[i].name) == 0) continue;
    printf("%32s ",prps[i].name);
    if ((i+1)%4 == 0) printf("\n");
  }
  return 0;
}
int tproplist(char *dev)
{
  int i,cc,n = 1024,hasQryFcn=FALSE;
  NAME32 *prps = n32val;
  if ((cc=GetDeviceProperties(dev,prps,&n)) != 0)
  {
    if (cc & has_query_function) hasQryFcn = TRUE;
    if (cc & CE_SENDDATA) cc = 0;
  }
  if (cc) return cc;
  printf("Property List for %s (%d properties)\n\n",dev,n);
  for (i=0; i<n; i++) 
  {
    if (strlen(prps[i].name) == 0) continue;
    printf("%s\n",prps[i].name);
  }
  return 0;
}
int tdevsEx(char *dev,char *prp)
{
  int i,cc,n = 1024,hasQryFcn=FALSE;
  NAME16 *devs = n16val;
  char devname[32];
  if ((cc=GetDeviceNamesEx(dev,prp,devs,&n)) != 0)
  {
    if (cc & has_query_function) hasQryFcn = TRUE;
    if (cc & CE_SENDDATA) cc = 0;
  }
  if (cc) return cc;
  if (n == 2 && !strcmp(devs[1].name,"XYZHASQUERYFCN"))
  {
    n = 1;
    if (!strcmp(devs[0].name,"#0")) strncpy(devs[0].name,prp,16);
  }
  printf("%d devices found :\n",n);
  for (i=0; i<n; i++) 
  {
    strncpy(devname,devs[i].name,16);
    if (devname[15] == '&')
    {
      i++; devname[15] = 0;
      strncat(devname,devs[i].name,16);
    }
    if (strlen(devname) == 0) continue;
    printf("%s\n",devname);
  }
  return 0;
}
int tdevs(char *dev)
{
  int i,cc,n = 1024,hasQryFcn=FALSE;
  NAME16 *devs = n16val;
  char devname[32];
  if ((cc=GetDeviceNames(dev,devs,&n)) != 0)
  {
    if (cc & has_query_function) hasQryFcn = TRUE;
    if (cc & CE_SENDDATA) cc = 0;
  }
  if (cc) return cc;
  if (n == 2 && !strcmp(devs[1].name,"XYZHASQUERYFCN")) hasQryFcn = TRUE;
  if (hasQryFcn)
  {
    printf("Device List depends on targetted Property: use tdevsEx()\n",n);
    return ambiguous;
  }
  printf("%d devices found :\n",n);
  for (i=0; i<n; i++) 
  {
    strncpy(devname,devs[i].name,16);
    if (devname[15] == '&')
    {
      i++; devname[15] = 0;
      strncat(devname,devs[i].name,16);
    }
    if (strlen(devname) == 0) continue;
    printf("%s\n",devname);
  }
  return 0;
}
int tdevlist(char *dev)
{
  int i,cc,n = 1024,hasQryFcn=FALSE;
  NAME16 *devs = n16val;
  char devname[32];
  if ((cc=GetDeviceNames(dev,devs,&n)) != 0)
  {
    if (cc & has_query_function) hasQryFcn = TRUE;
    if (cc & CE_SENDDATA) cc = 0;
  }
  if (cc) return cc;
  printf("DEVICE_NAME,DEVICE_NUMBER\n",n);
  for (i=0; i<n; i++)
  {
    strncpy(devname,devs[i].name,16);
    if (devname[15] == '&')
    {
      i++; devname[15] = 0;
      strncat(devname,devs[i].name,16);
    }
    if (strlen(devname) == 0) continue;
    printf("%s,%d\n",devname,i);
  }
  return 0;
}
int tsendstring(char *dev,char *prp,char *str)
{
  if (str == NULL || strlen(str) == 0) return invalid_data;
  return _tput(str,dev,prp,CF_TEXT,strlen(str),1000);
}
/* #define OLDWAY */
extern int NGdebug;
BYTE *parseInputDataString(char *str,int *fmt,int *siz)
{
  BYTE *data=NULL;
  char *s, *c, *tok = " ,;\t\n";
  int i, n, len, base = 10;
  if (str == NULL || fmt == NULL || siz == NULL) return data;
  *fmt = CF_NULL; *siz = 0;
  if (str[0] == '\"')
  {
    if ((c=strchr(&str[1],'\"')) == NULL) return data; /* badly formed */
    *c = 0;
    data = (BYTE *)&str[1];
    *siz = strlen((char *)data);
    *fmt = CF_TEXT;
  }
  else
  {
    len = strlen(str); n = 0;
    data = (BYTE *)calloc(len,32); /* worst case ? */
    for (s=strtok(str,tok); s != NULL; s = strtok(NULL,tok))
    {
      if (*fmt == CF_NULL)
      {
        len = strlen(s);
        for (i=0; i<len; i++) if (!isdigit(s[i])) break;
        if (i == len) /* all digits ! */
        {
          *fmt = CF_LONG;
          base = 10;
        }
        else if (strncmp(s,"0x",2) == 0 || strncmp(s,"0X",2) == 0)
        {
          *fmt = CF_LONG;
          base = 16;
        }
        else if (s[i] == '.')
        {
          *fmt = CF_DOUBLE;
        }
        else
        {
          if (len < 16) *fmt = CF_NAME16;
          else if (len < 32) *fmt = CF_NAME32;
          else *fmt = CF_TEXT;
        }
      }
      switch (*fmt)
      {
        case CF_LONG:
          ((UINT32 *)data)[n] = strtol(s,NULL,base); break;
        case CF_DOUBLE:
          ((double *)data)[n] = strtod(s,NULL); break;
        case CF_NAME16:
          strncpy(((NAME16 *)data)[n].name,s,16); break;
        case CF_NAME32:
          strncpy(((NAME32 *)data)[n].name,s,16); break;
        case CF_TEXT:
          strncpy((char *)data,s,16); 
          *siz = len;
          return data;
        default:
          break;
      }
      n++;
    }
    if ((*fmt == CF_NAME16 || *fmt == CF_NAME32) && n == 1)
    {
      *fmt = CF_TEXT;
      n = len;
    }
    *siz = n;
  }
  return data;
}
int tput(char *dev,char *prp,char *datastr)
{
  BYTE *data = NULL;
  int fmt=CF_NULL,siz=0;
  data = parseInputDataString(datastr,&fmt,&siz);
  return _tput(data,dev,prp,fmt,siz,1000);
}
void dumpReturnedData(void)
{
  int i;
  int fmt,siz;
  char s[65];
  fmt = GetCompletionDataType(-1);
  siz = GetCompletionDataSize(-1);
  switch (fmt)
  {
    case CF_BYTE:
      for (i=0; i<siz; i++) printf("%d ",bval[i]); break;
    case CF_SHORT:
      for (i=0; i<siz; i++) printf("%d ",sval[i]); break;
    case CF_LONG:
      for (i=0; i<siz; i++) printf("%d ",lval[i]); break;
    case CF_FLOAT:
      for (i=0; i<siz; i++) printf("%g ",fval[i]); break;
    case CF_DOUBLE:
      for (i=0; i<siz; i++) printf("%g ",dval[i]); break;
    case CF_TEXT:
      printf("%s",cval); break;
    case CF_NAME16:
      for (i=0; i<siz; i++) 
      {
        strncpy(s,n16val[i].name,16);
        printf("%s ",s);
      }
      break;
    case CF_NAME32:
      for (i=0; i<siz; i++)
      {
        strncpy(s,n32val[i].name,32);
        printf("%s ",s);
      }
      break;
    default: printf("output format type %d not implemented\n",fmt);
  }
}
int tgetFromRepeater(char *dev,char *prp,int fmt_,int siz_)
{
  int cc;
  int siz;
  struct in_addr ip;
  UINT32 haddr;
  char ipaddr[16];
  char rpdev[100],tgt[100],rpfn[100];
  DTYPE dout,din;
  struct stat ipcstat;

  sprintf(rpfn,"%s%s-server.ipc",IPC_PATH,STD_TRPTR_FEC);
  if (stat(rpfn,&ipcstat) != 0) return file_error;
  haddr = getmyipaddr();
  memcpy(&ip,&haddr,4);
  inet_ntoa_b(ip,ipaddr);
  appendAddressToCache(STD_TRPTR_FEC,ipaddr,RPTR_DEFAULT_PORT);
  appendServerToCache(STD_TRPTR_FEC,"REPEATER",STD_TRPTR_SRV,LSNEQM);

  sprintf(tgt,"%s[%s]",dev,prp);
  sprintf(rpdev,"/REPEATER/%s/#0",STD_TRPTR_SRV);
  din.dFormat = CF_TEXT;
  din.dArrayLength = strlen(tgt);
  din.dTag[0] = 0;
  din.data.cptr = tgt;

  siz = siz_; 
  if (siz > sizeof(blob)) siz = sizeof(blob);
  dout.dFormat = fmt_ == CF_DEFAULT ? CF_FLOAT : fmt_;
  dout.dArrayLength = fmt_ == CF_DEFAULT ? 1 : siz;
  dout.dTag[0] = 0;
  dout.data.vptr = blob;
  if ((cc=ExecLink(rpdev,"MONITOR",&dout,&din,CA_READ)) != 0) 
  {
    while (cc == invalid_transport_size || cc == use_stream_transport)
    {
      siz /= 2; 
      if (siz < 1000) return cc;
      dout.dArrayLength = siz;
      if ((cc=ExecLink(rpdev,"MONITOR",&dout,&din,CA_READ)) == 0) goto get_done;
    }
    return cc;
  }
get_done:
  dumpReturnedData();
  return 0;
}
int tgetFormatAndSize(char *dev,char *prp,int *fmt,int *siz)
{
  int cc;
  if (dev == NULL || prp == NULL || fmt == NULL || siz == NULL) return argument_list_error;
  if ((cc=GetPropertyInformation(dev,prp,&prpinfo)) != 0) return cc;
  *fmt = LFMT(prpinfo.prpFormat);
  *siz = prpinfo.prpSize;
  return 0;
}
int tgetEx(char *dev,char *prp,int fmt_,int siz_)
{
  int cc,retry_cnt=0;
  int fmt,siz;

  if (dev == NULL || prp == NULL) return argument_list_error;
  if (fmt_ != CF_NULL && siz_ != 0)
  {
    fmt = fmt_;
    siz = siz_;
    if (siz > 8000) siz = 8000;
    if ((cc=_tget(blob,dev,prp,fmt,siz,5000)) != 0) return cc;
  }
  else
  {
    siz = sizeof(blob);
get_retry:
    if ((cc=_tget(blob,dev,prp,CF_DEFAULT,siz,500)) != 0) 
    {
      while (cc == invalid_transport_size || cc == use_stream_transport)
      {
        siz /= 2; 
        if (siz < 1000) return cc;
        if ((cc=_tget(blob,dev,prp,CF_DEFAULT,siz,500)) == 0) goto get_done;
      }
      if ((cc == link_not_open || cc == connection_timeout) && retry_cnt++ < 2) goto get_retry;
      return cc;
    }
  }
get_done:
  dumpReturnedData();
  return 0;
}
int tget(char *dev,char *prp)
{
  return tgetEx(dev,prp,CF_NULL,0);
}
int tmonitorEx(char *dev,char *prp,int fmt_,int siz_)
{
  int cc;
  int fmt,siz;

  if (fmt_ != CF_NULL && siz_ != 0)
  {
    fmt = fmt_;
    siz = siz_;
    prpinfo.prpFormat = BFMT(fmt);
    prpinfo.prpSize = siz;
  }
  else
  {
    if ((cc=GetPropertyInformation(dev,prp,&prpinfo)) != 0) return cc;
    fmt = LFMT(prpinfo.prpFormat);
    siz = prpinfo.prpSize;
  }
  if (siz > 8000) siz = 8000;
  if ((cc=_tmonitor(blob,dev,prp,fmt,siz,1000)) != 0) return cc;
  return 0;
}
int tmonitor(char *dev,char *prp)
{
  return tmonitorEx(dev,prp,CF_NULL,0);
}
time_t parseTimeString(char *ts)
{
  time_t t=time(NULL);
  int i, len;
  char *c,day[16],mon[64],year[16];
  char hr[64],mn[32],sec[16];
  struct tm tmts, *tnow;

  hr[0] = 0;
  tnow = localtime(&t);
  len = strlen(ts);
  strupr(ts);
  for (i=0; i<len && isdigit(ts[i]); i++);  
  if (i == len) t = atoi(ts);
  else if (!strcmp(ts,"NOW")) t = time(NULL);
  else if ((c=strstr(ts,".")) != NULL) /* day.month.year format */
  {
    *c++ = 0;
    strncpy(day,ts,16);
    strncpy(mon,c,64);
    if ((c=strstr(mon,".")) != NULL)
    { 
      *c++ = 0;
      strncpy(year,c,16);
      if ((c=strstr(year," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    else
    {
      sprintf(year,"%d",1900+tnow->tm_year);
      if ((c=strstr(mon,"_")) != NULL || (c=strstr(mon," ")) != NULL)
      {
        *c++ = 0;
        strncpy(hr,c,64);
      }
    }
    if (strlen(hr) > 0)
    {
      mn[0] = 0; sec[0] = 0;
      if ((c=strstr(hr,":")) != NULL)
      {     
        *c++ = 0;
        strncpy(mn,c,32);
        if ((c=strstr(mn,":")) != NULL)
        {
          *c++ = 0;
          strncpy(sec,c,16);
        }
      }
      tmts.tm_hour = atoi(hr);
      tmts.tm_min =  atoi(mn);
      tmts.tm_sec =  atoi(sec);
    }
    else
    {
      memset(&tmts,0,sizeof(struct tm));
    }
    tmts.tm_mon = (atoi(mon)+11)%12;
    tmts.tm_mday = atoi(day);
    tmts.tm_year = atoi(year);
    if (tmts.tm_year > 1900) tmts.tm_year -= 1900;
    tmts.tm_isdst = -1;
    t = mktime(&tmts);
  }
  return t;
}
int thistory(char *context,char *keyword,char *devname,char *stoptime,char *depth)
{
  char devsrv[256],srvnam[64];
  long stop = time(NULL), start, mf = 1, dp = 0, isLocalHistory = FALSE;
  int i, num = 5000, cc, offset = 0, fmt;
  FLTINT *fiDataArray = (FLTINT *)blob;
  NAME32I *niDataArray = (NAME32I *)blob;
  float *fDataArray = (float *)blob;

  if (keyword == NULL) return argument_list_error;
  if (devname == NULL || strlen(devname) == 0) devname = "#0";
  if (context == NULL) context = "HERA";
  else strupr(context);
  if (!strcmp(context,"HERA")) sprintf(devsrv,"/HERA/HEHISTORY/");
  else if (!strcmp(context,"PETRA")) sprintf(devsrv,"/PETRA/PEHISTORY/");
  else if (!strncmp(context,"DESY",4)) sprintf(devsrv,"/DESY/D3HISTORY/");
  else if (!strncmp(context,"LINAC",5)) sprintf(devsrv,"/DESY/D3HISTORY/");
  else if (!strcmp(context,"DORIS")) sprintf(devsrv,"/DORIS/DOHISTORY/");
  else if (!strcmp(context,"TTF")) sprintf(devsrv,"/TTF/TTFHISTORY/");
  else strncpy(devsrv,context,36);

  strncpy(srvnam,devsrv,64);
  if (strstr(devsrv,"HISTORY") == NULL) isLocalHistory = TRUE;

  if (devsrv[strlen(devsrv)-1] != '/') strcat(devsrv,"/");
  strcat(devsrv,devname);
  strcat(devsrv,"/");
  strcat(devsrv,keyword);
  if (isLocalHistory && strstr(keyword,".HIST") == NULL) strcat(devsrv,".HIST");
 
  stop = parseTimeString(stoptime);

  if (!stricmp(depth,"SNAPSHOT"))
  { /* snapshot request ? */
    tgetFormatAndSize(srvnam,keyword,&fmt,&num);
    if ((cc=GetArchivedTraceDataAsFloat(devsrv,stop,stop,fDataArray,num)) == 0)
    {
      stop = (time_t)gDataTimeStamp;
      printf("snapshot of %s %s/%s at %s",context,keyword,devname,ctime(&stop));
      for (i=0; i<num; i++)
      {
        printf("%g\n",fDataArray[i]);
      }
    }
    return cc;
  }  
  
  strupr(depth);
  if (strstr(depth,"DAY")) mf = 24 * 3600;
  else if (strstr(depth,"HOUR")) mf = 3600;
  else if (strstr(depth,"MINUTE")) mf = 60;
  else if (strstr(depth,"WEEK")) mf = 24 * 3600 * 7;
  dp = mf * atoi(depth);
  start = stop - dp;
  printf("%s/%s, Timestamp\n",keyword,devname);
  if ((cc=GetArchivedDataAsFloat(devsrv,start,stop,fiDataArray,&num)) == 0) 
  {
    if ((int)fiDataArray[0].ival <= num) offset = 1;
    for (i=offset; i<num; i++)
    {
      printf("%g, %s",fiDataArray[i].fval,ctime((time_t *)&fiDataArray[i].ival));
    }
    return 0;
  }
  if (cc != illegal_format) return cc;
  if ((cc=GetArchivedDataAsText(devsrv,start,stop,niDataArray,&num)) == 0) 
  {
    if ((int)niDataArray[0].ival <= num) offset = 1;
    for (i=offset; i<num; i++)
    {
      printf("%s, %s",niDataArray[i].name,ctime((time_t *)&niDataArray[i].ival));
    }
    return 0;
  }
  
  return cc;
}
  
