#include "project.def"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <stddef.h>
#include <string.h>
#include <time.h>
#include "unixcmd.h"

#define CLN_PATH "/tmp/"
#ifndef CLN_PATH
#  define CLN_PATH "/var/tmp/"
#endif
#define CLN_PERM S_IRWXU

int CMDcln(const char *name)
{
  int fd,len;
  struct sockaddr_un unix_addr;
#ifndef SVR4
  if ((fd=socket(AF_UNIX,SOCK_STREAM,0)) < 0) return -1;

  memset(&unix_addr,0,sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  
  sprintf(unix_addr.sun_path, 
          strlen(CLN_PATH) == 9 ? "%s%05d" : "%s%09d", 
          CLN_PATH, getpid());

# ifdef XSCM_RIGHTS
  len = sizeof(unix_addr.sun_len) + sizeof(unix_addr.sun_family) +
        strlen(unix_addr.sun_path) + 1;
  unix_addr.sun_len = len;
# else
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
  if (len != 16) return -5;
# endif

  unlink(unix_addr.sun_path);
  if (bind(fd, (struct sockaddr *)&unix_addr, len) < 0) return -2;
  if (chmod(unix_addr.sun_path, CLN_PERM) < 0) return -3;

  memset(&unix_addr, 0, sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  strcpy(unix_addr.sun_path, name);

# ifdef XSCM_RIGHTS
  len = sizeof(unix_addr.sun_len) + sizeof(unix_addr.sun_family) +
        strlen(unix_addr.sun_path) + 1;
  unix_addr.sun_len = len;
# else
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
# endif
  if (connect(fd, (struct sockaddr *)&unix_addr,len) < 0) return -4;
#else
  if ((fd=open(name,O_RDWR)) < 0) return -1;
  if (isastream(fd) == 0) return -2;
#endif /* SVR4 */

  return fd;
}

#define MAXLINES 1024

#define quit(txt) { perror(txt); return -1; }
char cmd[256];

int cleanup(void)
{
  static int step = 1;
  
  switch( step )
  {
    case 1: strcpy(cmd,"set debug=0"); break;
    case 2: strcpy(cmd,"set logdbg=0"); break;
    default: return -1;
  }
  return step++;
}
#define signoff(s) {\
  printf("\nThanks for using REMOTE\n");\
  close(csfd);\
  exit(0);\
}
     
int sndCommand(char *ipcfn)
{
  char buf[MAXLINES];
  int len, cleanupFlag = 0,sendFlag = 0;
  static int csfd = -1;
  static struct timeval to;
  fd_set rset,allset;
  int i,n,maxfd,maxi,lsnfd,clnfd,nread;
  int cnt = 0;

  if (csfd < 0 && (csfd=CMDcln(ipcfn)) < 0)
  { 
    printf("CMDcln returned %d\n",csfd); 
    quit("CMD client error");
  }
  printf("Remote session established\n>");

  FD_ZERO(&allset);
  
  FD_SET(STDIN_FILENO,&allset);
  FD_SET(csfd,&allset);
  maxfd = csfd;

  for (;;)
  {
    rset = allset;
    to.tv_sec = 1; to.tv_usec = 0;
    if ((n=select(maxfd+1,&rset,NULL,NULL,&to)) < 0) quit("select error");
    if (FD_ISSET(csfd,&rset)) 
    {
      if ((nread = read(csfd,buf,MAXLINES)) <= 0) 
      {
        if (nread < 0) perror("read");
        else printf("server no longer responding\n");
        exit(1);
      }
      else
      {
        buf[nread] = 0;  
        printf("%s",buf);
      }
    }
    if (FD_ISSET(STDIN_FILENO,&rset))
    {
      if ((len=read(STDIN_FILENO,cmd,sizeof(cmd))) < 0) return -1;
      if (cmd[0] == '\n') { printf(">"); continue; }
      sendFlag = 1;
      if (!strncmp(cmd,"quit",4)) cleanupFlag = 1;
      if (!strncmp(cmd,"kill",4)) cleanupFlag = 2;
    }
    switch (cleanupFlag)
    {
      case 0: break;
      case 1: sendFlag = 1;
              if (cleanup() < 0) cleanupFlag = 3; 
              break;
      case 2: sendFlag = 1;
              printf("Aborting Server Task\n");
              strcpy(cmd,"quit\n");
              cleanupFlag = 3;
              break;
      case 3: signoff(csfd);
              break;
    }
    if (sendFlag)
    {
      if (write(csfd,cmd,strlen(cmd)) < 0) quit("CMD write error");
      cmd[0] = 0;
      sendFlag = 0;
    }
  }
  
  return 0;
}
    
