/** \file cprolog.h
 *
 * client-side specific macros ...
 *
 */
#ifndef CPROLOG_H_
#define CPROLOG_H_

#include "project.def"
#include "prolog.h"

#ifndef NODAL
#include "errors.h"
#endif

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

void closeIPClients(void);
#if defined(IPXNW) || defined(IPXBSD)
   void closeIPXClients(void);
   int initIPXClient(int socket);
   int SendSPXData(WORD ID,int socket);
#  define ShutdownClients(i) {\
     closeIPXClients();\
     closeIPClients();\
   }
#  define ConsumerCheck(i) {\
     if (i == 0) {ReturnToServerCycle++; saveClientCycle = ReturnToClientCycle;}\
     doIPXConsumer();\
     doUDPConsumer();\
     if (i == 0) {ReturnToServerCycle--; ReturnToClientCycle = saveClientCycle;}\
   }
#else
#  define SendSPXData(i,j) ipx_not_supported
#  define ShutdownClients(i) {\
	   closeIPClients();\
   }
#  define ConsumerCheck(i) {\
     if (i == 0) {ReturnToServerCycle++; saveClientCycle = ReturnToClientCycle;}\
     doUDPConsumer();\
     if (i == 0) {ReturnToServerCycle--; ReturnToClientCycle = saveClientCycle;}\
   }
#endif  /* IPXNW || IPXBSD */

#ifdef MSDOS   /* definitions */
#  define clnWorkArea     WorkArea
#  define clnWorkAreaSize WorkAreaSize
#  define whoami(c) ;
#else
   extern BYTE *clnWorkArea;
   extern UINT32 clnWorkAreaSize;
   char *whoami(char *user);
   void ClientDelay(int sec);  /* delay(0) never returns */
#endif

/**
 * \internal
 * \brief Increments the connection table
 * 
 * Trivial macro wrapper which is the same as n++.
 * 
 * \todo consider ways of dynamically incrementing connection table past MaxNumConnections
 * (at least for some platforms).
 */
#define IncrementConnections(n) (n)++
/**
 * \internal
 * \brief Empty macro
 * 
 * A place-holder in clnip.c and clnipx.c, in tandem with ReleaseWorkArea().  As 
 * client cycles and server cycles generally use different work areas on all OSes
 * capable of using threads, we should never have to worry about synchronizing 
 * a common work area.  Note that DOS and NIOS both use a common work area to save
 * memory but are inherently single threaded.
 * 
 * \todo get rid of these lines of code.
 * 
 */
#define WaitForWorkArea(i) ;
/**
 * \internal
 * \brief Empty macro
 * 
 * A place-holder in clnip.c and clnipx.c, in tandem with WaitForWorkArea().  As 
 * client cycles and server cycles generally use different work areas on all OSes
 * capable of using threads, we should never have to worry about synchronizing 
 * a common work area.  Note that DOS and NIOS both use a common work area to save
 * memory but are inherently single threaded.
 * 
 * \todo get rid of these lines of code.
 * 
 */
#define ReleaseWorkArea(i) ;

/**
 * Allocates memory for a connection table entry
 *
 * \param n Index in connection table
 * \todo replace macro by a function.
 * \todo ??? means int addConnectionTableEntry() { allocate, increment_counter }
 * \bug does not increment counter (see getConnectionEntry())
 */
#define AssignConnTableMemory(n) \
  ( n >= MaxNumConnections ? out_of_client_memory :\
    (connTable[n]=(ConnTableEntry *)SystemCalloc(1,sizeof(ConnTableEntry))) == NULL ? out_of_client_memory :\
   0 )
#define CheckSubscriptions(i) \
   while ((i=PrepSubBuffer(clnWorkArea)) >= 0);
#define prepNameServerEntry(i) {\
   connTable[i] = &connNameServer;\
   strcpy(connTable[i]->deviceName,devname);\
   strcpy(connTable[i]->prop,property);\
   if (dout != NULL) {\
     connTable[i]->sizeOut = dout->dArrayLength;\
     connTable[i]->formatOut = dout->dFormat;\
     connTable[i]->dataOut = (char *)dout->data.vptr;\
   } else {\
     connTable[i]->sizeOut = 0;\
     connTable[i]->formatOut = CF_NULL;\
     connTable[i]->dataOut = NULL;\
   }\
   if (din != NULL) {\
     connTable[i]->sizeIn = din->dArrayLength;\
     connTable[i]->formatIn = din->dFormat;\
     connTable[i]->dataIn = (char *)din->data.vptr;\
     connTable[i]->BytesTotal = connTable[i]->sizeIn*fmtsizeof(connTable[i]->formatIn);\
     connTable[i]->datamtu = UDP_BUFFER_SIZE-SUBSCRIPTION_SIZE-HEADERSTRUCT_SIZE;\
     if (connTable[i]->BytesTotal < connTable[i]->datamtu) {\
       connTable[i]->numblksIn = 1;\
       connTable[i]->BytesRemaining = connTable[i]->BytesTotal;\
       connTable[i]->nextDataSize = connTable[i]->BytesTotal;\
     } else { connTable[i]->BytesTotal = 0; }\
   } else {\
     connTable[i]->sizeIn = 0;\
     connTable[i]->formatIn = CF_NULL;\
     connTable[i]->dataIn = NULL;\
   }\
   connTable[i]->EqNo = EqNo;\
   connTable[i]->pending = 1;\
   if (connTable[i]->bufOut) { SystemFree(connTable[i]->bufOut); connTable[i]->bufOut = NULL; }\
   if (connTable[i]->bufOutBlks) { SystemFree(connTable[i]->bufOutBlks); connTable[i]->bufOutBlks = NULL; }\
   if (connTable[i]->bufCmp) { SystemFree(connTable[i]->bufCmp); connTable[i]->bufCmp = NULL; }\
   if (connTable[i]->dtf) { SystemFree(connTable[i]->dtf); connTable[i]->dtf = NULL; }\
}

#define WaitForSyncCount(i) doConsumer()

#define ListenForMulticasts(i) ListenForIPMulticasts()
#define ListenForNetGlobals(i) ListenForIPNetGlobals()


#if defined(MSDOS)             /* MSDOS: initTransport */
#  define initTransport(i)\
   ( (cc=initIPXClient(i)) != 0 && cc != ipx_not_supported ? cc :\
     (cc=initIPClient(i))  != 0 && cc != tcp_not_supported ? cc : 0)
#else                         /* NOT MSDOS  */
/**
 * \internal
 * \brief Initializes clnWorkArea
 *
 * Allocates memory for the input buffer.
 * \todo implement as a function
 * \todo handle MSDOS specials in that function - don't use overwriting #defines
 */
#  define initClnWorkArea(i) \
   ( (clnWorkArea=(BYTE *)SystemCalloc(1,clnWorkAreaSize=i)) == NULL ?\
     out_of_client_memory : 0 )
#  ifdef VXWORKS              /* VXWORKS */
     void RPCConsumer(void);
#    define initConsumerProcess(c) \
     (taskSpawn(c,100,0,10000,RPCConsumer,0,0,0,0,0,0,0,0,0,0) == ERROR ?\
      not_running : 0)
#    define initTransport(i)\
     ( (cc=initControlStructs()) ? cc :\
       (cc=initIPClient(i)&~tcp_not_supported) ? cc :\
       (cc=initClnWorkArea(clnWorkAreaSize))\
     )
#  else                       /* NOT VXWORKS */
/**
 * \internal
 * \brief Initializes transport protocols
 *
 * This macro depends on the operating system.
 * It calls initControlStructs() (NOT for MSDOS), initIPClient(),
 * initIPXClient (only for MSDOS or idf IPXBSD is defined),
 * initClnWorkArea() (NOT for MSDOS).
 *
 * \todo Better implement as a function.
 */
#    ifdef IPXBSD
#    define initTransport(i)\
     ( (cc=initControlStructs()) ? cc :\
       (cc=initIPClient(i)&~tcp_not_supported) ? cc :\
       (cc=initIPXClient(i)&~ipx_not_supported) ? cc :\
       initClnWorkArea(clnWorkAreaSize) )
#    else                     /* NO IPX */
#    define initTransport(i)\
     ( (cc=initControlStructs()) ? cc :\
       (cc=initIPClient(i)&~tcp_not_supported) ? cc :\
       initClnWorkArea(clnWorkAreaSize) )
#    endif
#  endif
#endif                       /* MSDOS initTransport */

#define CompIncomingData(out,cmp,sz,f,t) (\
     cmp == NULL ? FALSE :\
     t   != NULL ? !_datcmp(out,cmp,sz,f,t) :\
     !memcmp(out,cmp,sz)\
  )
#ifdef INTEL  /* data preparation */
#  define AssignClnStructDataPtr(dst,src,dir) {\
     dst = (BYTE *)src;\
   }
#  define cpyClnStructData(dst,src,siz) {\
     memcpy(connTable[slot]->dataIn,din->data.vptr,siz);\
   }
#  define AssignOutgoingHeader(h,buf) {\
     if (!ForceStructAlignment) h = (HeaderStruct *)buf;\
     memset(h,0,sizeof(HeaderStruct));\
   }
#  define AssignOutgoingBrdcstData(h,buf) h = (GlobalDataHdr *)buf; memset(h,0,sizeof(GlobalDataHdr));
#  define PrepIncomingData(dst,src,dsiz,rsiz,fmt) {\
     if (fmt == CF_STRUCT && c != NULL && (s=getStructFromTag(c->strTagOut))) {\
       StructSwapEx(c->strTagOut,(BYTE *)src,(BYTE *)dst,RTOL,(int)rsiz/s->rawlen,s->size);\
     } else {\
       memcpy(dst,src,rsiz);\
       if ((dsiz) > (rsiz)) memset(&((char *)(dst))[rsiz],0,dsiz-rsiz);\
     }\
   }
#  define PrepOutgoingHeader(h,dat) {\
     if (ForceStructAlignment) memcpy(dat,h,HEADERSTRUCT_SIZE);\
   }
#  define PrepOutgoingGCastData(h,dat) ;
#  define PrepIncomingGCastData(h,dat) h = (GlobalDataHdr *)dat;
#  define PrepOutgoingData(d,s,siz,fmt) memcpy(d,s,siz)
#  define AssignOutgoingSubscription(p,hndl,cte,eqnm,buf) {\
     if (p == 4) {\
       s = (struct subscription4 *)buf;\
       s->msgsize = (short)(SUBSCRIPTION4_SIZE + dsize);\
       s->PollingRate = (UINT16)(cte)->pollRate;\
       s->id = hndl;\
       s->mode = (short)BASEMODE((cte)->mode);\
       s->starttime = (cte)->starttime;\
       strncpy(s->contract.EqpName,eqnm,8);\
       strncpy(s->contract.Property,(cte)->prop,12);\
       s->contract.EqpSize = (short)(cte)->sizeOut;\
       if ((cte)->access & CA_WRITE) s->contract.EqpSize *= -1;\
       s->contract.EqpFormat = (cte)->formatOut;\
       s->contract.EqpNumber = (cte)->EqNo;\
     } else {\
       sEx = ForceStructAlignment ? &sTmpEx : (struct subscription *)buf;\
       sEx->msgsize = SUBSCRIPTION_SIZE + dsize;\
       sEx->PollingRate = (cte)->pollRate;\
       sEx->id = hndl;\
       sEx->blknum = (UINT8)(cte)->blknumIn;\
       sEx->numblks = (UINT8)(cte)->numblksIn;\
       sEx->blkid = (cte)->blkidIn;\
       sEx->mtu = (cte)->datamtu;\
       sEx->mode = (short)TARGETMODE((cte)->mode);\
       sEx->starttime = (cte)->starttime;\
       strncpy(sEx->contract.EqpName,eqnm,8);\
       strncpy(sEx->contract.EqpProperty,(cte)->prop,32);\
       strncpy(sEx->contract.EqpDeviceName,(cte)->deviceName,16);\
       if (strlen((cte)->deviceName) > 16 && strlen((cte)->prop) <= 16) {\
         sEx->contract.EqpDeviceName[15] = '&';\
         sEx->contract.EqpProperty[16] = 0;\
         strncpy(&sEx->contract.EqpProperty[17],&(cte)->deviceName[15],15);\
       }\
       sEx->contract.EqpSizeOut = (cte)->sizeOut;\
       sEx->contract.EqpSizeIn = (cte)->sizeIn;\
       sEx->contract.EqpAccess = (BYTE)(cte)->access;\
       sEx->contract.EqpFormatOut = (BYTE)(BFMT((cte)->formatOut));\
       sEx->contract.EqpFormatIn = (BYTE)(BFMT((cte)->formatIn));\
       strncpy(sEx->contract.strTagIn,(cte)->strTagIn,8);\
       strncpy(sEx->contract.strTagOut,(cte)->strTagOut,8);\
       if (ForceStructAlignment) memcpy(buf,&sTmpEx,sizeof(struct subscription));\
     }\
   }
#  define PrepOutgoingSubscription(s,dat) ;
#  define PrepIncomingSubscription(p,dat) {\
     if (p == 4) {\
       Pheader = (ProducerHeaderStruct4 *)dat;\
       PheaderEx->msgsize = Pheader->msgsize;\
       PheaderEx->SubscriptionID = Pheader->SubscriptionID;\
       PheaderEx->numblks = 1;\
       PheaderEx->blknum  = 1;\
       PheaderEx->mtu = UDP_BUFFER_SIZE;\
       PheaderEx->timestampMSEC = (UINT16)(nConnectionTableEntries ? connTable[Pheader->SubscriptionID]->sizeOut : 0);\
       PheaderEx->CompletionCode = Pheader->CompletionCode;\
       PheaderEx->EqpFormat = (BYTE)(nConnectionTableEntries ? BFMT(connTable[Pheader->SubscriptionID]->formatOut) : 0);\
       PheaderEx->counter = (BYTE)Pheader->counter;\
       PheaderEx->timestamp = Pheader->timestamp;\
       PheaderEx->ClnStarttime = Pheader->starttime;\
     }\
     else {\
       if (ForceStructAlignment) {\
         memcpy(&pTmpEx,dat,sizeof(ProducerHeaderStruct));\
         PheaderEx = &pTmpEx;\
       } else {\
         PheaderEx = (ProducerHeaderStruct *)dat;\
       }\
     }\
   }
#  define prepFecaddr(fecstruct,buf)  memcpy(fecstruct,buf,RPCFECSTRUCT_SIZE)
#  define prepEqpaddr(datastruct,buf) memcpy(datastruct,buf,RPCDATASTRUCT_SIZE)
#else
#  define AssignClnStructDataPtr(dst,src,dir) {\
     if (dir == RTOL) {\
       StructSwapEx(din->dTag,(BYTE *)src,(BYTE *)s->sdata,RTOL,din->dArrayLength,s->size);\
       dst = s->sdata;\
     }\
   }
#  define cpyClnStructData(dst,src,siz) {\
     if (din->dFormat == CF_STRUCT && (s=getStructFromTag(din->dTag)) != NULL)\
       StructSwapEx(din->dTag,(BYTE *)dst,(BYTE *)src,LTOR,din->dArrayLength,s->size);\
     else memcpy(connTable[slot]->dataIn,din->data.vptr,siz);\
   }
#  define AssignOutgoingHeader(h,buf) memset(h,0,sizeof(HeaderStruct))
#  define AssignOutgoingBrdcstData(h,buf) memset(h,0,sizeof(GlobalDataHdr))
#  define PrepOutgoingHeader(h,dat) {\
     _StructSwap(HeaderStructStruct,(unsigned char *)dat,(unsigned char *)h,LTOR);\
   }
#  define PrepOutgoingGCastData(h,dat) {\
     _StructSwap(GlobalDataHdrStruct,(unsigned char *)dat,(unsigned char *)h,LTOR);\
   }
#  define PrepIncomingGCastData(h,dat) {\
     _StructSwap(GlobalDataHdrStruct,(unsigned char *)dat,(unsigned char *)h,RTOL);\
   }
#  define PrepIncomingData(dst,src,dsiz,rsiz,fmt) {\
     if (fmt == CF_STRUCT && c != NULL && (s=getStructFromTag(c->strTagOut))) {\
       StructSwapEx(c->strTagOut,(BYTE *)src,(BYTE *)dst,RTOL,rsiz/s->rawlen,s->size);\
     } else {\
       memcpy(dst,src,rsiz);\
       SwapBytes((char *)dst,rsiz/getFormatSize(fmt),fmt,RTOL);\
     }\
     if ((dsiz) > (rsiz)) memset(&((char *)(dst))[rsiz],0,dsiz-rsiz);\
   }
#  define PrepOutgoingData(d,s,siz,fmt) {\
     memcpy(d,s,siz);\
     SwapBytes((char *)d,siz/getFormatSize(fmt),fmt,LTOR);\
   }
#  define AssignOutgoingSubscription(p,hndl,cte,eqnm,buf) {\
     if (p == 4) {\
       memset(s,0,sizeof(struct subscription4));\
       s->msgsize = (short)(SUBSCRIPTION4_SIZE + dsize);\
       s->PollingRate = (UINT16)(cte)->pollRate;\
       s->id = hndl;\
       s->mode = BASEMODE((cte)->mode);\
       s->starttime = (cte)->starttime;\
       strncpy(s->contract.EqpName,eqnm,8);\
       strncpy(s->contract.Property,(cte)->prop,12);\
       s->contract.EqpSize = (short)(cte)->sizeOut;\
       if ((cte)->access & CA_WRITE) s->contract.EqpSize *= -1;\
       s->contract.EqpFormat = (cte)->formatOut;\
       if (s->contract.EqpFormat == CF_TEXT) s->contract.EqpFormat = 1;\
       s->contract.EqpNumber = (cte)->EqNo;\
     } else {\
       memset(sEx,0,sizeof(struct subscription));\
       sEx->msgsize = SUBSCRIPTION_SIZE + dsize;\
       sEx->PollingRate = (cte)->pollRate;\
       sEx->id = hndl;\
       sEx->blknum = (UINT8)(cte)->blknumIn;\
       sEx->numblks = (UINT8)(cte)->numblksIn;\
       sEx->blkid = (cte)->blkidIn;\
       sEx->mtu = (cte)->datamtu;\
       sEx->mode = TARGETMODE((cte)->mode);\
       sEx->starttime = (cte)->starttime;\
       strncpy(sEx->contract.EqpName,eqnm,8);\
       strncpy(sEx->contract.EqpProperty,(cte)->prop,32);\
       strncpy(sEx->contract.EqpDeviceName,(cte)->deviceName,16);\
       if (strlen((cte)->deviceName) > 16 && strlen((cte)->prop) <= 16) {\
         sEx->contract.EqpDeviceName[15] = '&';\
         sEx->contract.EqpProperty[16] = 0;\
         strncpy(&sEx->contract.EqpProperty[17],&(cte)->deviceName[15],15);\
       }\
       sEx->contract.EqpSizeOut = (cte)->sizeOut;\
       sEx->contract.EqpSizeIn = (cte)->sizeIn;\
       sEx->contract.EqpAccess = (BYTE)(cte)->access;\
       sEx->contract.EqpFormatOut = (BYTE)(BFMT((cte)->formatOut));\
       sEx->contract.EqpFormatIn = (BYTE)(BFMT((cte)->formatIn));\
       strncpy(sEx->contract.strTagIn,(cte)->strTagIn,8);\
       strncpy(sEx->contract.strTagOut,(cte)->strTagOut,8);\
     }\
   }

#  define PrepOutgoingSubscription(p,dat) {\
     if (p == 4) {\
       _StructSwap(SUBSCRIPTION4Struct,(unsigned char *)dat,(unsigned char *)s,LTOR);\
     } else {\
       _StructSwap(SUBSCRIPTIONStruct,(unsigned char *)dat,(unsigned char *)sEx,LTOR);\
     }\
   }
#  define PrepIncomingSubscription(p,dat) {\
     if (p == 4) {\
       _StructSwap(ProducerHeaderStruct4Struct,(unsigned char *)dat,(unsigned char *)Pheader,RTOL);\
       PheaderEx->msgsize = Pheader->msgsize;\
       PheaderEx->SubscriptionID = Pheader->SubscriptionID;\
       PheaderEx->numblks = 1;\
       PheaderEx->blknum  = 1;\
       PheaderEx->mtu = UDP_BUFFER_SIZE;\
       PheaderEx->timestampMSEC = (UINT16)(nConnectionTableEntries ? connTable[Pheader->SubscriptionID]->sizeOut : 0);\
       PheaderEx->CompletionCode = Pheader->CompletionCode;\
       PheaderEx->EqpFormat = (BYTE)(nConnectionTableEntries ? BFMT(connTable[Pheader->SubscriptionID]->formatOut) : 0);\
       PheaderEx->counter = (BYTE)Pheader->counter;\
       PheaderEx->timestamp = Pheader->timestamp;\
       PheaderEx->ClnStarttime = Pheader->starttime;\
     }\
     else _StructSwap(ProducerHeaderStructStruct,(unsigned char *)dat,(unsigned char *)PheaderEx,RTOL);\
   }
#  define prepFecaddr(fecstruct,buf)  {\
	   _StructSwap(RPCFecStructStruct,(unsigned char *)buf,(unsigned char *)fecstruct,RTOL);\
   }
#  define prepEqpaddr(datastruct,buf)  {\
	   _StructSwap(RPCDataStructStruct,(unsigned char *)buf,(unsigned char *)datastruct,RTOL);\
   }
#endif  /* INTEL data preparation */

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* CPROLOG_H_ */

