#include "prolog.h"
#include "cprolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "servdef.h"
#include "srvip.h"
#include "brdcast.h"
#include "errors.h"
#include "database.h"

/**
 * \internal
 * \brief UDP server port.
 * 
 */
int gtUDPPort = UDP_PORT;

/**
 * \internal .
 * \brief UDP client port.
 */
int gtTCPPort = TCP_PORT;
                   
/**
  *\internal
 * \brief Port for multicast network services.
 */
int gtNetSrvPort = NETSRV_PORT;
          
/**
 * \internal
 * \brief Client UDP port for globals multicast.
 */
int gtGCastPort = GCAST_PORT;
          
/**
 * \internal
 * \brief Client UDP multicast port.
 */
int gtMCastPort = MCAST_PORT;
          

/**
 * \internal
 * \brief UDP client port.
 * 
 **/
int gtUDPClnPort = UDP_CLN_PORT;

/**
 * \internal
 * \brief Address of network services.
 */                                     
char *gtNetSrvAddr = NETSRV_ADDR;


/**
 * \internal
 * \brief Local broadcast address.
 */
char *gtNetLclBcast = NETBCST_ADDR;

/**
 * \internal
 * \brief Multicast address
*/
char *gtMCastAddr = MCAST_ADDR;

/**
 * \internal
 * \brief Globals multicast addres
 */
char *gtGCastAddr = GCAST_ADDR;

/**
 * \internal
 * \brief Time to live for multicasts (no. of router hops)
 */
int gtMCastTTL = MCAST_TTL;

UINT32 TCPcount = 0, UDPcount = 0,
	     SPXcount = 0, IPXcount = 0,
	     BRDcount = 0;
char clientaddr[30];
int TCPloaded = 0, IPXloaded = 0;
struct sockaddr_in prod_name;
fd_set ext_set;
#ifdef EXT_FDSET
int externalFdSets = -1;
#else
int externalFdSets = 0;
#endif
int maxFdSets = MAX_FDSETS;
void FD_MERGE(fd_set fdsrc,fd_set *fddst)
{
  int i;
  for (i=0; i<(int)sizeof(fd_set); i++)
    ((BYTE *)fddst)[i] |= ((BYTE *)&fdsrc)[i];
}
#ifdef WINNT
int WinsockLoaded(void)
{
  static WSADATA wsaData;
  static int WSAinitialized = -1;
  static WORD wVersionRequested;

  wVersionRequested = MAKEWORD( 2, 0 );
 
	if (WSAinitialized == -1) WSAinitialized = WSAStartup(wVersionRequested,&wsaData) ? 0 : 1;
  return WSAinitialized;
}
int GetErrno(void) { return WSAGetLastError();}
#endif
#define MTU 1460

#ifndef MAX_RCVBUF
/**
 * \internal
 * \brief Default maximal number of receive buffers for socket options.
 */
# define MAX_RCVBUF 0x10000
#endif

#ifndef MAX_SNDBUF
/**
 * \internal
 * \brief Default maximal number of send buffers for socket options.
 *
 */
# define MAX_SNDBUF (MAX_RCVBUF / 2)
#endif

/**
 * \internal
 * \brief Maximal number of receive buffers for socket options.
 */
int max_rcvbuf = MAX_RCVBUF;

/**
 * \internal
 * \brief Maximal number of send buffers for socket options.
 */
int max_sndbuf = MAX_SNDBUF;

EXPORT int _SystemAssignBufferSpace(UINT32 rcvBufferSpace,UINT32 sndBufferSpace)
{
  if (rcvBufferSpace > 0)
  {
    if (rcvBufferSpace < 0x1000) return buffer_too_small; 
    max_rcvbuf = (int)rcvBufferSpace;
  }
  if (sndBufferSpace > 0)
  {
    if (sndBufferSpace < 0x1000) return buffer_too_small; 
    max_sndbuf = (int)sndBufferSpace;
  }

  return 0;
}

TCPBCKT *bcktList = NULL;

TCPBCKT *findBucket(int sck)
{
  TCPBCKT *bckt=NULL;
  for (bckt=bcktList; bckt != NULL && bckt->sck != sck; bckt=bckt->nxt);
  return bckt;
}
TCPBCKT *getBucket(int sck)
{
  TCPBCKT *bckt=NULL;
  for (bckt=bcktList; bckt != NULL; bckt=bckt->nxt)
  if (bckt->sck == sck) return bckt;
  if ((bckt=(TCPBCKT *)SystemCalloc(1,sizeof(TCPBCKT))) != NULL)
  {
    bckt->sck = sck;
    bckt->nxt = bcktList;
    bcktList = bckt;
  }
  return bckt;
}
void freeBucket(int sck)
{
  TCPBCKT *bckt,*b;
  if ((bckt=findBucket(sck)) == NULL) return;
  if (bckt == bcktList) bcktList = bckt->nxt;
  else
  {
    for (b=bcktList; b != NULL && b->nxt != bckt; b = b->nxt);
    if (b == NULL || b->nxt == NULL) return;
    b->nxt = bckt->nxt;
  }
  SystemFree(bckt);
}
#if defined(UNIX) && !defined(NO_TINE_PIPE)
void ClosePipeSocket(int sck)
{
  int i;
  unsigned int usck = (unsigned int)sck;

  FD_CLR(usck,&rwset); FD_CLR(usck,&conset);
  close(sck);
  freeBucket(sck);

  for (i=0; i<nconsumer; i++) 
  { /* am I a server sending data to a client ? */
    if (Consumers[i]->sck == sck) Consumers[i]->sck = 0;
  }
  for (i=0; i<RPCNumFecsRead; i++)
  { /* am I a client trying to contact a server ? */
    if (!memcmp(&sck,RPCFec[i].Network,4))
    {
      RPCFec[i].Prot = 0;
      memset(RPCFec[i].Network,0,4);
      break;
    }
  }
}
#endif
void CloseConnectionSocket(int sck)
{
  int i;
  unsigned int usck = (unsigned int)sck;

  FD_CLR(usck,&rwset); FD_CLR(usck,&conset);
  freeBucket(sck);

  for (i=0; i<nTCPsck; i++)              /* socket on the server side ? */
  {
    if (sck == TCPsck[i])
    {
      if (sck != 0) soclose(sck);
      TCPsck[i] = 0;
      nTCPsck--;
      break;
    }
  }
  for (; i<nTCPsck; i++) TCPsck[i] = TCPsck[i+1];
  for (i=0; i<nconsumer; i++) if (Consumers[i]->sck == sck) Consumers[i]->sck = 0;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (connTable[i]->TCPsocket == sck) /* socket on the client side ? */
    {
      closeIPConnectionToFec(connTable[i]->FECindex);
      return;
    }
  }
}
#ifdef INCLUDE_LEGACY_CODE
int SendTCPData(int socket,BYTE *buf,UINT32 buflen)
{
  unsigned short msgsize = PeekWord(*(unsigned short *)buf);
  fd_set fdwset;
  struct timeval wait_to;
  int d,bytes2send,cc=0,cnt=0;
  time_t interim = time(NULL);

  if (!TCPloaded) return tcp_not_supported;
  if (msgsize >= buflen) ccerr(buffer_too_small);
  wait_to.tv_sec=0;  wait_to.tv_usec=0;
  while (cnt<msgsize)
  {
    bytes2send = MIN(msgsize-cnt,MTU);
    FD_ZERO(&fdwset); FD_SET(socket,&fdwset);
    switch (select(maxFdSets,NULL,&fdwset,NULL,&wait_to))
    {
      case 1: 
        if ((d=sowrite(socket,(char *)&buf[cnt],bytes2send)) > 0) cnt += d;
	      break;
      case 0: 
        if (NGdebug > 1) dbglog("TCP: socket %d: write set not ready",socket);
	      break;
      default:
        if (NGdebug > 1) soperror("select");
	      return tcp_socket_error;
    }
    if (time(NULL)>(interim+SERVER_TIMEOUT+msgsize/MTU)) ccerr(connection_timeout);
  }
err:if (NGdebug > 1)
  {
    sprintf(dbgbuf,"TCP: returned %d of %d bytes",cnt,msgsize);
    if (cc) sprintf(&dbgbuf[strlen(dbgbuf)],"incomplete: %s",erlst[cc]);
    dbglog(dbgbuf);
  }
  return cc;
}
#endif
UINT32 sendtoErrcnt = 0;
int SendIPData(CLN *cln,BYTE *data)
{
  fd_set fdwset;
  struct timeval wait_to;
  int sck, cc = 0, retry = 0;
  char *strprot = getProtocolAsString(cln->transport);
  unsigned short msgsize = PeekWord(*(unsigned short *)data);
  IDLE_CONNECTION *ic;

  if (!TCPloaded) return tcp_not_supported;
  if (msgsize > UDP_BUFFER_SIZE) return use_stream_transport;
  if ((sck=cln->sck) <= 0) return tcp_socket_error;
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  if (isUnixSocket(sck)) goto SendIPretry;
# endif
SelectIPretry:
  FD_ZERO(&fdwset); FD_SET(sck,&fdwset);
  wait_to.tv_sec=0;  wait_to.tv_usec=1000;
  if (select(maxFdSets,NULL,&fdwset,NULL,&wait_to) != 1)
  {
    sendtoErrcnt++;
    if (retry++ > 3) ccerr(cannot_select);
    goto SelectIPretry;
  }
SendIPretry:
  if (writeto(sck,(char *)data,msgsize,0,
             (struct sockaddr *)&cln->IPaddress,sizeof(cln->IPaddress)) < 0)
  {
    sendtoErrcnt++;
    if (retry++ > 3) ccerr(net_write_error);
    goto SendIPretry;
  }
err:
  if (NGdebug)
  {
    inet_ntoa_b(cln->IPaddress.sin_addr,clientaddr);
    if (cc)
    {
      soperror("sendto"); printf(">");
      dbglog("%s error: %s",strprot,erlst[cc]);
    }
    else if (NGdebug > 2)
    {
      dbglog("%s: %d bytes to %s [%s]",strprot,msgsize,cln->userName,clientaddr);
    }
  }
  if (cc != 0)
  {
    if (cln->transport == TCP) 
    {
      CloseConnectionSocket(sck);
      while ((ic=getIdleConnectionBySocket(sck)) != NULL)
      {
        RemoveIdleConnection(ic);
      }
    }
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    /* if (cln->transport == PIPE) ClosePipeSocket(sck); */ /* handled in SendSubBuffer() */
#   endif
  }
  return cc;
}
struct in_addr tMCastAddr;
struct in_addr tGCastAddr;
struct in_addr tNetSrvAddr;
struct sockaddr_in tMCastAddrSck;
struct sockaddr_in tGCastAddrSck;
struct sockaddr_in tNetSrvAddrSck;


/**
 * \internal
 * \brief Binds an IP socket for UDP Communication and initializes some globals.
 *
 * Only binds a socket. Does <b>not</b> listen, as the name suggests.
 * If port equals one of the reserved port numbers gtMCastPort,
 * gtGCastPort, or gtNetSrvPort, the socket is added to a multicast group.
 * On UNIX and Win32 Systems, the send and receive buffers are set according
 * to the global variables max_sndbuf and max_rcvbuf.
 * 
 * Initializes the following globals when called for the first time:
 * <ul>
 * <li>tMCastAddr</li>
 * <li>tGCastAddr</li>
 * <li>tNetSrvAddr</li>
 * <li>tGCastAddrSck</li>
 * </ul>
 *
 * \param s Socket number (output)
 * \param port Port number. If set to 0, take default TINE client port.
 *
 * \todo separate initializer for globals
 * \todo rename (e.g. bindUdpSocket) (???)
 */
int listenUDP(int *s,unsigned short port)
{
  struct sockaddr_in sck_name;
  int mode = 1,optval,optlen = sizeof(int), cc = 0;
  char txt[50];
  unsigned short udpport;
  static UINT16 nextFreePort = 0;
  static int lInitDone = FALSE;
  SINT32 ip;
# if defined(MSDOS) || defined(VXWORKS)
  int reuseFlg = TRUE;
# else
  int reuseFlg = FALSE;
# endif
# ifndef NOT_MULTICASTING
  struct ip_mreq mreq;
# endif

  if (!lInitDone) /* always fill this in at initialization time for use by server code */
  {
    lInitDone = TRUE;
#   if !defined(VXWORKS) && !defined(VAX_VMS) && !defined(NIOS) && !defined(MSDOS)
    ip = (SINT32)getmyipaddr();
    ip |= 0xff000000;
#   else
    ip = (SINT32)(0xffffffff);
#   endif
#   ifdef NOT_MULTICASTING
    memcpy(&tNetSrvAddr,&ip,sizeof(struct in_addr));  /* service port */
#   else
    ip = inet_addr(gtMCastAddr);
    memcpy(&tMCastAddr,&ip,sizeof(struct in_addr));  /* producer-subscriber */
    ip = inet_addr(gtGCastAddr);
    memcpy(&tGCastAddr,&ip,sizeof(struct in_addr));  /* producer-consumer   */
    ip = inet_addr(gtNetSrvAddr);
    memcpy(&tNetSrvAddr,&ip,sizeof(struct in_addr));  /* service port */
    tGCastAddrSck.sin_addr = tGCastAddr;
    tGCastAddrSck.sin_family = AF_INET;
    tGCastAddrSck.sin_port = htons((UINT16)gtGCastPort);
#   endif
  }
  if (s == NULL) ccerr(argument_list_error);
  if (nextFreePort == 0) nextFreePort = gtUDPClnPort;
  sprintf(txt,"UDP PORT %d: ",port);
  if ((*s=(int)socket(PF_INET,SOCK_DGRAM,0)) < 0) soerr("socket");
  if (port == gtMCastPort || port == gtGCastPort || port == gtNetSrvPort) reuseFlg = TRUE;
  if (reuseFlg)
  {
    optval = 1;
    if (setsockopt(*s,SOL_SOCKET,SO_REUSEADDR,(char *)&optval,optlen)) soerr("REUSEADDR");
  }
# if defined(UNIX) || defined(WINNT)
  if (max_rcvbuf > 0)
  {
    optval = max_rcvbuf;
    if (setsockopt(*s,SOL_SOCKET,SO_RCVBUF,(char *)&optval,optlen)) soerr("RCVBUF");
  }
  if (max_sndbuf)
  {
    optval = max_sndbuf;
    if (setsockopt(*s,SOL_SOCKET,SO_SNDBUF,(char *)&optval,optlen)) soerr("SNDBUF");
  }
# endif
  memset(&sck_name,0,sizeof(sck_name));
  sck_name.sin_addr.s_addr = 0;
  sck_name.sin_family = AF_INET;

  udpport = port ? port : nextFreePort;
  if (udpport == 0) goto portbound;
bindport:
  sck_name.sin_port = htons(udpport);
  if (bind(*s,(struct sockaddr *)&sck_name,sizeof(sck_name)))
  {
    if (port == 0 && udpport++ < gtUDPClnPort + UDP_CLN_RANGE) goto bindport;
    soerr("bind");
  }
  if (port == 0) nextFreePort = udpport + 1;
portbound:
# ifndef NOT_MULTICASTING
  if (port == gtMCastPort) /* join multicast group */
  {
    mreq.imr_multiaddr.s_addr = tMCastAddr.s_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);
    cc = setsockopt(*s,IPPROTO_IP,IP_ADD_MEMBERSHIP,(char *)&mreq,sizeof(mreq));
    feclog("joining multicast group : %s",cc ? "failed" : "success");
  }
  else if (port == gtGCastPort) /* join multicast group */
  {
    mreq.imr_multiaddr.s_addr = tGCastAddr.s_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);
    cc = setsockopt(*s,IPPROTO_IP,IP_ADD_MEMBERSHIP,(char *)&mreq,sizeof(mreq));
    feclog("joining globals multicast group : %s",cc ? "failed" : "success");
  }
  else if (port == gtNetSrvPort) /* join multicast group */
  {
    mreq.imr_multiaddr.s_addr = tNetSrvAddr.s_addr;
    mreq.imr_interface.s_addr = htonl(INADDR_ANY);
    cc = setsockopt(*s,IPPROTO_IP,IP_ADD_MEMBERSHIP,(char *)&mreq,sizeof(mreq));
    feclog("joining net services multicast group : %s",cc ? "failed" : "success");
  }
# endif
  if (ServerCycleMode == POLLING && soioctl(*s,FIONBIO,(char *)&mode)) soerr("NONBLOCKING");
  feclog("%s bound to %d (socket %d %sblocking)",
          txt,ntohs(sck_name.sin_port),*s,ServerCycleMode == POLLING?" non-":" ");
err: 
  if (cc != 0)
  {
    soperror(txt); printf(">");
    if (s != NULL)
    {
      if (*s != 0) soclose(*s);
      *s = 0;
    }
    feclog("%s socket error",txt);
  }
  return cc;
}

UINT16 GetPortOffset(RPCFecStruct *f)
{
  UINT16 poff;
  poff = (f->Port[1] << 8) | f->Port[0];
  if (poff == 0) poff = f->PortOffset;
  return poff;
}
#if !defined(VXWORKS) && !defined(VAX_VMS) && !defined(NIOS) && !defined(MSDOS)
UINT32 getmyipaddr(void)
{
  static char hname[30];
  static struct hostent *hent;
  static struct sockaddr_in stlclAddr;
  static struct sockaddr_in strmtAddr;
  static int done = 0;
  static UINT32 myip = 0;
  int sck;
  unsigned int nRet=sizeof(struct sockaddr);

  if (done) return myip;
# if defined(WINNT)
  /* do this here, just in case this is called prior to SystemInit() ! */
  if (!WinsockLoaded()) return -1;
# endif
  stlclAddr.sin_addr.s_addr = INADDR_ANY;
# if !defined(MSDOS) 
  if (gethostname(hname,30) != -1)
  {
    if ((hent=gethostbyname(hname)) != NULL)
    {
      stlclAddr.sin_addr.s_addr = *((UINT32 *)(hent->h_addr));
    }
  }
# endif
  if (stlclAddr.sin_addr.s_addr == INADDR_ANY)
  {
    if ((sck=(int)socket(AF_INET, SOCK_DGRAM, 0)) > 0)
    {
      strmtAddr.sin_family = AF_INET;
      strmtAddr.sin_port = htons(IPPORT_ECHO);
      strmtAddr.sin_addr.s_addr = inet_addr("128.127.50.1");
      if (connect(sck,(struct sockaddr *)&strmtAddr,sizeof(struct sockaddr)) != -1)
      {
        getsockname(sck,(struct sockaddr *)&stlclAddr,(int *)&nRet);
      }
      soclose(sck);
    }
  }
  myip = stlclAddr.sin_addr.s_addr;
  done = TRUE;
  return stlclAddr.sin_addr.s_addr;  
}
#endif
/* trivial RPC call: */
int tUDPCall(int fecindex,BYTE *buf,int timout)
{
  fd_set tset;
  struct timeval t_to;
  static struct sockaddr_in IPfec;
  static CLN cln;
  static UINT16 ltid = 0xffff;
  UINT16 tid;
  int cc,n,poff;
  unsigned int len;
  char s[64];
  BYTE lclbuf[IPX_BUFFER_SIZE];
  RPCFecStruct *rpcfec = fecindex ? &RPCFec[fecindex] : &NameServer[curNameServer];

  if (SyncSocket == 0)
  {
    if (listenUDP(&SyncSocket,0) != 0) 
    {
      soperror("Listen SyncSocket");
      return udp_socket_error;
    }
  }
  /* flush any input bytes that are somehow there ... */
  FD_ZERO(&tset); FD_SET(SyncSocket,&tset);
  t_to.tv_sec=0; t_to.tv_usec=0;
  while (select(32,&tset,NULL,NULL,&t_to) > 0)
  {
    len = sizeof(IPfec);
    recvfrom(SyncSocket,(char *)lclbuf,IPX_BUFFER_SIZE,0,(struct sockaddr *)&IPfec,(int *)&len);
  }
  memset(&cln,0,sizeof(CLN));
  strcpy(cln.userName,rpcfec->Name);
  if (!rpcfec->IPh_addr) return non_existent_fec;
  memset(&prod_name,0,sizeof(prod_name));
  memcpy(&prod_name.sin_addr,&rpcfec->IPh_addr,sizeof(struct in_addr));
  poff = GetPortOffset(rpcfec);
  prod_name.sin_family=AF_INET;
  prod_name.sin_port = htons((UINT16)(gtUDPPort + poff));
  memcpy(&cln.IPaddress,&prod_name,sizeof(prod_name));
  cln.sck = SyncSocket;

  if ((cc=SendIPData(&cln,buf)) != 0) return cc;

lsnForDataGram:
  FD_ZERO(&tset); FD_SET(SyncSocket,&tset);
  t_to.tv_sec=timout/1000;
# ifndef MSDOS
  t_to.tv_usec=((SINT32)timout%1000)*(SINT32)1000;
# else
  t_to.tv_usec=((SINT32)timout%1000);
# endif
  len = sizeof(IPfec);
  switch (select(maxFdSets,&tset,NULL,NULL,&t_to))
  {
    case -1: 
      if (NGdebug) 
      {
        sprintf(s,"SyncSocket (%d) select (to : %d)",SyncSocket,timout);
        soperror(s);
      }
      return tcp_socket_error;
    case 0:  
      return link_not_open;
    default: 
      n = recvfrom(SyncSocket,(char *)buf,IPX_BUFFER_SIZE,0,(struct sockaddr *)&IPfec,(int *)&len);
  }
  if (n != PeekWord(*(short *)buf)) return remitted_data_lost;
  if ((cc=PeekWord(*(short *)&buf[6])) == illegal_protocol) return cc;
  if ((tid=PeekWord(*(UINT16 *)&buf[4])) == ltid) /* phantom block ! */
  {
    goto lsnForDataGram;
  }
  ltid = tid;
  return 0;
}
#ifndef RDATE_TIMEOUT
# define RDATE_TIMEOUT 1
#endif
#define IPPORT_TIMESERVER	37              /* well known internet port number for rdate	*/
#define SEVENTY_YEARS		2208988800UL     /* midnight of 1.1.1900 to midnight of 1.1.1970	*/
char TimeServerHost[64];
#ifdef VXWORKS
struct hostent *gethostbyname(const char *hstname)
{
  static struct hostent hst;
  static char addr[4];
  int haddr;

  if ((haddr = hostGetByName((char *)hstname)) == ERROR) return NULL;

  hst.h_name = (char *)hstname;
  hst.h_aliases = 0;
  hst.h_addrtype = AF_INET;
  hst.h_length = strlen(hstname);
  memcpy(addr,&haddr,4);
  hst.h_addr = addr;

  return &hst;
}
struct hostent *gethostbyaddr(const char *hstaddr, int hstlen, int hsttype)
{
  static struct hostent hst;
  static char haddr[4];
  static char *addr; 
  char *c, hstaddrBuffer[32];
  int i = 0;

  strncpy(hstaddrBuffer,hstaddr,16); hstaddrBuffer[16] = 0;
  hst.h_name = (char *)hstaddr;
  hst.h_aliases = 0;
  hst.h_addrtype = hsttype;
  hst.h_length = hstlen;

  for (c=strtok(hstaddrBuffer,"."); c != NULL; c=strtok((char *)NULL,".")) 
  {
    haddr[i++] = atoi(c);
  }
  addr = (char *) &(haddr[0]);
  hst.h_addr = addr;

  return &hst;
}
#endif
time_t getUnixTime(char *tsHost)
{
  int sockfd=-1;																				/* socket fd's */
  static struct sockaddr_in  serverAddr;
  UINT32 tmptime;
  static struct hostent *host = NULL;
  static int firstCallLogged = 0;
  int cc = 0,mode = 1,done = 0;
  time_t interim;

  if (firstCallLogged == 0)
  {
    firstCallLogged++;
    feclog("using TIME Server %s", tsHost == NULL ? "local host" : tsHost);
  }
  if (tsHost == NULL) return time(NULL);
  if (host == NULL)
  {
    memset(&serverAddr,0,sizeof(serverAddr));
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_port = htons(IPPORT_TIMESERVER);
    if ((host=gethostbyname(tsHost)) == NULL &&
        (host=gethostbyaddr(tsHost,(int)strlen(tsHost),AF_INET)) == NULL)
    {
      if (NGdebug) soperror("time svr gethost");
      ccerr(non_existent);
    }
    memcpy(&serverAddr.sin_addr,host->h_addr,sizeof(struct in_addr));
  }

  if ((sockfd=(int)socket(AF_INET,SOCK_STREAM,0)) < 0 )
  {
    if (NGdebug) soperror("time svr socket");
    ccerr(tcp_socket_error);
  }
  if (soioctl(sockfd,FIONBIO,(char *)&mode))
  {
    if (NGdebug) soperror("time svr soioctl");
    ccerr(tcp_socket_error);
  }
  for (time(&interim);
       (connect(sockfd,(struct sockaddr *)&serverAddr,sizeof(serverAddr))<0);)
  {
    if (soerrno == EISCONN) break;
    if (time(NULL) > interim + RDATE_TIMEOUT)
    {
      if (NGdebug) soperror("time svr connect");
      ccerr(tcp_connect_error);
    }
  }

  while (time(NULL) < interim + RDATE_TIMEOUT && !done)
  {
    if (soread(sockfd,(char *)&tmptime,sizeof(UINT32)) == sizeof(UINT32) )
      done = 1;
  }
  if (!done)
  {
    if (NGdebug > 1) soperror("soread");
    ccerr(connection_timeout);
  }
  if (firstCallLogged == 1)
  {
    firstCallLogged++;
    feclog("got time from %s",tsHost);
  }

err:
  if (sockfd > 0) soclose(sockfd);
  if (cc)
  {
    if (firstCallLogged < 3)
    {
      firstCallLogged++;
      feclog("get time from %s : %s",tsHost,erlst[cc]);
    }
    return 0;
  }
  return (ntohl(tmptime) - SEVENTY_YEARS);  /* convert to unix time	*/
}
