#include "cprolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stddef.h>
#include <malloc.h>
#include "servdef.h"
#include "database.h"
#include "nioscmd.h"

// command relevant entries:
void closeClients(void);
void ExitEQPModules(void);
int FecSettime(void);

struct CommandListStruct {
  char cmd[10];
  int (*fcn)(char *,char *);
};
struct CommandListStruct CommandList[] = {
  { "quit",quit},
  { "exit",quit},
  { "help",help},
  { "set",setVal},
  { "get",getVal}
};
int NumCommands = sizeof(CommandList)/sizeof(struct CommandListStruct);
int CommandLevel = 0;
char cmdbuf[CMDSIZE];
char strbuf[CMDSIZE];

USRCMD *usrcmd = NULL;
int numUsrCmd = 0;

USRFCN *usrfcn = NULL;
int numUsrFcn = 0;

int RegisterUserCommand(
    char *cmd,
    int (*fcn)(int,int,int,int),
    int *iparam,
    float *fparam,
    int access)
{
  USRCMD *uc;

  if ((uc=(USRCMD *)calloc(1,sizeof(USRCMD))) == NULL) return out_of_local_memory;
  if (access == 0) access = CA_READ;
  strncpy(uc->cmd,cmd,32);
  uc->fcn = fcn;
  uc->iparam = iparam;
  uc->fparam = fparam;
  uc->access = access;
  uc->next = usrcmd;
  usrcmd = uc;
  numUsrCmd++;
  return 0;
}
int RegisterUserFunction(
    char *name,
    int (*fcn)(int argc,char *argv[]))
{
  USRFCN *uf;

  if ((uf=(USRFCN *)SystemCalloc(1,sizeof(USRFCN))) == NULL) return out_of_local_memory;
  strncpy(uf->name,name,32);
  uf->fcn = fcn;
  uf->next = usrfcn;
  usrfcn = uf;
  numUsrFcn++;
  return 0;
}

int InterpretCommand(char *buf)
{
  char cmd[CMDSIZE];
  char *c,*ce=NULL,*cp=NULL,*arg1=NULL,*arg2=NULL;
  int cc=0,i,len;
  int argc;
  char *argv[10],vstr[32],sstr[128];
  USRFCN *uf;

  if (strlen(buf) <= 1) ccerr(0);
  strncpy(cmd,buf,CMDSIZE-1); cmd[CMDSIZE-1] = 0;
  for (i=0; i<NumCommands; i++)
   if (!strncmp(cmd,CommandList[i].cmd,strlen(CommandList[i].cmd))) break;

  if (i < NumCommands) 
  {
    if ((c=strchr(cmd,'=')) != NULL)
    {
      arg2 = c+1;
      *c = 0;
    }
    arg1 = &cmd[strlen(CommandList[i].cmd)+1];
    ccerr(CommandList[i].fcn(arg1,arg2));
  }
  for (i=0; i<32 && cmd[i] == ' '; i++);
  strncpy(vstr,&cmd[i],32);
  if ((c=strchr(vstr,'(')) != NULL) *c++ = 0;
  for (uf=usrfcn; uf != NULL; uf=uf->next)
  {
    if (uf->name != NULL && !strncmp(vstr,uf->name,32))
    {
      argc = 0;
      len = strlen(uf->name);
      if (c != NULL && (ce=strchr(c,')')) != NULL)
      {
        *ce = 0; strncpy(sstr,c,62); sstr[63] = 0;
        for (i=0,c=strtok_r(sstr,",",&cp); c && i<10; i++,c=strtok_r(NULL,",",&cp))
        {
          argv[i] = c;
        }
        argc = i;
      }
      i = uf->fcn(argc,argv);
      sprintf(strbuf,"result is %d",i);
      ttyoutput(strbuf);
      ccerr(0);  
    }
  }
err: 
  if (cc) ttyoutput("\n>Syntax: set name = value or get name");
  ttyoutput("\n>");
  return cc;
}
int putch(int c)
{
# ifdef NIOS_I
  nr_txchar(c);
# endif
  return c;
}
int getch(void)
{
# ifdef NIOS_I
  return nr_rxchar();
# else
  return 0;
# endif
}
void GetCommand(void)
{
  static char prvcmd[CMDSIZE];
  static int len = 0, prvlen = 0, clen = 0;
  int c;

  switch (c=getch())
  {
    case 27:
    case 0:
      if ((c=getch()) != '[') break;
    //case '[':
    //case 0: // extended (cursor movement)
      switch (c=getch())
      {
        case 'C': // right
                  if (clen > len)
                  {
                    putch(cmdbuf[len++]);
                  }
                  break;
        case 'D': // left
                  if (len > 0)
                  {
                    len--;
                  }
                  break;
        case 'A': // up
                  strncpy(cmdbuf,prvcmd,len=prvlen);
                  cmdbuf[len] = 0;
                  printf(cmdbuf);
                  break;
        case 'B': // down
                  if (clen > len)
                  {
                    cmdbuf[clen] = 0;
                    printf(&cmdbuf[len]);
                    len = clen;
                  }
                  break;
      }
      break;
    case 9: // tab
      break;
    case 8: // backspace
      if (len > 0)
      {
        len--;
      }
      break;
    case 10:
    case 13:
      ttyoutput("\n>");
      cmdbuf[len] = 0;
      InterpretCommand(cmdbuf);
      CMDcount++;
      strncpy(prvcmd,cmdbuf,len);
      prvlen = len;
      len = 0;
      break;
    default:
      if (c < 0) return;
      putch(c);
      cmdbuf[len++] = c;
      if (len > CMDSIZE-2) len = 0;
      clen = MAX(clen,len);
      break;
  }
  return;
}
int help(char *s1,char *s2)
{
  USRCMD *uc;
  char sstr[64];

  s1 = s1; s2 = s2;
  ttyoutput("Currently available commands:\n>");
  ttyoutput("\tquit            - terminates program gracefully\n>");
  if (NrOfEqpTypes)
  {
    ttyoutput("\tget clients          - displays the current consumer list\n>");
    ttyoutput("\tget contracts        - displays the current contract list\n>");
    ttyoutput("\tget contract = <#>   - displays contract Nr. <#> information\n>");
  }
  if (nglobals)
  {
    ttyoutput("\tget globals     - displays the current globals list\n>");
  }
  if (nConnectionTableEntries)
  {
    ttyoutput("\tget connections      - displays the current connection list\n>");
    ttyoutput("\tget connection = <#> - displays connection Nr. <#> information\n>");
  }
  ttyoutput("\tget BurstLimit       - displays Burst Limit in packets\n>");
  ttyoutput("\tset BurstLimit       - sets Burst Limit to num packets input\n>");
  ttyoutput("\tget CycleDelay       - displays Cycle Delay in msec\n>");
  ttyoutput("\tset CycleDelay       - sets Cycle Delay to msec input\n>");
  ttyoutput("\tget memory           - displays available memory\n>");
  ttyoutput("\tget time             - displays local time\n>");
  ttyoutput("\tget version          - displays  TINE version number\n>");
  ttyoutput("\tget users            - displays the users with WRITE permission\n>");
  ttyoutput("\tget nets             - displays the networks with WRITE access\n>");
  ttyoutput("\tget stats            - displays operational statistics\n>");
  ttyoutput("\tget filter           - displays current debug output filter string\n>");
  ttyoutput("\tset filter           - sets debug output filter string\n>");
  ttyoutput("\tset debug = 0        - turns debug printing off\n>");
  ttyoutput("\tset debug = 1        - sets debug level 1 (trace RPC commands)\n>");
  ttyoutput("\tset debug = 2        - sets debug level 2 (trace network activity)\n>");
  ttyoutput("\tset debug = 3        - sets debug level 2 (trace data exchange)\n>");
  ttyoutput("\tset debug = 4        - sets debug level 3 (full diagnostic trace )\n>");
  ttyoutput("\tset logdbg = 0       - turns debug logging off\n>");
  ttyoutput("\tset logdbg = 1       - turns debug logging on\n>");
  ttyoutput("\thelp                 - display this list\n>");
  ttyoutput("\n>Extra commands:\n>");
  for (uc=usrcmd; uc != NULL; uc=uc->next)
  {
    sstr[0] = 0;
    if (uc->access & CA_READ) strcpy(sstr,"get ");
    if (uc->access & CA_WRITE)
    {
      if (strlen(sstr)) strcat(sstr," or set ");
      else strcpy(sstr,"set ");
    }
    if (uc->iparam != NULL) strcat(sstr,"integer value");
    else if (uc->fparam != NULL) strcat(sstr,"float value");
    else if (uc->fcn != NULL) strcat(sstr,"function value");
    sprintf(strbuf,"\t%s\t - %s\n>",uc->cmd,sstr);
    ttyoutput(strbuf);
  }
  return 0;
}
int quit(char *s1,char *s2)
{
  s1 = s1; s2 = s2;
  ExitEQPModules();
  quitServer();                   /* non-servers must cover this routine */
  if (nConnectionTableEntries) closeClients();      /* non-clients must cover this routine */
  feclog("stopped");
  exit(0);
  return 0;
}
int getVal(char *var,char *s)
{
  USRCMD *uc;
  char *c,*cp=NULL,sstr[128],vstr[32];
  int i,len,ival[4];

  s = s;
  if (var == NULL) return -1;
  for (i=0; i<32 && var[i] == ' '; i++);
  strncpy(vstr,&var[i],32);
  if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
  if ((c=strchr(vstr,' ')) != NULL) *c = 0;
  if ((c=strchr(vstr,'(')) != NULL) *c = 0;

  for (uc=usrcmd; uc != NULL; uc=uc->next)
  {
    if (!strncmp(vstr,uc->cmd,32) && uc->access & CA_READ)
    {
      if (uc->iparam != NULL)
      {
        sprintf(strbuf,"%s is %d",uc->cmd,*uc->iparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fparam != NULL) 
      {
        sprintf(strbuf,"%s is %g",uc->cmd,*uc->fparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fcn != NULL) 
      {
        len = strlen(uc->cmd);
        memset(ival,0,sizeof(ival));
        if (var[len] == '(' && (c=strchr(var,')')) != NULL)
        {
          *c = 0; strncpy(sstr,&var[len+1],62); sstr[63] = 0;
          for (i=0,c=strtok_r(sstr,",",&cp); i<4 && c; i++,c=strtok_r(NULL,",",&cp))
          {
            ival[i] = atoi(c);
          }
        }
        i = uc->fcn(ival[0],ival[1],ival[2],ival[3]);
        sprintf(strbuf,"result is %d",i);
        ttyoutput(strbuf);
        return 0;
      }
    }
  }
  if (strstr(var,"version")) return dumpVersion();
  if (strstr(var,"debug"))
  {
    sprintf(strbuf,"Debug level %d",NGdebug);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"time")) return dumpServerTime();
  if (strstr(var,"users")) { dumpRegisteredUsers(); return 0; }
  if (strstr(var,"nets"))
  {
    dumpIPNets();
    return 0;
  }
  if (strstr(var,"globals")) { dumpGlobals(); return 0; }
  if (strstr(var,"memory"))
  {
    return 0;
  }
  if (strstr(var,"CommandLevel"))
  {
     sprintf(strbuf,"Command Level %d entered",CommandLevel); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"filter"))
  {
    sprintf(strbuf,"debug name filter : %s",tagNameFilter);
    ttyoutput(strbuf);
    return 0;
  }
  if (NrOfEqpTypes)
  {
    if (strstr(var,"clients")) return dumpConsumerList(); 
    if (strstr(var,"contract") && s != NULL)
    {
      i = atoi(s);
      dumpContractStats(i);
      return 0;
    }
    if (strstr(var,"contracts")) return dumpContractList();
  }
  if (nConnectionTableEntries)
  {
    if (strstr(var,"connection") && s != NULL)
    {
      i = atoi(s);
      dumpConnectionStats(i);      /* non-clients must cover this routine */
      return 0;
    }
    if (strstr(var,"connections"))
    {
      dumpConnectionList();      /* non-clients must cover this routine */
      ttyoutput("\n");
      return 0;
    }
  }
  if (strstr(var,"BurstLimit"))
  {
    sprintf(strbuf,"Burst Limit set at %d",gBurstLimit);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"CycleDelay"))
  {
    sprintf(strbuf,"Cycle Delay set at %d",gCycleDelay);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"stats")) return dumpServerStats();
  sprintf(strbuf,"Variable unavailable: %s",var); ttyoutput(strbuf);
  return -1;
}
int setVal(char *var,char *val)
{
  int v;
  USRCMD *uc;
  char *c,*cp=NULL,sstr[64],vstr[32];
  int i,len,ival[4];

  if (var == NULL || val == NULL) return -1;
  for (i=0; i<32 && var[i] == ' '; i++);
  strncpy(vstr,&var[i],32);
  if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
  if ((c=strchr(vstr,' ')) != NULL) *c = 0;
  if ((c=strchr(vstr,'(')) != NULL) *c = 0;

  for (uc=usrcmd; uc != NULL; uc=uc->next)
  {
    if (!strncmp(vstr,uc->cmd,32) && uc->access & CA_WRITE)
    {
      if (uc->iparam != NULL) 
      {
        *uc->iparam = atoi(val);
        sprintf(strbuf,"%s set to %d",uc->cmd,*uc->iparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fparam != NULL) 
      {
        *uc->fparam = (float)strtod(val,NULL);
        sprintf(strbuf,"%s is %g",uc->cmd,*uc->fparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fcn != NULL) 
      {
        len = strlen(uc->cmd);
        memset(ival,0,sizeof(ival));
        if (var[len] == '(' && (c=strchr(var,')')) != NULL)
        {
          *c = 0; strncpy(sstr,&var[len+1],62); sstr[63] = 0;
          for (i=0,c=strtok_r(sstr,",",&cp); i<4 && c; i++,c=strtok_r(NULL,",",&cp))
          {
            ival[i] = atoi(c);
          }
          if (i<4) ival[i] = atoi(val);
        }
        i = uc->fcn(ival[0],ival[1],ival[2],ival[3]);
        sprintf(strbuf,"result is %d",i);
        ttyoutput(strbuf);
        return 0;
      }
    }
  }
  if (strstr(var,"time"))
  {
     FecSettime();
     return 0;
  }
  if (strstr(var,"BurstLimit"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     v = SetBurstLimit(v);
     sprintf(strbuf,"Burst Limit set to %d",gBurstLimit); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"CycleDelay"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     v = SetCycleDelay(v);
     sprintf(strbuf,"Cycle Delay set to %d",gCycleDelay); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"debug"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     NGdebug = v;
     sprintf(strbuf,"Debug level %d",NGdebug); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"logdbg"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     if (v > 0)
     {
       if ((dbgfp=fopen("debug.log","w")) == NULL)
         sprintf(strbuf,"Couldn't open debug log file");
       else sprintf(strbuf,"debug logging ON");
       ttyoutput(strbuf);
     }
     else
     {
       if (dbgfp) fclose(dbgfp);
       dbgfp = NULL;
       sprintf(strbuf,"debug logging OFF");
       ttyoutput(strbuf);
     }
     return 0;
  }
  if (strstr(var,"almdbg"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     ALMdebug = v;
     sprintf(strbuf,"Alarm Debug level %d",ALMdebug); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"filter"))
  {
     memset(tagNameFilter,0,64);
     if ((c=strchr(val,'\n')) != NULL) *c = 0;
     if (strlen(val) > 0) strncpy(tagNameFilter,val,63);
     if (NGdebug)
     {
       sprintf(strbuf,"debug name filter %s entered",tagNameFilter);
       ttyoutput(strbuf);
     }
     return 0;
  }
  if (strstr(var,"CommandLevel"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     CommandLevel = v;
     sprintf(strbuf,"Command Level %d entered",CommandLevel); ttyoutput(strbuf);
     return 0;
  }
  sprintf(strbuf,"Variable %s unavailable",var); ttyoutput(strbuf);
  return -1;
}
