/** \file ntcmd.c
 * \brief TINE console command line API
 *
 * The API routines offered here allow the developer to register his own command-line 
 * calling routines to be used for interacting with a running server at the command line.
 *
 */

#ifndef NO_DOXYGEN
#include "cprolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stddef.h>
#include <conio.h>
#include <malloc.h>
#include <windows.h>
#include "servdef.h"
#include "database.h"
#include "ntcmd.h"

extern int ALMdebug;

/* command relevant entries: */
void closeClients(void);
void ExitEQPModules(void);

struct CommandListStruct {
  char cmd[10];
  int (*fcn)(char *,char *);
};
struct CommandListStruct CommandList[] = {
  { "quit",quit},
  { "exit",quit},
  { "help",help},
  { "set",setVal},
  { "get",getVal}
};
int NumCommands = sizeof(CommandList)/sizeof(struct CommandListStruct);
int CommandLevel = 0;

USRCMD *usrcmd = NULL;
int numUsrCmd = 0;

USRFCN *usrfcn = NULL;
int numUsrFcn = 0;

int InterpretCommand(char *buf)
{
  char cmd[CMDSIZE],strbuf[CMDSIZE];
  char *c,*ce=NULL,*cp=NULL,*arg1=NULL, *arg2=NULL;
  int cc=0,i,len;
  int argc;
  char *argv[10],vstr[32],sstr[128];
  USRFCN *uf;

  if (strlen(buf) <= 1) ccerr(0);
  strncpy(cmd,buf,CMDSIZE-1); cmd[CMDSIZE-1] = 0;
  for (i=0; i<NumCommands; i++)
   if (!strncmp(cmd,CommandList[i].cmd,strlen(CommandList[i].cmd))) break;

  if (i < NumCommands) 
  {
    if ((c=strchr(cmd,'=')) != NULL)
    {
      arg2 = c+1;
      *c = 0;
    }
    arg1 = &cmd[strlen(CommandList[i].cmd)+1];
    ccerr(CommandList[i].fcn(arg1,arg2));
  }
  for (i=0; i<32 && cmd[i] == ' '; i++);
  strncpy(vstr,&cmd[i],32);
  if ((c=strchr(vstr,'(')) != NULL) *c++ = 0;
  for (uf=usrfcn; uf != NULL; uf=uf->next)
  {
    if (uf->name != NULL && !strncmp(vstr,uf->name,32))
    {
      argc = 0;
      len = (int)strlen(uf->name);
      if (c != NULL && (ce=strchr(c,')')) != NULL)
      {
        *ce = 0; strncpy(sstr,c,62); sstr[63] = 0;
        for (i=0,c=strtok_r(sstr,",",&cp); c && i<10; i++,c=strtok_r(NULL,",",&cp))
        {
          argv[i] = c;
        }
        argc = i;
      }
      i = uf->fcn(argc,argv);
      sprintf(strbuf,"result is %d",i);
      ttyoutput(strbuf);
      ccerr(0);  
    }
  }
err: 
  if (cc) ttyoutput("\n>Syntax: set name = value or get name");
  ttyoutput("\n>");
  return cc;
}
void GetCommand(void)
{
  static char prvcmd[CMDSIZE],cmdbuf[CMDSIZE];
  static int len = 0, prvlen = 0, clen = 0;
  int i;
  char c;

  if (kbhit())
  {
    switch (c=getch())
    {
      case -32: /* extended (cursor movement) */
        switch (c=getch())
        {
          case 'M': /* right */
                    if (clen > len)
                    {
                      putch(cmdbuf[len++]);
                    }
                    break;
          case 'S': /* Delete */
          case 'K': /* left */
                    if (len > 0)
                    {
                      cmdbuf[len--] = 0;
                      putch('\r'); putch('>'); 
                      for (i=0; i<len; i++) putch(cmdbuf[i]);
                    }
                    break;
          case 'H': /* up */
                    strncpy(cmdbuf,prvcmd,len=prvlen);
                    cmdbuf[len] = 0;
                    putch('\r'); putch('>'); 
                    printf(cmdbuf);
                    break;
          case 'P': /* down */
                    if (clen > len)
                    {
                      cmdbuf[clen] = 0;
                      putch('\r'); putch('>'); 
                      printf(&cmdbuf[len]);
                      len = clen;
                    }
                    break;
        }

        break;
      case 9: /* tab */
        break;
      case 8: /* backspace */
        if (len > 0)
        {
          cmdbuf[len--] = 0;
          putch('\r'); putch('>'); 
          for (i=0; i<len; i++) putch(cmdbuf[i]);
        }
        break;
      case 10:
      case 13:
        ttyoutput("\n>");
        cmdbuf[len] = 0;
        InterpretCommand(cmdbuf);
        CMDcount++;
        strncpy(prvcmd,cmdbuf,len);
        prvlen = len;
        len = 0;
        break;
      default:
        putch(c);
        cmdbuf[len++] = c;
        if (len > CMDSIZE-2) len = 0;
        clen = MAX(clen,len);
        break;
    }
  }
  return;
}
int help(char *s1,char *s2)
{
  USRCMD *uc = NULL;
  USRFCN *uf = NULL;
  char sstr[64],strbuf[CMDSIZE];

  ttyoutput("Currently available commands:\n>");
  ttyoutput("\tquit                    - terminates program gracefully\n>");
  if (NrOfEqpTypes)
  {
    ttyoutput("\tget clients             - displays the current consumer list\n>");
    ttyoutput("\tget contracts           - displays the current contract list\n>");
    ttyoutput("\tget contract = <#>      - displays contract Nr. <#> information\n>");
  }
  if (nglobals)
  {
    ttyoutput("\tget globals             - displays the current globals list\n>");
  }
  if (nConnectionTableEntries)
  {
    ttyoutput("\tget connections - displays the current connection list\n>");
    ttyoutput("\tget connection = <#>     - displays connection Nr. <#> information\n>");
  }
  ttyoutput("\tget BurstLimit           - displays Burst Limit in packets\n>");
  ttyoutput("\tset BurstLimit           - sets Burst Limit to num packets input\n>");
  ttyoutput("\tget CycleDelay           - displays Cycle Delay in msec\n>");
  ttyoutput("\tset CycleDelay           - sets Cycle Delay to msec input\n>");
  ttyoutput("\tget time                 - displays local time\n>");
  ttyoutput("\tget version              - displays TINE version number\n>");
  ttyoutput("\tget users                - displays WRITE access user list\n>");
  ttyoutput("\tget nets                 - displays WRITE access net list\n>");
  ttyoutput("\tget stats                - displays operational statistics\n>");
  ttyoutput("\tget users                - displays the users with WRITE permission\n>");
  ttyoutput("\tget nets                 - displays the networks with WRITE access\n>");
  ttyoutput("\tset debug = 0            - turns debug printing off\n>");
  ttyoutput("\tset debug = 1            - sets debug level 1 (trace RPC commands)\n>");
  ttyoutput("\tset debug = 2            - sets debug level 2 (trace network activity)\n>");
  ttyoutput("\tset debug = 3            - sets debug level 2 (trace data exchange)\n>");
  ttyoutput("\tset debug = 4            - sets debug level 3 (full diagnostic trace )\n>");
  ttyoutput("\tset logdbg = 0           - turns debug logging off\n>");
  ttyoutput("\tset logdbg = 1           - turns debug logging on\n>");
  ttyoutput("\thelp                     - display this list\n>");
  ttyoutput("\n>Extra commands:\n>");
  for (uc=usrcmd; uc != NULL; uc=uc->next)
  {
    sstr[0] = 0;
    if (uc->access & CA_READ) strcpy(sstr,"get ");
    if (uc->access & CA_WRITE)
    {
      if (strlen(sstr)) strcat(sstr," or set ");
      else strcpy(sstr,"set ");
    }
    if (uc->iparam != NULL) strcat(sstr,"integer value");
    else if (uc->fparam != NULL) strcat(sstr,"float value");
    else if (uc->fcn != NULL) strcat(sstr,"function value");
    sprintf(strbuf,"\t%s\t - %s\n>",uc->cmd,sstr);
    ttyoutput(strbuf);
  }
  for (uf=usrfcn; uf != NULL; uf=uf->next)
  {
    sstr[0] = 0;
    sprintf(strbuf,"\t%s()\n>",uf->name);
    ttyoutput(strbuf);
  }
  return 0;
}
int quit(char *s1,char *s2)
{
  ExitEQPModules();
  quitServer();                   /* non-servers must cover this routine */
  if (nConnectionTableEntries) closeClients();      /* non-clients must cover this routine */
  feclog("stopped");
  exit(0);
  return 0;
}
int getVal(char *var,char *s)
{
  USRCMD *uc;
  char *c,*cp=NULL,sstr[128],vstr[32],strbuf[CMDSIZE];
  int i,len,ival[4];

  if (var == NULL) return -1;
  for (i=0; i<32 && var[i] == ' '; i++);
  strncpy(vstr,&var[i],32);
  if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
  if ((c=strchr(vstr,' ')) != NULL) *c = 0;
  if ((c=strchr(vstr,'(')) != NULL) *c = 0;

  for (uc=usrcmd; uc != NULL; uc=uc->next)
  {
    if (!strncmp(vstr,uc->cmd,32) && uc->access & CA_READ)
    {
      if (uc->iparam != NULL)
      {
        sprintf(strbuf,"%s is %d",uc->cmd,*uc->iparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fparam != NULL) 
      {
        sprintf(strbuf,"%s is %g",uc->cmd,*uc->fparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fcn != NULL) 
      {
        len = (int)strlen(uc->cmd);
        memset(ival,0,sizeof(ival));
        if (var[len] == '(' && (c=strchr(var,')')) != NULL)
        {
          *c = 0; strncpy(sstr,&var[len+1],62); sstr[63] = 0;
          for (i=0,c=strtok_r(sstr,",",&cp); i<4 && c; i++,c=strtok_r(NULL,",",&cp))
          {
            ival[i] = atoi(c);
          }
        }
        i = uc->fcn(ival[0],ival[1],ival[2],ival[3]);
        sprintf(strbuf,"result is %d",i);
        ttyoutput(strbuf);
        return 0;  
      }
    }
  }
  if (strstr(var,"version")) return dumpVersion();
  if (strstr(var,"semaphores")) return dumpSemaphores();
  if (strstr(var,"debug"))
  {
    sprintf(strbuf,"Debug level %d",NGdebug);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"time")) return dumpServerTime();
  if (strstr(var,"users")) { dumpRegisteredUsers(); return 0; }
  if (strstr(var,"nets")) { dumpIPNets(); return 0; }
  if (strstr(var,"globals")) { dumpGlobals(); return 0; }
  if (strstr(var,"CommandLevel"))
  {
     sprintf(strbuf,"Command Level %d entered",CommandLevel); ttyoutput(strbuf);
     return 0;
  }
  if (NrOfEqpTypes)
  {
    if (strstr(var,"clients")) return dumpConsumerList(); 
    if (strstr(var,"contract") && s != NULL)
    {
      i = atoi(s);
      dumpContractStats(i);
      return 0;
    }
    if (strstr(var,"contracts")) return dumpContractList();
  }
  if (nConnectionTableEntries)
  {
    if (strstr(var,"connection") && s != NULL)
    {
      i = atoi(s);
      dumpConnectionStats(i);      /* non-clients must cover this routine */
      return 0;
    }
    if (strstr(var,"connections"))
    {
      dumpConnectionList();      /* non-clients must cover this routine */
      ttyoutput("\n");
      return 0;
    }
  }
  if (strstr(var,"BurstLimit"))
  {
    sprintf(strbuf,"Burst Limit set at %d",gBurstLimit);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"CycleDelay"))
  {
    sprintf(strbuf,"Cycle Delay set at %d",gCycleDelay);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"stats")) return dumpServerStats();
  sprintf(strbuf,"Variable unavailable: %s",var); ttyoutput(strbuf);
  return -1;
}
int setVal(char *var,char *val)
{
  USRCMD *uc = NULL;
  char *c,*cp=NULL,sstr[64],vstr[32],strbuf[CMDSIZE];
  int i,len,ival[4];
  int v;

  if (var == NULL || val == NULL) return -1;
  for (i=0; i<32 && var[i] == ' '; i++);
  strncpy(vstr,&var[i],32);
  if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
  if ((c=strchr(vstr,' ')) != NULL) *c = 0;
  if ((c=strchr(vstr,'(')) != NULL) *c = 0;

  for (uc=usrcmd; uc != NULL; uc=uc->next)
  {
    if (!strncmp(vstr,uc->cmd,32) && uc->access & CA_WRITE)
    {
      if (uc->iparam != NULL) 
      {
        *uc->iparam = atoi(val);
        sprintf(strbuf,"%s set to %d",uc->cmd,*uc->iparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fparam != NULL) 
      {
        *uc->fparam = (float)strtod(val,NULL);
        sprintf(strbuf,"%s is %g",uc->cmd,*uc->fparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fcn != NULL) 
      {
        len = (int)strlen(uc->cmd);
        memset(ival,0,sizeof(ival));
        if (var[len] == '(' && (c=strchr(var,')')) != NULL)
        {
          *c = 0; strncpy(sstr,&var[len+1],62); sstr[63] = 0;
          for (i=0,c=strtok_r(sstr,",",&cp); i<4 && c; i++,c=strtok_r(NULL,",",&cp))
          {
            ival[i] = atoi(c);
          }
          if (i<4) ival[i] = atoi(val);
        }
        i = uc->fcn(ival[0],ival[1],ival[2],ival[3]);
        sprintf(strbuf,"result is %d",i);
        ttyoutput(strbuf);
        return 0;
      }
    }
  }
  if (strstr(var,"BurstLimit"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     v = SetBurstLimit(v);
     sprintf(strbuf,"Burst Limit set to %d",gBurstLimit); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"CycleDelay"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     v = SetCycleDelay(v);
     sprintf(strbuf,"Cycle Delay set to %d",gCycleDelay); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"debug"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     NGdebug = v;
     sprintf(strbuf,"Debug level %d",NGdebug); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"logdbg"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     if (v > 0)
     {
       if ((dbgfp=fopen("debug.log","w")) == NULL)
         sprintf(strbuf,"Couldn't open debug log file");
       else sprintf(strbuf,"debug logging ON");
       ttyoutput(strbuf);
     }
     else
     {
       if (dbgfp) fclose(dbgfp);
       dbgfp = NULL;
       sprintf(strbuf,"debug logging OFF");
       ttyoutput(strbuf);
     }
     return 0;
  }
  if (strstr(var,"almdbg"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     ALMdebug = v;
     sprintf(strbuf,"Alarm Debug level %d",ALMdebug); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"CommandLevel"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     CommandLevel = v;
     sprintf(strbuf,"Command Level %d entered",CommandLevel); ttyoutput(strbuf);
     return 0;
  }
  sprintf(strbuf,"Variable %s unavailable",var); ttyoutput(strbuf);
  return -1;
}
size_t IPCSocket = 0;
void resetPipeListener(PIPEINST *pipeInst)
{
  if (pipeInst == NULL) return;
  DisconnectNamedPipe(pipeInst->hPipeInst);
  pipeInst->oOverlap.hEvent = pipeInst->hPipeConnectEvent;
  ConnectNamedPipe(pipeInst->hPipeInst,&pipeInst->oOverlap);
  pipeInst->isConnected = FALSE; 
}
size_t CMDlisten(const char *name)
{
  PIPEINST *pipeInst=NULL; 
  HANDLE hPipeConnectEvent,hPipeReadEvent;
  int rc, cc = 0;
  int openMode = PIPE_ACCESS_DUPLEX|FILE_FLAG_OVERLAPPED;
  int pipeMode = PIPE_TYPE_MESSAGE|PIPE_READMODE_MESSAGE|PIPE_WAIT;

  /* no security attribute, manual-reset event, initial state = signaled, unnamed event object */
  hPipeConnectEvent = CreateEvent(NULL,TRUE,TRUE,NULL);
  if (hPipeConnectEvent == NULL) ccerr(code_failure); 
  hPipeReadEvent = CreateEvent(NULL,TRUE,TRUE,NULL);
  if (hPipeReadEvent == NULL) ccerr(code_failure); 
  if ((pipeInst=(PIPEINST *)calloc(1,sizeof(PIPEINST))) == NULL) ccerr(out_of_server_memory);
  pipeInst->hPipeConnectEvent = hPipeConnectEvent; 
  pipeInst->hPipeReadEvent = hPipeReadEvent;     
  pipeInst->oOverlap.hEvent = pipeInst->hPipeConnectEvent; 
  pipeInst->hPipeInst = 
    CreateNamedPipe(name,openMode,pipeMode,MAXIPCCLIENTS,IPCBUFSIZE,IPCBUFSIZE,1000,NULL);
  if (pipeInst->hPipeInst == INVALID_HANDLE_VALUE) ccerr(code_failure);
  if (ConnectNamedPipe(pipeInst->hPipeInst,&pipeInst->oOverlap) != 0) ccerr(code_failure);
  pipeInst->isConnected = FALSE; 
err:
  if (cc != 0) 
  {
    rc = GetLastError();
    feclog("CMD listen : %s (windows err %d)",erlst[cc],rc);
  }
  return (size_t)pipeInst;
}
int GetCommandEx(size_t fd,char *buf,int bufsiz)
{
  int rc, siz, len=0;
  PIPEINST *pipeInst = (PIPEINST *)fd;

  if ((rc=PeekNamedPipe(pipeInst->hPipeInst,NULL,0,NULL,&siz,NULL)) == 0)
  {
    resetPipeListener(pipeInst);
    return -1;
  }
  siz = MIN(siz,bufsiz);
  if ((rc=ReadFile(pipeInst->hPipeInst,pipeInst->chBuf,siz,&len,&pipeInst->oOverlap)) == 0) 
  {
    rc = GetLastError();
    if (rc == ERROR_IO_PENDING) return 0;
    resetPipeListener(pipeInst);
    return -1;
  }
  strncpy(buf,pipeInst->chBuf,len);
  memset(pipeInst->chBuf,0,bufsiz);
  buf[len] = 0;
  if (len == 0) 
  {
    return 0;
  }
  CMDcount++;
  return len;
}
size_t CMDaccept(PIPEINST *pipeInst)
{
  if (pipeInst == NULL) return 0;
  /* just test the wait status */
  if (WaitForSingleObject(pipeInst->oOverlap.hEvent,0) != 0) return 0;
  if (pipeInst->isConnected == FALSE)
  {
    pipeInst->isConnected = TRUE;
    pipeInst->oOverlap.hEvent = pipeInst->hPipeReadEvent;     
  }
  return (size_t)pipeInst;
}
int dbglog(char *text,...)
{
# ifndef NO_REMOTE_PIPE
  PIPEINST *pipeInst = (PIPEINST *)IPCSocket;
  int rc,len=0;
# endif
  int cc=0;
  char str[256];
  va_list args;

  if (strlen(text) > 80) ccerr(string_too_long);
  va_start(args,text);
  vsprintf(str,text,args);
  if (str[strlen(str)-1] == '\n') strcat(str,">");
  else sprintf(&str[strlen(str)]," @%s\n>",getDataTimeString((double)time(NULL),FALSE));
  va_end(args);
  if (strlen(tagNameFilter) && !strstr(str,tagNameFilter)) return 0;
  printf(str);
  if (dbgfp != NULL) fprintf(dbgfp,str);
# ifndef NO_REMOTE_PIPE
  if (IPCSocket != 0) 
  {
    rc = WriteFile(pipeInst->hPipeInst,str,(UINT32)strlen(str),&len,NULL);
    if (rc == 0) 
    {
      rc = GetLastError();
      if (rc != ERROR_IO_PENDING) resetPipeListener(pipeInst);
    }
  }
# endif
err:
  return cc;
}
int dbgoutput(char *str,time_t *ts)
{
  char strbuf[CMDSIZE];
# ifndef NO_REMOTE_PIPE
  PIPEINST *pipeInst = (PIPEINST *)IPCSocket;
  int rc, len=0;
# endif

  if (strlen(tagNameFilter) && !strstr(str,tagNameFilter)) return 0;
  if (ts != NULL)
  {
    if (strlen(str) > CMDSIZE-29) str[CMDSIZE-29] = 0; /* just in case */
    sprintf(strbuf,"%s @%d\n>",str,*ts);
  }
  else sprintf(strbuf,"%s\n>",str);

  printf(strbuf);
  if (dbgfp != NULL) fprintf(dbgfp,strbuf);
# ifndef NO_REMOTE_PIPE
  if (IPCSocket != 0) 
  {
    rc = WriteFile(pipeInst->hPipeInst,str,(UINT32)strlen(str),&len,NULL);
    if (rc == 0) 
    {
      rc = GetLastError();
      if (rc != ERROR_IO_PENDING) resetPipeListener(pipeInst);
    }
  }
# endif
  return 0;
}
int ttyoutput(char *str)
{
# ifndef NO_REMOTE_PIPE
  PIPEINST *pipeInst = (PIPEINST *)IPCSocket;
  int rc,len=0;
# endif
  printf(str);
# ifndef NO_REMOTE_PIPE
  if (IPCSocket != 0) 
  {
    rc = WriteFile(pipeInst->hPipeInst,str,(UINT32)strlen(str),&len,NULL);
    if (rc == 0) 
    {
      rc = GetLastError();
      if (rc != ERROR_IO_PENDING) resetPipeListener(pipeInst);
    }
  }
# endif
  return 0;
}
int rmtoutput(char *str)
{
  PIPEINST *pipeInst = (PIPEINST *)IPCSocket;
  int rc, len=0;
  if (IPCSocket != 0) 
  {
    rc = WriteFile(pipeInst->hPipeInst,str,(UINT32)strlen(str),&len,NULL);
    if (rc == 0) 
    {
      rc = GetLastError();
      if (rc != ERROR_IO_PENDING) resetPipeListener(pipeInst);
    }
  }
  return 0;
}
#endif /* NO_DOXYGEN */
EXPORT int RegisterUserCommand(
    char *cmd,
    int (*fcn)(int,int,int,int),
    int *iparam,
    float *fparam,
    int access)
{
  USRCMD *uc;

  if ((uc=(USRCMD *)SystemCalloc(1,sizeof(USRCMD))) == NULL) return out_of_local_memory;
  if (access == 0) access = CA_READ;
  strncpy(uc->cmd,cmd,32);
  uc->fcn = fcn;
  uc->iparam = iparam;
  uc->fparam = fparam;
  uc->access = access;
  uc->next = usrcmd;
  usrcmd = uc;
  numUsrCmd++;
  return 0;
}
EXPORT int RegisterUserFunction(
    char *name,
    int (*fcn)(int argc,char *argv[]))
{
  USRFCN *uf;

  if ((uf=(USRFCN *)SystemCalloc(1,sizeof(USRFCN))) == NULL) return out_of_local_memory;
  strncpy(uf->name,name,32);
  uf->fcn = fcn;
  uf->next = usrfcn;
  usrfcn = uf;
  numUsrFcn++;
  return 0;
}

