#ifndef NTCMD_H_
#define NTCMD_H_

#include <windows.h>
#include <signal.h>
#include <sys/types.h>
#include "rpcdefs.h"
#define CMDSIZE 256

#define MAXIPCCLIENTS  5
#define IPCBUFSIZE 1024
typedef struct 
{ 
   OVERLAPPED oOverlap; 
   HANDLE hPipeInst;
   HANDLE hPipeConnectEvent,hPipeReadEvent;
   CHAR chBuf[IPCBUFSIZE]; 
   DWORD cbToWrite; 
   DWORD dwState; 
   BOOL fPendingIO;
   BOOL isConnected;
} PIPEINST;
extern size_t IPCSocket;
size_t CMDlisten(const char *name);
size_t CMDaccept(PIPEINST *pipeInst);
int GetCommandEx(size_t fd,char *buf,int bufsiz);
int ttyoutput(char *str);
typedef struct usrfcnStruct
{
  char name[32];
  int (*fcn)(int argc,char *argv[]);
  struct usrfcnStruct *next;
} USRFCN;
extern USRFCN *usrfcn;
extern int numUsrFcn;

typedef struct usrcmdStruct
{
  char cmd[32];
  int (*fcn)(int,int,int,int);
  int *iparam;
  float *fparam;
  int access;
  struct usrcmdStruct *next;
} USRCMD;
extern USRCMD *usrcmd;
extern int numUsrCmd;

/* RegisterUserCommand() */
/**
 * \brief Registers a user-defined command or variable which can be accessed via the TINE 
 * command line interpreter.
 * 
 * As a console application, a running tine server (or client) offers a variety of services
 * at the command line (just type 'help').  For instance, you can turn debugging on or off,
 * get the current server statistics, examine the connection tables, etc.  This interface is
 * avaible via the 'remote' accessor even if the server is running in the background.
 * It it often frequently desireable to examine or alter server-specific variables or call a
 * server specific routine.  This can be partially achieved by making use of this registration
 * function.
 *
 * \param cmd is the command string to be parsed by the command interpreter.
 * \param fcn is an optional user defined function taking four integer arguments.
 * \param iparam is a pointer to an optional global integer variable.
 * \param fparam is a pointer to an optional global float variable.
 * \param access is a TINE access code (either CA_READ or CA_READ|CA_WRITE).  To allow 'set'
 * commands at the command line, you should include the CA_WRITE flag.
 *
 * \note Only one of fcn, iparam, or fparm, should be non-NULL when making this call. If the
 * users types "get <cmd>" without arguments at the command line, then either iparam or
 * fparam is displayed on the console, depending on which parameter was non-NULL at registration
 * time. Likewise "set <cmd> = <val>" can set this variable to a value entered at the console.  
 * If a function is registered, then "get <cmd>(arg1,arg2,arg3,arg4)" will call the function
 * registered with the arguments supplied.  "set <cmd>(arg1,arg2,arg3,arg4) = <val>" will set
 * all arguments equal to <val> and then call the function.  If fewer arguments than 4 are
 * supplied then the function will be called with '0' for the missing arguments.
 *
 * \note A more reasonable command line function registration call is underway.  However the
 * routine so presented is generally sufficient for checking hardware readouts etc. and checking
 * and setting the contents of global variables.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
EXPORT int RegisterUserCommand(
    char *cmd,
    int (*fcn)(int,int,int,int),
    int *iparam,
    float *fparam,
    int access);

/* RegisterUserFunction() */
/**
 * \brief Registers a user-defined function which can be accessed via the TINE 
 * command line interpreter.
 * 
 * As a console application, a running tine server (or client) offers a variety of services
 * at the command line (just type 'help').  For instance, you can turn debugging on or off,
 * get the current server statistics, examine the connection tables, etc.  This interface is
 * avaible via the 'remote' accessor even if the server is running in the background.
 * It it often frequently desireable to examine or alter server-specific variables or call a
 * server specific routine.  This can be partially achieved by making use of this registration
 * function.
 *
 * \param name is a string identifying the function to be parsed by the command interpreter.
 * \param fcn is the user defined function taking the argc and argv.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
EXPORT int RegisterUserFunction(char *name,int (*fcn)(int argc,char *argv[]));

int getVal(char *var,char *s);
int setVal(char *var,char *val);
int InterpretCommand(char *cmd);
int quit(char *s1,char *s2);
int help(char *s1,char *s2);
void GetCommand(void);
int getControlSystemInfo(void);
void FecSynchronize(void);
void FecShowtime(void);
void initWatchDog(void);
int initSAPWatchDog(void);
void netWatchDog(void);
void watchDog(void);
void enableWatchDog(void);
void disableWatchDog(void);
void dumpGlobals(void);
int NetHandler(int errval,int ax,int bp,int si);
extern int NumCommands;
extern int NumEquipmentFunctions;
extern WORD watchdogBaseAddress;
extern int watchdogEnabled;
extern int CommandLevel;

#endif /* NTCMD_H_ */
