#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "errors.h"
#include "rpctypes.h"
#include "srvip.h"
#include "toolkit.h"
#include "listener.h"

#define RPTREQM "RPTEQM"
#define RPTR_DEFAULT_PORT 101
#define RPTR_MAX_NDEVICES 5000
#define RPTR_POLLRATE 1000

int gTruncateDeviceNames = 0;
int rpPollingRate = RPTR_POLLRATE;
NAME16 *rpDeviceNames = NULL;
PropertyQueryStruct *rpProperties = NULL;
int rpteqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
void rptini(void);
void rptbkg(void);
void rptexi(void);

#define PRP_MONITOR 1
#define LSNEQM "LSNEQM"
#define STD_TRPTR_FEC "TINERPTRFEC"
#define STD_TRPTR_SRV "TINERPTRSRV"
int lsneqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);
void lsnini(void);
void lsnbkg(void);
void lsnexi(void);

char gTargetDeviceServer[96];
int GetNumberOf(char *srv,char *stkprp)
{
  DTYPE dout;
  int cc = 0;
  short n;
  dout.dFormat = CF_SHORT;
  dout.dArrayLength = 1;
  dout.data.sptr = (UINT16 *)&n;
  if ((cc=ExecLinkEx(srv,stkprp,&dout,NULL,CA_READ,800)) != 0 && !(cc & CE_SENDDATA)) return -cc;
  return n;
}
int rpPreSystemInit(char *tgtContext,char *tgtServer) 
{
  int cc=0;
  char *ptr = NULL;
  char thisFec[32], thisContext[32], thisServer[32], thisDescription[64], thisHost[32], thisUser[32];
  UINT16 portOffset = RPTR_DEFAULT_PORT;

  if ((ptr=getenv("TINE_PORT")) != NULL) portOffset = (UINT16)atoi(ptr);
  if ((ptr=getenv("TINE_CONTEXT")) == NULL) 
  {
    strncpy(thisContext,tgtContext,16);
  }
  else
  {
    strncpy(thisContext,ptr,16); 
  }
  thisContext[16] = 0;
  if ((ptr=getenv("TINE_SERVER")) == NULL) 
  {
    strncpy(thisServer,tgtServer,12); thisServer[12] = 0;
    strcat(thisServer,".RPT");
  }
  else
  {
    strncpy(thisServer,ptr,16);
  }
  thisServer[16] = 0;
  strncpy(thisFec,thisServer,16);
  if ((ptr=getenv("HOSTNAME")) == NULL && (ptr=getenv("COMPUTERNAME")) == NULL) ptr = "local.host";
  strncpy(thisHost,ptr,32);
  sprintf(thisDescription,"/%s/%s repeater",tgtContext,tgtServer);
  if ((ptr=getenv("USER")) == NULL && (ptr=getenv("USERNAME")) == NULL) ptr = "admin";
  strncpy(thisUser,ptr,16); 
  thisUser[16] = 0;

  if ((cc=RegisterFecInformation(thisFec,"RPT",thisContext,thisDescription,thisHost,"none",thisUser,portOffset)) != 0)
  {
    feclog("could not register repeater FEC name : %s",erlst[cc]);
  }
  return 0;
}
int rpPostSystemInit(char *tgtContext,char *tgtServer)
{
  int cc=0,i,n,ndevices,nproperties;
  char *ptr = NULL;
  char thisServer[32], thisDescription[64];
  char tgt[128], dname[36];
  UINT16 portOffset = RPTR_DEFAULT_PORT;

  sprintf(tgt,"/%s/%s/#0",tgtContext,tgtServer);
  sprintf(gTargetDeviceServer,"/%s/%s/",tgtContext,tgtServer);
  if ((ndevices=GetNumberOf(tgt,"NDEVICES")) < 0)
  {
    dbglog("cannot get number of devices from target %s: %s",tgt,erlst[-ndevices]);
    exit(1);
  }
  if ((rpDeviceNames=(NAME16 *)calloc(ndevices,sizeof(NAME16))) == NULL)
  {
    dbglog("cannot allocate space for device names!");
    exit(1);
  }
  n = ndevices;
  if ((cc=GetDeviceNames(tgt,rpDeviceNames,&n)) != 0)
  {
    dbglog("cannot get device list from target %s: %s",tgt,erlst[cc]);
    exit(1);
  }
  for (i=0; i<ndevices; i++)
  {
    if (strchr(rpDeviceNames[i].name,'&') != NULL)
    {
      longDeviceNames = TRUE;
      feclog("setting longDeviceNames to TRUE!");
      break;
    }
  }
  if ((nproperties=GetNumberOf(tgt,"NPROPS")) < 0)
  {
    dbglog("cannot get number of properties from target %s: %s",tgt,erlst[-ndevices]);
    exit(1);
  }
  if ((rpProperties=(PropertyQueryStruct *)calloc(nproperties,sizeof(PropertyQueryStruct))) == NULL)
  {
    dbglog("cannot allocate space for property list!");
    exit(1);
  }
  n = nproperties;
  if ((cc=GetDevicePropertyInformation(tgt,rpProperties,&n)) != 0)
  {
    dbglog("cannot get property list from target %s: %s",tgt,erlst[cc]);
    exit(1);
  }

  if ((ptr=getenv("TINE_PORT")) != NULL) portOffset = (UINT16)atoi(ptr);
  if ((ptr=getenv("TINE_CONTEXT")) == NULL) ptr = tgtContext;
  if ((ptr=getenv("TINE_SERVER")) == NULL) 
  {
    strncpy(thisServer,tgtServer,12); thisServer[12] = 0;
    strcat(thisServer,".RPT");
  }
  else
  {
    strncpy(thisServer,ptr,16);
  }
  thisServer[16] = 0;
  sprintf(thisDescription,"/%s/%s repeater",tgtContext,tgtServer);

  if (gTruncateDeviceNames > 0) ndevices = gTruncateDeviceNames;
  n = longDeviceNames ? ndevices/2 : ndevices;
  if ((cc=RegisterEquipmentModule(thisServer,RPTREQM,n,rpteqm,rptini,rptbkg,1000,rptexi)) != 0)
  {
    feclog("could not register repeater equipment module : %s",erlst[cc]);
  }

  dname[0] = 0;
  for (i=0,n=0; i<ndevices; i++)
  {
    strncat(dname,rpDeviceNames[i].name,16);
    if (dname[15] == '&')
    {
      dname[15] = 0;
      continue;
    }
    if (strlen(dname) == 0) continue;
    if ((cc=RegisterDeviceName(RPTREQM,dname,n++)) != 0)
    {
      dbglog("cannot register device name %s : %s",dname,erlst[cc]);
    }
    dname[0] = 0;
  }

  for (i=0; i<nproperties; i++)
  {
    if (strstr(rpProperties[i].prpName,".HIST") != NULL) continue;
    cc = RegisterProperty(RPTREQM,rpProperties[i].prpName,rpProperties[i].prpSize,
           rpProperties[i].prpFormat,rpProperties[i].prpAccess,rpProperties[i].prpDescription);
    if (cc < 0)
    {
      dbglog("cannot register property name %s : %s",rpProperties[i].prpName,erlst[-cc]);
    }
  }  
  return 0;
}
int rpteqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  char lclTargetDeviceName[96];
  DTYPE *dinptr=NULL;
  int rate;
  strcpy(lclTargetDeviceName,gTargetDeviceServer);
  strcat(lclTargetDeviceName,devName);
  if (din != NULL && din->dArrayLength > 0) dinptr = din;
  rate = access & CA_LAST ? 800 : rpPollingRate;
  return alsnExecLink(lclTargetDeviceName,devProperty,dout,dinptr,access,rate);
}
int lsnPreSystemInit(void)
{
  int cc=0;
  char *ptr = NULL;
  char thisHost[32], thisUser[32];
  UINT16 portOffset = RPTR_DEFAULT_PORT;

  if ((ptr=getenv("TINE_PORT")) != NULL) portOffset = (UINT16)atoi(ptr);
  if ((ptr=getenv("HOSTNAME")) == NULL && (ptr=getenv("COMPUTERNAME")) == NULL) ptr = "local.host";
  strncpy(thisHost,ptr,32);
  if ((ptr=getenv("USER")) == NULL && (ptr=getenv("USERNAME")) == NULL) ptr = "admin";
  strncpy(thisUser,ptr,16); 
  thisUser[16] = 0;

  if ((cc=RegisterFecInformation(STD_TRPTR_FEC,"RPT","REPEATER","local repeater",thisHost,"none",thisUser,portOffset)) != 0)
  {
    feclog("could not register repeater FEC name : %s",erlst[cc]);
  }
  return 0;
}

int lsnPostSystemInit()
{
  int cc=0;

  if ((cc=RegisterEquipmentModule(STD_TRPTR_SRV,LSNEQM,1,lsneqm,lsnini,lsnbkg,1000,lsnexi)) != 0)
  {
    feclog("could not register repeater equipment module : %s",erlst[cc]);
  }

  if ((cc=RegisterDeviceName(LSNEQM,"REPEATER",0)) != 0)
  {
    dbglog("cannot register device name : %s",erlst[cc]);
  }
  cc = RegisterPropertyEx(LSNEQM,"MONITOR",96,CF_TEXT,CA_READ,"monitor target provided",PRP_MONITOR);
  if (cc < 0)
  {
    dbglog("cannot register property : %s",erlst[-cc]);
  }
  return 0;
}
int parseInputDtype(DTYPE *dt,char *devName,char *devProp)
{
  int len = 0;
  char *c, str[100];
  if (dt == NULL || devProp == NULL) return argument_list_error;
  if (dt->dFormat != CF_TEXT) return illegal_format;
  if ((len=dt->dArrayLength) > 96) len = 96;
  strncpy(str,dt->data.cptr,len); str[96] = 0;
  if ((c=strchr(str,'[')) == NULL) return illegal_string_set; 
  *c++ = 0;
  strncpy(devName,str,96);
  strncpy(devProp,c,32);
  if ((c=strchr(devProp,']')) != NULL) *c = 0;
  return 0;
}
int lsneqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  char lclTgtDevNam[96],lclTgtDevPrp[48];
  int cc,rate;
  int prpId = GetPropertyId("LSNEQM",devProperty);
  if ((cc=parseInputDtype(din,lclTgtDevNam,lclTgtDevPrp)) != 0) return cc;
  rate = access & CA_LAST ? 800 : rpPollingRate;
  switch (prpId)
  {
    case PRP_MONITOR:
      return alsnExecLink(lclTgtDevNam,lclTgtDevPrp,dout,NULL,access,rate);
    default:
      return illegal_property;
  }
}
void rptini(void)
{
}
void rptbkg(void)
{
}
void rptexi(void)
{
}
void lsnini(void)
{
}
void lsnbkg(void)
{
}
void lsnexi(void)
{
}
