/** \file server.c
 * \brief TINE server engine plus API routines
 *
 */
#include "prolog.h"
#include "cprolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "servdef.h"
#include "threader.h"
#include "brdcast.h"
#include "errors.h"
#include "database.h"
#include "almsrv.h"
#include "history.h"
#include "srvip.h"
#include "srvipx.h"
#if defined(UNIX) && !defined(NO_REMOTE_PIPE)
# define UNIX_TTY
# include "unixcmd.h"
#endif
#if defined(MAKE_TINE32_DLL)
# include "shmalloc.h"
#endif

int BroadcastServer = UDP_BROADCAST;
int ngcasts = 0, nglobals = 0;
int longDeviceNames = FALSE;
int gBurstLimit = BURST_LIMIT;
int gCycleDelay = CYCLE_DELAY;
EXPORT int SystemPresetMemory = TRUE;
int gDelayInitRPCServices = FALSE;

int RPCprotocol;
#ifndef srvWorkArea
BYTE *srvWorkArea;
UINT32 srvWorkAreaSize = WORK_SIZE;
#else
BYTE *WorkArea;
UINT32 WorkAreaSize = WORK_SIZE;
#endif
BYTE *tmpWorkArea = NULL;
UINT32 tmpWorkAreaSize = WORK_SIZE;

void SurveyAlarms(struct timeval *tv);
void doGCast(struct timeval *thisTime);
void DataDump(char *txt,char *data,int siz,short fmt,short dir);
int datacmp(BYTE *d1,BYTE *d2,int siz,int fmt,double t);
void CloseLink(int i);
void CleanupStrandedDatagrams(void);
void _SystemEngine(void);
int initRPCServices(void);

int gEqpFcnBusySemaphore = 0;
int gEqpFcnAccessor = 0;
ContractListStruct *ContractListItem = NULL;
char *SRVDataBuffer;          /* global Buffer pointers for data exchange */
UINT32 SysPoll = DEFAULT_POLLRATE;
EXPORT UINT16 MinPollingRate = MIN_POLLRATE;
EXPORT UINT16 MaxPollingRate = DEFAULT_POLLRATE;
EXPORT UINT32 MaxRPCTransportSize = MAX_TRANSPORTSIZE;
int gBrdcstBufferSize = BRDCST_BUFFER_SIZE; 
int AveBusyTime = 0;
struct timeval thisCycleTime;
int RequireAcknowledgments = 1; /* global flag to force acknowledge in REFRESH mode */
short NrEQPBckTsks = 0;
int nExpiredContracts = 0;

AccessLockStruct currentAccessLock;
GlobalListStruct **GCastList = NULL;
IPXAddress IPXFecAddress;
int ExportsRegistered = 0;
int FecNameRegistered = 0;
RPCFecExStruct RPCFecEx;

int SystemRunning = 0;
EXPORT int gRetardSingleContractRemoval = TRUE;
SINT32 findpoll(SINT32 p1,SINT32 p2)
{
  SINT32 p3;

  if (MinPollingRate > MaxPollingRate) MinPollingRate = MaxPollingRate;
  if (p1 < MinPollingRate) return p2 <= MinPollingRate ? MinPollingRate : p2;
  if (p2 < MinPollingRate || p1 == p2) return p1;
  if (p2 < p1) while ((p3 = p1%p2) >= MinPollingRate) { p1 = p2; p2=p3; }
  else if (p1 < p2) while ((p3 = p2%p1) >= MinPollingRate) { p2 = p1; p1=p3; }
  return p1 > p2 ? p2 : p1;
}
EXPORT char RPCLastErrorString[RPCERR_SIZE];
ExportListStruct *ExportList = NULL;
int NrOfEqpTypes = 0;
void InitializeEQPModules(void)
{
  ExportListStruct *el;

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (el->inidone) continue;
    if (el->eqpIni) el->eqpIni();
    historyInit(el->tagName,el->EqpName);
    el->inidone = TRUE;
  }
  return;
}
void devNamesInitialization(void)
{
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  int plen, idx;
  char prpName[64];

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (el->hasPropertyDeviceNames)
    {
      for (idx=0; idx<PROPERTY_HASH_SIZE; idx++)
        for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
      {
        if ((plen=(int)strlen(prp->prpName)) < 5) continue;
        if (strcmp(&prp->prpName[plen-4],".NAM")) continue;
        strncpy(prpName,prp->prpName,plen-4); prpName[plen-4] = 0;
        AssignDeviceListFromProperty(el->EqpName,prpName);
      }
    }
  }
  acquireDeviceNamesFromProperties = FALSE;
}
int LastMinuteInitialization(void)
{
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  int idx;
  char tag[32];

  SystemRunning = 1;
  InitializeEQPModules();

  if (gDelayInitRPCServices) initRPCServices();

  if (StartupDebug)
  {
    printf("Exporting the following properties:\n>");
    for (el=ExportList; el != NULL; el=el->next)
      for (idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
        for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
    {
      strncpy(tag,el->EqpName,16); tag[16] = 0;
      printf("%s (%s):  %s (%s) -> %s <-\n>",el->tagName,
             el->EqpName,prp->prpName,
             itom(prp->prpAccessMode),prp->prpDescription);
    }
  }
  return 1;  /* init done */
}
int RegisterLocalServiceModule(void)
{
  ExportListStruct *el;
  int cc;
  if ((cc=RegisterEquipmentModule(SRVEXP_NAME,SRVEQM_NAME,0,NULL,NULL,NULL,0,NULL)) != 0) return cc;
  if ((el=getExportListItem(SRVEQM_NAME)) == NULL) return non_existent_elem;
  el->registered = REGISTERED_ALL; /* don't allow this to be sent to the ENS ! */
  return 0;
}
void ExitEQPModules(void)
{
  ExportListStruct *el;

  for (el=ExportList; el != NULL; el=el->next)
  {
#   ifdef MULTITHREADED
    if (el->hEqpBkg != 0) threadExit(el->hEqpBkg);
    if (hEqpFcnThread != 0) threadExit(hEqpFcnThread);
#   endif
    if (el->eqpExi) el->eqpExi();  /* call exit routines */
  }
  return;
}
#ifndef SMALL_TINE_LIB
EXPORT void FreeEQPModules(void)
{
  ExportListStruct *el, *nxtel=NULL;
  ExportPropertyListStruct *prp=NULL, *nxtp=NULL;
  HashXRefStruct *dev=NULL,*nxtd=NULL;
  ALARM *alm=NULL, *nxta=NULL;
  int i;

  ExitEQPModules();

  for (el=ExportList; el != NULL; )
  {
    for (i=0; i<PROPERTY_HASH_SIZE+1; i++)
    {
      for (prp=el->EqpProperty[i]; prp != NULL; )
      {
        nxtp = prp->next;
        SystemFree(prp);
        prp = nxtp;
      }
    }
    if (el->EqpModuleXRef != NULL)
    {
      for (i=0; i<MODULE_HASH_SIZE; i++)
      {
        for (dev=el->EqpModuleXRef[i]; dev != NULL; )
        {
          nxtd = dev->next;
          SystemFree(dev);
          dev = nxtd;
        }
      }
    }
    if (el->EqpModuleName != NULL) free(el->EqpModuleName);
    if (el->EqpModuleRdr != NULL) free(el->EqpModuleRdr);
    if (el->EqpModulePrpList != NULL) 
    {
      for (i=0; i<el->EqpNumModulesSpace; i++)
      {
        if (el->EqpModulePrpList[i].prpList != NULL) SystemFree(el->EqpModulePrpList[i].prpList);
      }
      SystemFree(el->EqpModulePrpList);
    }
    if (el->EqpModuleRdr != NULL) SystemFree(el->EqpModuleRdr);
    if (el->adsTable != NULL) SystemFree(el->adsTable);
    if (el->almLst != NULL) 
    {
      for (i=0; i<el->EqpNumModulesSpace; i++)
      {
        for (alm=el->almLst[i]; alm !=NULL; )
        {
          nxta = alm->next;
          SystemFree(alm);
          alm = nxta;
        }
      }
      SystemFree(el->almLst);
    }
    nxtel = el->next;
    SystemFree(el);
    el = nxtel;
  }
  ExportList = NULL;
  return;
}
#endif /* SMALL_TINE_LIB */
void AssignDotask(void (*tsk)(void)) /* for backward compatibility of NODAL FECs */
{
  if (ExportList == NULL) return;
  ExportList->eqpBkg  = tsk;
  ExportList->rateBkg = 0;
}
void AssignExittask(void (*tsk)(void)) /* for backward compatibility of NODAL FECs */
{
  if (ExportList == NULL) return;
  ExportList->eqpExi  = tsk;
}
void RegisterEQPtasks(char *name,void (*ini)(void),void (*tsk)(void),int rate,void (*exi)(void))
{
  ExportListStruct *el;

  if ((el=getExportListItem(name)) == NULL) return;

  if ((el->eqpBkg=tsk) != NULL) NrEQPBckTsks++;
  el->rateBkg = rate;
  if (ServerCycleMode != POLLING && rate < MaxPollingRate)
  {
    SysPoll = rate;
    MaxPollingRate = rate;
    if (rate < MinPollingRate) MinPollingRate = rate;
  }
  el->eqpIni = ini;
  el->eqpExi = exi;

  if (StartupDebug)
    printf("%12s task registration: %s %s %s\n>",name,
            ini ? "INIT" : "----",
            tsk ? "BCKG" : "----",
            exi ? "EXIT" : "----");

  return;
}
void checkTasklist(void)
{
  SINT32 msec;
  struct timeval tv;
  ExportListStruct *el;

  gettimeofday(&tv,(struct timezone *)NULL);

  for (el=ExportList; el != NULL; el=el->next) 
  {
    if (el->eqpBkg)
    {
      if ((msec=MSECS(tv,el->lastBkg)) < el->rateBkg) continue;
      if (msec == 0 && el->rateBkg != 0) continue;
      if (el->bkgbusy) continue;
      el->bkgbusy = 1;
      el->lastBkg = tv;
      el->eqpBkg();
      el->bkgbusy = 0;
      if (el->next != NULL) _SystemEngine();
    }
  }
  return;
}
extern char tRedirectedServer[64];
int WRITEaccessFromAllNets = 0; /* leave this undocumented! */
short IsIllegalProperty(char *eqpname,char *prop,char *dev,CLN *cln,short access,ExportListStruct **listEntry)
{
  int k,plen=0,ps;
  short prpType,cc=0,rc;
  ExportPropertyListStruct *prp=NULL;
  ExportListStruct *el;
  char *c,Property[32];

  strncpy(Property,prop,32);
  for (k=0; k<NrStockModifiers; k++)
  {
    if ((c=strstr(Property,StockModifier[k].modName)) != NULL)
    {
      *c = 0;
      break;
    }
  }
  prpType = EMPTY_PROP;
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqpname,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL)
  {
    for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqpname,el->EqpAlias,EQP_NAME_SIZE)) break;
  }
  if (el == NULL) ccerr(non_existent_elem);
  if (el->EqpPropertyQueryFcn == NULL && el->XEqpPropertyQueryFcn == NULL) /* use default property list */
  {
    if ((rc=lookupPropertyName(el,prop,dev)) == -server_redirection) ccerr(-rc);
    if (rc < 0 && (rc=lookupPropertyName(el,Property,dev)) < 0) /* failed first test */
    {
      if (rc == -server_redirection) ccerr(-rc);
      /* check for wild cards */
      for (prp=el->EqpProperty[PROPERTY_HASH_SIZE]; prp; prp=prp->next)
      {
        for (ps = 0; ps < (int)strlen(prp->prpName) && prp->prpName[ps] == '*'; ps++);
        plen = (ps < (int)strlen(prp->prpName)) ? (int)strcspn(&prp->prpName[ps],"*") : 0;
        if (plen && !strncmp(&Property[ps],&prp->prpName[ps],plen)) break;
      }
      if (prp == NULL) /* Maybe it's a query */
      {
        if ((prpType=IsStockProperty(Property)) != STOCK_PROP) ccerr(illegal_property);
      }
      else prpType = WILDCARD_PROP;
    }
    else /*  is registered normally */
    {
      prpType = TARGET_PROP;
    }
  }
  else  /* uses a property query function, but check a few things up front ... */
  {
    if (lookupPropertyName(el,prop,dev) == -server_redirection) ccerr(server_redirection);
    if (strncmp(prop,Property,32) && lookupPropertyName(el,Property,dev) == -server_redirection) 
      ccerr(server_redirection);
    if (lookupRedirectionNameStub != NULL) 
    {
      if (lookupRedirectionNameStub(el->EqpName,prop,dev) == server_redirection) ccerr(server_redirection);
      if (strncmp(prop,Property,32) && 
          lookupRedirectionNameStub(el->EqpName,Property,dev) == server_redirection) ccerr(server_redirection);
    }
    prpType = IsStockProperty(Property);
  }
  if ((access&CA_WRITE) && strcmp("MESSAGE",Property)) /* not a MESSAGE */
  {
    for (k=0;!WRITEaccess && k<NrOfUsers;k++)
      if (!strnicmp(UserList[k].UserName,cln->userName,16)) break;
    if (!WRITEaccess && k == NrOfUsers) 
    {
      feclog("COMMAND %s called by %s : user refused access",Property,cln->userName);
      ccerr(access_denied);
    }
    if (!WRITEaccessFromAllNets &&
        !IsMemberControlNets(cln)) 
    {
      inet_ntoa_b(cln->IPaddress.sin_addr,clientaddr);
      feclog("COMMAND %s called by %s from addr %s: network refused access",
        Property,cln->userName,clientaddr);
      ccerr(access_denied);
    }
    if (strcmp("ACCESSLOCK",Property) && !IsMemberLockSet(cln))
    {
      inet_ntoa_b(cln->IPaddress.sin_addr,clientaddr);
      feclog("COMMAND %s called by %s from addr %s: access lock refused access",
        Property,cln->userName,clientaddr);
      ccerr(access_denied);
    }
    if (putCommandsInFeclog) feclog("COMMAND %s called by %s",Property,cln->userName);
  }
err:
  *listEntry = el;
  return cc;
}
int renegotiateContract(ContractListStruct *cl,int cc)
{
  CONSUMER *c;
  UINT32 doutsize = 0,fmtoutsize;

  if ((fmtoutsize=fmtsizeof(cl->contract.EqpFormatOut)) != 0)
  {
    doutsize = cl->dataSizeOut*fmtoutsize;
    if (cc) doutsize += RPCERR_SIZE;
    if (LFMT(cl->contract.EqpFormatOut) == CF_SPECTRUM) doutsize += 96;
  }
  for (c=cl->consumer; c != NULL; c=c->next)
  {
    c->status.BytesRemaining = (UINT32)doutsize;
    c->status.BytesTotal = (UINT32)doutsize;
    c->status.nextDataSize = c->status.BytesRemaining;
    if (c->status.BytesRemaining > c->status.mtu) c->status.nextDataSize = c->status.mtu;
    c->status.numblks = (UINT16)(doutsize/c->status.mtu + (doutsize%c->status.mtu ? 1 : 0));
    if (doutsize == 0) c->status.numblks = 1;
  }
  return 0;
}
void gateOutputData(BYTE *data,int size,short format,UINT32 gate,int outputBoolean)
{
  int i;
  UINT16 *sval = (UINT16 *)data;
  UINT32 *ival = (UINT32 *)data;
  if (data == NULL) return;
  switch (format)
  {
    case CF_BYTE:
      for (i=0; i<size; i++) 
      {
        data[i] &= gate;
        if (outputBoolean) data[i] = data[i] == gate ? 1 : 0;
      }
      break;
    case CF_SHORT:
      for (i=0; i<size; i++) 
      {
        sval[i] &= gate;
        if (outputBoolean) sval[i] = sval[i] == gate ? 1 : 0;
      }
      break;
    case CF_LONG:
      for (i=0; i<size; i++) 
      {
        ival[i] &= gate;
        if (outputBoolean) ival[i] = ival[i] == gate ? 1 : 0;
      }
      break;
    default:
      return;
  }
}
int gEqpFcnContractTableId = -1;
EXPORT int getLastContractId(void)
{
  if (gEqpFcnContractTableId == -1) return -1;
  return ContractList[gEqpFcnContractTableId]->contractID;
}
EXPORT int getLastContractTableId(void)
{
  return gEqpFcnContractTableId;
}
EXPORT int getContractAccessRate(int id)
{
  if (id < 0) id = gEqpFcnContractTableId;
  if (id < 0 || id >= ncontract) return 0;
  return ContractList[id]->PollingRate;
}
#define isWildcardContract(s) (strchr(s,'*') != NULL)
int EqpFcnWithWildcardDevice(ExportListStruct *el,char *DevName,char *DevProperty,DTYPE *dout,DTYPE *din,short access)
{
  DTYPE drdb;
  char sDevName[32];
  int start = 0, len = 0, isCont = FALSE, isNotReady = FALSE;
  int cc = not_allowed, i, k;
  int fmt, fmtlen, nmlen, noff, voff, soff;
  UINT32 devlistsize = 0;
  ExportPropertyListStruct *prp = GetProperyListStruct(el->EqpName,DevProperty);
  char beginToken[32], endToken[32], midToken[32];
  int blen = 0, elen = 0, mlen = 0;
  NAME32 *devlist=NULL;
  BYTE *rb;
  float *frb;
  UINT32 *lrb;
  NAME16 *nrb = NULL;
  UINT32 nlen;
  double wcTimestamp = 0;

  if (prp == NULL) return illegal_property;
  if (isSpectrumArray(el->EqpName,DevProperty)) return not_allowed;
  if (dout == NULL || dout->dArrayLength < 1) return not_allowed;
  if (!isLocal(el->EqpName,DevName,DevProperty)) return data_not_local;
# ifdef INTEL
  rb = (BYTE *)alloca(dout->dArrayLength*sizeof(float));
# else
  rb = tmpWorkArea;
# endif
  frb = (float *)rb; lrb = (UINT32 *)rb;
  if (prp->devNames != NULL)
  {
	  devlist = prp->devNames;
    devlistsize = prp->numDevices;
  }
# if !defined(SMALL_TINE_LIB)
  else if (el->hasPropertyDeviceNames)
  {
    nlen = prp->prpSize;
#   ifndef VXWORKS
    nrb = (NAME16 *)alloca(2*nlen*sizeof(NAME16));
#   else
    if ((len=dout->dArrayLength*sizeof(float)) < tmpWorkAreaSize - 2*nlen*sizeof(NAME16))
      nrb = (NAME16 *)&tmpWorkArea[dout->dArrayLength*sizeof(float)];
#   endif
    if (GetDeviceListFromProperty(el->EqpName,DevProperty,nrb,&nlen) == 0)
    {
      for (i=nlen-1; i>=0; i--) 
      {
        strncpy(nrb[i*2].name,nrb[i].name,EQP_DEVICE_SIZE);
        nrb[i*2].name[EQP_DEVICE_SIZE] = 0;
      }
      devlist = (NAME32 *)nrb;
      devlistsize = nlen;
    }
  }
# endif /* not SMALL_TINE_LIB */
  if (devlistsize == 0)
  {
    devlist = el->EqpModuleName;
    devlistsize = el->EqpNumModules;
  }
  if ((prp->prpArrayType & AT_CHANNEL) == AT_CHANNEL)
  {
    len = devlistsize;
    isCont = isContiguous(el->EqpName,DevName,devlist,&start,&len);
  }
  switch (dout->dFormat)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      fmtlen = fmtsizeof(dout->dFormat);
      nmlen = fmtlen - 16;
      fmt = prp->prpFormat == CF_FLOAT || prp->prpFormat == CF_DOUBLE ? CF_FLOAT : CF_LONG;
      drdb.dFormat = fmt;
      drdb.dTag[0] = 0;
      if (isCont)
      {
        drdb.data.fptr = frb;
        drdb.dArrayLength = MIN(dout->dArrayLength,(UINT32)len);
        strncpy(sDevName,devlist[start].name,16);
        cc = (*el->EqpFcn)(sDevName,DevProperty,&drdb,din,access);
        if (cc != 0) return cc;
        wcTimestamp = gDataTimeStamp;
        for (i=0; i<(int)drdb.dArrayLength; i++)
        {
          *(double *)(dout->data.bptr + i*fmtlen + nmlen) = fmt == CF_FLOAT ? (double)frb[i] : (double)lrb[i];
          *(double *)(dout->data.bptr + i*fmtlen + nmlen + 8) = 0;
          strncpy(dout->data.cptr + i*fmtlen,devlist[start+i].name,nmlen);
        }
        dout->dArrayLength = drdb.dArrayLength;
      }
      else
      {
        getWildcardTokens(DevName,beginToken,endToken,midToken);
        /* calculate these once outside the loop ... */
        blen = (int)strlen(beginToken);
        mlen = (int)strlen(midToken);
        elen = (int)strlen(endToken);
        for (i=0, k=0; i<(int)devlistsize && k<(int)dout->dArrayLength; i++)
        {
          if (!wildcardMatches(devlist[i].name,beginToken,blen,endToken,elen,midToken,mlen)) continue;
          strncpy(sDevName,devlist[i].name,16);
          drdb.data.fptr = &frb[k];
          drdb.dArrayLength = 1;
          cc = (*el->EqpFcn)(sDevName,DevProperty,&drdb,din,access);
          if (cc == server_redirection) return data_not_local;
          if (cc == not_ready) isNotReady = TRUE;
          if (wcTimestamp < gDataTimeStamp) wcTimestamp = gDataTimeStamp;
          *(double *)(dout->data.bptr + k*fmtlen + nmlen) = fmt == CF_FLOAT ? (double)frb[k] : (double)lrb[k];
          *(double *)(dout->data.bptr + k*fmtlen + nmlen + 8) = (double)cc;
          strncpy(dout->data.cptr + k*fmtlen,devlist[start+i].name,nmlen);
          k++;
        }
        dout->dArrayLength = k;
      }
      gDataTimeStamp = wcTimestamp;
      return isNotReady ? not_ready : 0;
    case CF_USTRING:
    case CF_NAME16II:
    case CF_NAME16FI:
      fmt = dout->dFormat == CF_NAME16II ? CF_LONG : CF_FLOAT;
      fmtlen = fmtsizeof(dout->dFormat);
      if (dout->dFormat == CF_USTRING)
      {
        nmlen = 80;
        noff = 16; voff = 4; soff = 0;
      }
      else
      {
        nmlen = fmtlen - 8;
        noff = 0; voff = nmlen; soff = nmlen + 4;
      }
      drdb.dFormat = fmt;
      drdb.dTag[0] = 0;
      if (isCont)
      {
        drdb.data.fptr = frb;
        drdb.dArrayLength = MIN(dout->dArrayLength,(UINT32)len);
        strncpy(sDevName,devlist[start].name,16);
        cc = (*el->EqpFcn)(sDevName,DevProperty,&drdb,din,access);
        if (cc != 0) return cc;
        wcTimestamp = gDataTimeStamp;
        for (i=0; i<(int)drdb.dArrayLength; i++)
        {
          memset(dout->data.bptr + i*fmtlen,0,fmtlen);
          memcpy(dout->data.bptr + i*fmtlen + voff,&frb[i],4);
          strncpy(dout->data.cptr + i*fmtlen + noff,devlist[start+i].name,16);
        }
        dout->dArrayLength = drdb.dArrayLength;
      }
      else
      {
        getWildcardTokens(DevName,beginToken,endToken,midToken);
        /* calculate these once outside the loop ... */
        blen = (int)strlen(beginToken);
        mlen = (int)strlen(midToken);
        elen = (int)strlen(endToken);
        for (i=0, k=0; i<(int)devlistsize && k<(int)dout->dArrayLength; i++)
        {
          if (!wildcardMatches(devlist[i].name,beginToken,blen,endToken,elen,midToken,mlen)) continue;
          strncpy(sDevName,devlist[i].name,16);
          drdb.data.fptr = &frb[k];
          drdb.dArrayLength = 1;
          cc = (*el->EqpFcn)(sDevName,DevProperty,&drdb,din,access);
          if (cc == server_redirection) return data_not_local;
          if (cc == not_ready) isNotReady = TRUE;
          if (wcTimestamp < gDataTimeStamp) wcTimestamp = gDataTimeStamp;
          memset(dout->data.bptr + k*fmtlen,0,fmtlen);
          memcpy(dout->data.bptr + k*fmtlen + voff,&frb[k],4);
          *(UINT32 *)(dout->data.bptr + k*fmtlen + soff) = (UINT32)cc;
          strncpy(dout->data.cptr + k*fmtlen + noff,devlist[i].name,nmlen);
          k++;
        }
        dout->dArrayLength = k;
      }
      gDataTimeStamp = wcTimestamp;
      return isNotReady ? not_ready : 0;
    default:
      if (dout->dFormat <= CF_NAME32 && strcmp(DevName,"*") == 0)
      { /* primitive format allowed in this case */
        if (isCont) 
        {
          if (dout->dArrayLength > (UINT32)len) dout->dArrayLength = (UINT32)len;
          strncpy(sDevName,devlist[0].name,16);
          return (*el->EqpFcn)(sDevName,DevProperty,dout,din,access);
        }
        drdb.dFormat = dout->dFormat;
        drdb.dTag[0] = 0;
        for (i=0; i<(int)devlistsize && i<(int)dout->dArrayLength; i++)
        {
          strncpy(sDevName,devlist[i].name,16);
          drdb.data.bptr = &dout->data.bptr[i*fmtsizeof(dout->dFormat)];
          drdb.dArrayLength = 1;
          if ((cc=(*el->EqpFcn)(sDevName,DevProperty,&drdb,din,access)) != 0)
          {
            if (cc == not_ready) isNotReady = TRUE; else break;
          }
          if (wcTimestamp < gDataTimeStamp) wcTimestamp = gDataTimeStamp;
        }
        if (cc == server_redirection) return data_not_local;
        dout->dArrayLength = i;
        gDataTimeStamp = wcTimestamp;
        return isNotReady ? not_ready : cc;
      }
  }
  return cc;
}
NAME32 *gPrpList = NULL;
int EqpFcnWithWildcardProperty(ExportListStruct *el,char *DevName,char *DevProperty,DTYPE *dout,DTYPE *din,short access)
{
  DTYPE drdb;
  char sDevProp[32];
  int isNotReady = FALSE;
  int cc = not_allowed, i, k, devnr;
  int fmt, fmtlen, nmlen, noff, voff, soff;
  int prplistsize = 0;
  char beginToken[32], endToken[32], midToken[32];
  int blen = 0, elen = 0, mlen = 0;
  NAME32 *prplist=NULL;
  float frb;
  double wcTimestamp = 0;

  if (dout == NULL || dout->dArrayLength < 1) return not_allowed;
  if ((devnr=GetDeviceNumber(el->EqpName,DevName)) < 0) return illegal_equipment_number;

  if (el->EqpModulePrpList != NULL && 
      el->EqpModulePrpList[devnr].prpList != NULL &&
      el->EqpModulePrpList[devnr].prpListSize > 0)
  {
    prplist = el->EqpModulePrpList[devnr].prpList;
    prplistsize = (int)el->EqpModulePrpList[devnr].prpListSize;
  }
# if !defined(SMALL_TINE_LIB)
  else 
  {
    if (gPrpList == NULL)
    {
      if ((gPrpList=(NAME32 *)SystemCalloc(el->EqpNumProperties,sizeof(NAME32))) == NULL) return out_of_server_memory;
      prplistsize = el->EqpNumProperties;
      if ((cc=GetRegisteredPropertyList(el->EqpName,gPrpList,&prplistsize)) != 0) return cc;
    }
    if (dout->dArrayLength > (UINT32)el->EqpNumProperties) dout->dArrayLength = el->EqpNumProperties;
    prplistsize = el->EqpNumProperties;
    prplist = gPrpList;
  }
# endif /* SMALL_TINE_LIB */
  if (prplistsize == 0) return illegal_property;
  switch (dout->dFormat)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      fmtlen = fmtsizeof(dout->dFormat);
      nmlen = fmtlen - 16;
      fmt = CF_FLOAT;
      drdb.dFormat = fmt;
      drdb.dTag[0] = 0;
      getWildcardTokens(DevProperty,beginToken,endToken,midToken);
      /* calculate these once outside the loop ... */
      blen = (int)strlen(beginToken);
      mlen = (int)strlen(midToken);
      elen = (int)strlen(endToken);
      for (i=0, k=0; i<prplistsize && k<(int)dout->dArrayLength; i++)
      {
        if (!wildcardMatches(prplist[i].name,beginToken,blen,endToken,elen,midToken,mlen)) continue;
        strncpy(sDevProp,prplist[i].name,32);
        if (isSpectrumArray(el->EqpName,sDevProp)) continue;
        drdb.data.fptr = &frb;
        drdb.dArrayLength = 1;
        cc = (*el->EqpFcn)(DevName,sDevProp,&drdb,din,access);
        if (cc == illegal_format || cc == illegal_equipment_number || cc == dimension_error) continue;
        if (cc == not_ready) isNotReady = TRUE;
        if (wcTimestamp < gDataTimeStamp) wcTimestamp = gDataTimeStamp;
        *(double *)(dout->data.bptr + k*fmtlen + nmlen) = fmt == CF_FLOAT ? (double)frb : (double)(*(UINT32 *)&frb);
        *(double *)(dout->data.bptr + k*fmtlen + nmlen + 8) = (double)cc;
        strncpy(dout->data.cptr + k*fmtlen,prplist[i].name,nmlen);
        k++;
      }
      dout->dArrayLength = k;
      gDataTimeStamp = wcTimestamp;
      return isNotReady ? not_ready : 0;
    case CF_USTRING:
    case CF_NAME16II:
    case CF_NAME16FI:
      fmtlen = fmtsizeof(dout->dFormat);
      fmt = dout->dFormat == CF_NAME16II ? CF_LONG : CF_FLOAT;
      if (dout->dFormat == CF_USTRING)
      {
        nmlen = 80;
        noff = 16; voff = 4; soff = 0;
      }
      else
      {
        nmlen = fmtlen - 8;
        noff = 0; voff = nmlen; soff = nmlen + 4;
      }
      drdb.dFormat = fmt;
      drdb.dTag[0] = 0;
      getWildcardTokens(DevProperty,beginToken,endToken,midToken);
      /* calculate these once outside the loop ... */
      blen = (int)strlen(beginToken);
      mlen = (int)strlen(midToken);
      elen = (int)strlen(endToken);
      for (i=0, k=0; i<prplistsize && k<(int)dout->dArrayLength; i++)
      {
        if (!wildcardMatches(prplist[i].name,beginToken,blen,endToken,elen,midToken,mlen)) continue;
        strncpy(sDevProp,prplist[i].name,32);
        if (isSpectrumArray(el->EqpName,sDevProp)) continue;
        drdb.data.fptr = &frb;
        drdb.dArrayLength = 1;
        cc = (*el->EqpFcn)(DevName,sDevProp,&drdb,din,access);
        if (cc == illegal_format || cc == illegal_equipment_number || cc == dimension_error) continue;
        if (cc == not_ready) isNotReady = TRUE;
        if (wcTimestamp < gDataTimeStamp) wcTimestamp = gDataTimeStamp;
        memset(dout->data.bptr + k*fmtlen,0,fmtlen);
        memcpy(dout->data.bptr + k*fmtlen + voff,&frb,4);
        *(UINT32 *)(dout->data.bptr + k*fmtlen + soff) = (UINT32)cc;
        strncpy(dout->data.cptr + k*fmtlen + noff,prplist[i].name,nmlen);
        k++;
      }
      dout->dArrayLength = k;
      gDataTimeStamp = wcTimestamp;
      return isNotReady ? not_ready : 0;
    default:
      break;
  }
  return cc;
}
int doContract(ContractListStruct *cl)
{
  CONTRACT *con = &cl->contract;
  ExportListStruct *el = cl->exp;
  BYTE *dataIn = cl->DataIn;
  BYTE *dataOut = cl->DataBuf;
  int i,fmtsiz,cc=0,rw=1,isGated=FALSE;
  UINT32 doutlen,doutsize,dp,gate=0;
  struct structStruct *s;
  char txtbuf[128],tDevProperty[36];
  double dts = gServerDataTimeStamp; /* gServerDataTimeStamp is set prior to call */
  static DTYPE din,dout;
  static char sDevName[36],sDevProperty[36];

  memset(RPCLastErrorString,0,RPCERR_SIZE);
  doutlen = con->EqpSizeOut; fmtsiz = fmtsizeof(con->EqpFormatOut);
  doutsize = doutlen*fmtsiz; if (LFMT(con->EqpFormatOut) == CF_SPECTRUM) doutsize += 96;

  if (NGdebug > 3)
  {
    strncpy(txtbuf,con->EqpName,EQP_NAME_SIZE); txtbuf[EQP_NAME_SIZE] = 0; strcat(txtbuf," ");
    strncat(txtbuf,con->EqpProperty,EQP_PROP_SIZE); strcat(txtbuf," ");
    strncat(txtbuf,con->EqpDeviceName,EQP_DEVICE_SIZE);
    if (con->EqpSizeIn != 0)
      DataDump(txtbuf,(char *)dataIn,(int)con->EqpSizeIn,(short)LFMT(con->EqpFormatIn),(short)RTOL);
  }

  if (cl->expired)
  {
    if (NGdebug > 1) dbglog("%s %s %s -> returning expired data",
      cl->contract.EqpName,cl->contract.EqpDeviceName,cl->contract.EqpProperty);
    con->EqpSizeOut = cl->dataSizeOut;
    ccerr(cl->returnCode);
  }

  /* check stock properties: */
  /* copy contract's property into temporary variable */
  if (strchr(con->EqpProperty,'*') != NULL && isSimpleFormat(LFMT(con->EqpFormatOut))) strcpy(tDevProperty,"PROPERTIES");
  else strncpy(tDevProperty,con->EqpProperty,EQP_PROP_SIZE);
  if ((i=GetStockPropertyId(tDevProperty)) >= 0)
  {
    dts = putDataTimeStamp(gDataTimeStampOffset,startuptime,0);
    setDataTimeStamp(dts);
    if (doutsize) memset(dataOut,0,(size_t)doutsize);
    ccerr((*StockProperty[i].fcn)(con,(char *)dataIn,(char *)dataOut));
  }
  /* copy contract's property into temporary variable */
  strncpy(tDevProperty,con->EqpProperty,EQP_PROP_SIZE);
  /* check stock modifiers: (allow overloads) */
  if (GetPropertyId(con->EqpName,con->EqpProperty) < 0)
  { /* not a registered property */
    if (el->EqpPropertyQueryFcn == NULL && el->XEqpPropertyQueryFcn == NULL)
    { /* there's no registered property query function */
      for (i=0; i<NrStockModifiers; i++)
      { /* run through the list of stock modifiers */
        if (strstr(con->EqpProperty,StockModifier[i].modName))
        { /* found a match */
          if (doutsize) memset(dataOut,0,(size_t)doutsize);
          if ((isGated=IsGatedModifier(con,tDevProperty,&gate)) != 0) break;
          if ((cc=StockModifier[i].fcn(con,(char *)dataIn,(char *)dataOut)) == 0 ||
               cc == server_redirection ||
               cc == call_redirection) ccerr(cc);
        }
      }
    }
  }

  if (el == NULL || (el->EqpFcn == NULL && el->NdlFcn == NULL)) return non_existent;

  if (cl->compStatus == not_signalled && gEqpFcnBusySemaphore) 
  {
    if (NGdebug > 2) dbglog("%s %s %s is %s (busy count %d)",
      cl->contract.EqpName,cl->contract.EqpDeviceName,cl->contract.EqpProperty,
      cl->compStatus == not_signalled ? "not signalled" : "busy", gEqpFcnBusySemaphore);
    return gEqpFcnBusySemaphore > 1 ? cl->compStatus = operation_busy : not_ready;
  }

  /* copy contract's property into local static variable */
  strncpy(sDevProperty,tDevProperty,EQP_PROP_SIZE);
  /* copy contract's device name into local static variable */
  strncpy(sDevName,con->EqpDeviceName,16); sDevName[16] = 0;
  if (strlen(sDevName) == 16 && sDevName[15] == '&') /* long device name requested */
  {
    sDevName[15] = 0;
    strncat(sDevName,&con->EqpProperty[17],15);
    sDevName[30] = 0;
  }

  /* call equipment function: */
  if (el->EqpFcn != NULL)      /* Non-Nodal EQM function Prototype */
  {
    dout.dTimeStamp = dts;
    dout.dArrayLength = con->EqpSizeOut;
    dout.dFormat = LFMT(con->EqpFormatOut);
    dout.data.vptr = dataOut;
    strncpy(dout.dTag,con->strTagOut,8);
    if (dout.dFormat == CF_STRUCT)
    {
      AssignStructDataPtr(dout.data.vptr,dataOut,NONE);
      dout.dArrayLength = i;
    }
    if (SystemPresetMemory && doutsize && cl->compStatus == not_signalled) memset(dataOut,0,(size_t)doutsize);
    din.dArrayLength = con->EqpSizeIn;
    din.dFormat = LFMT(con->EqpFormatIn);
    din.data.vptr = dataIn;
    strncpy(din.dTag,con->strTagIn,8);
    if (din.dFormat == CF_STRUCT)
    {
      AssignStructDataPtr(din.data.vptr,dataIn,RTOL);
      din.dArrayLength = i;
    }
    cl->lngDeviceName = sDevName;
    cl->lngPropertyName = sDevProperty;
    cl->dout = &dout;
    cl->din = &din;

    if (isWildcardContract(sDevName))
    {
      gEqpFcnBusySemaphore = 1;
      gEqpFcnContractTableId = cl->listId;
      cc = EqpFcnWithWildcardDevice(el,sDevName,sDevProperty,&dout,&din,con->EqpAccess);
      gEqpFcnBusySemaphore = 0;
    }
    else if (isWildcardContract(sDevProperty))
    {
      gEqpFcnBusySemaphore = 1;
      gEqpFcnContractTableId = cl->listId;
      cc = EqpFcnWithWildcardProperty(el,sDevName,sDevProperty,&dout,&din,con->EqpAccess);
      gEqpFcnBusySemaphore = 0;
    }
    else
#   ifdef MULTITHREADED
    if (useMultiThreadedEquipmentFunctions)
    {
      if (cl->compStatus == not_signalled) 
      {
        if (NGdebug > 2) thread_dbglog("release mutex");
        gEqpFcnContractTableId = cl->listId;
        ContractListItem = cl;
        gEqpFcnBusySemaphore = 1;
        ReleaseMutex(hEqpThreadMutex);
      }
      cc = WaitForMutex(hEqpThreadMutex,gSystemTick);
      if (cc != 0)
      {
        if (NGdebug > 2) thread_dbglog("timed out waiting for mutex");
        gEqpFcnBusySemaphore++;
        cc = not_ready; 
      } 
      else
      {
        if (NGdebug > 2) thread_dbglog("got mutex");
        gEqpFcnBusySemaphore = 0;
        cc = cl->compStatus == not_signalled ? not_ready : cl->compStatus;
      }
    }
    else
#   endif
    {
      gEqpFcnBusySemaphore = 1;
      gEqpFcnContractTableId = cl->listId;
      cc = (*el->EqpFcn)(sDevName,sDevProperty,&dout,&din,con->EqpAccess);
      gEqpFcnBusySemaphore = 0;
    }
    if (cc != 0 && !(cc & CE_SENDDATA)) ccerr(cc);
    if (dout.dTimeStamp != dts) gServerDataTimeStamp = dout.dTimeStamp;
    if (dout.dFormat == CF_STRUCT)
    {
      AssignStructDataPtr(dataOut,dataOut,LTOR);
    }
    if (dout.dArrayLength < con->EqpSizeOut)
    {
      if (!SystemPresetMemory) /* zero out the remainder for later data comparisons */
      {
        dp = dout.dArrayLength*fmtsiz;
        memset(&dataOut[(int)dp],0,(size_t)(con->EqpSizeOut-dout.dArrayLength)*fmtsiz);
      }
      con->EqpSizeOut = dout.dArrayLength;
    }
  }
  else if (el->NdlFcn != NULL) /* Nodal EQM function Prototype */
  {
    if (con->EqpSizeIn != 0 &&
        con->EqpSizeIn*fmtsizeof(con->EqpFormatIn) <= con->EqpSizeOut*fmtsizeof(con->EqpFormatOut))
    {
      if (dataOut != dataIn) memcpy(dataOut,dataIn,(size_t)con->EqpSizeIn*fmtsizeof(con->EqpFormatIn));
      if (PrepFormattedData(dataOut,con->EqpSizeIn,CF_DOUBLE,LFMT(con->EqpFormatIn))) return illegal_format;
      rw = -1;
    }
    i = sDevName[0] == '#' ? atoi(&sDevName[1]) : 0;
    if ((cc=(*el->NdlFcn)((double *)dataOut,
                          (short)(con->EqpSizeOut*rw),
                          (short)i,
#   ifndef NODAL
                          (short)LFMT(con->EqpFormatOut),
#   endif
                          sDevProperty)) != 0) ccerr(cc);
    if (PrepFormattedData(dataOut,con->EqpSizeOut,LFMT(con->EqpFormatOut),CF_DOUBLE)) return illegal_format;
  }
err:
  if (isGated) gateOutputData(dataOut,(int)con->EqpSizeOut,(short)LFMT(con->EqpFormatOut),gate,isGated == GATE_BOOLEAN);
  if (con->EqpAccess&CA_FIRST) con->EqpAccess &= ~(CA_FIRST);
  if (cc < 0) cc = ambiguous;
  if (NGdebug > 3 && (cc == 0 || cc & CE_SENDDATA))
    DataDump(txtbuf,(char *)dataOut,(int)con->EqpSizeOut,(short)LFMT(con->EqpFormatOut),(short)LTOR);

  cl->dataSizeOut = con->EqpSizeOut;    /* record possible new data output length */
  if (doutlen != con->EqpSizeOut)
  {
    con->EqpSizeOut = doutlen;   /* don't disrupt the contract itself */
    cc |= CE_RENEGOTIATE;
  }
  return cc;
}
#ifdef INCLUDE_LEGACY_CODE
unsigned short execRequest(CLN *cln)
{
  HeaderStruct tmpH,*h = &tmpH;
  short cc,sval,i,len,doffset=0,fmtsize=0,dsize=0;
  short eqpidx=-1,eqpsize=0;
  unsigned short msgsize=2;
  UINT32 worksize=0;
  UINT32 datasize;
  time_t ts;
  char eqpmodname[8];
  BYTE *Values=NULL;
  BYTE *tWorkArea;
  EqpInfoStruct *EqpInfo=NULL;
  ExportListStruct *el;
  ContractListStruct cl;
  short access = CA_SYNC;

  if (SRVDataBuffer == NULL) ccerr(out_of_server_memory);
  PrepIncomingHeader(h,SRVDataBuffer); eqpsize = abs(h->EqpSize);

  for (i=0,len=0; i<abs(h->EqpSize); i++)
  {
    PrepIncomingInfo(EqpInfo,&SRVDataBuffer[HEADERSTRUCT_SIZE+EQPINFOSTRUCT_SIZE*i]);
    worksize += 3*sizeof(short) + EqpInfo->DataSize*fmtsizeof(EqpInfo->Format) + 8;
    if (EqpInfo->DataSize > len) len = EqpInfo->DataSize;
  }
  if (worksize > (UINT32)srvWorkAreaSize)
  {
   if ((tWorkArea=(BYTE *)realloc((char *)srvWorkArea,(size_t)worksize)) == NULL) ccerr(out_of_server_memory);
   srvWorkArea = tWorkArea;
   srvWorkAreaSize = worksize;
  }
  if ((Values=(BYTE *)SystemCalloc(len,sizeof(double))) == NULL) ccerr(out_of_server_memory);

  memset(eqpmodname,0,8);
  if (h->EqpSize < 0) access |= CA_WRITE;
  doffset = HEADERSTRUCT_SIZE+EQPINFOSTRUCT_SIZE*abs(h->EqpSize);
  RPCprotocol = h->Protocol;

  strncpy(cln->userName,h->UserName,16);
  if (NGdebug) sprintf(dbgbuf,"%s : ",cln->userName);

  for (eqpidx=0; eqpidx<eqpsize; eqpidx++)
  {
    cc=0;
    EqpInfo=(EqpInfoStruct *)&SRVDataBuffer[HEADERSTRUCT_SIZE+EQPINFOSTRUCT_SIZE*eqpidx];
    strncpy(cl.contract.EqpName,EqpInfo->EqpModName,EQP_NAME_SIZE);
    strncpy(cl.contract.EqpProperty,EqpInfo->Property,EQP_PROP_SIZE);
    cl.contract.EqpSizeIn = access & CA_WRITE ? EqpInfo->DataSize : 0;
    cl.contract.EqpSizeOut = EqpInfo->DataSize;
    cl.contract.EqpAccess = access & CA_WRITE ? CA_WRITE|CA_READ : CA_READ;
    cl.contract.EqpFormatIn = BFMT(access & CA_WRITE ? EqpInfo->Format : CF_NULL);
    cl.contract.EqpFormatOut = (BYTE)EqpInfo->Format;
    cl.contract.strTagIn[0] = 0;
    cl.contract.strTagOut[0] = 0;
    sprintf(cl.contract.EqpDeviceName,"#%d",EqpInfo->EqpModNo);

    fmtsize = fmtsizeof(EqpInfo->Format);
    dsize = EqpInfo->DataSize*fmtsize;

    if ((cc=IsIllegalProperty(cl.contract.EqpName,
            cl.contract.EqpProperty,cl.contract.EqpDeviceName,
            cln,access,&el)) != 0) ccerr(cc);

    if (access&CA_WRITE)                                   /* WRITE command */
    {
      PrepStreamData(Values,&SRVDataBuffer[doffset],-EqpInfo->DataSize,EqpInfo->Format);
      doffset += dsize;
    }
    if (strcmp(eqpmodname,EqpInfo->EqpModName)) strncpy(eqpmodname,EqpInfo->EqpModName,6);
    gDataTimeStamp = makeDataTimeStamp();
    ts = RPCtimestamp;

    datasize = EqpInfo->DataSize;

    cl.dataSizeOut = datasize;
    cl.exp = el;
    cl.DataOut = Values;
    cl.DataIn = Values;
    cl.DataBuf = srvWorkArea;
    if ((cc=doContract(&cl)) != 0) goto err;

err:
    PrepWord(cc,cc);
    memcpy(&srvWorkArea[msgsize],&cc,2); msgsize += 2;
    if (!cc)
    {
      PrepWord(sval,EqpInfo->Format); memcpy(&srvWorkArea[msgsize],&sval,2); msgsize += 2;
      PrepWord(sval,dsize); memcpy(&srvWorkArea[msgsize],&sval,2); msgsize += 2;
      PrepStreamData(&srvWorkArea[msgsize],Values,EqpInfo->DataSize,EqpInfo->Format);
      msgsize += EqpInfo->DataSize*fmtsize;

      if (RPCprotocol > 1)
      {
        PrepDWord(ts,ts);
        memcpy(&srvWorkArea[msgsize],&ts,4);
        msgsize += 8;
      }
    }
    if (NGdebug) sprintf(&dbgbuf[strlen(dbgbuf)],"%s %s <%d> ",cl.contract.EqpName,cl.contract.EqpProperty,cc);
  }
  if (NGdebug) dbglog(dbgbuf);
  if (Values != NULL) SystemFree((char *)Values);
  PrepWord(*(unsigned short *)srvWorkArea,msgsize);
  return msgsize;
}
#endif
int initBroadcastList(void)
{
  int cc=0,i,n=1,fmtsize,dsize;
  int GlobalPacketSize = 0;
  static int initialized = 0;

  if (initialized) return 0; else initialized = 1;
  GlobalPacketSize = 0;
  if (StartupDebug) printf("Sending the following global parameters :\n");
  for (i = 0, GlobalPacketSize = sizeof(HeaderStruct);
       i < ngcasts && GlobalPacketSize < gBrdcstBufferSize; i++)
  {
    if (StartupDebug) printf("%s ",GCastList[i]->keyword);
    GlobalPacketSize += sizeof(GlobalDataHdr);
    fmtsize = fmtsizeof(GCastList[i]->format);
    if (!fmtsize) ccerr(illegal_format);
    if ((dsize=GCastList[i]->size*fmtsize) >
         (int)(gBrdcstBufferSize - sizeof(HeaderStruct) - sizeof(GlobalDataHdr)))
    {
      if (StartupDebug) printf("Broadcast %s exceeds %d bytes\n>",
	      GCastList[i]->keyword,gBrdcstBufferSize);
      ccerr(buffer_too_small);
    }
    GlobalPacketSize += dsize;
    if (GlobalPacketSize > gBrdcstBufferSize)
    {
      n++;
      GlobalPacketSize = sizeof(HeaderStruct);
    }
  }
  if (StartupDebug) printf("\n>Total packets required : %d\n>",n);
err:
  if (cc)
  {
    if (StartupDebug) printf("\n>initBroadcastList : %s\n>",erlst[cc]);
    ngcasts = 0;
  }
  return 0;
}
int initRPCServices(void)
{
  static int init_done = FALSE;
  int cc = 0;
  if (init_done) return 0;
  if (gRespondToServiceRequests) cc = RegisterLocalServiceModule();
  initServers(0);
  initAlarmServer();
  makeStockPropXRefTable();
  memset(&currentAccessLock,0,sizeof(currentAccessLock));
  SysPoll = MaxPollingRate;
  if (cdiSystemInitHook != 0) cdiSystemInitHook(); /* if CDI is in use ... */
# ifdef MAKE_TINE32_DLL
  if (CreateShmWindow()) feclog("Cannot create shm window!");
# endif
  init_done = TRUE;
  return cc;
}
int initRPC(void)
{
  int cc;
  BYTE *ver = SystemVersion();

  feclog("VERSION : %d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);
  feclog("OS      : %s",OS_TAG);
  if (srvWorkAreaSize >= WORK_SIZE && 
      MaxRPCTransportSize > MAX_TRANSPORTSIZE &&
      MaxRPCTransportSize > srvWorkAreaSize) srvWorkAreaSize = MaxRPCTransportSize;
  tmpWorkAreaSize = srvWorkAreaSize;
# ifdef MULTITHREADED
  srvWorkAreaSize = MIN_WORK_SIZE;
  feclog("Work Area: %u",srvWorkAreaSize);
  feclog("Temp Size: %u",tmpWorkAreaSize);
  feclog("FEC: is running multi-threaded");
# else
  if (srvWorkAreaSize <= WORK_SIZE && 
      MaxRPCTransportSize > srvWorkAreaSize) MaxRPCTransportSize = srvWorkAreaSize;
  feclog("Work Area: %u",(unsigned int)srvWorkAreaSize);
  feclog("Temp Size: %u",(unsigned int)tmpWorkAreaSize);
  feclog("Max Transport Size: %u",(unsigned int)MaxRPCTransportSize);
# ifdef INTEL
  tmpWorkArea = srvWorkArea;
# endif
# endif
  feclog("FEC HOME : [%s]",FecDBpath);
  startuptime = time(NULL);
  nconsumer = ncontract = 0;
  if ((srvWorkArea = (BYTE *)SystemCalloc(1,(size_t)srvWorkAreaSize)) == NULL)
  {
   feclog("Work Space %u bytes: insufficient memory",(unsigned int)srvWorkAreaSize);
   return out_of_server_memory;
  }
  if (tmpWorkArea == NULL && (tmpWorkArea=(BYTE *)SystemCalloc(1,tmpWorkAreaSize)) == NULL)
  {
    feclog("Temp Work Space: insufficient memory");
    return out_of_server_memory;
  }
  if (MaxNumClients <= 0) return 0; /* no server functionality required */
  if ((ContractList=(ContractListStruct **)SystemCalloc(MaxNumContracts,sizeof(ContractListStruct *))) == NULL)
  {
   feclog("Contract List: insufficient memory");
   return out_of_server_memory;
  }
  if ((Consumers=(CLN **)SystemCalloc(MaxNumClients,sizeof(CLN *))) == NULL)
  {
   feclog("Client List: insufficient memory");
   return out_of_server_memory;
  }
  if (BroadcastServer)
  {
    if ((GCastList=(GlobalListStruct **)SystemCalloc(MaxNumGlobals,sizeof(GlobalListStruct *))) == NULL)
    {
      feclog("Globals List: insufficient memory");
      return out_of_server_memory;
    }
  }
  GetAliasTable();
  memset(&IPXFecAddress,0,sizeof(IPXAddress));
  cc = getRegisteredUsers(&UserList,USERS_FILE,&NrOfUsers,"WRITE",&WRITEaccess);
  cc = getRegisteredUsers(&DeniedUserList,DENYUSER_FILE,&NrOfDeniedUsers,"DENIED",&DENYaccess);
  cc = getRegisteredExports(EXPORT_FILE);
  cc = getFecName();
  if (!FecNameRegistered)
  {
    gDelayInitRPCServices = TRUE;
    feclog("delaying system initialization");
  }
  else
  {
    cc = initRPCServices();
  }
  return cc;
}
int ServerExitCondition = 0;
void quitServer(void)                         /* allow a clean exit */
{
  ServerExitCondition = 1;
  return;
}
char gLastWriteAccessUsr[32] = "";
char gLastWriteAccessAddr[32] = "";
char gLastWriteAccessPrp[36] = "";
char gLastWriteAccessDev[36] = "";
double gLastWriteAccessTime = 0;
UINT32 gBurstLimitReachedCount = 0;
UINT32 gTimeServerCycle = 0;
UINT32 gTimeServerIdleLastCycle = 0;
int gLastMinuteInitializationDone = FALSE;
EXPORT void resetLastMinuteInitialization(void) 
{ 
  gLastMinuteInitializationDone = FALSE; 
}
int gCycleCounts = 0, gMaxCycleCounts = 1;
int dorpc(void)
{
  static struct timeval lastCycleTime;
  static int busy = 0;
  static int CycleCounts = 0;
  int i;

  if (MaxNumClients <= 0) return 0; /* no server functionality required */
  if (ServerExitCondition) { if (ServerExitCondition++ > 3) haltServer(); }
  if (!gLastMinuteInitializationDone) 
  {
    gettimeofday(&thisCycleTime,(struct timezone *)NULL);/* Cycle time */
    gLastMinuteInitializationDone = LastMinuteInitialization();
    lastCycleTime = thisCycleTime;
    if (useGlobalSynchronization) SystemStartGlobalSynchronization();
  }
  if (busy) return not_ready; else busy = 1;
  gettimeofday(&thisCycleTime,(struct timezone *)NULL);/* Cycle time */
  gTimeServerCycle += MSECS(thisCycleTime,lastCycleTime);
  lastCycleTime = thisCycleTime;
  AcceptRequest(&thisCycleTime);/* Accept Consumer subscriptions and renewals   */
  gTimeServerIdleLastCycle += MSECS(thisCycleTime,lastCycleTime);
  ProduceData(&thisCycleTime);  /* Is it time to service consumers?   */
  for (i=0; StaleData && i<gBurstLimit; i++)
  {
    DeliverData(&thisCycleTime);
  }
  if (StaleData)
  {
    if (gCycleDelay) millisleep(gCycleDelay);
    gBurstLimitReachedCount++;
  }
  if (ngcasts) doGCast(&thisCycleTime);
  if (acquireDeviceNamesFromProperties) devNamesInitialization();
  Cleanup(&thisCycleTime);
  SurveyAlarms(&thisCycleTime);
  if (!ExportsRegistered) sndRegisteredExports(&thisCycleTime);
  if (currentAccessLock.lockType) checkAccessLocks((time_t)thisCycleTime.tv_sec);
  gTimeServerCycle += MSECS(thisCycleTime,lastCycleTime);
  lastCycleTime = thisCycleTime;
  CycleCounts++;
  if (gTimeServerCycle >= 1000)
  {
    if (CycleCounts > gMaxCycleCounts) gMaxCycleCounts = CycleCounts;
    gCycleCounts = CycleCounts;
    if (ServerCycleMode != POLLING)
    {
      AveBusyTime = (int)((gTimeServerCycle-gTimeServerIdleLastCycle)*100/gTimeServerCycle);
    }
    else
    {
      AveBusyTime = (gMaxCycleCounts-CycleCounts)*100/gMaxCycleCounts;
    }
    gTimeServerCycle = gTimeServerIdleLastCycle = 0;
    CycleCounts = 0;
  }
  return busy = 0;
}

ContractListStruct **ContractList;
CLN **Consumers;                /* registered consumer list */
int nconsumer = 0, ncontract = 0;

int StaleData = 0;

char *SubType[] = {"CANCEL","SINGLE","REFRESH","POLL"};
int RemoveConsumerFromList(CONSUMER *c,int j)          /* from contract j */
{
  CONSUMER *p=ContractList[j]->consumer;
  int wasPending = FALSE;

  if (j < 0 || j>=ncontract || c == NULL) return -2;

  if (NGdebug > 2) dbglog("remove consumer %s from contract %d",c->client->userName,j);
  if (c->status.blknum < c->status.numblks) wasPending = TRUE;
  c->client->count = MAX(c->client->count-1,0);
  StaleData++;                                  /*  Force RemoveConsumer() */
  if (p == c) ContractList[j]->consumer = c->next;
  else
  {
    while (p->next != c) p=p->next;
    p->next = c->next;
  }
  if (p == p->next) p->next = NULL; /* just in case ! */
  SystemFree((char *)c);
  ContractList[j]->nconsumer = MAX(ContractList[j]->nconsumer-1,0);
  if (ContractList[j]->nconsumer == 0) return 0;
  if (wasPending) ContractList[j]->pending--;
  p = ContractList[j]->consumer;
  ContractList[j]->PollingRate = p->status.PollingRate;
  for (p=p->next; p != NULL; p=p->next)
   ContractList[j]->PollingRate =
     findpoll(ContractList[j]->PollingRate,p->status.PollingRate);
  return 0;
}
int RemoveConsumer(int i)  /* return memory to heap */
{
  if (i < 0 || i >= nconsumer) return -2;
  if (Consumers[i] == NULL) return -3;
  if (NGdebug > 2) dbglog("remove consumer %s",Consumers[i]->userName);
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  if (Consumers[i]->transport == PIPE && Consumers[i]->sck > 0) 
  {
    close(Consumers[i]->sck);
    if (nipcconsumers > 0) nipcconsumers--;
  }
# endif  
  SystemFree((char *)Consumers[i]);
  for (;i<nconsumer-1; i++) Consumers[i] = Consumers[i+1];
  Consumers[--nconsumer] = NULL;
  return 0;
}
int RemoveContract(int j)  /* return memory to heap */
{
  CONSUMER *c=ContractList[j]->consumer,*p;

  if (j < 0 || j>=ncontract) return argument_list_error;
  if (ContractList[j]->compStatus == not_posted) return ContractList[j]->compStatus;
  if (NGdebug > 2) dbglog("remove contract %d",j);
  for (p=c; p != NULL; c=p)                /* decrement consumer count: */
  {
    c->client->count = MAX(c->client->count-1,0);
    StaleData++;                            /*  Force RemoveConsumer()   */
    p = c->next;
    SystemFree((char *)c);
  }
  if (ContractList[j]->expired == CONTRACT_ACTIVE) nExpiredContracts++;
  ContractList[j]->expired = CONTRACT_AVAILABLE;
  if (ContractList[j]->DataIn) 
  {
    SystemFree((char *)ContractList[j]->DataIn);
    ContractList[j]->DataIn = NULL;
  }
  if (ContractList[j]->DataOut) 
  {
    SystemFree((char *)ContractList[j]->DataOut);
    ContractList[j]->DataOut = NULL;
  }
  for (j=0, SysPoll = MaxPollingRate; j<ncontract; j++)
  {
    if (ContractList[j]->expired) continue;
    SysPoll = findpoll(SysPoll,ContractList[j]->PollingRate);
  }
  return 0;
}
UINT32 gContractMisses = 0;
UINT32 gContractDelays = 0;
UINT32 gClientMisses = 0;
UINT32 gClientReconnects = 0;
UINT32 gClientRetries = 0;
UINT32 gSystemTick = SYSTEM_TICK;
UINT32 gSynchronizeContracts = (UINT32)CONTRACT_SYNCHRONIZE;
CONSUMER *EQPconsumerlist = NULL;               /* global reference  */
int EQPnconsumers = 0;
CONTRACT *EQPcontract = NULL;
int ProduceData(struct timeval *tv)             /* Production system */
{
  static int busy = 0;
  ContractListStruct *cl;       /* improve readability by assignment */
  CONSUMER *c;
  short cc,rc,i,dirty=0,Stale,fmtoutsize;
  UINT32 doutsize;
  SINT32 delta_t;
  double dts;
  int isScheduled, msec;
  struct timeval tv_stamp;                

  if (busy) return not_ready; else busy = TRUE;
  for (i=0; i<ncontract; i++)
  {
    if ((cl=ContractList[i])->expired == CONTRACT_AVAILABLE) continue;/* just a place-holder */
    if (cl->nconsumer == 0 && cl->expired == CONTRACT_ACTIVE) 
    {                                                   /* no clients, not marked as expired */
      if (cl->contract.EqpAccess & CA_LAST) doContract(cl);            /* going out of scope */
      RemoveContract(i);
      continue;
    }
    cc = Stale = isScheduled = 0;
    if ((fmtoutsize=fmtsizeof(cl->contract.EqpFormatOut)) != 0)
    {
      doutsize = cl->contract.EqpSizeOut*fmtoutsize;
      if (LFMT(cl->contract.EqpFormatOut) == CF_SPECTRUM) doutsize += 96;
    }
    else doutsize = 0;
    if ((delta_t=MSECS(*tv,cl->lasttime)) < 0) cl->lasttime = *tv; /* system clock set back ! */
    if (delta_t <= cl->PollingRate - (SINT32)gSystemTick) continue;
    if (cl->expired == CONTRACT_EXPIRED && cl->nconsumer == 0)
    {
      if (delta_t > 3*cl->PollingRate) RemoveContract(i);
      continue;
    }
    if (delta_t < MAXALLOWED_POLLRATE && delta_t >= 2*cl->PollingRate)
    {
      if (NGdebug > 1)
      {
        dbglog("[%d] (%s) %s %s missed last schedule (%d vs %d)",i,
          cl->contract.EqpName,
          cl->contract.EqpDeviceName,
          cl->contract.EqpProperty,(int)delta_t,(int)(2*cl->PollingRate));
      }
      cl->misses++;
      gContractMisses++;
    }
    if (cl->pending > 0) 
    {
      if (NGdebug > 2)
      {
        dbglog("[%d] (%s) %s %s still pending",i,
          cl->contract.EqpName,
          cl->contract.EqpDeviceName,
          cl->contract.EqpProperty);
      }
      cl->delays++;
      gContractDelays++;
      continue;           /* still sending out the last round */
    }
    if (cl->starttime.tv_sec == 0 && cl->starttime.tv_usec == 0) cl->starttime = *tv;
    tv_stamp = *tv; 
    msec = (int)(tv->tv_usec/1000);
    /* Synchronize on 1000 msec boundaries if possible */
    if (!(cl->PollingRate%1000) && 
        gSynchronizeContracts   && 
        msec < (1000 - (int)gSystemTick)) tv_stamp.tv_usec = 0;
    dts = putDataTimeStamp(gDataTimeStampOffset,tv->tv_sec,msec);
    setDataTimeStamp(dts);
    EQPnconsumers = cl->nconsumer;
    EQPconsumerlist = cl->consumer;
    EQPcontract = &cl->contract;
    if (cl->nconsumer == 1 && cl->consumer->status.counter == 1)
      cl->contract.EqpAccess |= CA_LAST; /* dangling client going out of scope */
    switch (cc=doContract(cl))
    {
      case 0: break;
      case not_ready:
        msec = MSECS(*tv,cl->starttime);
        if (!gEqpFcnBusySemaphore ||   /* call returned not_ready or still not_signalled */
            (msec < cl->PollingRate + 500 && gEqpFcnBusySemaphore))
        {
          for (c=cl->consumer; c != NULL; c=c->next) c->status.Stale = 0;
          cl->pending = 0;
          continue;      /* don't send out */
        }
        if (NGdebug > 2) dbglog("contract [%d] busy for %d msecs (busy count %d)",i,msec,gEqpFcnBusySemaphore);
        goto err;
      case call_redirection:
      case server_redirection:
        memcpy(cl->compString,tRedirectedServer,64);
      default: 
        if (cc & 0xf000) /* CE_RENEGOTIATE or CE_SENDDATA */
        {
          if ((rc=renegotiateContract(cl,(cc & 0xfff))) != 0) cc = rc; else cc &= ~CE_RENEGOTIATE;
        }
        else goto err;
    }
    if (cl->contract.EqpAccess & CA_WRITE)
    {
      if (cl->consumer != NULL && cl->consumer->client != NULL) 
      {
        strncpy(gLastWriteAccessUsr,cl->consumer->client->userName,16);
        strncpy(gLastWriteAccessAddr,inet_ntoa(cl->consumer->client->IPaddress.sin_addr),16);
      }
      strncpy(gLastWriteAccessPrp,cl->contract.EqpProperty,32);
      strncpy(gLastWriteAccessDev,cl->contract.EqpDeviceName,32);
      gLastWriteAccessTime = gDataTimeStamp;
    }
    if (doutsize && cl->expired == CONTRACT_ACTIVE && memcmp(cl->DataOut,cl->DataBuf,(size_t)doutsize))
    {
      memcpy(cl->DataOut,cl->DataBuf,(size_t)doutsize);
      Stale = dirty = 1;                       /* Data have changed: */
    }                                          /* the cached data are dirty */
    else
    {
      if (MSECS(*tv,cl->lastmarked) > (HEARTBEAT-1)*1000L) Stale = 1;
      dirty = 0;                               /* Cached data have not changed */
    }

err:if (cc) Stale = 1;
    if (Stale) cl->lastmarked = tv_stamp;      /* Last marked as Stale */
    if (cl->lasttime.tv_sec == PRP_SCHEDULE_SIGNAL) isScheduled = TRUE;
    if (cc == not_ready)
    {
      cc = operation_busy;
    }
    else
    {
      cl->lasttime = tv_stamp;                   /* Last polled */
      cl->compStatus = not_signalled;
    }
    cl->starttime.tv_sec = cl->starttime.tv_usec = 0;
    cl->dtimestamp = (SINT32)gServerDataTimeStamp;
    cl->dtimestampMSEC = (int)((gServerDataTimeStamp - (SINT32)gServerDataTimeStamp) * 1000); 
    cl->returnCode = cc;

    for (c=cl->consumer; c != NULL; c=c->next)
    {
      if (c->status.counter == 0)              /* scheduled for removal */
      {
        StaleData++;                           /* Ensure pass thru deliverData */
        continue; 
      }
      c->status.statusCode = cc;
      if (isScheduled) c->status.lasttime.tv_sec = c->status.lasttime.tv_usec = PRP_SCHEDULE_SIGNAL;
      if (cc) c->status.Stale = 1;             /* Stale because of error code */
      if (c->status.mode == CM_REFRESH)
      {
        if (Stale)                             /* Stale because of data change */
        {                                      /* or system heartbeat          */
          c->status.Stale = 1;
          if (dirty && RequireAcknowledgments)
            c->status.counter = ACK_REQUEST;   /* Force acknowledgment */
        }
      }
      else                                     /* Honor polling rate   */
      {
        if (MSECS(*tv,c->status.lasttime) > c->status.PollingRate - (SINT32)gSystemTick)
          c->status.Stale = 1;                 /* Stale because of polling rate */
      }
      if (c->status.Stale)                     /* Consumer for this contract */
      {                                        /* is to be refreshed */
        cl->pending++;
        c->status.lasttime = tv_stamp;
        StaleData++;                           /* Ensure pass thru deliverData */
      }
      if (c->status.counter == ACK_REQUEST)    /* Get serious until an */
      {                                        /* acknowledgment comes */
        c->status.mode = ACK_PENDING;          /* Reset by renewal */
        c->status.PollingRate = cl->PollingRate;
      }
    }
  }
  gettimeofday(tv,(struct timezone *)NULL);
  busy = FALSE;
  return 0;
}
#define maxDeadTime(m,p) ((m) == CM_REFRESH ? HEARTBEAT + (p) : 2 * (p))
int DeliverData(struct timeval *tv)            /* Delivery system */
{
  static int busy = 0;
  ProducerHeaderStruct4 tmpP,*Pheader = &tmpP;
  ProducerHeaderStruct tmpPEx,*PheaderEx = &tmpPEx;
  CONSUMER *c;
  ContractListStruct *cl;       /* improve readability by assignment */
  short i,k,n,dsize,ssize,fsize,passed,snddata,basemode,actionTaken=FALSE;
  SINT32 gapsize, gapptr;
  UINT16 msgsize,totalsize=0,packetmtu;
  UINT32 dptr,delta_t;
  short stepsize,fmt;
  BYTE *buf = srvWorkArea;
  int swallowIt = FALSE;

  if (busy) return not_ready; else busy = TRUE;
  for (k=0; k<nconsumer; k++)
  {
    if (Consumers[k] == NULL) continue;
    if (Consumers[k]->count == 0)
    {
      RemoveConsumer(k--);
      continue;
    }
    passed = 1; totalsize = sizeof(totalsize);
    for (i=0,n=0; i<ncontract && n<Consumers[k]->count; i++)
    {
      cl = ContractList[i];
      if (cl->expired == CONTRACT_AVAILABLE) continue;
      for (c=cl->consumer; c != NULL; c=c->next) if (c->client == Consumers[k]) break;
      if (c == NULL) continue;
      if (Consumers[k]->transport == TCP && Consumers[k]->sck == 0)
        c->status.counter = 0;    /*  TCP connection error -> remove it */
      n++; snddata = TRUE; gapptr = gapsize = 0;
      basemode = c->status.mode;
      if (isServiceRequest(&cl->contract) && c->status.statusCode) swallowIt = TRUE;
      if (c->status.Stale)
      {
        actionTaken = TRUE;
        stepsize = c->client->protocol == 4 ? PRODUCER4_SIZE : PRODUCER_SIZE;
        packetmtu = sizeof(UINT16) + stepsize + c->status.mtu;
        snddata = (c->status.statusCode == 0) ? TRUE : 
                  (c->status.statusCode & CE_SENDDATA) ? TRUE : FALSE;
        if (snddata)
        {
          dsize = (short)c->status.nextDataSize;
          dptr  = c->status.nextDataPtr;
        }
        else
        {
          gapsize = (c->client->protocol == 4 ? 0 : RPCERR_SIZE);
          dsize = (short)gapsize;
          dptr  = 0;
        }
        if ((msgsize=dsize+stepsize) > packetmtu) { c->status.counter = 0; continue; }
        if (msgsize % 2 != 0) msgsize++;  /* make sure of even boundaries */
        PrepPCHeader(&buf[totalsize],c,cl->contract,cl->dtimestamp,cl->dtimestampMSEC);
        if (snddata && (fsize=fmtsizeof(cl->contract.EqpFormatOut)) != 0)
        {
          if ((c->status.statusCode & 0xfff))
          {
            gapsize = (SINT32)(c->status.nextDataPtr + c->status.nextDataSize -
                             cl->contract.EqpSizeOut*fmtsizeof(cl->contract.EqpFormatOut));
            if (gapsize > RPCERR_SIZE) gapsize = RPCERR_SIZE; /* shouldn't be possible */
            if (gapsize > dsize)                              /* err string remainder  */
            {
              gapsize = dsize;
              gapptr  = RPCERR_SIZE - gapsize;
            }
            if (gapsize > 0) dsize -= (short)gapsize; else gapsize = 0;
          }
          if (dsize > 0)
          {
            memcpy(&buf[totalsize+stepsize],&cl->DataOut[(int)dptr],(size_t)dsize);
            ssize = dsize / fsize; fmt = cl->contract.EqpFormatOut;
            if (c->status.blknum > 1 &&        /* long CF_SPECTRUM contract */
                LFMT(cl->contract.EqpFormatOut) == CF_SPECTRUM)
            fmt = BFMT(CF_FLOAT);            /* remaing data is CF_FLOAT */
            PrepSubData(&buf[totalsize+stepsize],ssize,fmt,LTOR);
          }
        }
        if ((totalsize + msgsize) > packetmtu) { passed = 0; goto bok; }
        totalsize += msgsize;

        if (NGdebug > 2 || (NGdebug && c->status.mode <= CM_SINGLE))
        {
          dbglog("[%d]: %s %s %s <%d>(%d b %d of %d) to %s (+%d)",
                  i,cl->contract.EqpName,
                  cl->contract.EqpDeviceName,
                  cl->contract.EqpProperty,c->status.statusCode,msgsize,
                  c->status.blknum,c->status.numblks,
                  c->client->userName,c->status.counter);
        }
        if (gapsize > 0)                      /* insert err string */
        {
          if (c->status.statusCode == server_redirection ||
              c->status.statusCode == call_redirection) 
          {
            memcpy(RPCLastErrorString,cl->compString,RPCERR_SIZE);
            c->status.counter = 1;
          }
          else if ((c->status.statusCode & 0xfff) < numErr) 
          {
            SetRPCCompletion(erlst[c->status.statusCode & 0xff]);
            if (c->status.statusCode == data_not_local) c->status.counter = 1;
          }
          memcpy(&buf[totalsize-(int)gapsize],&RPCLastErrorString[(int)gapptr],(size_t)gapsize);
        }
        if (snddata == 0 || c->status.numblks == c->status.blknum)
        {                                     /* datablock finished and/or error code */
          c->status.Stale = 0;
          if (c->status.counter > 0 ) c->status.counter--;
          if (c->status.mode == CM_CANCEL) c->status.counter = 0;
          c->status.blknum = 1;
          c->status.nextDataPtr = 0;
          c->status.BytesRemaining = c->status.BytesTotal;
          ReturnToServerCycle = 0;
          cl->pending--;
          if ((delta_t=MSECS(*tv,c->status.lasttime)) > (UINT32)maxDeadTime(c->status.mode,c->status.PollingRate))
          {
            if (NGdebug > 1)
            {
              dbglog("[%d]: %s %s %s -> returning late data (%d > %d ms)",
                  i,cl->contract.EqpName,
                  cl->contract.EqpDeviceName,
                  cl->contract.EqpProperty,(int)delta_t,c->status.PollingRate);
            }
            c->status.misses++;
            gClientMisses++;
          }
        }
        else
        {
          ReturnToServerCycle = c->status.numblks - c->status.blknum;
          c->status.blknum++;
          c->status.nextDataPtr += c->status.nextDataSize;
          c->status.BytesRemaining = c->status.BytesTotal - c->status.nextDataPtr;
          StaleData++;
        }
        c->status.nextDataSize = MIN(c->status.mtu,c->status.BytesRemaining);
      }
      if (c->status.counter == 0) 
      {
        actionTaken = TRUE;
        RemoveConsumerFromList(c,i); 
        n--;
      }
      if (!cl->nconsumer) 
      {
        actionTaken = TRUE;
        if (gRetardSingleContractRemoval && basemode == CM_SINGLE) 
        {
          cl->expired = CONTRACT_EXPIRED;
          cl->contract.EqpAccess &= ~(CA_FIRST|CA_LAST);
          nExpiredContracts++;
        }
        else 
        {
          RemoveContract(i);
        }
      }
    }

bok:if (totalsize > sizeof(totalsize) && Consumers[k])
    {
      actionTaken = TRUE;
      PrepWord(*(short *)srvWorkArea,totalsize);
      if (!swallowIt) SendToPeer(Consumers[k],srvWorkArea,SRV_RSP_);
    }
    if (!passed) k--;
  }
  gettimeofday(tv,(struct timezone *)NULL);
  busy = FALSE;
  if (!actionTaken) StaleData = 0;
  return StaleData = --StaleData > 0 ? StaleData : 0;
}
int SendToPeer(CLN *c,BYTE *payload,int msgtype)
{
  switch (c->transport)
  {
    case UDP:
    case TCP:
    case PIPE:
      return SendIPData(c,payload);
#   if defined(IPXNW) || defined(IPXBSD)
    case IPX:
    case SPX:
      return SendIPXData(c,payload);
#   endif
#   if defined(MAKE_TINE32_DLL)
    case MMF:
      return SendShmData(c,payload,(WORD)(msgtype == SRV_RSP_ ? WM_RCV_LCL : WM_REQ_LCL));
#   endif
  }
  return illegal_protocol;
}

ContractProgress cp[MAXLONGGRAMS];
void CleanupStrandedDatagrams(void)
{
  int i, n;

  for (i=0; i<MAXLONGGRAMS; i++)
    if (cp[i].busy && thisCycleTime.tv_sec > (time_t)cp[i].blktime + 5)
  {
    SystemFree((char *)cp[i].data);
    memset(&cp[i],0,sizeof(ContractProgress));
    if (NGdebug > 1) dbglog("stale long datagram %d",i);
  }
  for (i=0, n=0; i<ncontract; i++)
  {
    if (ContractList[i]->expired == CONTRACT_AVAILABLE) continue;
    if (ContractList[i]->compStatus == not_signalled && gEqpFcnBusySemaphore) n++;
  }
  if (n == ncontract) /* this is a lock-out and should never happen !*/
  {
    gEqpFcnBusySemaphore = 0;
  }
}
int GetDefaultFormatAndSize(CONTRACT *con)
{
  int i, num, idx, cc = illegal_property;
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  char *eqm, *eqpprp;
  BYTE *fmt;
  UINT32 *siz;
  PropertyQueryStruct *pqs;
  XPropertyQueryStruct *xpqs;

  if (con == NULL) ccerr(code_failure);

  eqm = con->EqpName; eqpprp = con->EqpProperty;
  fmt = &con->EqpFormatOut; siz = &con->EqpSizeOut;

  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);

  if (el->XEqpPropertyQueryFcn != NULL)        /* registered ext query fcn */
  {
    num = el->XEqpPropertyQueryFcn(con->EqpDeviceName,&xpqs);
    for (i=0; i<num; i++) if (!strncmp(xpqs->prpName,eqpprp,EQP_PROP_SIZE))
    {
      *fmt = BFMT(xpqs->prpFormat);
      *siz = xpqs->prpSize;
      ccerr(0);
    }
  }
  else if (el->EqpPropertyQueryFcn != NULL)         /* registered query fcn */
  {
    num = el->EqpPropertyQueryFcn(con->EqpDeviceName,&pqs);
    for (i=0; i<num; i++) if (!strncmp(pqs->prpName,eqpprp,EQP_PROP_SIZE))
    {
      *fmt = BFMT(pqs->prpFormat);
      *siz = pqs->prpSize;
      ccerr(0);
    }
  }
  else                                                    /* handle it here */
  {  
    idx = (int)(ElfHash((unsigned char *)eqpprp) % PROPERTY_HASH_SIZE);
    for (prp=el->EqpProperty[idx]; prp; prp=prp->next) if (!strncmp(prp->prpName,eqpprp,EQP_PROP_SIZE))
    {
      *fmt = BFMT(prp->prpFormat);
      *siz = prp->prpSize;
      ccerr(0);
    }
  }
  if ((i=GetStockPropertyId(eqpprp)) >= 0)
  {
    *fmt = BFMT(StockProperty[i].prpFormat);
    *siz = StockProperty[i].prpSize;
    ccerr(0);
  }
err:
  return cc;
}
int SendMessageToCaller(BYTE *pkthdr,CLN *cln,struct subscription *s,int cc,BYTE *data,int len)
{
  BYTE dbuf[HEADERSTRUCT_SIZE+PRODUCER_SIZE+RPCERR_SIZE], *buf;
  CONSUMER tmpC,*c = &tmpC;
  HeaderStruct tmpH,*h = &tmpH;
  ProducerHeaderStruct4 tmpP,*Pheader = &tmpP;
  ProducerHeaderStruct tmpPEx,*PheaderEx = &tmpPEx;
  UINT16 msgsize,prot = 5;
  double dts;

  PrepIncomingHeader(h,pkthdr); /* prepare the packet header */
  switch (cc)
  {
    case get_subscription_id:
     if (h->Protocol !=  5) ccerr(illegal_protocol);
     break;
    case call_redirection:
    case server_redirection:
      len = RPCERR_SIZE; /* = 2 * NAME16 */
      data = (BYTE *)tRedirectedServer;
      break;
    default:
      if (cc > 0 && cc < numErr) SetRPCCompletion(erlst[cc]);
      break;
  }
  if (NGdebug) 
  {
    switch (cc)
    {
      case illegal_property:
      case illegal_equipment_name:
      case non_existent_elem:
      case call_redirection:
      case server_redirection:
      case access_denied:
        dbglog("MSG: (%s) %s -> %s",s->contract.EqpName,s->contract.EqpProperty,erlst[cc]);
        break;
      default: 
        dbglog("MSG: -> %s",erlst[cc]);
        break;
    }
  }
  (c = &tmpC)->status.starttime = s->starttime;
  c->status.statusCode = cc;
  c->status.id = s->id;
  c->status.numblks = 1;
  c->client = cln;
  c->client->protocol = prot;
  dts = putDataTimeStamp(gDataTimeStampOffset,thisCycleTime.tv_sec,(int)(thisCycleTime.tv_usec/1000));
  msgsize = prot == 4 ? PRODUCER4_SIZE : PRODUCER_SIZE+RPCERR_SIZE;
  buf = dbuf;
  PrepWord(*(short *)buf,msgsize+sizeof(msgsize));
  PrepPCHeader(&buf[sizeof(msgsize)],c,s->contract,(SINT32)dts,(int)((dts - (SINT32)dts) * 1000));
  if (data != NULL && len > 0)
  {
    memcpy(&buf[sizeof(msgsize)+PRODUCER_SIZE],data,len);
  }
  else if (cc)
  {
    memcpy(&buf[sizeof(msgsize)+PRODUCER_SIZE],RPCLastErrorString,RPCERR_SIZE);
  }
  SendToPeer(cln,buf,SRV_RSP_);
err:
  return cc;
}
int getNextFreeContractID(void)
{
  static UINT16 cid = 0;
  int i;
  for (i=0,cid++; i<ncontract; i++) 
  {
    if (ContractList[i]->expired == CONTRACT_AVAILABLE) continue;
    if (ContractList[i]->contractID == cid) cid++;
  }
  return cid;
}
UINT32 gSingleLinkCount = 0;
#define replaceContractAliasNames(cn) { if (nAliasTableEntries > 0) {\
  if ((i=getAliasTableIndex((cn)->EqpProperty)) >= 0) strncpy((cn)->EqpProperty,gAliasTable[i].name,EQP_PROP_SIZE);\
  if ((i=getAliasTableIndex((cn)->EqpDeviceName)) >= 0) strncpy((cn)->EqpDeviceName,gAliasTable[i].name,EQP_DEVICE_SIZE);\
  }\
}
#define ccout(o) { cc = o; goto out; }
int getConsumerRequest(CLN *caller,BYTE *buf,int packetsize)
{
  BYTE *dbuf;
  HeaderStruct tmpH,*h = &tmpH;
  CONSUMER tmpC,*c = &tmpC;
  CLN *nwcln, *cln=caller;
  int i,j = -1, slot = -1;
  short subid,fmtinsize,fmtoutsize,prot,cc=0,isTrueMCast=0,renewalMultiplier,swallowOnErr=FALSE;
  UINT16 totalsize,msgsize,datasize,bytesread=HEADERSTRUCT_SIZE;
  static UINT32 dinsize,doutsize,defsize,dataPtr;
  struct subscription4 tmpReg,*reg = &tmpReg;
  struct subscription tmpRegEx,*regEx = &tmpRegEx;
  static UINT16 datamtu;

  msgsize = 0; /* avoid compiler warning */
  PrepIncomingHeader(h,buf); totalsize = h->TotalSize;

  if (totalsize <= 0) ccout(out_of_range);
  if ((prot=h->Protocol) < 4 || prot > 5) ccout(illegal_protocol);
  if (NGdebug) sprintf(&dbgbuf[strlen(dbgbuf)],"%s -> ",h->UserName);
  strncpy(cln->userName,h->UserName,15);                  /* always reset   */

  PrepSubHeader(prot,&buf[bytesread]);
  replaceContractAliasNames(&regEx->contract);
  if (isServiceRequest(&regEx->contract)) swallowOnErr = TRUE;
  while (bytesread < totalsize && bytesread < packetsize) /* unpack request */
  {
    cln = caller; /* reset this in case of packed NETWORK request ! */
    if (NGdebug && regEx->numblks < 4) /* output to screen can influence io buffering */
    {
      sprintf(&dbgbuf[strlen(dbgbuf)],"%s %s %s %d msec (%d of %d)[%s]",
              regEx->contract.EqpName,regEx->contract.EqpDeviceName,
              regEx->contract.EqpProperty,(int)regEx->PollingRate,
              regEx->blknum,regEx->numblks,SubType[(regEx->mode)&0x03]);
      dbglog(dbgbuf);
      sprintf(dbgbuf,"\t\t ->");
    }
    if (NrOfDeniedUsers > 0 && isDeniedUser(h->UserName)) ccerr(access_denied);
    if ((fmtoutsize=fmtsizeof(regEx->contract.EqpFormatOut)) != 0)
    {
      if ((doutsize=regEx->contract.EqpSizeOut*fmtoutsize) <= 0) ccerr(illegal_data_size);
      if (LFMT(regEx->contract.EqpFormatOut) == CF_SPECTRUM) doutsize += 96; /* CF_SPECTRUM HEADER */
      if (doutsize > (UINT32)MaxRPCTransportSize) ccerr(invalid_transport_size);
    }
    else doutsize = 0;

    if ((fmtinsize=fmtsizeof(regEx->contract.EqpFormatIn)) != 0) 
    { /* data input */
      if (regEx->blknum > regEx->numblks) ccerr(link_error);
      dinsize = regEx->contract.EqpSizeIn*fmtinsize;
      if (LFMT(regEx->contract.EqpFormatIn) == CF_SPECTRUM) dinsize += 96; /* CF_SPECTRUM HEADER */
      dbuf = &buf[bytesread + (prot == 4 ? SUBSCRIPTION4_SIZE : SUBSCRIPTION_SIZE)];
      dataPtr = ((UINT32)(regEx->blknum-1))*regEx->mtu;
      datasize = MIN((UINT16)(dinsize-dataPtr),regEx->mtu);
      if (dinsize > (UINT32)MaxRPCTransportSize) ccerr(invalid_transport_size);
      if (dinsize > regEx->mtu) /* incoming long datagram */
      {
        for (i=0; i<MAXLONGGRAMS; i++)
          if (cp[i].haddr == (UINT32)cln->IPaddress.sin_addr.s_addr &&
              cp[i].SubscriptionID == regEx->id &&
              cp[i].SubsStarttime  == regEx->starttime) break;
        if (i == MAXLONGGRAMS) for (i=0; cp[i].busy && i<MAXLONGGRAMS; i++);
        if (i == MAXLONGGRAMS) ccerr(resources_exhausted);
        if (++(cp[i].blksin) <= regEx->numblks)
        {
          if (cp[i].blksin == 1) cp[i].blkid = regEx->blkid;
          else if (cp[i].blkid < regEx->blkid ||
                   cp[i].blkid > regEx->blkid + 500)
          {
            if (NGdebug > 1)
              dbglog("re-sync long datagram id %d to %d",cp[i].blkid,regEx->blkid);
            cp[i].blksin = 1;
            cp[i].blkid = regEx->blkid;
            cp[i].blktime = (UINT32)time(NULL);
            cp[i].SubsStarttime = regEx->starttime;
          }
          if (!cp[i].busy)
          {
            if (NGdebug > 1) dbglog("new long datagram (%d) sub id %d %d bytes",i,regEx->id,(int)dinsize);
            cp[i].busy = 1;
            cp[i].haddr = cln->IPaddress.sin_addr.s_addr;
            cp[i].SubscriptionID = regEx->id;
            cp[i].blktime = (UINT32)time(NULL);
            cp[i].SubsStarttime = regEx->starttime;
            if ((cp[i].data=(BYTE *)SystemCalloc(1,(size_t)dinsize)) == NULL) ccerr(out_of_server_memory);
          }
          memcpy(&cp[i].data[(int)dataPtr],dbuf,(size_t)datasize);
          ReturnToServerCycle = regEx->numblks - cp[i].blksin;
          if (cp[i].blksin < regEx->numblks) goto bow;
          cp[i].busy = 2;
          dbuf = cp[i].data;
          if (NGdebug > 1) dbglog("in-coming long datagram (%d blocks) complete",regEx->numblks);
        }
        else ReturnToServerCycle = 0;
      }
      PrepSubData(dbuf,regEx->contract.EqpSizeIn,regEx->contract.EqpFormatIn,RTOL);
    }
    else 
    { /* no input data -> make sure the input tag is zeroed out */
      memset(regEx->contract.strTagIn,0,8);
      regEx->contract.EqpSizeIn = 0;
      dinsize = 0; 
      dbuf = NULL;
    }

    if (regEx->PollingRate < MinPollingRate) regEx->PollingRate = MinPollingRate;
    if (regEx->PollingRate > 1000) regEx->PollingRate = (regEx->PollingRate/1000)*1000;

    if (LFMT(regEx->contract.EqpFormatOut) == CF_DEFAULT)
    {                                      /* caller wants default settings */
      if ((cc=GetDefaultFormatAndSize(&regEx->contract)) != 0) ccerr(cc);
      if ((fmtoutsize=fmtsizeof(regEx->contract.EqpFormatOut)) != 0)
      {
        if ((defsize=regEx->contract.EqpSizeOut*fmtoutsize) <= 0) ccerr(illegal_data_size);
        if (LFMT(regEx->contract.EqpFormatOut) == CF_SPECTRUM) defsize += 96; /* CF_SPECTRUM HEADER */
        if (defsize > (UINT32)MaxRPCTransportSize) ccerr(invalid_transport_size);
        if (defsize > doutsize)           /* caller can't buffer default */
        {
          regEx->contract.EqpSizeOut = doutsize/fmtoutsize;
          defsize = regEx->contract.EqpSizeOut*fmtoutsize;
          if (LFMT(regEx->contract.EqpFormatOut) == CF_SPECTRUM) 
          {
            if (regEx->contract.EqpSizeOut > 24) regEx->contract.EqpSizeOut -= 24; /* CF_SPECTRUM HEADER */
          }
        }
      }
      else defsize = 0;
      doutsize = defsize;                  /* for allocation purposes */
    }
    slot = -1;                             /* start search for available slot */
    for (j=0; j<ncontract; j++)            /* is contract in list ? */
    {
      if (ContractList[j]->expired == CONTRACT_AVAILABLE) 
      {
        if (slot < 0) slot = j;
        continue;
      }
      if (!memcmp(&ContractList[j]->contract,&regEx->contract,sizeof(CONTRACT)))
      {
        if (!dbuf && !ContractList[j]->DataIn) break;          /* no data input */
        if (dbuf && ContractList[j]->DataIn) /* input buffer has been allocated */
        {
          if (!memcmp(ContractList[j]->DataIn,dbuf,(size_t)dinsize)) break;
        }
      }
    }
    if (j == ncontract)                    /* no, add to list  */
    {
      if (slot >= 0) 
      {
        j = slot;
        memset(ContractList[j],0,sizeof(ContractListStruct));
        if (nExpiredContracts > 0) nExpiredContracts--;
      }
      else
      {
        if (j >= MaxNumContracts) ccerr(resources_exhausted);
        if ((ContractList[j]=(ContractListStruct *)SystemCalloc(1,sizeof(ContractListStruct))) == NULL)
          ccerr(out_of_server_memory);
        ncontract++;
      }
      memcpy(&ContractList[j]->contract,&regEx->contract,sizeof(CONTRACT));
      ContractList[j]->compStatus = not_signalled;
      ContractList[j]->returnCode = not_posted;
      ContractList[j]->listId = j;

      if ((cc=IsIllegalProperty(ContractList[j]->contract.EqpName,
                                ContractList[j]->contract.EqpProperty,
                                ContractList[j]->contract.EqpDeviceName,
                                cln,
                                ContractList[j]->contract.EqpAccess,
                                &ContractList[j]->exp)) != 0) ccerr(cc);

      if (dinsize && (ContractList[j]->DataIn = (BYTE *)SystemCalloc(1,(size_t)dinsize)) == NULL)
      {
        ccerr(out_of_server_memory);
      }
      if (doutsize && (ContractList[j]->DataOut = (BYTE *)SystemCalloc(nDataOutBuffers,(size_t)doutsize)) == NULL)
      {
        if (dinsize) SystemFree((char *)ContractList[j]->DataIn);
        ccerr(out_of_server_memory);
      }
      ContractList[j]->DataBuf = nDataOutBuffers > 1 ? &ContractList[j]->DataOut[doutsize] : srvWorkArea;
      if (dbuf) memcpy(ContractList[j]->DataIn,dbuf,(size_t)dinsize);
      ContractList[j]->contract.EqpAccess |= CA_FIRST; /* turn on FIRST bit */
      ContractList[j]->contractID = getNextFreeContractID();
      ContractList[j]->expired = CONTRACT_ACTIVE;
      ContractList[j]->PollingRate = 0; /* reset this too ! */
      ContractList[j]->lasttime.tv_sec = PRP_SCHEDULE_SIGNAL;
    }
    /* is a new link attaching to an expired contract ? */
    if (ContractList[j]->expired == CONTRACT_EXPIRED && !(regEx->mode & CM_RETRY)) 
    {
      ContractList[j]->expired = CONTRACT_ACTIVE;
      ContractList[j]->contract.EqpAccess |= CA_FIRST; /* turn on FIRST bit */
      if (nExpiredContracts > 0) nExpiredContracts--;
    }
#   if defined(MAKE_TINE32_DLL)
    if (cln->transport == MMF && regEx->mode & (CM_BCAST|CM_MCAST)) 
    { /* a local client is asking for a network subscription */
      if (gForceShmDatagrams || getNetworkClientFromList() == NULL)
      { /* there's no network client going at the moment */ 
        regEx->mode &= ~(CM_BCAST|CM_MCAST); /* strip this off for a more efficient MMF transfer */
      }
    }
#   endif
    /* is broadcast or multicast request ? */
    if (regEx->mode & (CM_BCAST|CM_MCAST))
    {
      isTrueMCast = regEx->mode & CM_MCAST;
      regEx->mode &= 0x0ff; /* strip off */
      if (regEx->mode == CM_CANCEL) ccerr(0); /* don't do anything */
      if (IsMemberBcastNets(cln) || isTrueMCast)
      {
        /* send subscription ID to caller */
        subid = ISWAP(ContractList[j]->contractID);
        SendMessageToCaller(buf,cln,regEx,get_subscription_id,(BYTE *)&subid,2);
        if ((nwcln=getNetworkClientFromList()) != NULL) 
        {
          cln = nwcln;
        }
        else
        { /* Network client not yet in list */
          if (cln->transport == PIPE || cln->transport == MMF)
          { 
            cln->transport = UDP; 
            cln->IPaddress.sin_family = AF_INET;
            cln->IPaddress.sin_addr.s_addr = 0x010101ee;
            cln->sck = ProducerSocket; /* reset this in case of windows MMF */
          }
          /* add network to client list and replace cln */
          strncpy(cln->userName,"NETWORK",16);
          if (cln->IPaddress.sin_addr.s_addr != 0)
          {
            cln->IPaddress.sin_port = htons((UINT16)gtMCastPort);
            if (isTrueMCast) cln->IPaddress.sin_addr.s_addr = tMCastAddr.s_addr;
            else cln->IPaddress.sin_addr.s_addr |= 0xff000000L;
            cln = LocateConsumerInList(&cln->IPaddress,NULL,NULL);
          }
          else if (*(UINT32 *)cln->IPXaddress.network != 0)
          {
            memset(&cln->IPXaddress.network[10],MCAST_SOCKET,2);
            memset(cln->IPXaddress.node,0xff,6);
            cln = LocateConsumerInList(NULL,&cln->IPXaddress,cln->node);
          }
        }
        regEx->id = ContractList[j]->contractID;
        regEx->starttime = BCAST_ID;
      }
    }

    /* is client in contractList?    */
    for (c=ContractList[j]->consumer; c!=NULL && c->client!=cln; c=c->next);

    if (c == NULL)  /* no, add to list */
    {
      if ((c=(CONSUMER*)SystemCalloc(1,sizeof(CONSUMER))) == NULL) ccerr(out_of_server_memory);
      c->next = ContractList[j]->consumer;
      ContractList[j]->consumer = c;
      c->client = cln;
      ContractList[j]->nconsumer++;
      c->client->count++;
      c->client->protocol = prot;
      c->status.statusCode = ContractList[j]->returnCode;
      c->status.counter = 1;
    }
    c->status.PollingRate = regEx->PollingRate;
    c->status.id = regEx->id;
    c->status.starttime = regEx->starttime;
    c->status.blknum = 1;
    c->status.nextDataPtr = 0;
    c->status.BytesRemaining = doutsize;
    c->status.BytesTotal = doutsize;
    datamtu = PMTU(*(short *)c->client->IPXaddress.socket);
    if (fmtoutsize > 0) datamtu -= datamtu%fmtoutsize;
    c->status.mtu = datamtu;
    c->status.nextDataSize = c->status.BytesRemaining;
    if (c->status.BytesRemaining > c->status.mtu) c->status.nextDataSize = c->status.mtu;
    c->status.numblks = doutsize/((UINT32)c->status.mtu) + (doutsize%((UINT32)c->status.mtu) ? 1 : 0);
    if (doutsize == 0) c->status.numblks = 1;
    renewalMultiplier = 1;
    if (regEx->PollingRate < 500) renewalMultiplier = 3;
    if (c->status.mode != ACK_PENDING)         /* Force Contract execution */
    {
      if ((c->status.counter == 0 && regEx->starttime != BCAST_ID) || regEx->mode == CM_REFRESH)
      {
        if (ContractList[j]->returnCode != not_posted) 
        { /* contract return code can be trusted */
          c->status.Stale = 1; 
          c->status.lasttime = thisCycleTime;
        }
      }
      if (regEx->mode != CM_CANCEL && regEx->mode != CM_REFRESH &&
          c->status.counter > ACK_REQUEST*renewalMultiplier &&
          c->status.counter < RENEWAL*renewalMultiplier)
      {
        if (NGdebug > 1)
        {
          dbglog("reconnecting to %s %s %s from %s (+%d)",
            regEx->contract.EqpName,regEx->contract.EqpDeviceName,
            regEx->contract.EqpProperty,h->UserName,c->status.counter);
        }
        c->status.reconnects++;
        gClientReconnects++;
      }
      StaleData++;
    }
    if (regEx->mode & CM_RETRY)
    {
      if (NGdebug > 1)
      {
        dbglog("retrying %s %s %s by %s",
          regEx->contract.EqpName,regEx->contract.EqpDeviceName,
          regEx->contract.EqpProperty,h->UserName);
      }
      c->status.Stale = 1;                     
      StaleData++;   /* if expired, DeliverData() will send the last payload again */
      c->status.isRetry = TRUE;
      gClientRetries++;
      regEx->mode = BASEMODE(regEx->mode);
    }
    if (ContractList[j]->nconsumer == 0)       /* just turned off, turn it back on */
    {
      ContractList[j]->nconsumer++;
      ContractList[j]->contract.EqpAccess &= ~(CA_LAST);
    }
    switch (c->status.mode = regEx->mode)
    {
      case CM_SINGLE :
        gSingleLinkCount++;
        c->status.counter = 1;
        if (!(ContractList[j]->expired == CONTRACT_EXPIRED && c->status.isRetry))
        {
          ContractList[j]->lasttime.tv_sec = PRP_RETRY_SIGNAL;  
        }
        else  /* a RETRY marked as expired should not be called */
        {
          ContractList[j]->lasttime.tv_sec = (UINT32)time(NULL);
          if (ContractList[j]->dataSizeOut != ContractList[j]->contract.EqpSizeOut)
            renegotiateContract(ContractList[j],ContractList[j]->returnCode);
        }
        if (ContractList[j]->PollingRate == 0) ContractList[j]->PollingRate = c->status.PollingRate;
        if (ContractList[j]->nconsumer == 1)
        {
          ContractList[j]->contract.EqpAccess |= CA_LAST; /* turn on LAST bit */
        }
        break;
      case CM_POLL   :
      case CM_REFRESH:
        c->status.counter = RENEWAL * renewalMultiplier;
        ContractList[j]->PollingRate = findpoll(regEx->PollingRate,ContractList[j]->PollingRate);
        break;
      default     :
      case CM_CANCEL :
        if (!strcmp(c->client->userName,"NETWORK")) break;
        if (ContractList[j]->nconsumer == 1)
        {
          if (c->status.counter)                            /* was active */
            ContractList[j]->contract.EqpAccess |= CA_LAST; /* turn on LAST bit */
          ContractList[j]->lasttime.tv_sec = PRP_CANCEL_SIGNAL;
          ContractList[j]->pending = 0;
        }
        c->status.counter = 0;
        break;
    }
    if (NGdebug > 2)
      dbglog("link counter set to %d for %s (mode %d)",c->status.counter,c->client->userName,c->status.mode);
err:
    if (cc)
    {
      if (!swallowOnErr) SendMessageToCaller(buf,cln,regEx,cc,NULL,0);
      if (j >= 0 && j < MaxNumContracts) RemoveContract(j);
      cc = 0;
    }
    else
    {
      SysPoll = findpoll(SysPoll,ContractList[j]->PollingRate);
    }
    for (i=0; i<MAXLONGGRAMS; i++) if (cp[i].busy == 2)
    {
      SystemFree((char *)cp[i].data);
      memset(&cp[i],0,sizeof(ContractProgress));
      if (NGdebug > 1) dbglog("discarding long datagram %d",i);
    }
bow:
    if (regEx->msgsize <= 0) break;                  /* appears to be super corrupt ! */
    if ((bytesread += regEx->msgsize) < totalsize)
    {
      if (regEx->msgsize % 2 != 0) break;
      PrepSubHeader(prot,&buf[bytesread]);
      replaceContractAliasNames(&regEx->contract);
    }
  }                                                           /* continue unpacking  */
out:
  if (cc && !swallowOnErr) SendMessageToCaller(buf,cln,regEx,cc,NULL,0);
  return cc;
}
void doGCast(struct timeval *thisTime)
{
  BYTE *buf = srvWorkArea;
  int cc = 0,i,n,setsize,GlobalPacketSize,datasize;
  char *strptr;
  GlobalDataHdr tmpBD,*gd = &tmpBD;
  HeaderStruct tmpH,*GlobalHdr = &tmpH;
  int delta_t;
  time_t interim = thisTime->tv_sec;
  static int done = FALSE;

  if (!done)
  {
    if (initBroadcastList()) return;
    done = TRUE;
  }
  GlobalPacketSize = HEADERSTRUCT_SIZE;
  for (i=0, n=0; i<ngcasts; i++)
  {
    if (GCastList[i]->tv.tv_sec > 0 &&                         /* not a scheduled broadcast */
        (GCastList[i]->minPeriod <= 0 || GCastList[i]->maxPeriod <= 0)) continue;
    delta_t = (int)MSECS(*thisTime,GCastList[i]->tv);
    if (delta_t < GCastList[i]->minPeriod) continue; /* too soon */
    if (delta_t < GCastList[i]->maxPeriod) /* within heartbeat */
    {
      if (!datacmp(GCastList[i]->refdata,
            GCastList[i]->data,
            GCastList[i]->size,
            GCastList[i]->format,
            GCastList[i]->tolerance)) continue;
    }
    datasize = GCastList[i]->size*fmtsizeof(GCastList[i]->format);
    setsize = GLOBALDATA_HEADER_SIZE + datasize;
    if (GlobalPacketSize + setsize > gBrdcstBufferSize)
    {                                          /* send what's there  */
      AssignOutgoingHeader(GlobalHdr,buf);  /* fill in the header */
      GlobalHdr->EqpSize = n;               /* number of keywords */
      strncpy(GlobalHdr->UserName,gUserName,16);
      GlobalHdr->Protocol = 3;              /* broadcast protocol */
      GlobalHdr->TotalSize = GlobalPacketSize;
      PrepOutgoingHeader(GlobalHdr,buf);
      cc = SendGCast(buf,GlobalPacketSize);
      GlobalPacketSize = HEADERSTRUCT_SIZE;
      n = 0;
    }
    AssignOutgoingBrdcstData(gd,&buf[GlobalPacketSize]);
    strncpy(gd->name,GCastList[i]->keyword,16);
    gd->size = GCastList[i]->size;
    gd->format = GCastList[i]->format;
    gd->time = (SINT32)(GCastList[i]->time = interim);
    GCastList[i]->tv = *thisTime;
    PrepOutgoingGCastData(gd,&buf[GlobalPacketSize]);
    GlobalPacketSize += GLOBALDATA_HEADER_SIZE;
    if (GCastList[i]->data)
    {
      memcpy(GCastList[i]->refdata,GCastList[i]->data,datasize);
      PrepOutgoingData(&buf[GlobalPacketSize],GCastList[i]->data,datasize,GCastList[i]->format);
    }
    if (NGdebug > 1)
    {
      sprintf(dbgbuf,"GCAST: %s ",GCastList[i]->keyword);
      strptr = &dbgbuf[strlen(dbgbuf)];
      if (cc) sprintf(strptr,"(ERR: %s)",erlst[cc]);
      dbglog(dbgbuf);
      if (cc == 0 && NGdebug > 2)
      {
        DataDump("GCAST: ",(char *)GCastList[i]->data,
                  GCastList[i]->size,GCastList[i]->format,LTOR);
      }
    }
    if (GCastList[i]->fcn) (GCastList[i]->fcn)(i,cc);
    GlobalPacketSize += datasize; n++;
  }
  if (n > 0)                                 /* something left     */
  {
    AssignOutgoingHeader(GlobalHdr,buf);  /* fill in the header */
    GlobalHdr->EqpSize = n;               /* number of keywords */
    strncpy(GlobalHdr->UserName,gUserName,16);
    GlobalHdr->Protocol = 3;              /* broadcast protocol */
    GlobalHdr->TotalSize = GlobalPacketSize;
    PrepOutgoingHeader(GlobalHdr,buf);
    cc = SendGCast(buf,GlobalPacketSize);
  }
  return;
}

char tExecLinkReturnedCompletion[36];
int tExecLinkReturnedDataSize = 0;
int tExecLink(int fecindex,char *dat,char *eqm,char *dev,char *prp,short siz,short fmt,short timout)
{
  short access=CA_READ,sizIn=0,fmtIn=CF_NULL;
  char *datIn=NULL;
  if (siz < 0)
  {
    access = CA_WRITE|CA_READ;
    siz = sizIn = abs(siz);
    fmtIn = fmt;
    datIn = dat;
  }
  return tExecLinkEx(fecindex,eqm,dev,prp,siz,fmt,dat,sizIn,fmtIn,datIn,access,timout);
}
int tExecLinkEx(int fecindex,char *eqm,char *dev,char *prp,
	   short siz,short fmt,char *dat,short sizIn,short fmtIn,char *datIn,
     short access,short timout)
{
  int fmtsize,fmtsizeIn,retry = 2;
  short cc=0,dsize,dsizeIn,dretsiz=0,sstepsize,pstepsize;
  static BYTE lclWorkArea[IPX_BUFFER_SIZE];
  static short ServiceProtocol = 5; /* trivial calls used for services only */
  static UINT16 tid = 0;            /* trivial subscription id */
  HeaderStruct hTmp,*h = &hTmp;
  struct subscription4 sTmp, *s = &sTmp;
  struct subscription  sTmpEx, *sEx = &sTmpEx;
  ProducerHeaderStruct4 pTmp, *Pheader = &pTmp;
  ProducerHeaderStruct pTmpEx, *PheaderEx = &pTmpEx;
  ConnTableEntry c;

  if (NGdebug) dbglog("trivial RPC: %s %s %s",eqm,prp,dev);
  fmtsize = fmtsizeof(fmt);
  fmtsizeIn = fmtsizeof(fmtIn);
  if (sizIn > 0 && datIn == NULL) return argument_list_error;
  if (strlen(eqm) > 8 || strlen(prp) > 32) return string_too_long;
  if ((dsize = siz*fmtsize) < 0) return out_of_range;
  if ((dsizeIn = sizIn*fmtsizeIn) < 0) return out_of_range;
  if ((2 + PRODUCER_SIZE + dsize) > IPX_BUFFER_SIZE) return use_stream_transport;
  if ((HEADERSTRUCT_SIZE + SUBSCRIPTION_SIZE + dsizeIn) > IPX_BUFFER_SIZE) return use_stream_transport;
  if (timout < 1) timout = 1;
  if (ServiceProtocol != 5) return illegal_protocol;

  memset(&c,0,sizeof(ConnTableEntry));
  c.mode = CM_SINGLE; 
  c.starttime = (UINT32)time(NULL);
  strncpy(c.prop,prp,32);
  strncpy(c.deviceName,dev,16); 
  c.FECindex = fecindex;
  c.sizeIn = sizIn; 
  c.formatIn = fmtIn;
  c.sizeOut = siz; 
  c.formatOut = fmt;
  c.access = access;
  c.datamtu = IPX_BUFFER_SIZE-SUBSCRIPTION_SIZE-HEADERSTRUCT_SIZE;
  c.blkidIn = 0;
  c.blknumIn = 1;
  c.numblksIn = 1;

  sstepsize = SUBSCRIPTION_SIZE;
  pstepsize = PRODUCER_SIZE;

  AssignOutgoingHeader(h,lclWorkArea);

tRETRY:
  c.RPCprotocol = ServiceProtocol;
  AssignOutgoingSubscription(ServiceProtocol,tid,&c,eqm,&lclWorkArea[HEADERSTRUCT_SIZE]);
  s->msgsize = sstepsize + dsizeIn;
  if (sizIn > 0)
  {
    memcpy(&lclWorkArea[HEADERSTRUCT_SIZE+sstepsize],datIn,dsizeIn);
  }
  PrepOutgoingSubscription(ServiceProtocol,&lclWorkArea[HEADERSTRUCT_SIZE]);

  h->TotalSize = HEADERSTRUCT_SIZE + sstepsize + dsizeIn;
  memcpy(h->UserName,gUserName,16);
  h->Protocol = ServiceProtocol;
  h->EqpSize = 1;
  PrepOutgoingHeader(h,lclWorkArea);

  switch (cc = tCall(fecindex,lclWorkArea,timout))
  {
    case 0: 
      break;
    case link_not_open:
      if (--retry) goto tRETRY;
    default:
      ccerr(cc);
  }

  PrepIncomingSubscription(ServiceProtocol,&lclWorkArea[2]);
  if (PheaderEx->SubscriptionID != tid) ccerr(connection_timeout);
  if ((cc=PheaderEx->CompletionCode) == 0 || (cc & CE_SENDDATA))
  {
    if (dsize < (dretsiz=PheaderEx->msgsize-pstepsize)) ccerr(illegal_format);
    memcpy(dat,&lclWorkArea[2+pstepsize],dretsiz);
  }
  else
  {
    strncpy(tExecLinkReturnedCompletion,(char *)&lclWorkArea[2+pstepsize],RPCERR_SIZE);
  }
err:
  tid = (tid + 1) % 10000;
  if (cc && NGdebug)
  {
    dbglog("trivial RPC error: <%d> %s",cc, cc > 0 && cc < numErr ? erlst[cc] : "unknown");
  }
  tExecLinkReturnedDataSize = fmtsize > 0 ? dretsiz / fmtsize : 0;
  return cc;
}
