/** \file shmalloc.c
 * \brief shared memory routines for windows
 *
 */
#include "prolog.h"
#include <windows.h>
#include <windowsx.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <ctype.h>
#include <time.h>
#include "errors.h"
#include "database.h"
#include "servdef.h"
#include "shmalloc.h"

SECURITY_ATTRIBUTES mmfSA;
HANDLE hShmWnd = 0;
HANDLE hParent = 0;
HWND IncomingLocalFec = 0;
int gForceShmDatagrams = TRUE;

__declspec(dllexport) LRESULT CALLBACK shmWndProc(HWND hWnd, WORD iMessage, WORD wParam, LONG lParam)
{
  char shmTag[64];
  shmDataGram *dg;
  HANDLE hDG;
  int siz;
  int hClnWnd;
  CLN *cln;
  switch (iMessage)
  {
    case WM_REQ_LCL: /* server getting a request */
      hClnWnd = (int)lParam;
      if ((cln=LocateConsumerInList(NULL,NULL,(BYTE *)&hClnWnd)) == NULL) break;
      sprintf(shmTag,"_shmTContract:%d",lParam);
      cln->transport = MMF;
      cln->sck = lParam;
      siz = wParam;
      dg = (shmDataGram *)shmGetDataGram(shmTag,siz,&hDG);
      strncpy(cln->userName,dg->dataTag,16);
      getConsumerRequest(cln,dg->payload,dg->size);
      dg->dirty = 0;
      shmFreeDataGram(dg,hDG);
      break;
    case WM_RCV_LCL: /* client getting a reply */
      sprintf(shmTag,"_shmTData:%d",hShmWnd);
      siz = wParam;
      dg = (shmDataGram *)shmGetDataGram(shmTag,siz,&hDG);
      IncomingNetworkProtocol = MMF;
      IncomingLocalFec = (HWND)lParam;
      consumeData(dg->payload);
      dg->dirty = 0;
      shmFreeDataGram(dg,hDG);
      break;
    default: 
      return DefWindowProc (hWnd, iMessage, wParam, lParam);
  }
  return 0;
}
void DestroyShmWindow(void)
{
  static char  lpszWN[17];
  strncpy(lpszWN,gFecName,16); lpszWN[16] = 0;
  if (hShmWnd != 0)
  {
    if (DestroyWindow(hShmWnd) == FALSE) feclog("Cannot destroy shm window!");
    if (UnregisterClass(lpszWN,hParent) == FALSE) feclog("Cannot unregister shm window class!");
  }
  hShmWnd = 0;
}
int CreateShmWindow(void)
{
  static WNDCLASS wndclass;
  static char  lpszWN[17];
  char lpConsoleTitle[256];
  DWORD hConsole = 0;

  if ((hConsole=GetConsoleTitle(lpConsoleTitle,256)) != 0) return 0; /* is a console app */

  strncpy(lpszWN,gFecName,16); lpszWN[16] = 0;

  if (hShmWnd != 0 && FindWindow(lpszWN,lpszWN) != 0) return 0; /* still active */
  
  wndclass.style         = CS_GLOBALCLASS|CS_SAVEBITS; 
  wndclass.lpfnWndProc   = (WNDPROC) shmWndProc;
  wndclass.cbClsExtra    = 0;
  wndclass.cbWndExtra    = 0;
  wndclass.hInstance     = hParent;
  wndclass.hIcon         = LoadIcon (0, IDI_APPLICATION);
  wndclass.hCursor       = LoadCursor (0, IDC_ARROW);
  wndclass.hbrBackground = GetStockObject (WHITE_BRUSH);
  wndclass.lpszMenuName  = 0;
  wndclass.lpszClassName = lpszWN;

  if (!RegisterClass(&wndclass)) return -1;

  if (!(hShmWnd = CreateWindow (
			lpszWN, lpszWN,               
			WS_BORDER|WS_CAPTION,         
			CW_USEDEFAULT, CW_USEDEFAULT, 
			CW_USEDEFAULT, CW_USEDEFAULT, 
			0, 0,                 
			hParent,               
			(LPSTR)""                     
      ) ) ) return -1;


  return 0;
}

int SendShmData(CLN *cln,BYTE *data,WORD iMessage)
{
  char shmTag[64];
  HANDLE hShm = 0;
  int msgsize = *(short *)data;
  if (msgsize > SHM_BUFFER_SIZE) return use_stream_transport;
  sprintf(shmTag,"_shmTData:%d",cln->sck);
  return shmSndDataGram(shmTag,data,msgsize,&hShm,iMessage);
}
HANDLE mmfalloc(char *tag,int SizeInBytes,BYTE **lpView,int *mmfFlag)
{
  HANDLE hFileMap;
  int MMFattached = 0;
  char msg[256];
  static int done = 0;

  if (!done)
  {
    done = TRUE;
    mmfSA.nLength = sizeof(mmfSA);
    mmfSA.lpSecurityDescriptor = NULL;
    mmfSA.bInheritHandle = TRUE;
  }

  if (tag == NULL || strlen(tag) == 0) return NULL;
  if ((hFileMap=CreateFileMapping(hSHARE,&mmfSA,PAGE_READWRITE,0,SizeInBytes,tag)) != NULL)
  {
    MMFattached = (GetLastError() == ERROR_ALREADY_EXISTS);
    if ((*lpView=(BYTE *)MapViewOfFile(hFileMap,FILE_MAP_WRITE,0,0,0)) == NULL)
    {
      sprintf(msg,"mmfalloc shm error: %s (%x)",tag,GetLastError());
      MessageBox(NULL,msg,"",IDOK);
      CloseHandle(hFileMap);
      return NULL;
    }
  }
  if (mmfFlag) *mmfFlag = MMFattached;
  return hFileMap;
}

int mmfInit(void)
{
  mmfSA.nLength = sizeof(mmfSA);
  mmfSA.lpSecurityDescriptor = NULL;
  mmfSA.bInheritHandle = TRUE;
  return 0;
}
int shmSndDataGram(char *tag,BYTE *data,int size,HANDLE *h,WORD iMessage)
{
  shmDataGram *dg = NULL;
  int cc,rsize,MMFattached=0;
  char *c, msg[256];
  HANDLE hWnd;

  if ((c=strchr(tag,':')) == NULL) return -1; c++;
  hWnd = (HANDLE)atoi(c);
  rsize = sizeof(shmDataGram) + size;

  if ((*h=mmfalloc(tag,rsize,(BYTE **)&dg,&MMFattached)) == NULL)
  {
    sprintf(msg,"shmSndDataGram error: %s (%x)",tag,GetLastError());
    MessageBox(NULL,msg,"",IDOK);
    *h = 0;
    return -1;
  }
  dg->payload = (BYTE *)dg + 28;
  memcpy(dg->payload,data,size);
  dg->size = size;
  dg->dirty = TRUE;
  cc = SendMessage(hWnd,iMessage,size,(long)hShmWnd);
  if (MMFattached) shmFreeDataGram(dg,*h);
  return cc;
}
shmDataGram *shmGetDataGram(char *tag,int size,HANDLE *h)
{
  shmDataGram *dg = NULL;
  int rsize,MMFattached=0;
  char msg[256];

  rsize = sizeof(shmDataGram) + size;

  if ((*h=mmfalloc(tag,rsize,(BYTE **)&dg,&MMFattached)) == NULL)
  {
    sprintf(msg,"shmGetDataGram shm error: %s (%x)",tag,GetLastError());
    MessageBox(NULL,msg,"",IDOK);
    *h = 0;
    return NULL;
  }
  dg->payload = (BYTE *)dg + 28;
  return dg;
}
void shmFreeDataGram(shmDataGram *dg,HANDLE h)
{
  if (h == 0) return;
  UnmapViewOfFile(dg);
  CloseHandle(h);
}
BYTE *shmDataBlockAlloc(char *tag,int size,HANDLE *hBlock)
{
  BYTE *d = NULL;
  int MMFattached=0;
  char msg[256];

  if ((*hBlock=mmfalloc(tag,size,&d,&MMFattached)) == NULL)
  {
	 sprintf(msg,"shm error: %s",tag);
	 MessageBox(NULL,msg,"",IDOK);
	 *hBlock = 0;
	 return NULL;
  }
  return d;
}
void shmDataBlockFree(BYTE *d,HANDLE hBlock)
{
  if (hBlock == 0) return;
  UnmapViewOfFile(d);
  CloseHandle(hBlock);
}
EXPORT void setForceShmDatagrams(int value)
{
  gForceShmDatagrams = value;
}