/** \file srvip.c
 * \brief TCP/UDP routines and definitions
 * \internal
 *
 */
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "servdef.h"
#include "srvip.h"
#include "brdcast.h"
#include "errors.h"
#include "database.h"
#ifdef WINNT
# include "ntcmd.h"
#endif
#ifdef UNIX
# include "unixcmd.h"
int foreground = 1;
int ipcProducerSocket = 0;
int nipcconsumers = 0;
char gIpcTinePipe[64];
#endif

extern int BroadcastServer;

UINT16 PortOffset = PORTOFFSET;
int ReturnToServerCycle = 0;
int allowLegacyConnections = FALSE;
EXPORT int gRespondToServiceRequests = RESPOND_TO_SERVICE_REQUESTS;

int prepUDPBuffer(int msgsize,int cc);
int GetTCPStream(int socket,CLN *cln);
void CloseTCPStream(int socket);
int listenUDP(int *s,unsigned short port);

extern char FecDBpath[80];
struct sockaddr_in brdcst_name;
int TCPParent = 0, TCPProducerSocket = 0;                  /* TCP sockets */
int ProducerSocket = 0, SyncSocket = 0;                    /* UDP sockets */
int NetServiceSocket = 0;
int TCPsocket[NumTCPChan];
int *TCPsck = NULL;
int nTCPsck = 0;
int maxTCPConnections = MAX_TCPCONNECTIONS;
extern fd_set conset;
fd_set TCPrset, rwset, UDPset;
struct timeval TCPto, UDPto, fec_to;
void (*socketAcceptEntryFcn)(void) = NULL;
int initIP(void)
{
# ifdef MSDOS
  int i,s=0;
  char *c;
  if (!loaded()) return -1;
  if ((c=getenv("KEEPSOCKETS")) != NULL)
  {
    s = atoi(c);
  }
  for (i=63; i > s; i--)
  {
    soclose(i);
    shutdown(i,2); /* shutdown any open ports */
  }
# endif
# ifdef WINNT
  if (!WinsockLoaded()) return -1;
  MaxPollingRate = 10; MinPollingRate = 10;
# endif
  return 0;
}
struct sockaddr_in *IPNetList = NULL;
int NrOfIPNets = 0;
struct sockaddr_in *IPBcastList = NULL;
int NrOfIPBcastNets = 0;

int MaxNumAllowedIPNets = MAX_NUM_IPNETS;
EXPORT int appendRegisteredIPNetsList(char *ipaddr)
{
  return appendRegisteredNetsList(IPNETS_FILE,&NrOfIPNets,&IPNetList,ipaddr);
}
EXPORT int appendRegisteredBCastNetsList(char *ipaddr)
{
  return appendRegisteredNetsList(IPBCASTNETS_FILE,&NrOfIPBcastNets,&IPBcastList,ipaddr);
}
int appendRegisteredNetsList(char *fn,int *nr,struct sockaddr_in **list,char *ipaddr)
{
  int cc=0, i, nlist = *nr;
  DWORD haddr;
  struct sockaddr_in *lst = *list;
# ifndef FS_NONE /* normal case: there is a file system */
  FILE *fp = NULL;
  char netsfilename[128],ip[32];
  if (ipaddr != NULL && strlen(ipaddr) > 0)
  {
    for (i=0; i<nlist; i++) 
    {
      haddr = inet_addr(ipaddr);
      if (!memcmp(&lst[i].sin_addr,&haddr,4)) break;
    }
    if (i < nlist) return 0; /* already in list */
    sprintf(netsfilename,"%s%s",FecDBpath,fn);
    if ((fp=fopen(netsfilename,"a")) == NULL) ccerr(no_such_file);
    fseek(fp,0,SEEK_END);
    if (ftell(fp) == 0)
    {
      fprintf(fp,"SUBNET\n");  /* file did not exist before */
    }
    strncpy(ip,ipaddr,16); ip[16] = 0;
    fprintf(fp,"%s\n",ip);
    fclose(fp);
  }
  getIPControlNets(fn,nr,list);
# else
  if (lst == NULL)
  {
    if ((lst=(struct sockaddr_in *)SystemCalloc(MaxNumAllowedIPNets,sizeof(struct sockaddr_in))) == NULL)
      ccerr(out_of_server_memory);
    *list = lst;
    nlist = 0;
  }
  if (ipaddr == NULL || strlen(ipaddr) == 0) ccerr(argument_list_error);
  for (i=0; i<nlist; i++) 
  {
    haddr = inet_addr(ipaddr);
    if (!memcmp(lst[i].sin_addr,&haddr,4)) break;
  }
  if (i < nlist) return 0; /* already in list */
  if (nlist == MaxNumAllowedIPNets) ccerr(resources_exhausted);
  haddr = inet_addr(scratch);
  lst[i].sin_family = AF_INET;
  lst[i].sin_port = htons((UINT16)gtGCastPort); /* potential broadcasting */
  memcpy(&lst[i].sin_addr,&haddr,4);
  nlist++;
  *nr = nlist;
# endif
err:
  return cc;
}
int getIPControlNets(char *fn,int *nr,struct sockaddr_in **list)
{
  int size,n;
  FILE *fp = NULL;
  DWORD haddr;
  int cc=0,done=0,col=-1;
  char str[256],hdr[256],scratch[32],ipnetsfilename[128];
  struct sockaddr_in *lst;

  n = 0;
  if ((lst=*list) != NULL) SystemFree((char *)lst);

  /* look for registerd properties: (must be a local file) */
  sprintf(ipnetsfilename,"%s%s",FecDBpath,fn);
  if ((size=GetDataBaseSize(ipnetsfilename)) <= 0)
  {
    if (StartupDebug) printf("No registered IP nets!\n>");
  }
  else
  {
    lst = (struct sockaddr_in *)SystemCalloc(sizeof(struct sockaddr_in),size);
    if (lst == NULL) return out_of_server_memory;
    if ((fp=fopen(ipnetsfilename,"r")) == NULL) ccerr(no_such_file);
    for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
    {
      strtrm(str);
      if (strchr("%;#\n",str[0])) continue;    /* comment line(s) */
      if (!done)
      {
        strcpy(hdr,str);
        if ((col=findcol(hdr,"SUBNET")) < 0) ccerr(no_such_line);
        done = 1;
        continue;
      }
      getColumnStr(col,str,scratch,16); 
      haddr = inet_addr(scratch);
      lst[n].sin_family = AF_INET;
      lst[n].sin_port = htons((UINT16)gtGCastPort); /* potential broadcasting */
      memcpy(&lst[n].sin_addr,&haddr,4);
      n++;
    }
    if (n != size)
    {
      feclog("Illegal Number of IP nets!");
      n = 0;
      ccerr(database_not_loaded);
    }
  }
err:
  if (fp) fclose(fp);
  *nr = n;
  *list = lst;
  return cc;
}
int setupTCPServerSocket(int port)
{
  struct sockaddr_in sck_name;
  int optval = 1, optlen = sizeof(int), mode, cc = 0;
  int sck = 0;
  struct linger to;
  char txt[50];

  mode = 1;
  memset(&sck_name,0,sizeof(sck_name));
  sck_name.sin_addr.s_addr = 0;
  sck_name.sin_family = PF_INET;
  sck_name.sin_port = htons((UINT16)(port+PortOffset));
  sprintf(txt,"TCP PORT %d: ",port+PortOffset);

  if ((sck=(int)socket(PF_INET,SOCK_STREAM,0)) < 0) soerr("socket");
  if (setsockopt(sck,SOL_SOCKET,SO_REUSEADDR,(char *)&optval,optlen)) soerr("REUSEADDR");
  to.l_onoff = 0; to.l_linger = 0;
  if (setsockopt(sck,SOL_SOCKET,SO_LINGER,(char *)&to,sizeof(to))) soerr("LINGER");
  if (bind(sck,(struct sockaddr *)&sck_name,sizeof(sck_name))) soerr("bind");
  if (listen(sck,NumTCPChan) < 0) soerr("listen");
  if (ServerCycleMode == POLLING && soioctl(sck,FIONBIO,(char *)&mode))
    soerr("NONBLOCKING");

  feclog("%s initialized",txt);
err:
  if (cc != 0)
  {
    feclog("%s socket error",txt);
    soperror(txt);
    if (sck > 0) soclose(sck);
    sck = -1;
  }
  return sck;
}
static CLN IncomingTCPClient;
int initTCP(void)
{
  int optval = 1, optlen = sizeof(int), cc = 0;
# ifndef NOT_MULTICASTING
# ifdef WINNT
  int ttl = gtMCastTTL;
  char ipcfn[32];
# else
  BYTE ttl = gtMCastTTL;
# endif
# endif
  char txt[50];
# if defined(UNIX) && !defined(NO_REMOTE_PIPE)
  char ipcfn[32];
# endif

  if (initIP()) return -1;
  strcpy(txt,"initTCP() : ");
  feclog("Using Port Offset %d",PortOffset);
  TCPProducerSocket = setupTCPServerSocket(gtTCPPort);
  if (allowLegacyConnections)
  {
    memset(&IncomingTCPClient,0,sizeof(CLN));
    TCPParent = setupTCPServerSocket(TCP_LEGACYPORT);
  }
  FD_ZERO(&UDPset); FD_ZERO(&rwset); FD_ZERO(&conset); FD_ZERO(&ext_set);
  if (listenUDP(&ProducerSocket,(UINT16)(gtUDPPort+PortOffset)) != 0) soerr("Listen ProducerSocket");
# ifndef MSDOS
  if (setsockopt(ProducerSocket,SOL_SOCKET,SO_BROADCAST,(char *)&optval,optlen)) soerr("BROADCAST");
# endif /* MSDOS */
# ifndef NOT_MULTICASTING
  if (setsockopt(ProducerSocket,IPPROTO_IP,IP_MULTICAST_TTL,(char *)&ttl,sizeof(ttl))<0) soerr("MULTICAST");
# endif /* NOT_MULTICASTING */
  FD_SET(ProducerSocket,&UDPset);
  if (gRespondToServiceRequests)
  {
    if (listenUDP(&NetServiceSocket,(UINT16)(gtNetSrvPort)) != 0) soerr("Listen NetServiceSocket");
    FD_SET(NetServiceSocket,&UDPset);
  }
  UDPto.tv_sec=0;  UDPto.tv_usec=0;
  if (listenUDP(&SyncSocket,0) != 0) soerr("Listen SyncSocket");
  getIPControlNets(IPNETS_FILE,&NrOfIPNets,&IPNetList);
  getIPControlNets(IPBCASTNETS_FILE,&NrOfIPBcastNets,&IPBcastList);
  TCPloaded = 1;
# ifdef WINNT
  sprintf(ipcfn,"%s%s.ipc",IPC_PATH,gFecName);
  if ((IPCSocket=CMDlisten(ipcfn)) == 0) 
  { 
    feclog("Cannot start IPC Remote Server");
  }
  else 
  {
    feclog("IPC server listening on %s",ipcfn);
  }
# endif
# ifdef UNIX
# ifndef NOTTY
  if ((foreground=isatty(STDIN_FILENO)) != 0)
  {
    if (StartupDebug) printf("running in foreground\n");
  }
# else
  foreground = 0;
# endif /* NOTTY */
# ifndef NO_REMOTE_PIPE
  sprintf(ipcfn,"%s%s.ipc",IPC_PATH,gFecName);
  if ((IPCSocket=CMDlisten(ipcfn)) < 0) 
  { 
    perror("IPC listen"); printf("return %d\n>",IPCSocket); 
    feclog("Cannot start IPC Remote Server (error %d)",IPCSocket);
  }
  else 
  {
    feclog("IPC server listening on %s (fd %d)",ipcfn,IPCSocket);
  }
  if (bsdsignal(SIGPIPE, sig_pipe) == SIG_ERR) printf("signal error\n");
  if (bsdsignal(SIGHUP, sig_pipe) == SIG_ERR) printf("signal error\n");
# endif /* NO_REMOTE_PIPE */
# ifndef NO_TINE_PIPE
  sprintf(gIpcTinePipe,"%s%s-server.ipc",IPC_PATH,gFecName);
  if ((ipcProducerSocket=CMDlisten(gIpcTinePipe)) < 0)
  {
    perror("IPC Producer Socket"); printf("return %d\n",ipcProducerSocket);
    feclog("Cannot start IPC Server (error %d)",ipcProducerSocket);
  }
  else
  {
    feclog("IPC Producer socket listening on %s (fd %d)",gIpcTinePipe,ipcProducerSocket);
  }
# endif /* NO_TINE_PIPE */
# endif /* UNIX */
  if (maxTCPConnections > maxFdSets + MIN_FDSETS) maxFdSets = maxTCPConnections + MIN_FDSETS;
  if ((TCPsck=(int *)SystemCalloc(maxTCPConnections,sizeof(int))) == NULL)
  {
    feclog("TCP connection buffer: couldn't allocate %d connections",maxTCPConnections);
  }
err:
  if (cc != 0)
  {
    soperror(txt);
    feclog("%s socket error (%d)",txt,soerrno);
    exit(1); /* this is serious! don't try to start */
  }
  return cc;
}
#ifdef INCLUDE_LEGACY_CODE
void fulfillTCPStreamRequest(int socket)
{
  if (!allowLegacyConnections) return;
  SRVDataBuffer = NULL;
  TCPcount++;
  if (GetTCPStream(socket,&IncomingTCPClient)) goto out;
  execRequest(&IncomingTCPClient);
  SendTCPData(socket,srvWorkArea,srvWorkAreaSize);
  CloseTCPStream(socket);
out: if (SRVDataBuffer != NULL) SystemFree((char *)SRVDataBuffer);
  return;
}
#endif /* INCLUDE_LEGACY_CODE */
#ifdef MSDOS
#pragma argsused
#endif
int IPDataReady(int socket,time_t interim,int timeout) /* return TRUE if data are ready */
{
  int ntry=0,nready;
  fd_set fdrwset;

  if (ServerCycleMode == POLLING) return 1;

  FD_ZERO(&fdrwset); FD_SET(socket,&fdrwset);
  fec_to.tv_sec=timeout;  fec_to.tv_usec=0;
  while ((nready=select(maxFdSets,&fdrwset,NULL,NULL,&fec_to))!=1)
  {
    if (nready == 0) return 0;
    if ((soerrno == EINTR && (ntry++)<3)) continue; /* int. by signal */
    soperror("select");
    return 0;
  }
  return 1;
}
#ifdef INCLUDE_LEGACY_CODE
int GetTCPStream(int socket,CLN *cln)
{
  struct sockaddr_in addr;
  int TimeOut,addrlen = sizeof(addr);
  short cc=0,size=0,b=0,d;
  time_t interim = time(NULL);

  while (b<2)
  {
    if (!IPDataReady(socket,interim,SERVER_TIMEOUT)) ccerr(connection_timeout);
    if ((d=soread(socket,(char *)srvWorkArea,2-b)) > 0) b += d;
  }
  if ((size=PeekWord(*(short *)srvWorkArea)) <= 0) ccerr(invalid_data);
  if ((SRVDataBuffer=(char *)SystemCalloc(1,size))==NULL) ccerr(out_of_server_memory);
  *(short *)SRVDataBuffer = size;

  SetTimeOut(TimeOut);                            /* dynamic timeout */
  while (b<size)
  {
    if (!IPDataReady(socket,interim,TimeOut)) ccerr(connection_timeout);
    if ((d=soread(socket,(char *)&SRVDataBuffer[b],size-b)) > 0) b += d;
  }
err:
  if (getpeername(socket,(struct sockaddr *)&addr,&addrlen)) memset(&addr,0,sizeof(addr));
  memcpy(&cln->IPaddress,&addr,sizeof(addr));
  if (NGdebug)
  {
    inet_ntoa_b(addr.sin_addr,clientaddr);
    dbglog("TCP: [%s] (%d bytes)",clientaddr,size);
  }
  if (cc)
  {
    if (NGdebug)
    {
      if (soerrno) soperror("TCP: read stream");
      dbglog("RPC communication error: %s",erlst[cc]);
    }
    soclose(socket);
  }
  return cc;
}
#endif
int AssignConnectionSocket(void)
{
  int sck,len = sizeof(struct sockaddr);
  struct sockaddr addr;
  TCPBCKT *bckt;

  if (TCPProducerSocket == 0) return -1;
  if ((sck=(int)accept(TCPProducerSocket,&addr,&len)) > 0)
  {
    if (nTCPsck < maxTCPConnections)
    {
      if ((bckt=getBucket(sck)) != NULL)
      {
        bckt->idx = -1;
        TCPsck[nTCPsck] = sck;
        return nTCPsck++;
      }
    }
    soclose(sck);
  }
  return -1;
}
#ifdef INCLUDE_LEGACY_CODE
void AcceptTCPStream(void)
{
  int len = sizeof(struct sockaddr);
  static int i = -1;
  struct sockaddr addr;

  i = (++i)%NumTCPChan;

  if ((TCPsocket[i]=accept(TCPParent,&addr,&len)) == -1) return;
  fulfillTCPStreamRequest(TCPsocket[i]);

  return;
}
#endif
CLN *getNetworkClientFromList(void)
{
  int i;

  for (i=0; i<nconsumer; i++)
  {
    if (strcmp(Consumers[i]->userName,"NETWORK")) continue;
    if (Consumers[i]->IPaddress.sin_port != htons((UINT16)gtMCastPort)) continue;
    return Consumers[i];
  }
  return NULL;
}
CLN *LocateConsumerInList(struct sockaddr_in *ip,IPXAddress *nwAddr,BYTE *immAddr)
{
  int i, hWnd = 0;

  if (ip == NULL && nwAddr == NULL) 
  {
    if (immAddr == NULL) return NULL;
#   ifdef MAKE_TINE32_DLL
    hWnd = *(int *)immAddr;          /* reuse this parameter for local windows handle */
#   endif
  }
  for (i=0; i<nconsumer; i++)        /* is client in client list ?     */
  {
    if (ip != NULL && !memcmp(ip,&Consumers[i]->IPaddress,8)) break;
    if (nwAddr != NULL && !memcmp(nwAddr,&Consumers[i]->IPXaddress,12)) break;
    if (hWnd != 0 && Consumers[i]->sck == hWnd) break;
  }
  if (i < nconsumer) return Consumers[i];
  if (i >= MaxNumClients ||
     (Consumers[i]=(CLN *)SystemCalloc(1,sizeof(CLN))) == NULL)
    return NULL;
  nconsumer++;
  if (ip != NULL) 
  {
    memcpy(&Consumers[i]->IPaddress,ip,sizeof(struct sockaddr_in));
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    if (ip->sin_family == AF_UNIX) nipcconsumers++;
#   endif
  }
  if (nwAddr != NULL) memcpy(&Consumers[i]->IPXaddress,nwAddr,12);
  if (hWnd == 0 && immAddr != NULL) memcpy(Consumers[i]->node,immAddr,6);

  return Consumers[i];
}
#ifdef UNIX
struct sockaddr_in *whichUnixAddr(int sck)
{
  int i;
  for (i=0; i<nconsumer; i++)
  {
    if (Consumers[i]->sck == sck &&
        Consumers[i]->IPaddress.sin_family == AF_UNIX)
      return &Consumers[i]->IPaddress;      
  }
  return NULL;
}
int isUnixSocket(int sck)
{
  int i;
  if (sck == ipcProducerSocket) return TRUE;
  for (i=0; nipcconsumers > 0 && i<nconsumer; i++)
  {
    if (Consumers[i]->sck == sck &&
        Consumers[i]->IPaddress.sin_family == AF_UNIX) return TRUE;       
  }
  for (i=0; i<RPCNumFecsRead; i++) if (RPCFec[i].Prot == PIPE) 
  {
    if (!memcmp(RPCFec[i].Network,&sck,4)) return TRUE;
  }
  return FALSE;
}
int AcceptIPC(int sck)
{
  short cc = -not_implemented;
# ifndef NO_TINE_PIPE
  TCPBCKT *bckt;
  CLN *cln = NULL;
  struct sockaddr_in IPconsumer;
  int len = sizeof(IPconsumer);
  int s;
  static UINT32 addr = 0;
# ifndef SVR4
  struct sockaddr_un UXconsumer;
  struct stat statbuf;
# else
  struct strrecvfd rfd;
# endif /* SVR4 */
  if (addr == 0) addr = getmyipaddr();
  if (sck == ipcProducerSocket) 
  {
#   ifndef SVR4
    len = sizeof(UXconsumer);
    if ((s=accept(sck,(struct sockaddr *)&UXconsumer,&len)) < 0) ccerr(-not_accepted);
    strncpy(clientaddr,UXconsumer.sun_path,14); clientaddr[14] = 0;
    if (stat(clientaddr,&statbuf) < 0) ccerr(net_read_error);
    unlink(clientaddr);
#   else
    if (ioctl(sck,I_RECVFD,&rfd) < 0) ccerr(-1);
    s = rfd.fd;
#   endif /* SVR4 */
    memset(&IPconsumer,0,sizeof(IPconsumer));
    memcpy(&IPconsumer.sin_addr,&addr,4);
    IPconsumer.sin_family = AF_UNIX;
    if ((cln=LocateConsumerInList(&IPconsumer,NULL,NULL)) == NULL)
      ccerr(out_of_server_memory);
    cln->sck = s;
    cln->transport = PIPE;
    cc = 0;
  }
  if ((bckt=getBucket(s)) != NULL) bckt->idx = -1;
# endif /* NO_TINE_PIPE */
err:
  if (cc)
  {
    if (NGdebug)
    {
      sprintf(&dbgbuf[strlen(dbgbuf)],"IPC ERR: %s (sck %d)",erlst[cc],sck);
      dbglog(dbgbuf);
    }
  }
  return cc;
}
#endif
CLN *getConsumerIPAddress(int sck,int prot)
{
  CLN *cln = NULL;
  struct sockaddr_in IPconsumer;
  int retry=2,len = sizeof(IPconsumer);
  short nread,nleft,n,cc = -1;
  char *strprot = getProtocolAsString(prot);
  fd_set fds;
  struct timeval to;
  TCPBCKT *bckt;
  BYTE *bufptr=srvWorkArea;

  /* don't block if there are stale data : */
  /* POLLING mode bypasses getNextCycleTimeout() ! */
  FD_ZERO(&fds); FD_SET(sck,&fds); 
  to.tv_sec = 0;
  to.tv_usec = 0;

retry_select:
  if (select(maxFdSets,&fds,NULL,NULL,&to) == -1)
  {
    if (soerrno != EWOULDBLOCK) soperror("select");
    if (retry-- > 0) goto retry_select;
    ccerr(tcp_socket_error);
  }
  if (!FD_ISSET(sck,&fds)) goto err;

  cc = 0;
  if (prot == UDP) UDPcount++;
  if ((nread=readfrom(sck,(char *)srvWorkArea,srvWorkAreaSize,0,
             (struct sockaddr *)&IPconsumer,&len)) <= 0) ccerr(net_read_error);
  switch (prot)
  {
    case UDP:
      UDPcount++;
      inet_ntoa_b(IPconsumer.sin_addr,clientaddr);
      IPconsumer.sin_family = PF_INET;
      break;
    case TCP: 
      TCPcount++;      
      len = sizeof(IPconsumer);
      getpeername(sck,(struct sockaddr *)&IPconsumer,&len);
      inet_ntoa_b(IPconsumer.sin_addr,clientaddr);
      IPconsumer.sin_family = PF_INET;
      break;
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    case PIPE:
      strncpy(clientaddr,"localhost",16);
      IPconsumer = *whichUnixAddr(sck);
      IPconsumer.sin_family = AF_UNIX;
      break;
#   endif      
  }
  if (NGdebug) sprintf(dbgbuf,"%s: [%s] ",strprot,clientaddr);
  if (prot == TCP || prot == PIPE)
  {
    if ((bckt=findBucket(sck)) == NULL) ccerr(tcp_socket_error);
    nleft = nread;
    while (nleft > 0)
    {
      if (bckt->ptr == 0) bckt->siz = PeekWord(*(short *)bufptr);
      n = bckt->siz - bckt->ptr;
      if (nleft >= n)
      {
        memcpy(&bckt->buf[bckt->ptr],bufptr,n);
        if ((cln=LocateConsumerInList(&IPconsumer,NULL,NULL)) == NULL)
          ccerr(out_of_server_memory);
        cln->sck = sck;  cln->transport = prot;
        getConsumerRequest(cln,bckt->buf,UDP_BUFFER_SIZE);
        bufptr += n;
        nleft -= n;
        bckt->ptr = 0;
      }
      else
      {
        memcpy(&bckt->buf[bckt->ptr],bufptr,nleft);
        bckt->ptr += nleft;
        nleft = 0;
      }
    }
    ccerr(0);
  }
  if (nread < PeekWord(*(short *)srvWorkArea)) ccerr(net_read_error);
  if ((cln=LocateConsumerInList(&IPconsumer,NULL,NULL)) == NULL) ccerr(out_of_server_memory);
  cln->sck = sck;
  cln->transport = prot;
  getConsumerRequest(cln,srvWorkArea,UDP_BUFFER_SIZE);
err:
  if (cc > 0)
  {
    if ((prot == TCP) && (ServerCycleMode != POLLING || soerrno != EWOULDBLOCK))
      CloseConnectionSocket(sck);
    if (NGdebug)
    {
      sprintf(&dbgbuf[strlen(dbgbuf)],"%s ERR: %s (sck %d)",strprot,erlst[cc],sck);
      dbglog(dbgbuf);
    }
  }
  return cln;
}
void dumpIPNets(void)
{
  int n;

  dbglog("Registered IP Subnets (broadcasting %s):\n",
          BroadcastServer ? "enabled" : "disabled" );
  for (n=0; n<NrOfIPNets; n++)
  {
    inet_ntoa_b(IPNetList[n].sin_addr,clientaddr);
    dbglog("\t[%s]\n",clientaddr);
  }
}
int IsMemberIPNets(int num,struct sockaddr_in *lst,struct sockaddr_in *ipaddr)
{
  SINT32 targetnet,targetnode,cnode, sinaddr;
  int n;

  if (num == 0) return TRUE;  /* everything's open */
  if (ipaddr == NULL) return FALSE;
  sinaddr = LSWAP(ipaddr->sin_addr.s_addr);
  targetnet  = (sinaddr)&0xffffffL;
  targetnode = ((sinaddr) >> 24) & 0xff;
  for (n=0; n<num; n++)
  {
    sinaddr = LSWAP(lst[n].sin_addr.s_addr);
    if ((SINT32)(sinaddr&0xffffffL) != targetnet) continue;
    if ((cnode=((sinaddr >> 24)&0xff)) == 0xff) return TRUE;
    if (cnode == targetnode) return TRUE;
  }
  return FALSE;
}
int SendIPBroadcast(BYTE *buf,int size)
{
  int n;
  fd_set fdwset;
  struct timeval wait_to;
  struct sockaddr *sa;

  if (!TCPloaded) return tcp_not_supported;
  for (n=0; n<NrOfIPBcastNets+1; n++)
  {
    if (NGdebug > 1) 
    {
      if (n == 0) sprintf(dbgbuf,"UDP broadcast : ");
      if (n == NrOfIPBcastNets) sprintf(dbgbuf,"UDP multicast : ");
    }
    FD_ZERO(&fdwset); FD_SET(ProducerSocket,&fdwset);
    wait_to.tv_sec=0;  wait_to.tv_usec=0;
    switch (select(maxFdSets,NULL,&fdwset,NULL,&wait_to))
    {
      case 1: break;
      case 0: 
        if (NGdebug > 1) dbglog("socket %d: write set not ready",ProducerSocket);
        return tcp_socket_error;
      default:if (NGdebug > 1) soperror("select");
        return tcp_socket_error;
    }
    sa = n < NrOfIPBcastNets ? (struct sockaddr *)&IPBcastList[n] : (struct sockaddr *)&tGCastAddrSck;
    if (sendto(ProducerSocket,(char *)buf,size,0,sa,sizeof(struct sockaddr_in)) < 0)
    {
      dbglog("UDP: broadcast failed");
      return tcp_socket_error;
    }
    else if (NGdebug > 1)
    {
      inet_ntoa_b(((struct sockaddr_in *)sa)->sin_addr,clientaddr);
      sprintf(&dbgbuf[strlen(dbgbuf)],"%s ",clientaddr);
    }
  }
  if (NGdebug > 1) dbglog(dbgbuf);

  return 0;
}
#ifdef INCLUDE_LEGACY_CODE
void CloseTCPStream(int socket)
{
  if (!allowLegacyConnections) return;
  if (soclose(socket) && NGdebug > 1)
  {
    soperror("TCP: soclose"); printf("\n>");
  }
  return;
}
#endif
void closeIPServer(void)
{
  if (!TCPloaded) return;
# ifdef INCLUDE_LEGACY_CODE
  if (TCPParent && shutdown(TCPParent,2)) soperror("TCP shutdown"); 
  TCPParent = 0;
# endif
  if (TCPProducerSocket && shutdown(TCPProducerSocket,2)) soperror("TCP shutdown");
  TCPProducerSocket = 0;
  if (ProducerSocket && soclose(ProducerSocket)) soperror("UDP soclose");
  ProducerSocket = 0;
  if (NetServiceSocket && soclose(NetServiceSocket)) soperror("UDP soclose");
  NetServiceSocket = 0;
  if (SyncSocket && soclose(SyncSocket)) soperror("UDP soclose");
  SyncSocket = 0;
  return;
}
int gSelectOnSysPoll = TRUE;
struct timeval *getNextCycleTimeout(struct timeval *tv)
{
  unsigned int deadtime;
  static short firstCycle = 1;
  static struct timeval lastCycleTime;

/*  if (nconsumer == 0) return (struct timeval *)NULL; */

  if (firstCycle) { lastCycleTime = *tv; firstCycle = 0; }
  /* don't block if there are stale data : */
  if (StaleData || ReturnToClientCycle || ReturnToServerCycle)
  {
    fec_to.tv_sec = 0;
    fec_to.tv_usec = 0;
    ReturnToServerCycle = 0;
    return &fec_to;
  }
  deadtime = MSECS(*tv,lastCycleTime);
  if (gSelectOnSysPoll) deadtime = 0;
  lastCycleTime = *tv;
  if (deadtime >= (UINT32)SysPoll) deadtime = SysPoll;
  if ((fec_to.tv_sec=(SysPoll-deadtime)/1000) < 0)
  {
    fec_to.tv_sec = 0;
    fec_to.tv_usec = 1000;
  }
  else fec_to.tv_usec = ((SysPoll-deadtime)%1000)*1000;
  return &fec_to;
}
int IsInForeground(void)
{
# ifdef UNIX
  struct timeval to;
  fd_set fdrwset;

  if (ServerCycleMode == POLLING) return -1;
  if (!foreground) return 0;
  FD_ZERO(&fdrwset); FD_SET(STDIN_FILENO,&fdrwset);
  to.tv_sec=0;  to.tv_usec=0;
  if (select(maxFdSets,&fdrwset,NULL,NULL,&to) < 0) return 0;
# endif
  return -1;
}
void AcceptIP(struct timeval *tv)
{
  int cc = 0,retry=2,i,nready;
  struct timeval *to;
  char cmdbuf[CMDSIZE];
# if defined(UNIX) || defined(WINNT)
  int j, len;
# endif

  if (!TCPloaded) return;

  if (ServerCycleMode != POLLING)
  { /* fill in the rw select()ion set : */
    FD_ZERO(&rwset);
    rwset = conset;
#   ifdef INCLUDE_LEGACY_CODE
    /* to go to the chopping block in release 4.0 */
    if (TCPParent) FD_SET(TCPParent,&rwset);
#   endif
    if (TCPProducerSocket) FD_SET(TCPProducerSocket,&rwset);
    for (i=0; i<nTCPsck; i++) FD_SET(TCPsck[i],&rwset);
    FD_SET(ProducerSocket,&rwset);
    if (NetServiceSocket) FD_SET(NetServiceSocket,&rwset);
#   ifdef UNIX
    /* stdin is a normal fd in the unix world :*/
    if (foreground) FD_SET(STDIN_FILENO,&rwset);
#   ifndef NO_REMOTE_PIPE
    /* let remote process attach to a server running in the background */
    if (IPCSocket > 0) FD_SET(IPCSocket,&rwset);
    for (i=0; i<nipcclients; i++) FD_SET(IPCfds[i],&rwset);
#   endif /* NO_REMOTE_PIPE */
#   ifndef NO_TINE_PIPE
    /* client on the same machine as the server ? */
    if (ipcProducerSocket > 0) FD_SET(ipcProducerSocket,&rwset);
    for (i=0; nipcconsumers > 0 && i<nconsumer; i++) 
      if (Consumers[i]->transport == PIPE && Consumers[i]->sck > 0)
        FD_SET(Consumers[i]->sck,&rwset);
#   endif /* NO_TINE_PIPE */
#   endif /* UNIX */
#   ifdef IPXBSD
    /* anybody still using IPX ? */
    if (IPXProducerSocket) FD_SET(IPXProducerSocket,&rwset);
#   endif
    /* allow external socket 'hooks' beyond our control */
    if (externalFdSets) FD_MERGE(ext_set,&rwset);
    to = getNextCycleTimeout(tv);
#   if defined(WINNT) && !defined(NO_REMOTE_PIPE)
    /* Windows Pipes are not select()able, but handle them here anyway */
    /* Note: this can't work on 64bit windows as IPCSocket is a 4byte integer */
    if (CMDaccept((PIPEINST *)IPCSocket) != 0)
    {
      if ((len=GetCommandEx(IPCSocket,cmdbuf,CMDSIZE)) > 0)
        InterpretCommand(cmdbuf);
      j = 0;
    }
#   endif /* WINNT */
retry_select:
    /* anything fall through select() ? */
    nready = select(maxFdSets,&rwset,NULL,NULL,to);
    gSelectOnSysPoll = TRUE;
    /* is there a hook ? */
    if (socketAcceptEntryFcn) (socketAcceptEntryFcn)();
    if (nready < 0 && soerrno == EINTR && retry-- > 0)
    {
      to->tv_sec = 0; to->tv_usec = 0;
      goto retry_select;
    }
    if (nready <= 0) ccerr(nready);
    gSelectOnSysPoll = FALSE;
#   ifdef UNIX
#   ifndef NO_TINE_PIPE
    /* local client ? */
    if (ipcProducerSocket > 0 && FD_ISSET(ipcProducerSocket,&rwset)) 
      AcceptIPC(ipcProducerSocket);
    for (i=0; nipcconsumers > 0 && i<nconsumer; i++) 
      if (Consumers[i]->transport == PIPE)
    {
      if (FD_ISSET(Consumers[i]->sck,&rwset))
      {
        if (getConsumerIPAddress(Consumers[i]->sck,PIPE) == NULL)
        {
          close(Consumers[i]->sck);
          Consumers[i]->sck = 0;
        }
      }
    }
#   endif /* NO_TINE_PIPE */
    /* stdin ? */
    if (foreground && FD_ISSET(STDIN_FILENO,&rwset) &&
        GetCommand(STDIN_FILENO,cmdbuf,CMDSIZE) > 0) InterpretCommand(cmdbuf);
#   ifndef NO_REMOTE_PIPE
    /* attached process issuing a command ? */
    if (IPCSocket > 0 && FD_ISSET(IPCSocket,&rwset)) addIPCcln();
    for (i=0; i<nipcclients; i++)
    {
      if (FD_ISSET(IPCfds[i],&rwset))
      {
        if ((len=GetCommand(IPCfds[i],cmdbuf,CMDSIZE)) > 0)
          InterpretCommand(cmdbuf);
        if (len == 0)
        {
          close(IPCfds[i]);
          FD_CLR(IPCfds[i],&rwset);
          for (j=i; j<nipcclients-1; j++) IPCfds[j] = IPCfds[j+1];
          i--; nipcclients--;
          if (NGdebug > 1) printf("Remote session closed\n>");
        }
      }
    }
#   endif /* NO_REMOTE_PIPE */
#   endif /* UNIX */
  }
# ifdef INCLUDE_LEGACY_CODE
  /* hope we can get rid of this one day ... */
  if (allowLegacyConnections)
  {
    if (ServerCycleMode == POLLING || FD_ISSET(TCPParent,&rwset))
      AcceptTCPStream();
  }
# endif
  /* is someone trying to get a connection ? */
  if (ServerCycleMode == POLLING || FD_ISSET(TCPProducerSocket,&rwset))
  {
    if ((i=AssignConnectionSocket()) > 0) FD_SET(TCPsck[i],&rwset);
  }
  /* has something has come in on a connected socket ? */
  for (i=0; i<nTCPsck; i++)
    if (ServerCycleMode == POLLING || FD_ISSET(TCPsck[i],&rwset))
      getConsumerIPAddress(TCPsck[i],TCP);
  /* has something has come in on a UDP socket ? */
  if (ServerCycleMode == POLLING || FD_ISSET(ProducerSocket,&rwset))
    getConsumerIPAddress(ProducerSocket,UDP);
  /* has something has come in on the service socket ? */
  if (NetServiceSocket != 0)
    if (ServerCycleMode == POLLING || FD_ISSET(NetServiceSocket,&rwset))
      getConsumerIPAddress(NetServiceSocket,UDP);

err:
  if (cc < 0 && NGdebug > 1) soperror("select");
  /* make sure these descriptors carry over to the client side */
  FD_ZERO(&rwset);
# ifdef INCLUDE_LEGACY_CODE
  if (TCPParent) FD_SET(TCPParent,&rwset);
# endif
  if (TCPProducerSocket) FD_SET(TCPProducerSocket,&rwset);
  for (i=0; i<nTCPsck; i++) FD_SET(TCPsck[i],&rwset);
  FD_SET(ProducerSocket,&rwset);
  if (NetServiceSocket) FD_SET(NetServiceSocket,&rwset);
# if defined(UNIX) && !defined(NO_REMOTE_PIPE)
  if (IPCSocket > 0) FD_SET(IPCSocket,&rwset);
  for (i=0; i<nipcclients; i++) FD_SET(IPCfds[i],&rwset);
# endif
# ifdef IPXBSD
  if (IPXProducerSocket) FD_SET(IPXProducerSocket,&rwset);
# endif
  if (externalFdSets) FD_MERGE(ext_set,&rwset);

  gettimeofday(tv,(struct timezone *)NULL);
  return;
}
#ifdef MSDOS
#pragma argsused
#endif
void CleanupIP(struct timeval *tv)
{
  return;
}
