/** \file srvipx.c
 * \brief IPX routines and definitions
 * \internal
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "prolog.h"
#include "srvipx.h"
#include "brdcast.h"
#include "database.h"
#include "errors.h"
#ifdef IPXBSD
#include <netipx/ipx.h>
#include "srvip.h"
#endif

extern UINT16 PortOffset;
int GetSPXStream(WORD ID,CLN *cln);
void CloseSPXStream(WORD ID);
int SendSPXData(WORD ID,int socket);
void addIPXNet(BYTE *net,BYTE *mac,BYTE *node);
void dumpIPXNets(void);

struct IPXNet *IPXNets = NULL;
extern int BroadcastServer;
extern IPXAddress IPXFecAddress;
extern char FecDBpath[80];
extern int allowLegacyConnections;

BYTE rpc_socket[2] = {RPC_SOCKET,RPC_SOCKET};
const BYTE brdcst_socket[2] = {BRDCST_SOCKET,BRDCST_SOCKET};
BYTE reg_socket[2] = {REG_SOCKET,REG_SOCKET};
const BYTE mcast_socket[2] = {MCAST_SOCKET,MCAST_SOCKET};

int IPXProducerSocket = 0;
int IPXBroadcastSocket = 0;
#ifdef IPXNW
int RpcSendSocket;

IPXECB  *brdcstECB,
        *glbECB[NUM_GLB_ECB],
        *regECB[NUM_REG_ECB],
        *PCECB;                  /* Producer/Consumer IPX Send ECB */

SPXECB  *rcvECB[NUM_RCV_ECB],
        *lsnECB[NUM_LSN_ECB],
        *sndECB,
        *conECB;

WORD ISwap(WORD w) { return (w >> 8) | (w << 8); }

IPXECB *CreateIPXECB(int socket)
{
  IPXECB *ecb;

  if ((ecb=(IPXECB *)SystemCalloc(1,sizeof(IPXECB))) != NULL)
    initIPXECB(ecb,socket);
  return ecb;
}
SPXECB *CreateSPXECB(int socket)
{
  SPXECB *ecb;

  if ((ecb=(SPXECB *)SystemCalloc(1,sizeof(SPXECB))) != NULL)
    initSPXECB(ecb,socket);
  return ecb;
}
WORD PCSocket = 0;
#define nwAddr (ecb->hdr.source)
#define immAddr (ecb->ecb.immediateAddress)
#else
fd_set ipxset;

int ipxCreateSocket(UINT32 network, short port)
{
  int s,optval,optlen = sizeof(int),rc,nlen;
  struct sockaddr_ipx ipxaddr;
  char str[80];

  sprintf(str,"IPX: ERR socket init");
  if ((s=socket(AF_IPX, SOCK_DGRAM, 0)) < 0)
  {
    perror(str);
    feclog(str);
  }
  else
  {
    optval = 1;
    sprintf(str,"IPX: ERR setsockopt");
    if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&optval,optlen)) 
    {
      perror(str);
      feclog(str);
    }
    ipxaddr.sipx_family = AF_IPX;
    ipxaddr.sipx_network = htonl(network);
    ipxaddr.sipx_port = htons(port);
    ipxaddr.sipx_type = IPX_FRAME_ETHERII;
    sprintf(str,"IPX: ERR bind");
    if ((rc=bind(s, (struct sockaddr *)&ipxaddr,sizeof(ipxaddr))) < 0)
    {
      perror(str);
      feclog(str);
      close(s);
      s = -1;
    }
  }
  if (s >= 0)
  {
    sprintf(str,"IPX: open socket %d",s);
    feclog(str);
  }
  return s;
}
#endif

BYTE IPXbrdcstNode[6] = {0xff,0xff,0xff,0xff,0xff,0xff};

IPXNetListStruct *IPXNetList = NULL;
int NrOfIPXNets = 0;
IPXNetListStruct *IPXBcastList = NULL;
int NrOfIPXBcastNets = 0;

int getIPXControlNets(char *fn,int *nr,IPXNetListStruct **list)
{
  int i,j,n=0,size;
  char ipxnetsfilename[128];
  IPXNetListStruct *lst = NULL;
  FILE *fp = NULL;
  int cc=0,done=0;
  int net_col,nod_col,rou_col;
  char c,str[256],hdr[256],scratch[16],hexstr[64];

  n = 0; lst = *list;
  if (lst != NULL) SystemFree((char *)lst);

  sprintf(ipxnetsfilename,"%s%s",FecDBpath,fn);
  if ((size=GetDataBaseSize(ipxnetsfilename)) <= 0)
  {
    feclog("No registered IPX control nets\n>");
    ccerr(0);
  }
  if ((lst=(IPXNetListStruct *)SystemCalloc(sizeof(IPXNetListStruct),size)) == NULL)
      return out_of_server_memory;
  if ((fp=fopen(ipxnetsfilename,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((net_col=findcol(hdr,"NETWORK")) < 0) ccerr(no_such_column);
      if ((nod_col=findcol(hdr,"NODE")) < 0) ccerr(no_such_column);
      if ((rou_col=findcol(hdr,"ROUTER")) < 0) ccerr(no_such_column);
      done = 1;                              /* read header info */
      continue;
    }
    getColumnStr(net_col,str,scratch,16); 
    strupr(scratch); 
    for (i=0,j=0; i<4 && j<strlen(scratch); i++) 
    {
      if (i<4-strlen(scratch)/2) lst[n].net[i] = 0;
      else
      {
        sprintf(hexstr,"0x%c%c",scratch[j],scratch[j+1]); j += 2;
        lst[n].net[i] = (BYTE)strtol(hexstr,NULL,16);
      }
    }
    getColumnStr(nod_col,str,scratch,16); 
    strupr(scratch); 
    for (i=0,j=0; i<6 && j<strlen(scratch); i++) 
    {
      if (i<6-strlen(scratch)/2) lst[n].mac[i] = 0;
      else
      {
        sprintf(hexstr,"0x%c%c",scratch[j],scratch[j+1]); j += 2;
        lst[n].mac[i] = (BYTE)strtol(hexstr,NULL,16);
      }
    }
    getColumnStr(rou_col,str,scratch,16); 
    strupr(scratch); 
    for (i=0,j=0; i<6 && j<strlen(scratch); i++) 
    {
      if (i<6-strlen(scratch)/2) lst[n].node[i] = 0;
      else
      {
        sprintf(hexstr,"0x%c%c",scratch[j],scratch[j+1]); j += 2;
        lst[n].node[i] = (BYTE)strtol(hexstr,NULL,16);
      }
    }
    n++;
  }
err:
  if (cc) feclog("%s : %s",fn,erlst[cc]);
  *nr = n; *list = lst;
  return cc;
}
int setupIPX(void)
{
# ifdef IPXNW
  BYTE major, minor;
  WORD maxCon, availCon;

  if (IPXInitialize()) return -1;
  if (!SPXInitialize(&major,&minor,&maxCon,&availCon)) return -1;
# endif
  return 0;
}
static CLN IncomingSPXClient;
int initIPX(void)
{
  WORD socket;
  int i, cc = 0;

  if (setupIPX()) ccerr(ipx_not_supported);

  memset(&IncomingSPXClient,0,sizeof(CLN));

  reg_socket[0] = REG_SOCKET + PortOffset;
  reg_socket[1] = REG_SOCKET + PortOffset;
  rpc_socket[0] = RPC_SOCKET + PortOffset;
  rpc_socket[1] = RPC_SOCKET + PortOffset;

# ifdef IPXNW
# ifdef INCLUDE_LEGACY_CODE
  if (allowLegacyConnections)                 /* the old SPX sockets */
  {
    socket=0;                                 /*  Send socket */
    if (IPXOpenSocket((BYTE *)&socket,0)) ccerr(ipx_socket_error);
    RpcSendSocket = socket;

    if ((sndECB=CreateSPXECB(socket)) == NULL) ccerr(out_of_server_memory);
    if ((conECB=CreateSPXECB(0)) == NULL) ccerr(out_of_server_memory);

    socket = *(int *)&rpc_socket[0];          /* FEC RPC server socket */
    if (IPXOpenSocket((BYTE *)&socket,0)) ccerr(link_error);

    for (i=0; i<NUM_LSN_ECB; i++)             /* create listen ECBs for */
    {                                         /* Windows clients        */
      if ((lsnECB[i]=CreateSPXECB(socket)) == NULL) ccerr(out_of_server_memory);
      SPXListenForConnection((BYTE)3,(BYTE)0,&lsnECB[i]->ecb);
    }
    for (i=0; i<NUM_RCV_ECB; i++)             /* create sequenced packet ECBs */
    {
      if ((rcvECB[i]=CreateSPXECB(socket)) == NULL) ccerr(out_of_server_memory);
      SPXListenForSequencedPacket(&rcvECB[i]->ecb);
    }
  }
# endif
  IPXGetInternetworkAddress((BYTE *)&IPXFecAddress);

  /* Support consumer subscriptions:
     Producer/Consumer send ECB & Consumer Receive ECB
     (get immediated address from caller)
  */
  PCSocket = 0;
  if (IPXOpenSocket((BYTE *)&PCSocket,0)) ccerr(ipx_socket_error);
  if ((PCECB=CreateIPXECB(PCSocket)) == NULL) ccerr(out_of_server_memory);

  socket = *(int *)&reg_socket[0];
  if (IPXOpenSocket((BYTE *)&socket,0)) ccerr(ipx_socket_error);
  for (i=0; i<NUM_REG_ECB; i++)  /*/ create subscription registration ECBs */
  {
    if ((regECB[i]=CreateIPXECB(socket)) == NULL) ccerr(out_of_server_memory);
    IPXListenForPacket(&regECB[i]->ecb);
  }
# else  /* not IPXNW but IPXBSD */
  IPXProducerSocket = ipxCreateSocket(0L,*(short *)reg_socket);
  if (IPXProducerSocket < 0) IPXProducerSocket = 0;
# endif

  /* look for broadcasting information: */
  socket = 0;
  getIPXControlNets(IPXNETS_FILE,&NrOfIPXNets,&IPXNetList);
  getIPXControlNets(IPXBCASTNETS_FILE,&NrOfIPXBcastNets,&IPXBcastList);
  if (BroadcastServer)
  {
    socket = 0;
#   ifdef IPXNW
    if (IPXOpenSocket((BYTE *)&socket,0)) ccerr(ipx_socket_error);
    if ((brdcstECB=CreateIPXECB(socket)) == NULL) ccerr(out_of_server_memory);
    memset(brdcstECB->hdr.destination.node,0xff,6);      /*/ IPX broadcast */
    memset(brdcstECB->ecb.immediateAddress,0xff,6);
    memcpy(brdcstECB->hdr.destination.network,IPXFecAddress.network,4);
    *(WORD *)brdcstECB->hdr.destination.socket = *(WORD *)&brdcst_socket[0];
#   else  /* not IPXNW but IPXBSD */
    IPXBroadcastSocket = ipxCreateSocket(0L,socket);
    if (IPXBroadcastSocket < 0) IPXBroadcastSocket = 0;
#   endif
    for (i=0; i<NrOfIPXBcastNets; i++)  /* add them all to broadcast list */
      addIPXNet(IPXBcastList[i].net,IPXBcastList[i].mac,IPXBcastList[i].node);
    if (StartupDebug) dumpIPXNets();
  }
  else if (StartupDebug) printf("No IPX broadcasting from this FEC");
err:
  if (cc)
  {
    feclog("IPX failed to initialize : %s",erlst[cc]);
  }
  else
  {
    feclog("IPX initialized");
    IPXloaded = 1;
  }
  return cc;
}
void closeIPXServer(void)
{
  if (!IPXloaded) return;
# ifdef IPXNW
  IPXCloseSocket(lsnECB[0]->ecb.socketNumber);
  IPXCloseSocket(sndECB->ecb.socketNumber);
  IPXCloseSocket(PCSocket);
# else
  if (IPXProducerSocket) close(IPXProducerSocket);
  if (IPXBroadcastSocket) close(IPXBroadcastSocket);
# endif
  return;
}
void AcceptIPX(struct timeval *tv)
{
# ifndef IPXNW
  struct timeval ipxto;
  int retry = 2,nready;
# endif
  if (!IPXloaded) return;

# ifdef IPXNW
  while (getConsumerIPXAddress() != NULL);
# ifdef INCLUDE_LEGACY_CODE
  if (allowLegacyConnections) AcceptSPXStream();
# endif
# else
  if (IPXProducerSocket == 0) return;
  FD_ZERO(&ipxset);
  FD_SET(IPXProducerSocket,&ipxset);

  /* don't block if there are stale data : */
  /* POLLING mode bypasses getNextCycleTimeout() ! */
  ipxto.tv_sec = 0;
  ipxto.tv_usec = 0;

retry_select:
  if ((nready=select(maxFdSets,&ipxset,NULL,NULL,&ipxto)) == 0) goto err;
  if (nready < 0)
  {
    if (NGdebug > 1) soperror("select");
    if (retry-- > 0) goto retry_select;
    goto err;
  }
  getConsumerIPXAddress();
# endif
err:gettimeofday(tv, (struct timezone *) NULL);
  return;
}

CLN* getConsumerIPXAddress(void)
{
  int i, cc = 0;
  CLN *cln = NULL;
# ifdef IPXNW
  IPXECB *ecb;
  static int ecbcnt = -1;
# else
  int timeout = 1;
  int index;
  int bytes;
  IPXAddress nwAddr;
  BYTE *immAddr = NULL;
  struct sockaddr_ipx IPXconsumer;
  int len = sizeof(IPXconsumer);
# endif

# ifdef IPXNW
  ecbcnt = (++ecbcnt)%NUM_REG_ECB;
  for (i=0; i<NUM_REG_ECB; i++)  /* Cycle through listening ECBs*/
    if (!(ecb=regECB[(i+ecbcnt)%NUM_REG_ECB])->ecb.inUseFlag) break;

  if ((ecbcnt=i) == NUM_REG_ECB) return (CLN *)NULL;
  if (ecb->ecb.completionCode) ccerr(ipx_socket_error);
  memcpy(srvWorkArea,ecb->buf,IPX_BUFFER_SIZE);
# else
  if (IPXProducerSocket == 0) return (CLN *)NULL;
  if (!FD_ISSET(IPXProducerSocket,&ipxset)) return (CLN *)NULL;
  if ((i=recvfrom(IPXProducerSocket,(char *)srvWorkArea,srvWorkAreaSize,0,
          (struct sockaddr *)&IPXconsumer,&len)) < 0) ccerr(ipx_socket_error);
  if (i != PeekWord(*(short *)srvWorkArea)) ccerr(net_read_error);

  IPXconsumer.sipx_family = AF_IPX;
  memcpy(nwAddr.node,IPXconsumer.sipx_node,6);
  memcpy(nwAddr.network,&IPXconsumer.sipx_network,4);
  *(UINT16 *)nwAddr.socket = ntohs(IPXconsumer.sipx_port);
# endif
  IPXcount++;
  if (NGdebug)
  {
    sprintf(clientaddr,ipxnet_ntoa(nwAddr.node));
    sprintf(dbgbuf,"IPX: [%s] ",clientaddr);
  }
  if ((cln=LocateConsumerInList(NULL,&nwAddr,immAddr)) == NULL)
    ccerr(out_of_server_memory);
  cln->sck = IPXProducerSocket;
  cln->transport = IPX;

err: 
  if (cc && NGdebug) dbglog("ERR: %s",erlst[cc]);
# ifdef IPXNW
  ecb->ecb.socketNumber = *(int *)&reg_socket[0];
  IPXListenForPacket(&ecb->ecb);
# else
  FD_SET(IPXProducerSocket,&ipxset);
# endif
  if (cln != NULL) getConsumerRequest(cln,srvWorkArea,IPX_BUFFER_SIZE);
  return cln;
}

int SendIPXData(CLN *cln, BYTE *data)
{
  short cc;
  unsigned short msgsize = PeekWord(*(unsigned short *)data);
# ifndef IPXNW
  fd_set fdwset;
  struct timeval wait_to;
  int retry = 2,IPXSocket,bytesSent;
  static struct sockaddr_ipx adrCln;
# endif

# ifdef IPXNW
  if (msgsize > IPX_BUFFER_SIZE) return use_stream_transport;
  memcpy(&PCECB->hdr.destination.network,&cln->IPXaddress,12);
  memcpy(&PCECB->ecb.immediateAddress,cln->node,6);
  memcpy(PCECB->buf,data,msgsize);
  PCECB->ecb.fragmentDescriptor[1].size = msgsize;
  IPXSendPacket(&PCECB->ecb);
  while (PCECB->ecb.inUseFlag && !PCECB->ecb.completionCode);
  if ((cc=PCECB->ecb.completionCode) != 0)
  {
    if (NGdebug) printf("IPX: Failed packet delivery (error %x)",cc);
    return ipx_socket_error;
  }
# else
  memcpy(adrCln.sipx_node,cln->IPXaddress.node,6);
  memcpy(&adrCln.sipx_network,cln->IPXaddress.network,4);
  adrCln.sipx_port = htons(*(UINT16 *)cln->IPXaddress.socket);
  adrCln.sipx_family = AF_IPX;
  IPXSocket = cln->sck;
  if (msgsize > IPX_BUFFER_SIZE) return use_stream_transport;
  FD_ZERO(&fdwset); FD_SET(IPXSocket,&fdwset);
  wait_to.tv_sec=0;  wait_to.tv_usec=0;
retry_select:
  switch (select(maxFdSets,NULL,&fdwset,NULL,&wait_to))
  {
    case 1: 
      break;
    case 0: 
      if (NGdebug > 1) dbglog("IPX: socket %d: write set not ready",IPXSocket);
	    return 0;
    default:
      if (NGdebug > 1) soperror("select");
      if (retry-- > 0) goto retry_select;
	    return 0;
  }
  if (sendto(IPXSocket,(char *)data,msgsize,0,
             (struct sockaddr *)&adrCln,sizeof(adrCln)) < 0)
  {
    soperror ("IPX: sendto"); printf(">");
    return ipx_socket_error;
  }
# endif
  if (NGdebug > 1)
  {
    strcpy(clientaddr,ipxnet_ntoa(cln->IPXaddress.node));
    dbglog("IPX: %d bytes to %s [%s]",msgsize,cln->userName,clientaddr);
  }
  return 0;
}
#pragma argsused
void CleanupIPX(struct timeval *tv)
{
# ifdef IPXNW
  int i;
  for (i=0; i<NUM_RCV_ECB; i++)
  {
    if (rcvECB[i]->hdr.dataStreamType == 0xfe) clearStrandedPackets(rcvECB[i]);
  }
# endif
  return;
}
void addIPXNet(BYTE *net,BYTE *mac,BYTE *node)
{
  struct IPXNet *n;
  if ((n=SystemCalloc(1,sizeof(struct IPXNet))) == NULL) return;
  memcpy(n->network,net,4);
  memcpy(n->node,mac,6);
  memcpy(n->ImmediateAddress,node,6);
  n->nextNet = IPXNets;
  IPXNets = n;
  return;
}
void dumpIPXNets(void)
{
  int i,n;
  char s[40];

  dbglog("Registered IPX Subnets (broadcasting %s):\n",
          BroadcastServer ? "enabled" : "disabled" );
  for (n=0; n<NrOfIPXNets; n++)
  {
    for (i=0;i<4;i++) sprintf(&s[i*2],"%02X",IPXNetList[n].net[i]);
    sprintf(dbgbuf,"\t[%s] ",s);
    for (i=0;i<6;i++) sprintf(&s[i*2],"%02X",IPXNetList[n].mac[i]);
    sprintf(&dbgbuf[strlen(dbgbuf)],"[%s]",s);
    dbglog(dbgbuf);
  }
}
int IsMemberIPXNets(int num,IPXNetListStruct *lst,IPXAddress *ipxaddr)
{
  int n;

  if (num == 0) return TRUE;  /* everything's open */
  if (ipxaddr == NULL) return FALSE;
  for (n=0; n<num; n++)
  {
    if (memcmp(lst[n].net,ipxaddr->network,4)) continue;
    if (!memcmp(lst[n].mac,IPXbrdcstNode,6)) return TRUE;
    if (!memcmp(lst[n].mac,ipxaddr->node,6)) return TRUE;
  }
  return FALSE;
}

#ifdef INCLUDE_LEGACY_CODE
void AcceptSPXStream(void)
{
# ifdef IPXNW
  int i,cc=0;
  SPXECB *lecb;
  static int n=-1;

  if (!allowLegacyConnections) return;
  n = (++n)%NUM_LSN_ECB;
  for (i=0; i<NUM_LSN_ECB; i++)  /* Cycle through listening ECBs*/
    if (!(lecb=lsnECB[(i+n)%NUM_LSN_ECB])->ecb.inUseFlag) break;

  if ((n=i) == NUM_LSN_ECB) return;

  if (lecb->ecb.completionCode) ccerr(connection_timeout);

  fulfillSPXStreamRequest(*(WORD *)lecb->ecb.IPXWorkspace);

err: 
  if (cc && NGdebug) dbglog("ERR: %s",erlst[cc]);
  lecb->ecb.socketNumber = *(int *)&rpc_socket[0];
  SPXListenForConnection((BYTE)3,(BYTE)0,&lecb->ecb);
# endif
  return;
}
void fulfillSPXStreamRequest(WORD ID)
{
# ifdef IPXNW
  if (!allowLegacyConnections) return not_implemented;
  SRVDataBuffer = NULL;
  SPXcount++;
  if (GetSPXStream(ID,&IncomingSPXClient)) goto out;
  execRequest(&IncomingSPXClient);
  SendSPXData(ID,RpcSendSocket);
  CloseSPXStream(ID);
out: if (SRVDataBuffer != NULL) SystemFree((void *)SRVDataBuffer);
# endif
  return;
}
int GetSPXStream(WORD ID,CLN *cln)
{
# ifdef IPXNW
  int seqnr,actseq,dsize,size,done,i,x,b,cc=0;
  time_t interim = time(NULL);

  if (!allowLegacyConnections) return not_implemented;

  sprintf(clientaddr,"address unknown");
  for (done=0,actseq=0,interim=time(NULL); !done;)
  {
    for (i=0;i<NUM_RCV_ECB && !done;i++)
    {
      if (rcvECB[i]->ecb.inUseFlag) continue;
      if (rcvECB[i]->ecb.completionCode) goto resume;
      if (ID == rcvECB[i]->hdr.destConnectionID)
      {
        if ((seqnr=ISwap((WORD)rcvECB[i]->hdr.sequenceNumber)) == actseq)
	{
	  if (rcvECB[i]->hdr.dataStreamType) ccerr(wrong_type);
	  if (!seqnr)     /*/ get info from the first packet in sequence */
	  {
	    size = *(int *)rcvECB[i]->buf;
	    if ((SRVDataBuffer=(char *)SystemCalloc(1,size)) == NULL)
	      ccerr(out_of_server_memory);
	    for (x=0;x<6;x++) sprintf(&clientaddr[x*2],"%02X",rcvECB[i]->hdr.source.node[x]);
            memcpy(cln->IPXaddress.node,rcvECB[i]->hdr.source.node,6);
            memcpy(cln->IPXaddress.network,rcvECB[i]->hdr.source.network,4);
	    b = 0;
	  }
	  dsize = size-b > IPX_DATA_SIZE ? IPX_DATA_SIZE : size-b;
	  memcpy(&SRVDataBuffer[b],rcvECB[i]->buf,dsize);
	  b += dsize;
resume:	  rcvECB[i]->ecb.fragmentCount = 2;
          rcvECB[i]->hdr.dataStreamType = 0;
	  SPXListenForSequencedPacket(&rcvECB[i]->ecb); /*/ finished with packet */
	  if (size == b) done=1;
	  actseq=seqnr+1;
        }
      }
    }
    if (time(NULL) > interim+SERVER_TIMEOUT) ccerr(connection_timeout);
  }
err:
  if (NGdebug) dbglog("SPX: [%s] (%d bytes)",clientaddr,size);
  if (cc)
  {
    sndECB->ecb.fragmentCount=1;
    SPXTerminateConnection(ID,(ECB *)sndECB);
    while (sndECB->ecb.inUseFlag);
    if (NGdebug)
    {
      if (sndECB->ecb.completionCode) dbglog("SPX error: %x",sndECB->ecb.completionCode);
      dbglog("RPC error: %s",erlst[cc]);
    }
    for (i=0;i<NUM_RCV_ECB;i++)
    {
      if (!rcvECB[i]->ecb.inUseFlag && ID == rcvECB[i]->hdr.destConnectionID)
      {
        rcvECB[i]->ecb.fragmentCount = 2;           /*/ very important ! */
        rcvECB[i]->hdr.dataStreamType = 0;
        SPXListenForSequencedPacket(&rcvECB[i]->ecb);
      }
    }
  }
  return cc;
# else
  return not_implemented;
# endif
}
#endif 
void clearStrandedPackets(SPXECB *ecbptr)
{
# ifdef IPXNW
  int i,count;
  WORD RpcID;
  if (!allowLegacyConnections) return;
  if (ecbptr->ecb.inUseFlag) return;
  RpcID = ecbptr->hdr.destConnectionID;
  for (i=0,count=0; i<NUM_RCV_ECB; i++)
    if (rcvECB[i]->hdr.destConnectionID == RpcID) count++;
  if (count==1)
  {
    ecbptr->ecb.fragmentCount = 2;
    ecbptr->hdr.dataStreamType = 0;
    SPXListenForSequencedPacket(&ecbptr->ecb);
  }
# endif
  return;
}
int SendSPXData(WORD ID,int socket)
{
# ifdef IPXNW
  BYTE *buf = WorkArea;
  unsigned short msgsize = *(unsigned short *)WorkArea;
  int cc=0,cnt=0,dsize=0;
  time_t t0 = time(NULL);

  if (!allowLegacyConnections) return not_implemented;
  if (msgsize >= WorkAreaSize) ccerr(buffer_too_small);
  sndECB->hdr.dataStreamType = 0;
  sndECB->ecb.fragmentCount = 2;
  sndECB->ecb.socketNumber = socket;
  for (cnt=0; cnt<msgsize; cnt += dsize)
  {
    dsize = msgsize-cnt > IPX_DATA_SIZE ? IPX_DATA_SIZE : msgsize-cnt;
    memcpy(sndECB->buf,&buf[cnt],dsize);
    SPXSendSequencedPacket(ID,&sndECB->ecb);
    while (sndECB->ecb.inUseFlag && !sndECB->ecb.completionCode);
    if (time(NULL) > t0+SERVER_TIMEOUT) ccerr(connection_timeout);
  }
err:if (NGdebug)
  {
    sprintf(dbgbuf,"SPX: returned %d of %d bytes",cnt,msgsize);
    if (cc) sprintf(&dbgbuf[strlen(dbgbuf)],"incomplete: %s",erlst[cc]);
    dbglog(dbgbuf);
  }
  return cc;
# else
  return not_implemented;
# endif
}
int SendIPXBroadcast(BYTE *buf,int size)
{
  int i;
  struct IPXNet *n;
# ifdef IPXNW
  memcpy(brdcstECB->buf,buf,size);
  for (i=0,n=IPXNets; i<NrOfIPXBcastNets && n != NULL; i++,n=n->nextNet)
  {
    if (NGdebug > 1 && n == IPXNets)
      dbglog(dbgbuf,"IPX: broadcast on %d subnets",NrOfIPXBcastNets);
    memcpy(brdcstECB->hdr.destination.node,n->node,6);
    memcpy(brdcstECB->ecb.immediateAddress,n->ImmediateAddress,6);
    memcpy(brdcstECB->hdr.destination.network,n->network,4);

    IPXSendPacket(&brdcstECB->ecb);
    while (brdcstECB->ecb.inUseFlag);
    if (brdcstECB->ecb.completionCode && NGdebug > 1)
      dbglog("broadcast failed: %x",brdcstECB->ecb.completionCode);
  }
# else
  fd_set fdwset;
  struct timeval wait_to;
  int IPXSocket,bytesSent;
  struct sockaddr_ipx adrCln;

  for (i=0,n=IPXNets; i<NrOfIPXBcastNets && n != NULL; i++,n=n->nextNet)
  {
    if (NGdebug > 1 && n == IPXNets)
      dbglog("IPX: broadcast on %d subnets",NrOfIPXBcastNets);
    memcpy(adrCln.sipx_node,n->node,6);
    memcpy(&adrCln.sipx_network,n->network,4);
    adrCln.sipx_port = htons(*(UINT16 *)brdcst_socket);
    adrCln.sipx_family = AF_IPX;
    IPXSocket = IPXBroadcastSocket;
    if (size > IPX_BUFFER_SIZE) return use_stream_transport;
    FD_ZERO(&fdwset); FD_SET(IPXSocket,&fdwset);
    wait_to.tv_sec=0;  wait_to.tv_usec=0;
    switch (select(maxFdSets,NULL,&fdwset,NULL,&wait_to))
    {
      case 1: 
        break;
      case 0: 
        if (NGdebug > 1) dbglog("IPX: socket %d: write set not ready",IPXSocket);
	      return 0;
      default:
        if (NGdebug > 1) soperror("select");
	      return 0;
    }
    if (sendto(IPXSocket,(char *)buf,size,0,
               (struct sockaddr *)&adrCln,sizeof(adrCln)) < 0)
    {
      soperror ("IPX: sendto"); printf(">");
      return ipx_socket_error;
    }
  }
# endif
  return 0;
}
void CloseSPXStream(WORD ID)
{
# ifdef IPXNW
  if (!allowLegacyConnections) return;
  sndECB->ecb.fragmentCount=1;
  SPXTerminateConnection(ID,&sndECB->ecb);
  while (sndECB->ecb.inUseFlag && !sndECB->ecb.completionCode);
# endif
  return;
}
char *ipxnet_ntoa(BYTE *node)
{
  static char buf[16];
  int i;
  for (i=0; i<6; i++) sprintf(&buf[i * 2], "%02X", node[i]);
  return buf;
}

/* trivial RPC call: */
int tIPXCall(int fecindex,BYTE *buf,int timout)
{
# ifdef IPXNW
  static CLN cln;
  static BYTE reg_sock[2];
  int i,cc;
  time_t t0 = time(NULL);
  RPCFecStruct *rpcfec = fecindex ? &RPCFec[fecindex] : &NameServer[curNameServer];

  reg_sock[0] = REG_SOCKET + rpcfec->PortOffset;
  reg_sock[1] = REG_SOCKET + rpcfec->PortOffset;

  memset(&cln,0,sizeof(CLN));
  strcpy(cln.userName,rpcfec->Name);
  memcpy(&cln.IPXaddress,rpcfec->Network,10);
  memcpy(cln.IPXaddress.socket,reg_sock,2);
  memcpy(cln.node,rpcfec->ImmediateAddress,6);
  cln.sck = 0;    /* not used in NW API */

  if ((cc=SendIPXData(&cln,buf)) != 0) return cc;

  timout /= 1000;
  if (timout < 1) timout = 1;
  for (i=0; time(NULL) <= t0 + timout; i = (++i)%NUM_CNS_ECB)
    if (!cnsECB[i]->ecb.inUseFlag && !cnsECB[i]->ecb.completionCode)
    {
      if (((ProducerHeaderStruct *)&cnsECB[i]->buf[2])->SubscriptionID == 1)
      {
        memcpy(buf,cnsECB[i]->buf,IPX_BUFFER_SIZE);
        IPXListenForPacket(&cnsECB[i]->ecb);
        return 0;
      }
    }
  return link_not_open;
# else
  return not_implemented;
# endif
}
