/** \file srvquery.c
 * \brief query routines for the stock properties
 * \internal
 *
 */
#include "prolog.h"
#include "cprolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "servdef.h"
#include "brdcast.h"
#include "errors.h"
#include "database.h"
#include "almsrv.h"
#include "history.h"
#include "srvip.h"
#include "srvipx.h"
#if defined(UNIX) && !defined(NO_REMOTE_PIPE)
# define UNIX_TTY
# include "unixcmd.h"
#endif

int T_numQuery(CONTRACT *con,char *data,short num)
{
  if (con->EqpSizeOut != 1) return illegal_read_write;
  switch (LFMT(con->EqpFormatOut))
  {
    case CF_SHORT: *(short *)data = num; break;
    case CF_LONG: *(UINT32 *)data = (UINT32)num; break;
    case CF_FLOAT: *(float *)data = (float)num; break;
    case CF_DOUBLE: *(double *)data = (double) num; break;
    default: return illegal_format;
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int taggedStructQuery(CONTRACT *con,char *dataIn,char *data)
{
  int cc,k,len;
  char tag[9];

  if (con->EqpSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqpFormatOut) != CF_INTINT) return illegal_format;
  if ((len=(int)con->EqpSizeIn) == 0) return illegal_read_write;
  memset(tag,0,9);
  switch (LFMT(con->EqpFormatIn))
  {
    case CF_TEXT:
      if (len < 1 || len > 8) return illegal_read_write;
      strncpy(tag,dataIn,len);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      if (len != 1) return illegal_read_write;
      strncpy(tag,dataIn,8);
      break;
    default:  return illegal_format;
  }
  k = (int)con->EqpSizeOut;
  cc = queryStructFormat(tag,(INTINT *)data,&k);
  if (k < (int)con->EqpSizeOut) con->EqpSizeOut = k;
  return cc;
}
#ifdef MSDOS
#pragma argsused
#endif
int getServerAliasList(CONTRACT *con,char *dataIn,char *data)
{
  int len,i,j;
  if (con->EqpSizeOut == 0) return illegal_read_write;
  switch (LFMT(con->EqpFormatOut))
  {
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      len = MIN(fmtsizeof(con->EqpFormatOut),32);
      break;
    default:  return illegal_format;
  }
  for (i=0,j=0; i<(int)con->EqpSizeOut && j<nAliasTableEntries; i += 2,j++)
  {
     strncpy(((NAME32 *)data)[i].name,gAliasTable[j].name,len);
     strncpy(((NAME32 *)data)[i+1].name,gAliasTable[j].alias,len);
  }
  if (i < (int)con->EqpSizeOut) con->EqpSizeOut = i;
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int ndeviceQuery(CONTRACT *con,char *dataIn,char *data)
{
  ExportListStruct *el;
  short n;
/*
  // get this to work for an "ndevices for property" call !
  int pid;
  ExportPropertyListStruct *pl=NULL;
  if (con->EqpSizeIn > 0)
  {
    switch (LFMT(con->EqpFormatIn))
    {
      case CF_TEXT:
      case CF_NAME16:
      case CF_NAME32:
        if ((pid=GetPrpId(con->EqpName,dataIn)) < 0) break;
        pid = PROPERTY_HASH_INDEX(pid);
        for (pl=el->EqpProperty[pid]; pl != NULL && strncmp(pl->prpName,dataIn,32); pl=pl->next);
        break;
      default:  return illegal_format;
    }
  }
*/
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
/*
    if (pl != NULL && pl->numDevices > 0) n = pl->numDevices;
    else 
*/
    n = el->EqpNumModules;
    if (longDeviceNames) n *= 2;
    return T_numQuery(con,data,n);
  }
  return non_existent_elem;
}
#ifdef MSDOS
#pragma argsused
#endif
int deviceQuery(CONTRACT *con,char *dataIn,char *data)
{
  int i,k,slen,start;
  ExportListStruct *el;
  int cc = 0;
  char btok[32], etok[32], mtok[32];
  int blen = 0, elen = 0, mlen = 0, checkWildCard = FALSE;

  if (con->EqpSizeOut == 0) return illegal_read_write;
  start = con->EqpDeviceName[0] == '#' ? atoi(&con->EqpDeviceName[1])-1 : 0;
  if (start < 0) start = 0;

  if (strcmp(con->EqpDeviceName,"*") && strchr(con->EqpDeviceName,'*') != NULL)
  { /* device query with non-trivial wildcard string */
    getWildcardTokens(con->EqpDeviceName,btok,etok,mtok);
    blen = (int)strlen(btok); mlen = (int)strlen(mtok); elen = (int)strlen(etok);
    checkWildCard = TRUE;
  }

  switch (LFMT(con->EqpFormatOut))
  {
    case CF_NAME8: 
    case CF_NAME16:
    case CF_NAME32:slen = fmtsizeof(con->EqpFormatOut); break;
    default: return illegal_format;
  }
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) break;
  if (el == NULL) return non_existent_elem;
  for (i=0,k=0; i<(int)con->EqpSizeOut && i+start<el->EqpNumModules; i++)
  {
    if (el->EqpModuleName && strlen(el->EqpModuleName[i+start].name))
    {
      if (checkWildCard && 
          !wildcardMatches(el->EqpModuleName[i+start].name,btok,blen,etok,elen,mtok,mlen))
        continue;
      strncpy(&data[k*slen],el->EqpModuleName[i+start].name,slen);
      if (longDeviceNames && slen == 16 &&
          strlen(el->EqpModuleName[i+start].name) > 16)
      {
        data[k*slen+15] = '&';
        strncpy(&data[k*slen+16],&el->EqpModuleName[i+start].name[15],15);
        i++; k++;
      }
    }
    else 
    {
      if (checkWildCard) continue;
      sprintf(&data[k*slen],"#%d",i);
    }
    k++;
  }
  if (el->hasPropertyDeviceNames) 
  {
    cc = has_query_function | CE_SENDDATA;
  }
  if (k < (int)con->EqpSizeOut) con->EqpSizeOut = k;
  return cc;
}
#define xpqs2pqs(p) {\
  strncpy(pqs[p].prpName,xpqs[p].prpName,32);\
  strncpy(pqs[p].prpDescription,xpqs[p].prpDescription,32);\
  pqs[p].prpSize = (UINT16)xpqs[p].prpSize;\
  pqs[p].prpFormat = xpqs[p].prpFormat;\
  pqs[p].prpAccess = xpqs[p].prpAccess;\
}
#define xpqs2pqsx(p) {\
  if (p > 0) memcpy(&pqsx[p],&xpqs[p],sizeof(PropertyQueryStructEx));\
}
#define pqs2xpqs(p) {\
  memset(&xpqs[p],0,sizeof(XPropertyQueryStruct));\
  strncpy(xpqs[p].prpName,pqs[p].prpName,32);\
  strncpy(xpqs[p].prpDescription,pqs[p].prpDescription,32);\
  xpqs[p].prpSize = pqs[p].prpSize;\
  xpqs[p].prpFormat = pqs[p].prpFormat;\
  xpqs[p].prpAccess = pqs[p].prpAccess;\
}
#define pqsx2pqs(p) {\
  strncpy(pqs[p].prpName,pqsx[p].prpName,32);\
  strncpy(pqs[p].prpDescription,pqsx[p].prpDescription,32);\
  pqs[p].prpSize = (UINT16)pqsx[p].prpSize;\
  pqs[p].prpFormat = pqsx[p].prpFormat;\
  pqs[p].prpAccess = pqsx[p].prpAccess;\
}
#define pqs2pqsx(p) {\
  memset(&pqsx[p],0,sizeof(PropertyQueryStructEx));\
  strncpy(pqsx[p].prpName,pqs[p].prpName,32);\
  strncpy(pqsx[p].prpDescription,pqs[p].prpDescription,32);\
  pqsx[p].prpSize = pqs[p].prpSize;\
  pqsx[p].prpFormat = pqs[p].prpFormat;\
  pqsx[p].prpAccess = pqs[p].prpAccess;\
}
#ifdef MSDOS
#pragma argsused
#endif
int stockpropsQuery(CONTRACT *con,char *dataIn,char *data)
{
  int i,k,n,fmt,len;
  PropertyQueryStruct *pqs = (PropertyQueryStruct *)tmpWorkArea;
  PropertyQueryStructEx *pqsx = (PropertyQueryStructEx *)tmpWorkArea;
  XPropertyQueryStruct *xpqs = (XPropertyQueryStruct *)tmpWorkArea;
  int tgtprplen = 0;
  char *c,tgtprp[36];

  if ((len=(int)con->EqpSizeIn) > 0)
  {
    memset(tgtprp,0,33);
    switch (LFMT(con->EqpFormatIn))
    {
      case CF_TEXT:
        if (len < 1 || len > 32) return illegal_read_write;
        strncpy(tgtprp,dataIn,len);
        break;
      case CF_NAME16:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME16 *)dataIn)->name,16);
        break;
      case CF_NAME32:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME32 *)dataIn)->name,32);
        break;
      default:  return illegal_format;
    }
    if ((c=strchr(tgtprp,'*')) != NULL)
    {
      *c = 0;
      tgtprplen = (int)strlen(tgtprp);
    }
    else tgtprplen = 32;
  }
  switch (fmt=LFMT(con->EqpFormatOut))
  {
    case CF_NAME32:
      if (con->EqpSizeOut == 0) return illegal_read_write;
      for (i=0,k=0; i<(int)con->EqpSizeOut && i<NrStockProperties; i++)
      {
        if (tgtprplen && strncmp(StockProperty[i].prpName,tgtprp,tgtprplen)) continue;
        if (k > 0 && !strncmp(((NAME32 *)data)[k-1].name,StockProperty[i].prpName,32)) continue;
        strncpy(((NAME32 *)data)[k].name,StockProperty[i].prpName,32);
        k++;
      }
      if (k < (int)con->EqpSizeOut) con->EqpSizeOut = k;
      return 0;
    case CF_USTRING:
      if (con->EqpSizeOut == 0) return illegal_read_write;
      for (i=0,k=0; i<(int)con->EqpSizeOut && i<NrStockProperties; i++)
      {
        if (tgtprplen && strncmp(StockProperty[i].prpName,tgtprp,tgtprplen)) continue;
        if (k > 0 && !strncmp(((DUSTRING *)data)[k-1].str,StockProperty[i].prpName,32)) continue;
        strncpy(((DUSTRING *)data)[k].str,StockProperty[i].prpName,32);
        strncpy(&((DUSTRING *)data)[k].str[32],StockProperty[i].prpDescription,48);
        ((DUSTRING *)data)[k].ival = (UINT32)StockProperty[i].prpSize;
        ((DUSTRING *)data)[k].tm = (UINT32)StockProperty[i].prpFormat;
        k++;
      }
      if (k < (int)con->EqpSizeOut) con->EqpSizeOut = k;
      return 0;
    case CF_STRUCT:
      for (i=0,k=0; i<NrStockProperties; i++)
      {
        if (tgtprplen && strncmp(StockProperty[i].prpName,tgtprp,tgtprplen)) continue;
        memset(&xpqs[k],0,sizeof(XPropertyQueryStruct));
        strncpy(xpqs[k].prpName,StockProperty[i].prpName,32);
        strncpy(xpqs[k].prpDescription,StockProperty[i].prpDescription,64);
        xpqs[k].prpSize = StockProperty[i].prpSize;
        xpqs[k].prpSizeIn = StockProperty[i].prpSizeIn;
        xpqs[k].prpFormat = BFMT(StockProperty[i].prpFormat);
        xpqs[k].prpFormatIn = BFMT(StockProperty[i].prpFormatIn);
        xpqs[k].prpAccess = (BYTE)StockProperty[i].prpAccessMode;
        xpqs[k].prpArrayType = StockProperty[i].prpArrayType;
        strncpy(xpqs[k].prpTag,StockProperty[i].prpStructTag,8);
        k++;
      }
      if (tgtprplen) for (i=0; i<k; i++) xpqs[i].prpNumOverloads = k;
      if (con->EqpSizeOut%XPROPERTYQUERYSTRUCT_SIZE == 0 && !strncmp(con->strTagOut,"XPQS",8))
      {
        n = (UINT16)con->EqpSizeOut/XPROPERTYQUERYSTRUCT_SIZE;
        if (k > n) k = n;
        PrepOutgoingXPQS(xpqs,data,k);
        if (k < n) con->EqpSizeOut = k*XPROPERTYQUERYSTRUCT_SIZE;
        return 0;
      }
      if (con->EqpSizeOut%PROPERTYQUERYSTRUCTEX_SIZE == 0)
      {
        n = (int)(con->EqpSizeOut/PROPERTYQUERYSTRUCTEX_SIZE);
        if (k > n) k = n;
        for (i=0; i<n; i++) xpqs2pqsx(i);
        PrepOutgoingPQSX(pqsx,data,k);
        if (k < n) con->EqpSizeOut = k*PROPERTYQUERYSTRUCTEX_SIZE;
        return 0;
      }
      if (con->EqpSizeOut%PROPERTYQUERYSTRUCT_SIZE == 0)
      {
        n = (int)(con->EqpSizeOut/PROPERTYQUERYSTRUCT_SIZE);
        if (k > n) k = n;
        for (i=0; i<n; i++) xpqs2pqs(i);
        PrepOutgoingPQS(pqs,data,k);
        if (k < n) con->EqpSizeOut = k*PROPERTYQUERYSTRUCT_SIZE;
        return 0;
      }
      return illegal_format;
    default:
      break;
  }
  return illegal_format;
}
#ifdef MSDOS
#pragma argsused
#endif
int nstockpropsQuery(CONTRACT *con,char *dataIn,char *data)
{
  return T_numQuery(con,data,(short)NrStockProperties);
}
int historyRecordDepth(char *prpName,UINT16 *shortDepth,UINT16 *longDepth)
{
  int i;
  for (i=0; i<nHistoryRecords && strncmp(prpName,historyRecords[i].c.EqpProperty,32); i++);
  if (i < nHistoryRecords) 
  {
    if (shortDepth != NULL) *shortDepth = historyRecords[i].depthShort;
    if (longDepth != NULL) *longDepth = historyRecords[i].depthLong;
    return historyRecords[i].depthShort;
  }
  return 0;
}
int lclXEqpPropertyQueryFcn(ExportListStruct *el,char *dev,XPropertyQueryStruct *xpqs,int num,int start,int useAliasName)
{
  ExportPropertyListStruct *prp;
  int idx,j,k,l,n,devnr=-1;
  /* char *bp=NULL,*mp=NULL; */

  if (dev != NULL) devnr = GetDeviceNumber(el->EqpName,dev);
  for (l=0,j=0,idx=0; idx<PROPERTY_HASH_SIZE+1 && l<num; idx++)
    for (prp=el->EqpProperty[idx]; prp && l<num; prp=prp->next)
  {
    if (el->EqpModulePrpList != NULL && devnr >=0 && el->EqpModulePrpList[devnr].prpListSize > 0) 
    { /* there's a device-specific property list */
      for (k=0; k<el->EqpModulePrpList[devnr].prpListSize; k++) 
        if (!strncmp(el->EqpModulePrpList[devnr].prpList[k].name,prp->prpName,EQP_PROP_SIZE)) break;
      if (k == el->EqpModulePrpList[devnr].prpListSize) continue; /* not found in list */
    }
    if (j++ < start) continue;
    memset(&xpqs[l],0,sizeof(XPropertyQueryStruct));
    strncpy(xpqs[l].prpName,prp->prpName,32);
    strncpy(xpqs[l].prpDescription,prp->prpDescription,64);
    xpqs[l].prpSize = prp->prpSize;
    xpqs[l].prpSizeIn = prp->prpSizeIn;
    xpqs[l].prpFormat = BFMT(prp->prpFormat);
    xpqs[l].prpFormatIn = BFMT(prp->prpFormatIn);
    strncpy(xpqs[l].prpTag,prp->prpFormatTag.name,8);
    strncpy(xpqs[l].prpTagIn,prp->prpFormatTagIn.name,8);
    xpqs[l].prpGraphType = prp->egu.graph;
    xpqs[l].prpMaxValue = (float)prp->egu.max;
    xpqs[l].prpMinValue = (float)prp->egu.min;
    if (prp->prd != NULL) memcpy(xpqs[l].prpRedirection,prp->prd->prpRedirection->name,32);
    strncpy(xpqs[l].prpUnits,prp->egu.units,16);
    xpqs[l].prpNumOverloads++;
    xpqs[l].prpAccess = (BYTE)prp->prpAccessMode;
    historyRecordDepth(prp->prpName,&xpqs[l].prpHistoryDepthShort,&xpqs[l].prpHistoryDepthLong);
    xpqs[l].prpArrayType = prp->prpArrayType;
    xpqs[l].numRows = prp->numRows;
    xpqs[l].rowSize = prp->rowSize;
    xpqs[l].rngMaxValue = prp->xegu.max;
    xpqs[l].rngMinValue = prp->xegu.min;
    strncpy(xpqs[l].rngUnits,prp->xegu.units,16);
    n = l++;
    if (!useAliasName && prp->prpAlias[0] != 0  && l > 0)
    {
      xpqs[l] = xpqs[n];
      strncpy(xpqs[l].prpName,prp->prpAlias,32);
      l++;
    }
/*  // brute force scanning takes way too much time when there's a large alias table
    // only way to provide this information at browse time is to maintain a list
    // in the ExportPropertyListStruct which is filled in when an alias is registered
    // which is a meta property of the target property.
    for (i=0; i<nAliasTableEntries; i++)
    {
      if (IsMetaProperty(gAliasTable[i].name,&bp,&mp) && !strcmp(xpqs[n].prpName,bp))
      {
        xpqs[l] = xpqs[n];
        if (strlen(xpqs[l].prpDescription) < 48)
        {
          strcat(xpqs[l].prpDescription," ");
          strncat(xpqs[l].prpDescription,mp,15);
        }
        strncpy(xpqs[l].prpName,gAliasTable[i].alias,32);
        l++;
      }
    }
*/
  }
  return l;
}
#ifdef MSDOS
#pragma argsused
#endif
int getSrvAddressParams(CONTRACT *con,char *dataIn,char *data)
{
  DTYPE d;
  char *tgtcon,*tgteqm,*tgtexp;
  ExportListStruct *el;
  if (isServiceRequest(con))
  {
    if (con->EqpSizeIn != 3) return illegal_read_write;
    if (LFMT(con->EqpFormatIn) != CF_NAME16) return illegal_format;
    tgtcon = dataIn; tgteqm = dataIn+16; tgtexp = dataIn+32;
    if (strlen(tgteqm) == 0 && strlen(tgtexp) == 0) return invalid_data;
    if (strlen(tgtcon) && strncmp(tgtcon,gDeviceContext,EQP_CONTEXT_SIZE)) return invalid_data;
    for (el=ExportList; el != NULL; el=el->next)
    {
      if (strlen(tgteqm) && strncmp(tgteqm,el->EqpName,EQP_NAME_SIZE)) continue;
      if (strlen(tgtexp) && strncmp(tgtexp,el->tagName,ELEM_NAME_SIZE)) continue;
      break; /* this looks like me ! */
    }
    if (el == NULL) return invalid_data;
  }
  d.dArrayLength = con->EqpSizeOut;
  d.dFormat = LFMT(con->EqpFormatOut);
  d.dTag[0] = 0;
  d.data.cptr = data;
  return getSrvAddress(&d,el->EqpName,NULL);
}
#ifdef MSDOS
#pragma argsused
#endif
int getLastAccessList(CONTRACT *con,char *dataIn,char *data)
{
  int slen = 0;
  switch (LFMT(con->EqpFormatOut))
  {
    case CF_NAME16:slen = 16; break;
    case CF_NAME32:slen = 32; break;
    default: return illegal_format;
  }
  switch (con->EqpSizeOut)
  {
    case 5: strncpy(data + slen*4,getDataTimeString(gLastWriteAccessTime,-1),slen);
    case 4: strncpy(data + slen*3,gLastWriteAccessDev,slen);
    case 3: strncpy(data + slen*2,gLastWriteAccessPrp,slen);
    case 2: strncpy(data + slen*1,gLastWriteAccessAddr,16);
    case 1: strncpy(data,gLastWriteAccessUsr,16);
      break;
    default:
      return dimension_error;
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int propsQuery(CONTRACT *con,char *dataIn,char *data)
{
  static BYTE *tmpDataCache = (BYTE *)NULL;
  static int tmpDataCacheSize = 0;
  int idx,j=0,l,n,num,len;
  short fmt;
  ExportListStruct *el;
  PropertyQueryStruct *pqs = (PropertyQueryStruct *)tmpDataCache;
  PropertyQueryStructEx *pqsx = (PropertyQueryStructEx *)tmpDataCache;
  XPropertyQueryStruct *xpqs = (XPropertyQueryStruct *)tmpDataCache;
  int cc = 0,tgtprplen = 0,useAliasName=FALSE,usepqs=FALSE,usepqsx=FALSE,uselist=FALSE,hasdevgroup=FALSE;
  char tgtprp[36];
  char btok[32], etok[32], mtok[32];
  int blen = 0, elen = 0, mlen = 0, checkWildCard = FALSE;

  if (tmpDataCache == NULL)
  {
    tmpDataCache = tmpWorkArea;
    tmpDataCacheSize = tmpWorkAreaSize;
  }
  if (con->EqpSizeOut == 0) return illegal_read_write;

  if (strcmp(con->EqpProperty,"*") && strchr(con->EqpProperty,'*') != NULL)
  { /* property names query with non-trivial wildcard string */
    getWildcardTokens(con->EqpProperty,btok,etok,mtok);
    blen = (int)strlen(btok); mlen = (int)strlen(mtok); elen = (int)strlen(etok);
    checkWildCard = TRUE;
  }

  if ((len=(int)con->EqpSizeIn) > 0)
  {
    memset(tgtprp,0,33);
    switch (LFMT(con->EqpFormatIn))
    {
      case CF_TEXT:
        if (len < 1 || len > 32) return illegal_read_write;
        strncpy(tgtprp,dataIn,len);
        break;
      case CF_NAME16:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME16 *)dataIn)->name,16);
        break;
      case CF_NAME32:
        if (len != 1) return illegal_read_write;
        strncpy(tgtprp,((NAME32 *)dataIn)->name,32);
        break;
      default:  return illegal_format;
    }
    if (nAliasTableEntries > 0) /* is the incoming name an alias ? */
    {
      if ((j=getAliasTableIndex(tgtprp)) >= 0) 
      {
        if (!IsMetaProperty(gAliasTable[j].name,NULL,NULL))
        {
          strncpy(tgtprp,gAliasTable[j].name,EQP_PROP_SIZE);
          useAliasName = TRUE;
        }
      }
    }
    tgtprplen = (int)strlen(tgtprp);
  }
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
    if (strlen(el->group) > 0) hasdevgroup = TRUE;
    len = tmpDataCacheSize/XPROPERTYQUERYSTRUCT_SIZE;
    if (el->XEqpPropertyQueryFcn != NULL)        /* registered ext query fcn */
    {
      num = el->XEqpPropertyQueryFcn(con->EqpDeviceName,&xpqs);
      if (num > 0 && tmpDataCache != (BYTE *)xpqs)
      {
        if (num > len)
        {
          if ((tmpDataCache=(BYTE *)calloc(num,sizeof(XPropertyQueryStruct))) == NULL)
            return out_of_server_memory;
          tmpDataCacheSize = num * sizeof(XPropertyQueryStruct);
          feclog("re-alloc property query buffer space (%d bytes)",tmpDataCacheSize);
        }
        memcpy(tmpDataCache,xpqs,num*sizeof(XPropertyQueryStruct));
        xpqs = (XPropertyQueryStruct *)tmpDataCache;
      }
      usepqsx = TRUE;
    }
    else if (el->EqpPropertyQueryFcn != NULL)           /* registered query fcn */
    {
      num = el->EqpPropertyQueryFcn(con->EqpDeviceName,&pqs);
      if (num > len)
      {
        if ((tmpDataCache=(BYTE *)calloc(num,sizeof(XPropertyQueryStruct))) == NULL)
          return out_of_server_memory;
        tmpDataCacheSize = num * sizeof(XPropertyQueryStruct);
        feclog("re-alloc property query buffer space (%d bytes)",tmpDataCacheSize);
      }
      xpqs = (XPropertyQueryStruct *)tmpDataCache;
      for (l=0; l<num; l++) pqs2xpqs(l);
      usepqs = TRUE;
    }
    else                                                      /* handle it here */
    {
      if ((num=(el->EqpNumProperties*2)) > len) /* allow room for alias entries */
      {
        if ((tmpDataCache=(BYTE *)calloc(num,sizeof(XPropertyQueryStruct))) == NULL)
          return out_of_server_memory;
        tmpDataCacheSize = num * sizeof(XPropertyQueryStruct);
        feclog("re-alloc property query buffer space (%d bytes)",tmpDataCacheSize);
      }
      xpqs = (XPropertyQueryStruct *)tmpDataCache;
      num = lclXEqpPropertyQueryFcn(el,con->EqpDeviceName,xpqs,num,0,useAliasName);
      if (el->EqpModulePrpList != NULL) uselist = TRUE;
    }
    if (num <= 0) return -num;
    /* re-assign these just in case they've changed */
    pqs = (PropertyQueryStruct *)tmpDataCache;
    pqsx = (PropertyQueryStructEx *)tmpDataCache;
    xpqs = (XPropertyQueryStruct *)tmpDataCache;

    /* do we have to collapse the query results ? */
    if (checkWildCard)
    { /* checking a wildcard match */
      for (idx=0,l=0; idx<num && l<num; idx++)
      {
        if (!wildcardMatches(xpqs[idx].prpName,btok,blen,etok,elen,mtok,mlen))
          continue;
        if (idx != l) memcpy(&xpqs[l],&xpqs[idx],sizeof(XPropertyQueryStruct));
        l++;
      }
      num = l;
    }
    else if (tgtprplen > 0)
    { /* checking for input target property */
      for (idx=0,l=0; idx<num && l<num; idx++)
      {
        if (strncmp(xpqs[idx].prpName,tgtprp,EQP_PROP_SIZE)) continue;
        if (idx != l) memcpy(&xpqs[l],&xpqs[idx],sizeof(XPropertyQueryStruct));
        if (useAliasName) strncpy(xpqs[l].prpName,gAliasTable[j].alias,EQP_PROP_SIZE);
        l++;
      }
      for (j=0; j<l; j++) xpqs[j].prpNumOverloads = l;
      num = l;
    }
    switch (fmt=LFMT(con->EqpFormatOut))
    {
      case CF_STRUCT:
        if (num == 0) return non_existent;
        if (con->EqpSizeOut%XPROPERTYQUERYSTRUCT_SIZE == 0 && !strncmp(con->strTagOut,"XPQS",8))
        {
          n = (int)con->EqpSizeOut/XPROPERTYQUERYSTRUCT_SIZE;
          if (num > n) num = n;
          PrepOutgoingXPQS(xpqs,data,num);
          if (num < n) con->EqpSizeOut = num*XPROPERTYQUERYSTRUCT_SIZE;
          return 0;
        }
        if (con->EqpSizeOut%PROPERTYQUERYSTRUCTEX_SIZE == 0)
        {
          n = (int)con->EqpSizeOut/PROPERTYQUERYSTRUCTEX_SIZE;
          if (num > n) num = n;
          for (l=0; l<num; l++) xpqs2pqsx(l);
          PrepOutgoingPQSX(pqsx,data,num);
          if (num < n) con->EqpSizeOut = num*PROPERTYQUERYSTRUCTEX_SIZE;
          return 0;
        }
        if (con->EqpSizeOut%PROPERTYQUERYSTRUCT_SIZE == 0)
        {
          n = (int)con->EqpSizeOut/PROPERTYQUERYSTRUCT_SIZE;
          if (num > n) num = n;
          if (usepqs && (checkWildCard || tgtprplen > 0))
          {
            pqs = (PropertyQueryStruct *)tmpDataCache;
            usepqs = FALSE;
          }
          if (!usepqs) for (l=0; l<num; l++) xpqs2pqs(l);
          PrepOutgoingPQS(pqs,data,num);
          if (num < n) con->EqpSizeOut = num*PROPERTYQUERYSTRUCT_SIZE;
          return 0;
        }
        return illegal_format;
      case CF_NAME32:
        n = (int)con->EqpSizeOut;
        if (num > n) num = n;
        for (idx=0,l=0; idx<num && l<n; idx++)
        { /* check for up to 2 overloads */
          if (l > 0 && !strncmp(((NAME32 *)data)[l-1].name,xpqs[idx].prpName,32)) continue;
          if (l > 1 && !strncmp(((NAME32 *)data)[l-2].name,xpqs[idx].prpName,32)) continue;
          strncpy(((NAME32 *)data)[l].name,xpqs[idx].prpName,32);
          l++;
        }
        if (l == 0) return non_existent;
        if ((usepqsx || usepqs || uselist || hasdevgroup) && tgtprplen == 0)
        {
          cc = has_query_function | CE_SENDDATA;
        }
        if (l < n) con->EqpSizeOut = l;
        return cc;
      case CF_USTRING:
        n = (int)con->EqpSizeOut;
        if (num > n) num = n;
        for (idx=0,l=0; idx<num && l<n; idx++)
        { /* check for up to 2 overloads */
          if (l > 0 && !strncmp(((DUSTRING *)data)[l-1].str,xpqs[idx].prpName,32)) continue;
          if (l > 1 && !strncmp(((DUSTRING *)data)[l-2].str,xpqs[idx].prpName,32)) continue;
          strncpy(((DUSTRING *)data)[l].str,xpqs[idx].prpName,32);
          strncpy(&((DUSTRING *)data)[l].str[32],xpqs[idx].prpDescription,48);
          ((DUSTRING *)data)[l].ival = (UINT32)xpqs[idx].prpSize;
          ((DUSTRING *)data)[l].tm = (UINT32)xpqs[idx].prpFormat;
          l++;
        }
        if (l == 0) return non_existent;
        if ((usepqsx || usepqs || uselist || hasdevgroup) && tgtprplen == 0)
        {
          cc = has_query_function | CE_SENDDATA;
        }
        if (l < n) con->EqpSizeOut = l;
        return cc;
      default:
        return illegal_format;
    }
  }
  return non_existent_elem;
}
#ifdef MSDOS
#pragma argsused
#endif
int npropsQuery(CONTRACT *con,char *dataIn,char *data)
{
  ExportListStruct *el;
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
    return T_numQuery(con,data,el->EqpNumProperties);
  }
  return non_existent_elem;
}
#ifdef MSDOS
#pragma argsused
#endif
int alarmsQuery(CONTRACT *con,char *dataIn,char *data)
{
  AMS *ams;
  UINT16 n,nalm;
  int start;

  if (LFMT(con->EqpFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqpSizeOut%AMS_SIZE) return illegal_read_write;
  n = (UINT16)con->EqpSizeOut/AMS_SIZE;
  ams = (AMS *)tmpWorkArea;
  start = con->EqpDeviceName[0] == '#' ? atoi(&con->EqpDeviceName[1]) : 0;
  nalm = (UINT16)getAlarms(con->EqpName,ams,(short)n,(short)start);
  PrepOutgoingAMS(ams,data,n);
  if (nalm < n) con->EqpSizeOut = nalm*AMS_SIZE;
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int alarmsQueryEx(CONTRACT *con,char *dataIn,char *data)
{
  AMS *ams;
  AMRB amrb;
  int n,nalm,start;
  int size = (int)con->EqpSizeOut;
  int sizeIn = (int)con->EqpSizeIn;
  /* Note: allow input to be an array of UINT32s as well !! */
  if (LFMT(con->EqpFormatIn) != CF_STRUCT || LFMT(con->EqpFormatOut) != CF_STRUCT)
    return illegal_format;
  if (sizeIn != AMRB_SIZE) return dimension_error;
  if (size%AMS_SIZE) return illegal_read_write;
  if ((n=size/AMS_SIZE) == 0) return buffer_too_small;
  ams = (AMS *)tmpWorkArea;
  start = con->EqpDeviceName[0] == '#' ? atoi(&con->EqpDeviceName[1]) : 0;
  PrepIncomingAMRB(&amrb,(UINT32 *)dataIn);
  nalm = getAlarmsEx(con->EqpName,ams,(short)n,(short)amrb.almMinSeverity,
	             (time_t)amrb.almStartTime,(time_t)amrb.almStopTime,(short)start);
  PrepOutgoingAMS(ams,data,nalm);
  if (nalm < n) con->EqpSizeOut = nalm*AMS_SIZE;
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int nalarmsQuery(CONTRACT *con,char *dataIn,char *data)
{
  short sc,ntim,nsev;

  if (LFMT(con->EqpFormatOut) != CF_LONG) return illegal_format;
  switch (con->EqpSizeOut)
  {
    case 3:
      *(UINT32 *)data = numAlarmsEx(con->EqpName,&((UINT32 *)data)[1],&sc,&ntim,&nsev);
      ((UINT32 *)data)[2] = (UINT32)sc;
      break;
    case 5:
      *(UINT32 *)data = numAlarmsEx(con->EqpName,&((UINT32 *)data)[1],&sc,&ntim,&nsev);
      ((UINT32 *)data)[2] = (UINT32)sc;
      ((UINT32 *)data)[3] = (UINT32)ntim;
      ((UINT32 *)data)[4] = (UINT32)nsev;
      break;
    default: return illegal_read_write;
  }
  if (!gRequireReadFromCAS) 
  {
    if ((gRequireReadFromCAS=isCASInUserList()) != FALSE) 
      feclog("ALM: Central Alarm Server attached");
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int alarmDefsQuery(CONTRACT *con,char *dataIn,char *data)
{
  ADS *ads = NULL;
  UINT16 n;
  short num = 0;
  if (LFMT(con->EqpFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqpSizeOut%ADS_SIZE) return illegal_read_write;
  n = (UINT16)con->EqpSizeOut/ADS_SIZE;
  if ((num=getAlarmDefs(con->EqpName,&ads)) < 0) return -num;
  if (num == 0) return un_allocated;
  PrepOutgoingADS(ads,data,n);
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int nalarmDefsQuery(CONTRACT *con,char *dataIn,char *data)
{
  return T_numQuery(con,data,(short)numAlarmDefs(con->EqpName));
}
#ifdef MSDOS
#pragma argsused
#endif
int msgDisplay(CONTRACT *con,char *data,char *dataOut)
{
  int len = (int)con->EqpSizeIn;
  if (len == 0) return illegal_read_write;
  if (LFMT(con->EqpFormatIn) != CF_TEXT) return illegal_format;
  memset(dbgbuf,0,256);
  strncpy(dbgbuf,data,MIN(len,255));
  dbglog(dbgbuf);
  feclog(dbgbuf);
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int activityQuery(CONTRACT *con,char *dataIn,char *data)
{
  ActivityQueryStruct aqs;

  if (LFMT(con->EqpFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqpSizeOut != ACTIVITYQUERYSTRUCT_SIZE) return illegal_read_write;

  memset(&aqs.reserved,0,4);
  strncpy(aqs.FecName,gFecName,FEC_NAME_SIZE);
  aqs.localtime = (UINT32)time(NULL);
  aqs.starttime = (UINT32)startuptime;
  aqs.systemPollingRate = SysPoll;
  aqs.numBkgTsks = NrEQPBckTsks;
  aqs.numTotalContracts = ncontract - nExpiredContracts;
  aqs.numTotalClients = nconsumer;
  aqs.numTargetContracts = GetNumContracts(con->EqpName);
  aqs.numTargetClients = GetNumConsumers(con->EqpName);
  aqs.numConnections = nConnectionTableEntries;
  aqs.numConnectionTimeouts = nConnectionTimeouts;
  aqs.numConnectionArrivals = nArrivals;
  aqs.numUDPpkts = UDPcount;
  aqs.numTCPpkts = TCPcount;
  aqs.numIPXpkts = IPXcount;
  aqs.numSPXpkts = SPXcount;

  PrepOutgoingAQS(&aqs,data);
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int contractsQuery(CONTRACT *con,char *dataIn,char *data)
{
  ContractQueryStruct *cqs = (ContractQueryStruct *)data;
  ContractListStruct *cl;
  int i, n, ncon;
# if !defined(NONALIGNED) && !defined(INTEL)
  int tmp;
# endif
  if (LFMT(con->EqpFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqpSizeOut%CONTRACT_QUERY_SIZE) return illegal_read_write;
  if ((ncon=(int)con->EqpSizeOut/CONTRACT_QUERY_SIZE) == 0) return dimension_error;

  for (i=0,n=0; n<ncon && i<ncontract; i++)
  {
    cl = ContractList[i];
    if (cl->expired) continue;
    memcpy(&cqs[n].contract,&cl->contract,sizeof(CONTRACT));
    cqs[n].pollrate = cl->PollingRate;
    cqs[n].nconsumers = cl->nconsumer;
    PrepOutgoingCONQS(cqs[n]);
    n++;
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int rmvRegisteredUser(CONTRACT *con,char *dataIn,char *data)
{
  NAME16 usr, *usrlst = NULL;
  int i, len = (int)con->EqpSizeIn, lstsiz = 0;
  if (!(con->EqpAccess&CA_WRITE)) return illegal_read_write;
  if (!(con->EqpAccess&CA_LAST)) return illegal_mode; /* this must be a synchronous call */
  switch (LFMT(con->EqpFormatIn))
  {
    case CF_NAME16:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      break;
    case CF_NAME32:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      for (i=0; i<len; i++) strncpy(usrlst[i].name,((NAME32 *)dataIn)[i].name,16);
      break;
    case CF_TEXT:
      memset(usr.name,0,16);
      if (len < 1) return illegal_read_write;
      if (len > 16) len = 16;
      strncpy(usr.name,dataIn,len);
      usrlst = &usr;
      lstsiz = 1;
      break;
    default: return illegal_format;
  }
  return removeRegisteredUser(usrlst,lstsiz);
}
#ifdef MSDOS
#pragma argsused
#endif
int addRegisteredUser(CONTRACT *con,char *dataIn,char *data)
{
  NAME16 usr, *usrlst = NULL;
  int i, len = (int)con->EqpSizeIn, lstsiz = 0;
  /* If there's no list now, don't let someone make one over the net (dangerous!) */
  if (NrOfUsers == 0) return not_allowed; 
  if (!(con->EqpAccess&CA_WRITE)) return illegal_read_write;
  if (!(con->EqpAccess&CA_LAST)) return illegal_mode; /* this must be a synchronous call */
  switch (LFMT(con->EqpFormatIn))
  {
    case CF_NAME16:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      break;
    case CF_NAME32:
      usrlst = (NAME16 *)dataIn;
      lstsiz = len;
      for (i=0; i<len; i++) strncpy(usrlst[i].name,((NAME32 *)dataIn)[i].name,16);
      break;
    case CF_TEXT:
      memset(usr.name,0,16);
      if (len < 1) return illegal_read_write;
      if (len > 16) len = 16;
      strncpy(usr.name,dataIn,len);
      usrlst = &usr;
      lstsiz = 1;
      break;
    case CF_NULL:
      /* force a re-read of users.csv */
      break;
    default: 
      return illegal_format;
  }
  return appendRegisteredUserList(usrlst,lstsiz);
}
#ifdef MSDOS
#pragma argsused
#endif
int addRegisteredIPNet(CONTRACT *con,char *dataIn,char *data)
{
  char ip[32];
  int len = (int)con->EqpSizeIn;
  /* If there's no list now, don't let someone make one over the net (dangerous!) */
  if (NrOfIPNets == 0) return not_allowed; 
  memset(ip,0,17);
  if (!(con->EqpAccess&CA_WRITE)) return illegal_read_write;
  switch (LFMT(con->EqpFormatIn))
  {
    case CF_NAME16:
      if (len != 1) return illegal_read_write;
      strncpy(ip,((NAME16 *)dataIn)->name,16);
      break;
    case CF_NAME32:
      if (len != 1) return illegal_read_write;
      strncpy(ip,((NAME32 *)dataIn)->name,32);
      break;
    case CF_TEXT:
      if (len < 1 || len > 16) return illegal_read_write;
      strncpy(ip,dataIn,len);
      break;
    case CF_NULL:
      /* force a file re-read */
      break;
    default: 
      return illegal_format;
  }
  return appendRegisteredIPNetsList(ip);
}
#ifdef MSDOS
#pragma argsused
#endif
int getNumRegisteredIPNets(CONTRACT *con,char *dataIn,char *data)
{
  return T_numQuery(con,data,(short)NrOfIPNets);
}
#ifdef MSDOS
#pragma argsused
#endif
int getRegisteredIPNetList(CONTRACT *con,char *dataIn,char *data)
{
  int i;
  char ip[32];
  if (con->EqpSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqpFormatOut) != CF_NAME16) return illegal_format;
  for (i=0; i<(int)con->EqpSizeOut && i<NrOfIPNets; i++)
  {
    inet_ntoa_b(IPNetList[i].sin_addr,ip);
    strncpy(((NAME16 *)data)[i].name,ip,16);
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int getNumRegisteredIPXNets(CONTRACT *con,char *dataIn,char *data)
{
# if defined(IPXNW) || defined(IPXBSD)
  return T_numQuery(con,data,NrOfIPXNets);
# else
  return ipx_not_supported;
# endif
}
#ifdef MSDOS
#pragma argsused
#endif
int getRegisteredIPXNetList(CONTRACT *con,char *dataIn,char *data)
{
# if defined(IPXNW) || defined(IPXBSD)
  int i,j;
  char ipx[32];
  if (con->EqpSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqpFormatOut) != CF_NAME16) return illegal_format;
  for (i=0; i<con->EqpSizeOut && i<NrOfIPXNets; i++)
  {
    for (j=0;j<4;j++) sprintf(&ipx[j*2],"%02X",IPXNetList[i].net[j]);
    strncpy(((NAME16 *)data)[i].name,ipx,16);
  }
  return 0;
# else
  return ipx_not_supported;
# endif
}
#ifdef MSDOS
#pragma argsused
#endif
int getNumRegisteredUsers(CONTRACT *con,char *dataIn,char *data)
{
  return T_numQuery(con,data,(short)NrOfUsers);
}
#ifdef MSDOS
#pragma argsused
#endif
int getFecLog(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqpSizeOut > 0)
  {
    if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
    readFecLog(data,(int)con->EqpSizeOut);
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int getFecLogFile(CONTRACT *con,char *dataIn,char *data)
{
  char fn[128];
  int slen, cc;

  if (con->EqpDeviceName[0] != '#') 
  {
    strncpy(fn,con->EqpDeviceName,16); fn[16] = 0;
  }
  if (con->EqpSizeIn > 0)
  {
    if (con->EqpAccess & CA_WRITE)
    { /* write the text file submitted to disk */
      if (con->EqpSizeIn == 0) return illegal_read_write;
      if ((cc=writeFecLogFile(fn,dataIn,(int)con->EqpSizeIn)) != 0) return cc;
    }
    else
    { /* alternate way of supplying a (long) file name */
      switch (LFMT(con->EqpFormatIn))
      {
        case CF_NAME8:
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME48:
        case CF_NAME64:
          slen = fmtsizeof(con->EqpFormatIn);
          break;
        case CF_TEXT:
          slen = con->EqpSizeIn;
          break;
        default: return illegal_format;
      }
      strncpy(fn,dataIn,slen); fn[slen] = 0;
    }
  }
  if (con->EqpSizeOut > 0)
  {
    if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
    if ((cc=readFecLogFile(fn,data,(int)con->EqpSizeOut)) != 0) return cc;
    con->EqpSizeOut = (UINT32)strlen(data);
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int getFecLogFileList(CONTRACT *con,char *dataIn,char *data)
{
  int cc, n, slen = 0;
  char path[256];
  if (con->EqpSizeOut > 0)
  {
    switch (LFMT(con->EqpFormatIn))
    {
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        slen = fmtsizeof(con->EqpFormatIn);
        break;
      case CF_TEXT:
        slen = con->EqpSizeIn;
        break;
      default: 
        break;
    }
    if (slen > 0) strncpy(path,dataIn,slen); 
    path[slen] = 0; 
    applyFecLogPath(path,128);
    if (LFMT(con->EqpFormatOut) != CF_NAME64) return illegal_format;
    n = (int)con->EqpSizeOut;
    if ((cc=scanFecLogFiles(path,NULL,(NAME64 *)data,&n,0)) != 0) return cc;
    con->EqpSizeOut = (UINT32)n;
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int getSelfTestFile(CONTRACT *con,char *dataIn,char *data)
{
  int bytesread;
  if (con->EqpSizeOut > 0)
  {
    if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
    bytesread = readSelfTest(con->EqpName,data,(int)con->EqpSizeOut);
    if (bytesread == 0) return no_such_file;
    con->EqpSizeOut = bytesread;
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int setLogDepth(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqpSizeIn > 0 && !nofeclog)
  {
    if (!(con->EqpAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqpFormatIn) != CF_SHORT) return illegal_format;
    FeclogDepth = *(short *)dataIn;
  }
  if (con->EqpSizeOut > 0)
  {
    if (LFMT(con->EqpFormatOut) != CF_SHORT) return illegal_format;
    *(short *)data = (short)FeclogDepth;
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int setLogCommands(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqpSizeIn > 0)
  {
    if (!(con->EqpAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqpFormatIn) != CF_SHORT) return illegal_format;
    putCommandsInFeclog = *(short *)dataIn;
  }
  if (con->EqpSizeOut > 0)
  {
    if (LFMT(con->EqpFormatOut) != CF_SHORT) return illegal_format;
    *(short *)data = (short)putCommandsInFeclog;
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int setDebugLevel(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqpSizeIn > 0)
  {
    if (!(con->EqpAccess&CA_WRITE)) return illegal_read_write;
    if (LFMT(con->EqpFormatIn) != CF_SHORT) return illegal_format;
    NGdebug = *(short *)dataIn;
  }
  if (con->EqpSizeOut > 0)
  {
    if (LFMT(con->EqpFormatOut) != CF_SHORT) return illegal_format;
    *(short *)data = NGdebug;
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int getAppVersion(CONTRACT *con,char *dataIn,char *data)
{
  int len;
  if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
  if ((len=con->EqpSizeOut) < 1) return illegal_read_write;
  strncpy(data,appmkversion,len);
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int getTINEVersion(CONTRACT *con,char *dataIn,char *data)
{
  char s[32];
  int len;
  BYTE *ver = SystemVersion();
  if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqpSizeOut) < 1) return illegal_read_write;
  sprintf(s,"%d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);
  strncpy(data,s,len);
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int getServerOS(CONTRACT *con,char *dataIn,char *data)
{
  int len;
  if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqpSizeOut) < 1) return illegal_read_write;
  strncpy(data,OS_TAG,len);
  return 0;
}
int getServerTimeStr(CONTRACT *con,char *data,time_t t)
{
  int len;
  char s[32];
  if ((len=(int)con->EqpSizeOut) < 1) return illegal_read_write;

  switch (LFMT(con->EqpFormatOut))
  {
    case CF_TEXT:
      strncpy(s,ctime(&t),24); s[24] = 0;
      switch (len)
      {
        case 32: case 26: case 24: strncpy(data,s,24); break;
        case 16: strncpy(data,&s[4],15); data[15] = 0; break;
        default: return illegal_read_write;
      }
      break;
    case CF_LONG:
      *(UINT32 *)data = (UINT32)t;
      break;
    default:
      return illegal_format;
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int getServerStarttime(CONTRACT *con,char *dataIn,char *data)
{
  return getServerTimeStr(con,data,startuptime);
}
#ifdef MSDOS
#pragma argsused
#endif
int getAppDate(CONTRACT *con,char *dataIn,char *data)
{
  return getServerTimeStr(con,data,appmktime);
}
#ifdef MSDOS
#pragma argsused
#endif
int getServerDescription(CONTRACT *con,char *dataIn,char *data)
{
  int len;
  if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqpSizeOut) < 1) return illegal_read_write;
  strncpy(data,RPCFecEx.desc,len);
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int getServerLocation(CONTRACT *con,char *dataIn,char *data)
{
  int len;
  if (LFMT(con->EqpFormatOut) != CF_TEXT) return illegal_format;
  if ((len=(int)con->EqpSizeOut) < 1) return illegal_read_write;
  strncpy(data,RPCFecEx.loc,len);
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int getServerStats(CONTRACT *con,char *dataIn,char *data)
{
  UINT32 statsBuffer[16];
  int i;
  if (LFMT(con->EqpFormatOut) != CF_LONG) return illegal_format;
  statsBuffer[0] = AveBusyTime;
  statsBuffer[1] = gCycleCounts;
  statsBuffer[2] = gMaxCycleCounts;
  statsBuffer[3] = gSingleLinkCount;
  statsBuffer[4] = gClientMisses;
  statsBuffer[5] = gClientReconnects;
  statsBuffer[6] = gClientRetries;
  statsBuffer[7] = gContractMisses;
  statsBuffer[8] = gContractDelays;
  statsBuffer[9] = gBurstLimitReachedCount;
  statsBuffer[10] = (UINT32)(gDataTimeStampOffset * 1000);
  for (i=0; i<(int)con->EqpSizeOut && i<11; i++) ((UINT32 *)data)[i] = statsBuffer[i];
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int getDeniedUserList(CONTRACT *con,char *dataIn,char *data)
{
  int i;
  if (con->EqpSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqpFormatOut) != CF_NAME16) return illegal_format;
  for (i=0; i<(int)con->EqpSizeOut && i<NrOfDeniedUsers; i++)
  {
    strncpy(((NAME16 *)data)[i].name,DeniedUserList[i].UserName,16);
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int getRegisteredUserList(CONTRACT *con,char *dataIn,char *data)
{
  int i;
  if (con->EqpSizeOut == 0) return illegal_read_write;
  if (LFMT(con->EqpFormatOut) != CF_NAME16) return illegal_format;
  for (i=0; i<(int)con->EqpSizeOut && i<NrOfUsers; i++)
  {
    strncpy(((NAME16 *)data)[i].name,UserList[i].UserName,16);
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int clientsQuery(CONTRACT *con,char *dataIn,char *data)
{
  CLN *cln;
  int i, ncln;

  if (LFMT(con->EqpFormatOut) != CF_STRUCT) return illegal_format;
  if (con->EqpSizeOut%CLIENT_QUERY_SIZE) return illegal_read_write;
  if ((ncln=(int)con->EqpSizeOut/CLIENT_QUERY_SIZE) == 0) return dimension_error;

  for (i=0; i<ncln && i<nconsumer; i++)
  {
    if (Consumers[i] == NULL) continue;
    cln = Consumers[i];
    PrepOutgoingCLNQS(&data[i*CLIENT_QUERY_SIZE],cln);
  }
  return 0;
}

#ifdef MSDOS
#pragma argsused
#endif
int gateFunction(CONTRACT *con,char *dataIn,char *data)
{ /* keep up apperances ... */
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int EGUQueryFunction(CONTRACT *con,char *dataIn,char *data)
{
  int idx,num,len;
  ExportPropertyListStruct *prp,prpTmp;
  ExportListStruct *el;
  PropertyQueryStruct *pqs = NULL;
  XPropertyQueryStruct *xpqs = NULL;
  char *c, Property[32],dsc[32];
  DUSTRING *ustr;

  if ((len=(int)con->EqpSizeOut) == 0) return illegal_read_write;
  strncpy(Property,con->EqpProperty,32);
  if ((c=strstr(Property,".EGU")) == NULL) return illegal_property;
  *c = 0;

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
    if (el->XEqpPropertyQueryFcn != NULL)        /* registered query fcn */
    {
      prp = NULL;
      num = el->XEqpPropertyQueryFcn(con->EqpDeviceName,&xpqs);
      if (num <= 0) return -num;
      for (idx=0; idx<num; idx++)
      {
        /* Is .EGU query managed by query function ? */
        if (!strncmp(con->EqpProperty,xpqs[idx].prpName,32)) return ambiguous;
        if (strncmp(Property,xpqs[idx].prpName,32)) continue;
        strncpy(prpTmp.egu.units,xpqs[idx].prpUnits,32);
        prpTmp.egu.min = xpqs[idx].prpMinValue;
        prpTmp.egu.max = xpqs[idx].prpMaxValue;
        prpTmp.egu.graph = xpqs[idx].prpGraphType;
        prp = &prpTmp;
      }
      if (prp != NULL) goto checkformat;
      return illegal_property;
    }
    if (el->EqpPropertyQueryFcn != NULL)        /* registered query fcn */
    {
      prp = NULL;
      num = el->EqpPropertyQueryFcn(con->EqpDeviceName,&pqs);
      if (num <= 0) return -num;
      for (idx=0; idx<num; idx++)
      {
        /* Is .EGU query managed by query function ? */
        if (!strncmp(con->EqpProperty,pqs[idx].prpName,32)) return ambiguous;
        if (strncmp(Property,pqs[idx].prpName,32)) continue;
        strncpy(dsc,pqs[idx].prpDescription,32);
        ParsePropertyDescription(dsc,&prpTmp);
        prp = &prpTmp;
      }
      if (prp != NULL) goto checkformat;
      return illegal_property;
    }
    for (idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
      for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
    {
      if (strncmp(Property,prp->prpName,32)) continue;
      goto checkformat;
    }
    if (prp == NULL)
    {
      memset(&prpTmp,0,sizeof(ExportPropertyListStruct));
      prp = &prpTmp;
    }
checkformat:
    switch (LFMT(con->EqpFormatOut))
    {
      case CF_USTRING:
        ustr = (DUSTRING *)data;
        ustr->ival  = (UINT32)prp->egu.graph;
        ustr->f1val = (float)prp->egu.min;
        ustr->f2val = (float)prp->egu.max;
        ustr->tm = (UINT32)startuptime;
        strncpy(ustr->str,prp->egu.units,16);
        return 0;
      case CF_NAME16:
        if (len != 1) return illegal_read_write;
        strncpy((char *)data,prp->egu.units,16);
        return 0;
      case CF_NAME32:
        if (len != 1) return illegal_read_write;
        strncpy((char *)data,prp->egu.units,32);
        return 0;
      case CF_TEXT:
        if (len < 1 || len > 16) return illegal_read_write;
        strncpy((char *)data,prp->egu.units,len);
        return 0;
      case CF_FLOAT:
        if (len != 2) return illegal_read_write;
        ((float *)data)[0] = (float)prp->egu.min;
        ((float *)data)[1] = (float)prp->egu.max;
        return 0;
      default:
        return illegal_format;
    }
  }
  return non_existent_elem;
}

int AllowRemoteExit = FALSE;
#ifdef MSDOS
#pragma argsused
#endif
int srvExit(CONTRACT *con,char *dataIn,char *data)
{
  if (!AllowRemoteExit) return not_accepted;
  if (!(con->EqpAccess&CA_WRITE)) return illegal_read_write;
  feclog("remote exit from %s",GetCaller());
  quitServer();
  return 0;
}

#ifdef MSDOS
#pragma argsused
#endif
int lockAccess(CONTRACT *con,char *dataIn,char *data)
{
  short sval[2];
  int actionTaken = FALSE;
  int passed = FALSE;
  CLN *cln = EQPconsumerlist->client;

  if (con->EqpAccess&CA_WRITE)
  {
    if (con->EqpSizeIn != 2) return illegal_read_write;
    if (LFMT(con->EqpFormatIn) != CF_SHORT) return illegal_format;
    sval[0] = ((short *)dataIn)[0];
    sval[1] = ((short *)dataIn)[1];
    if (!putCommandsInFeclog)
    {
      feclog("ACCESSLOCK %d called from %s",sval[0],GetCaller());
    }
    if (sval[0] < 0 || sval[0] > LOCK_ABORT) return out_of_range;
    if (IsMemberLockSet(cln)) passed = TRUE;
    else if (currentAccessLock.lockType != LOCK_PERSISTENT) passed = TRUE;
    else if (sval[0] == LOCK_ABORT) passed = TRUE;
    if (!passed) return access_denied;
    if (sval[0] == LOCK_ABORT || sval[0] == LOCK_UNLOCKED)
    {
      memset(&currentAccessLock,0,sizeof(currentAccessLock));
    }
    else
    {
      memcpy(&currentAccessLock.cln,cln,sizeof(CLN));
      currentAccessLock.lockType = sval[0];
      currentAccessLock.lockDuration = sval[1];
      currentAccessLock.lockStart = (UINT32)time(NULL);
    }
    actionTaken = TRUE;
  } 
  if (con->EqpSizeOut == 2)
  { 
    if (LFMT(con->EqpFormatOut) != CF_NAME32) return illegal_format;
    strncpy(data,currentAccessLock.cln.userName,16);
    ntoa(&currentAccessLock.cln,&data[32]);
    actionTaken = TRUE;
  }

  return actionTaken ? 0 : illegal_read_write;
}
#ifdef MSDOS
#pragma argsused
#endif
int nhistoryRecordsQuery(CONTRACT *con,char *dataIn,char *data)
{
  int i,n=0;
  for (i=0; i<nHistoryRecords; i++)
  {
    if (strcmp(con->EqpName,historyRecords[i].c.EqpName)) continue;
    n++;
  }
  return T_numQuery(con,data,(short)n);
}
#ifdef MSDOS
#pragma argsused
#endif
int historyRecordsQuery(CONTRACT *con,char *dataIn,char *data)
{
  int i,n=0;

  if (LFMT(con->EqpFormatOut) != CF_NAME32) return illegal_format;
  for (i=0; i<nHistoryRecords && n<(int)con->EqpSizeOut; i++)
  {
    if (strcmp(con->EqpName,historyRecords[i].c.EqpName)) continue;
    strncpy(((NAME32 *)data)[n++].name,historyRecords[i].c.EqpProperty,32);
  }
  return 0;
}
#ifdef MSDOS
#pragma argsused
#endif
int NAMQueryFunction(CONTRACT *con,char *dataIn,char *data)
{
  int cc, i,slen;
  ExportPropertyListStruct *prp;
  char *c,tgtprp[64];
  strncpy(tgtprp,con->EqpProperty,32);
  if ((c=strstr(tgtprp,".NAM")) != NULL) *c = 0;
  if ((prp=AssignDeviceListFromProperty(con->EqpName,tgtprp)) != NULL)
  {
    if (con->EqpSizeOut == 0) return illegal_read_write;
    switch (LFMT(con->EqpFormatOut))
    {
      case CF_NAME16:
      case CF_NAME32:slen = fmtsizeof(con->EqpFormatOut); break;
      default: return illegal_format;
    }
    for (i=0; i<(int)con->EqpSizeOut && i<prp->numDevices; i++)
    {
      strncpy(&data[i*slen],prp->devNames[i].name,EQP_DEVICE_SIZE);
    }
    return 0;
  }
  cc = deviceQuery(con,dataIn,data);
  if (cc & CE_SENDDATA) cc = 0;
  return cc;
}
#ifdef MSDOS
#pragma argsused
#endif
int DESCQueryFunction(CONTRACT *con,char *dataIn,char *data)
{
  int idx,len,num;
  ExportPropertyListStruct *prp;
  ExportListStruct *el;
  XPropertyQueryStruct *xpqs = NULL;
  PropertyQueryStruct *pqs = NULL;
  char *c, Property[32];

  strncpy(Property,con->EqpProperty,32);
  if ((c=strstr(Property,".DESC")) == NULL &&
      (c=strstr(Property,".DSC")) == NULL) return illegal_property;
  *c = 0;
  len = (int)con->EqpSizeOut;
  switch (LFMT(con->EqpFormatOut))
  {
    case CF_TEXT:
      if (len < 1 || len > 64) return illegal_read_write;
      break;
    case CF_NAME16:
      if (len != 1) return illegal_read_write;
      len = 16;
      break;
    case CF_NAME32:
      if (len != 1) return illegal_read_write;
      len = 32;
      break;
    case CF_NAME48:
      if (len != 1) return illegal_read_write;
      len = 48;
      break;
    case CF_NAME64:
      if (len != 1) return illegal_read_write;
      len = 64;
      break;
    default:  return illegal_format;
  }

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
    if (el->XEqpPropertyQueryFcn != NULL)        /* registered query fcn */
    {
      num = el->XEqpPropertyQueryFcn(con->EqpDeviceName,&xpqs);
      if (num <= 0) return -num;
      for (idx=0; idx<num; idx++)
      {
        if (strncmp(xpqs[idx].prpName,Property,32)) continue;
        strncpy(data,xpqs[idx].prpDescription,len);
        return 0;
      }
    }
    if (el->EqpPropertyQueryFcn != NULL)        /* registered query fcn */
    {
      num = el->EqpPropertyQueryFcn(con->EqpDeviceName,&pqs);
      if (num <= 0) return -num;
      for (idx=0; idx<num; idx++)
      {
        if (strncmp(pqs[idx].prpName,Property,32)) continue;
        strncpy(data,pqs[idx].prpDescription,MIN(len,32));
        return 0;
      }
    }
    for (idx=0; idx<PROPERTY_HASH_SIZE+1; idx++)
      for (prp=el->EqpProperty[idx]; prp; prp=prp->next)
    {
      if (strncmp(prp->prpName,Property,32)) continue;
      strncpy(data,prp->prpDescription,len);
      return 0;
    }
    return illegal_property;
  }
  return non_existent_elem;
}
EXPORT int GetLocalHistoryData(char *eqp,char *prp, char *dev,DTYPE *dout,DTYPE *din)
{
  CONTRACT cn;
  strncpy(cn.EqpName,eqp,EQP_NAME_SIZE);
  strncpy(cn.EqpProperty,prp,EQP_PROP_SIZE);
  strncpy(cn.EqpDeviceName,dev,EQP_DEVICE_SIZE);
  cn.EqpSizeOut = dout->dArrayLength;
  cn.EqpFormatOut = BFMT(dout->dFormat);
  cn.EqpSizeIn = din->dArrayLength;
  cn.EqpFormatIn = BFMT(din->dFormat);
  return HISTQueryFunction(&cn,din->data.cptr,dout->data.cptr);
}
#ifdef MSDOS
#pragma argsused
#endif
int HISTQueryFunction(CONTRACT *con,char *dataIn,char *data)
{
  SINT32 numread, nr;
  char *c, Property[32];
  time_t stop=time(NULL)-1; /* default values */
  double dstop = (double)stop, dstart = 0;
  float lwr=0,upr=0;
  int n,fmt,cc=0,index=0,sample=0,recsreadFlag=0,isSnapshot=0;
  SINT32 *lptr;
  double *dptr;
  ExportListStruct *el;
  PropertyQueryStruct *pqs = NULL;
  XPropertyQueryStruct *xpqs = NULL;
  BYTE *dataptr = (BYTE *)data;

  /* check for server redirection from possible property query functions ... */
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(con->EqpName,el->EqpName,EQP_NAME_SIZE)) continue;
    if (el->XEqpPropertyQueryFcn != NULL &&
        (n=el->XEqpPropertyQueryFcn(con->EqpDeviceName,&xpqs)) <= 0) ccerr(-n);
    if (el->EqpPropertyQueryFcn != NULL &&
        (n=el->EqpPropertyQueryFcn(con->EqpDeviceName,&pqs)) <= 0) ccerr(-n);
    break;
  }
  /* proceed ...*/
  strncpy(Property,con->EqpProperty,32);
  if ((c=strstr(Property,".AR")) != NULL)  
  {
    *c = 0;
    return getRemoteHistoryServer(con->EqpName,Property);
  }
  if (strstr(Property,".HIST@") != NULL || strstr(Property,".HST@") != NULL)
    isSnapshot = TRUE;
  if ((c=strstr(Property,".HIST")) == NULL &&
      (c=strstr(Property,".HST")) == NULL) return illegal_property;
  *c = 0;
  if (con->EqpSizeOut == 0) return illegal_read_write;
  n = (int)con->EqpSizeOut;
  switch (fmt = LFMT(con->EqpFormatOut))
  {
    case CF_DOUBLE: case CF_FLOAT: case CF_LONG: case CF_INTFLTINT:
      recsreadFlag = 0;
      break;
    default:
      break;
  }
  if (con->EqpSizeIn > 0)
  {
    switch (LFMT(con->EqpFormatIn))
    {
      case CF_LONG:
        lptr = (SINT32 *)dataIn;
        switch (con->EqpSizeIn)
        {
          case 5: recsreadFlag = (int)lptr[4];
          case 4: sample = (int)lptr[3];
          case 3: index = (int)lptr[2];
          case 2: if (lptr[1] < stop) dstop = (double)lptr[1];
          case 1: dstart = (double)lptr[0];
          default: break;
        }
        if (sample < 0) sample = 0;
        break;
      case CF_DOUBLE:
        dptr = (double *)dataIn;
        switch (con->EqpSizeIn)
        {
          case 5: recsreadFlag = (int)dptr[4];
          case 4: sample = (int)dptr[3];
          case 3: index = (int)dptr[2];
          case 2: if (dptr[1] < stop) dstop = dptr[1];
          case 1: dstart = dptr[0];
          default: break;
        }
        if (sample < 0) sample = 0;
        break;
      case CF_FWINDOW:
        dstart = (double)((FWINDOW *)dataIn)[0].i1val;
        dstop = (double)((FWINDOW *)dataIn)[0].i2val;
        lwr = ((FWINDOW *)dataIn)[0].f1val;
        upr = ((FWINDOW *)dataIn)[0].f2val;
        if (lwr < upr) sample = -1;
        break;
      case CF_WINDOW:
        dstart = (double)((DWINDOW *)dataIn)[0].t1;
        dstop = (double)((DWINDOW *)dataIn)[0].t2;
        sample = ((DWINDOW *)dataIn)[0].i2val;
        break;
      default:
        return illegal_format;
    }
  }
  else
  {
    dstart = (double)getStartTimeFromDepth(con->EqpName,Property,n);
    if (dstart < 0) ccerr((int)(-dstart));
  }

  if (dstart < 0) dstart = 0;
  if (dstop < dstart) dstop = (double)time(NULL);
  if (isSnapshot) 
  { 
    if (dstart < 1.0) return argument_list_error;
    if (fmt > CF_FLOAT) return illegal_format;
    dstop = dstart;
    /* snapshot read doesn't use sample, index or recsread ... */
    sample = 0;
    index = 0;
    recsreadFlag = 0;
    lwr = upr = 0.0;
  }

  numread = getLTSData(con->EqpName,Property,con->EqpDeviceName,index,dstart,dstop,lwr,upr,
                       dataptr,fmt,n,sample,recsreadFlag);

  if (numread < 0) ccerr((int)(-numread));
  if (numread > 0) 
  {
    if (!isSnapshot && numread < (SINT32)con->EqpSizeOut - 1)
    {
      dataptr += numread * fmtsizeof(fmt);
      dstart = getDataTimeStamp() + 1.0;
      n -= numread;
    }
    else 
    {
      ccerr(0);
    }
  }

  nr = getSTSData(con->EqpName,Property,con->EqpDeviceName,index,dstart,dstop,lwr,upr,
                  dataptr,fmt,n,sample,recsreadFlag);
  if (nr < 0) ccerr((int)(-nr));
  numread += nr;
  if (numread < (SINT32)con->EqpSizeOut - 1) con->EqpSizeOut = numread;
err:
  return  cc;
}
#ifdef WINNT
int gCntNotPosted = 0;
int getNotPostedCount(CONTRACT *con,char *dataIn,char *data)
{
  if (con->EqpSizeOut != 1) return illegal_read_write;
  switch (LFMT(con->EqpFormatOut))
  {
    case CF_SHORT:
      *(short *)data = (short)gCntNotPosted; break;
    case CF_LONG:
      *(int *)data = gCntNotPosted; break;
    case CF_FLOAT:
      *(float *)data = (float)gCntNotPosted; break;
    case CF_DOUBLE:
      *(double *)data = (double)gCntNotPosted; break;
    default: return illegal_format;
  }
  return 0;

}
#endif
_StockPropertyStructEx StockProperty[] =
{
     {"PROPS",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE|AT_CHANNEL,"","Exported Properties",propsQuery},
     {"PROPS",100,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQS","Exported Properties",propsQuery},
     {"PROPS",20,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery},
     {"PROPS",20,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery},
     {"PROPS",10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery},
     {"PROPS",10,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery},
     {"NPROPS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Exported Properties",npropsQuery},
     {"PROPERTIES",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","Exported Properties",propsQuery},
     {"PROPERTIES",100,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQS","Exported Properties",propsQuery},
     {"PROPERTIES",20,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery},
     {"PROPERTIES",20,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"PQSX","Exported Properties",propsQuery},
     {"PROPERTIES",10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery},
     {"PROPERTIES",10,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"XPQS","Exported Properties",propsQuery},
     {"NPROPERTIES",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Exported Properties",npropsQuery},
     {"DEVICES",512,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Device Names",deviceQuery},
     {"DEVICES",512,1,CF_NAME16,CF_NAME16,CA_READ,AT_SINGLE,"","Device Names",deviceQuery},
     {"NDEVICES",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Devices",ndeviceQuery},
     {"ALARMS",512,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AMS","Total Current Alarm List",alarmsQuery},
     {"NALARMS",5,0,CF_LONG,CF_NULL,CA_READ,AT_SINGLE|AT_STRUCTURED,"","Current Alarm Information",nalarmsQuery},
     {"ALMDEFS",32,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"ADS","Defined Alarm Descriptions",alarmDefsQuery},
     {"NALMDEFS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Alarm Defined Alarms",nalarmDefsQuery},
     {"ALARMSEXT",512,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AMS","Specified Current Alarm List",alarmsQueryEx}, 
     {"MESSAGE",0,256,CF_NULL,CF_TEXT,CA_WRITE,AT_SINGLE,"","Accept Message",msgDisplay},
     {"ACTIVITY",68,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"","Current Server Activity",activityQuery},  
     {"ACTIVITY",1,0,CF_STRUCT,CF_NULL,CA_READ,AT_SINGLE,"AQS","Current Server Activity",activityQuery},  
     {"CONTRACTS",920,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"","Current Contract List",contractsQuery},
     {"CONTRACTS",10,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"CONQS","Current Contract List",contractsQuery},
     {"CLIENTS",540,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"","Current Client List",clientsQuery},
     {"CLIENTS",10,0,CF_STRUCT,CF_NULL,CA_READ,AT_UNKNOWN,"CLNQS","Current Client List",clientsQuery},
     {"SRVEXIT",0,0,CF_NULL,CF_NULL,CA_WRITE,AT_UNKNOWN,"","Stop Server (when enabled)",srvExit},
     {"ACCESSLOCK",0,2,CF_NULL,CF_SHORT,CA_WRITE,AT_UNKNOWN,"","lock type/lock duration",lockAccess},
     {"ACCESSLOCK",2,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","lock owner/address",lockAccess},
     {"ADDUSER",0,16,CF_NULL,CF_TEXT,CA_WRITE,AT_UNKNOWN,"","Add User to Access List",addRegisteredUser},
     {"DELUSER",0,16,CF_NULL,CF_TEXT,CA_WRITE,AT_UNKNOWN,"","Remove User from Access List",rmvRegisteredUser},
     {"USERS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read User Access List",getRegisteredUserList},
     {"DENIEDUSERS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read Denied User List",getDeniedUserList},
     {"NUSERS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Users with Access",getNumRegisteredUsers},
     {"STOCKPROPS",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery},
     {"STOCKPROPS",68*10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery},
     {"STOCKPROPS",188*10,1,CF_STRUCT,CF_NAME32,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery},
     {"STOCKPROPS",188*10,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"","Stock Properties",stockpropsQuery},
     {"STOCKPROPS",10,32,CF_STRUCT,CF_TEXT,CA_READ,AT_SINGLE,"XPQS","Stock Properties",stockpropsQuery},
     {"NSTOCKPROPS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of Stock Properties",nstockpropsQuery},
     {"IPNETS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read IP Access List",getRegisteredIPNetList},
     {"NIPNETS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of IP Nets with Access",getNumRegisteredIPNets},
     {"ADDIPNET",0,16,CF_NULL,CF_TEXT,CA_WRITE,AT_UNKNOWN,"","Add IP Net to Access List",addRegisteredIPNet},
     {"IPXNETS",256,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","Read IPX Access List",getRegisteredIPXNetList},
     {"NIPXNETS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of IPX Nets with Access",getNumRegisteredIPXNets},
     {"DEBUGLEVEL",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","get debug level",setDebugLevel},
     {"DEBUGLEVEL",1,1,CF_SHORT,CF_SHORT,CA_READ|CA_WRITE,AT_SCALAR,"","get/set debug level",setDebugLevel},
     {"LOGCOMMANDS",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","read en/dis-able command logging flag",setLogCommands},
     {"LOGCOMMANDS",1,1,CF_SHORT,CF_SHORT,CA_READ|CA_WRITE,AT_SCALAR,"","en/dis-able command logging",setLogCommands},
     {"LOGDEPTH",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","get log depth in lines",setLogDepth},
     {"LOGDEPTH",1,1,CF_SHORT,CF_SHORT,CA_READ|CA_WRITE,AT_SCALAR,"","get/set log depth in lines",setLogDepth},
     {"LOGFILE",8000,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","get log file",getFecLog},
     {"SRVVERSION",16,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","TINE Server Version Number",getTINEVersion},
     {"SRVOS",16,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","General Server OS",getServerOS},
     {"SRVDESC",64,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Description",getServerDescription},
     {"SRVLOCATION",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Location",getServerLocation},
     {"SRVSTATS",16,0,CF_LONG,CF_NULL,CA_READ,AT_SINGLE|AT_STRUCTURED,"","Server Statistics counters",getServerStats},
     {"SRVALIASLIST",512,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","Alias List (Name/Alias pairs)",getServerAliasList},
     {"NHISTORIES",1,0,CF_SHORT,CF_NULL,CA_READ,AT_SCALAR,"","Number of history properties",nhistoryRecordsQuery},
     {"HISTORIES",32,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","history properties",historyRecordsQuery},
     {"STRUCTFORMAT",512,1,CF_INTINT,CF_NAME16,CA_READ,AT_SINGLE,"","Registered structure information",taggedStructQuery},
     {"SRVSTARTTIME",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Server Startup Time",getServerStarttime},
     {"SRVSTARTTIME",1,0,CF_LONG,CF_NULL,CA_READ,AT_SCALAR,"","Server Startup Time",getServerStarttime},
     {"APPDATE",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Application Compile Time",getAppDate},
     {"APPDATE",1,0,CF_LONG,CF_NULL,CA_READ,AT_SCALAR,"","Application Compile  Time",getAppDate},
     {"APPVERSION",32,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","Application Version Number",getAppVersion},
     {"SRVSELFTEST",32000,0,CF_TEXT,CF_NULL,CA_READ,AT_SCALAR,"","get self-test file",getSelfTestFile},
     {"SRVLOGFILES",32,0,CF_NAME64,CF_NULL,CA_READ,AT_SCALAR,"","get log file list",getFecLogFileList},
     {"SRVLOGFILES",32,64,CF_NAME64,CF_TEXT,CA_READ,AT_SCALAR,"","get log file list",getFecLogFileList},
     {"SRVLOGFILE",32000,0,CF_TEXT,CF_TEXT,CA_READ|CA_WRITE,AT_SCALAR,"","get specific log file ",getFecLogFile},
     {"SRVLASTACCESS",5,0,CF_NAME32,CF_NULL,CA_READ,AT_SINGLE,"","get most recent access",getLastAccessList},
     {"SRVADDR",5,0,CF_NAME16,CF_NULL,CA_READ,AT_SINGLE,"","get server address parameters",getSrvAddressParams},
#    ifdef WINNT
     {"SRVNOTPOSTED",1,0,CF_LONG,CF_NULL,CA_READ,AT_SCALAR,"","Number Contracts not posted",getNotPostedCount},
#    endif
};
int NrStockProperties = (sizeof(StockProperty)/sizeof(StockPropertyStructEx));

HashXRefStruct **StockPropXRefTable = NULL;

int makeStockPropXRefTable(void)
{
  int k,idx;
  HashXRefStruct *xref;

  if ((StockPropXRefTable=(HashXRefStruct **)SystemCalloc(STOCKPROP_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
    return out_of_server_memory;
  for (k=0; k<NrStockProperties; k++)
  {
    idx = (int)(ElfHash((unsigned char *)StockProperty[k].prpName) % STOCKPROP_HASH_SIZE);
    for (xref=StockPropXRefTable[idx]; xref && xref->index != k; xref = xref->next)
    {
      if (!strncmp(StockProperty[k].prpName,StockProperty[xref->index].prpName,32)) break;
    }
    if (xref == NULL) 
    {
      if ((xref=(HashXRefStruct *)SystemCalloc(1,sizeof(HashXRefStruct))) == NULL) return out_of_server_memory;
      xref->index = k;
      xref->next = StockPropXRefTable[idx];
      StockPropXRefTable[idx] = xref;
    }
  }
  return 0;
}
_StockModifierStruct StockModifier[] =
{
  {".EGU",EGUQueryFunction},
  {".HIST",HISTQueryFunction},
  {".HST",HISTQueryFunction},
  {".HIST@",HISTQueryFunction},
  {".HST@",HISTQueryFunction},
  {".ARCH",HISTQueryFunction},
  {".ARC",HISTQueryFunction},
  {".AR",HISTQueryFunction},
  {".NAM",NAMQueryFunction},
  {".DESC",DESCQueryFunction},
  {".DSC",DESCQueryFunction},
  {".BIT.",gateFunction},
  {".MASK.",gateFunction},
  {".GATE.",gateFunction},
};
int NrStockModifiers = (sizeof(StockModifier)/sizeof(StockModifierStruct));

short IsStockProperty(char *Property)
{
  if (strchr(Property,'*') != NULL) return STOCK_PROP; /* maps to "PROPERTIES" */
  return GetStockPropertyId(Property) < 0 ? EMPTY_PROP : STOCK_PROP;
}
short IsMetaProperty(char *Property,char **baseProperty,char **metaProperty)
{
  int i;
  char *c;
  static char metaPropertyString[64];

  strncpy(metaPropertyString,Property,EQP_PROP_SIZE);
  for (i=0; i<NrStockModifiers; i++)
  {
    if ((c=strstr(metaPropertyString,StockModifier[i].modName)) != NULL)
    {
      *c++ = 0;
      if (baseProperty != NULL) *baseProperty = metaPropertyString;
      if (metaProperty != NULL) *metaProperty = c;
      return TRUE;
    }
  }
  return FALSE;
}

short IsGatedModifier(CONTRACT *con,char *baseProperty,UINT32 *gate)
{
  char *c;
  int bit;

  if (baseProperty == NULL || gate == NULL) return FALSE;
  if (LFMT(con->EqpFormatOut) != CF_BYTE &&
      LFMT(con->EqpFormatOut) != CF_SHORT &&
      LFMT(con->EqpFormatOut) != CF_LONG) return FALSE;
  strncpy(baseProperty,con->EqpProperty,EQP_PROP_SIZE);
  if ((c=strstr(baseProperty,".BIT.")) != NULL)
  {
    *c = 0;
    c += 5;
    bit = atoi(c);
    if (bit < 0 || bit >= 8*fmtsizeof(con->EqpFormatOut)) return FALSE;
    *gate = 0x01 << bit;
    return GATE_BOOLEAN;
  }
  if ((c=strstr(baseProperty,".MASK.")) != NULL)
  {
    *c = 0;
    c += 6;
    *gate = strtol(c,NULL,0);
    return GATE_BOOLEAN;
  }
  if ((c=strstr(baseProperty,".GATE.")) != NULL)
  {
    *c = 0;
    c += 6;
    *gate = strtol(c,NULL,0);
    return GATE_FILTER;
  }
  return FALSE;
}
