
#ifndef SWAPLIB_H_
#define SWAPLIB_H_
/**
 * \internal
 * \file swaplib.h
 * \brief Header for swaplib.c
 * \see swaplib.c
 */
#include "project.def"

#include "rpcdefs.h"
#include "rpctypes.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#undef NONE
#define NONE   -1
#define RTOL    0  /* Swap: Remote to Local */
#define LTOR    1  /* Swap: Local to Remote  */

/**
 * \internal
 * \brief Returns the address offset of a field in a structure.
 * 
 * \param struc  Structure name
 * \param field  Field name
 *
 */
#define OFFSETIN(struc,field)   ((size_t)&(((struc *)0L)->field))

#if defined(NONALIGNED) || defined(INTEL) || defined(VAX_VMS_IEEE)
#  define SwapBytes(dat,size,fmt,dir) ;
#  define ISWAP(w) (w)
#  define _ISWAP(w) (w)
#  define LSWAP(w) (w)
#  define _LSWAP(w) (w)
#else
#  ifdef VAX_VMS   /* ints unswapped, special swapping for floats, doubles */
#    define ISWAP(w) (w)
#    define _ISWAP(w) (w)
#    define LSWAP(w) (w)
#    define _LSWAP(w) (w)
#  else
#    define ISWAP(w) ((((w)&0xff00)>>8)|(((w)&0xff)<<8))
#    define _ISWAP(w) { tmp = ((BYTE *)&w)[0];\
                        ((BYTE *)&w)[0] = ((BYTE *)&w)[1];\
                        ((BYTE *)&w)[1] = tmp; }
#    define LSWAP(w) ((w>>24)|(w<<24)|((w&0xff00)<<8)|((w&0x00ff0000L)>>8))
#    define _LSWAP(w) { tmp = ISWAP(((short *)&w)[0]);\
	             ((short *)&w)[0] = ISWAP(((short *)&w)[1]);\
                     ((short *)&w)[1] = tmp; }
#  endif
#  define _DSWAP(w) { tmp = ISWAP(((short *)&w)[0]);\
		    ((short *)&w)[0] = ISWAP(((short *)&w)[3]);\
		    ((short *)&w)[3] = tmp;\
		    tmp = ISWAP(((short *)&w)[1]);\
		    ((short *)&w)[1] = ISWAP(((short *)&w)[2]);\
		    ((short *)&w)[2] = tmp; }
int SwapBytes(char *dat,int size,int fmt,int dir);
#endif

/**
 * \internal
 * \brief Format description of a structure element
 * 
 * To describe a whole structure, structFormat is used as
 * a linked list. In TINE, linked lists are mostly used in
 * reverse order.
 */
struct structFormat
{
  int   siz;   /**< Size in bytes. */
  int   fmt;   /**< TINE CF_ format. */
  int   off;   /**< ??? */
  int  addr;   /**< Byte offset within the structure. */
  struct structFormat *nxt; /**< Pointer to next element. */
};


/**
 * \internal
 * \brief Structure description for tagged structures
 * 
 * A structStruct contains a structFormat list of fields. 
 * It does not only describe a structure, but also describes
 * an array of data of that structure. 
 * At the same time, it is also a buffer.
 * It can also be used as an element of a linked list.
 * 
 * A structStruct is used for tagged structures.
 */
struct structStruct
{
  char name[12];           /**< Structure tag */
  int  rawlen;             /**< Raw/network data size of the described structure */
  int  size;               /**< Local structure size (memory size) of the described structure */
  int  num;                /**< Maximum number of structure sets buffered (array size) */
  BYTE *sdata;             /**< Buffer for data to be sent over the network */
  struct structFormat *s;  /**< structure definition */
  struct structStruct *next; /**< Pointer to the next element in a linked list.*/
};

/**
 * \internal
 * \brief Global list of tagged structures.
 */
extern struct structStruct *structList;

struct structStruct *getStructFromTag(char *tag);
struct structStruct *findstruct(char *tag);
int addtostruct(void *target,int addr,int size, int fmt);
EXPORT int fillstruct(char *tag,int addr,int size,int fmt);
EXPORT int GetStructSize(char *tag);
int _StructSwap(struct structFormat *s,unsigned char *dat,unsigned char *struc,int dir);
int StructSwap(char *tag,unsigned char *dat,unsigned char *struc,int dir);
int StructSwapEx(char *tag,unsigned char *dat,unsigned char *struc,int dir,int num,int structsize);
int _StructSwapEx(struct structFormat *s,unsigned char *dat,unsigned char *struc,int dir,int num,int structsize);
int initControlStructs(void);
void RegisterLongDataStructs(void);
int queryStructFormat(char *tag,INTINT *sfq,int *sfqsiz);
void freeStruct(struct structFormat *sf);

/**
 * \brief Field list of structure HeaderStruct. 
 */
extern struct structFormat *HeaderStructStruct;

/**
 * \brief Structure description for EqpInfoStruct.
 */
extern struct structFormat *EqpInfoStructStruct;

/**
 * \brief Field list of structure CONTRACT4
 */
extern struct structFormat *CONTRACT4Struct;

/**
 * \brief Field list of structure CONTRACT.
 */
extern struct structFormat *CONTRACTStruct;

/**
 * \brief Field list of structure ProducerHeaderStruct4.
 */
extern struct structFormat *ProducerHeaderStruct4Struct;

/**
 * \brief Field list of structure ProducerHeaderStruct.
 */
extern struct structFormat *ProducerHeaderStructStruct;

/**
 * \brief Field list of structure SUBSCRIPTION4
 */
extern struct structFormat *SUBSCRIPTION4Struct;

/**
 * \brief Field list of structure SUBSCRIPTION.
 */
extern struct structFormat *SUBSCRIPTIONStruct;

/**
 * \brief Field list of structure RPCDataStruct.
 */
extern struct structFormat *RPCDataStructStruct;

/**
 * \brief Field list of structure RPCFecStruct.
 */
extern struct structFormat *RPCFecStructStruct;

/**
 * \brief Field list of structure ADSStruct.
 */
extern struct structFormat *ADSStructStruct;

/**
 * \brief Field list of structure AMSStruct.
 */
extern struct structFormat *AMSStructStruct;

/**
 * \brief Field list of structure PropertyQueryStruct.
 */
extern struct structFormat *PropertyQueryStructStruct;

/**
 * \brief Field list of structure PropertyQueryStructEx.
 */
extern struct structFormat *PropertyQueryStructExStruct;

/**
 * \brief Field list of structure XPropertyQueryStruct.
 */
extern struct structFormat *XPropertyQueryStructStruct;

/**
 * \brief Field list of structure ActivityQueryStruct.
 */
extern struct structFormat *ActivityQueryStructStruct;

/**
 * \brief Field list of structure ServerStateStruct.
 */
extern struct structFormat *ServerStateStructStruct;

/**
 * \brief Field list of structure GlobalDataHdr.
 */
extern struct structFormat *GlobalDataHdrStruct;

/**
 * \brief Field list of structure ClientQueryStruct.
 */
extern struct structFormat *ClientQueryStructStruct;

/**
 * \brief Field list of structure SpectrumHeader.
 */
extern struct structFormat *SpectrumHeaderStruct;

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* SWAPLIB_H_ */
