/** \file bufsrv.c
 * \brief TINE buffered server documentation 
 *
 */
#include "prolog.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/timeb.h>
#include "rpctypes.h"
#include "errors.h"
#include "database.h"
#include "brdcast.h"
#include "servdef.h"
#include "toolkit.h"
#include "tbufsrv.h"

#ifdef _USRDLL
void TineStartCycler(void);
#endif

#define MAX_BUFFERED_PROPERTIES 512
char gServerWindow[64];
char gServerName[64];
char gEqpName[16];
int gNumDevices = 0;
int gSystemInitialized = FALSE;
int gTerminateServer = FALSE;

typedef struct
{
  char eqm[8];
  char nam[32];
  DTYPE dout;
  DTYPE din;
  DTYPE blk;
  float max;
  float min;
  double dataTimeStamp;
  int inputChanged;
  int interimCompletionCode;
  int arrayType;
  int (*callback)(void);
  void (*notifier)(int);
  int notifierId;
} BufferedProperty;

BufferedProperty bufferedProperties[MAX_BUFFERED_PROPERTIES];
int nBufferedProperties = 0;
NAME32 registeredProperties[MAX_BUFFERED_PROPERTIES];
void (*gBufferedNotifier)(int) = NULL;
int gBufferedNotifierId = 0;
char gPropertyNofified[32];
char gDeviceNotified[32];

int areValuesOutOfBounds(int id,DTYPE *din)
{
  int i;

  if (bufferedProperties[id].min == 0 && bufferedProperties[id].max == 0) return 0;
  switch (bufferedProperties[id].din.dFormat)
  {
    case CF_BYTE: 
      for (i=0; i<(int)bufferedProperties[id].din.dArrayLength; i++)
      {        
        if ((din->data.bptr[i] < bufferedProperties[id].min) ||
            (din->data.bptr[i] > bufferedProperties[id].max))
          return out_of_range;
      }
      break;
    case CF_SHORT: 
      for (i=0; i<(int)bufferedProperties[id].din.dArrayLength; i++)
      {        
        if ((din->data.sptr[i] < bufferedProperties[id].min) ||
            (din->data.sptr[i] > bufferedProperties[id].max))
          return out_of_range;
      }
      break;
    case CF_LONG: 
      for (i=0; i<(int)bufferedProperties[id].din.dArrayLength; i++)
      {        
        if ((din->data.lptr[i] < bufferedProperties[id].min) ||
            (din->data.lptr[i] > bufferedProperties[id].max))
          return out_of_range;
      }
      break;
    case CF_FLOAT: 
      for (i=0; i<(int)bufferedProperties[id].din.dArrayLength; i++)
      {        
        if ((din->data.fptr[i] < bufferedProperties[id].min) ||
            (din->data.fptr[i] > bufferedProperties[id].max))
          return out_of_range;
      }
      break;
    case CF_DOUBLE: 
      for (i=0; i<(int)bufferedProperties[id].din.dArrayLength; i++)
      {        
        if ((din->data.dptr[i] < bufferedProperties[id].min) ||
            (din->data.dptr[i] > bufferedProperties[id].max))
          return out_of_range;
      }
      break;
  }
  return 0;
}
int currentPropertyId = 0;
int currentDeviceNumber = 0;
int notifierCalled = 0;
int bufsrveqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
	int i,iin,p,id,cc=0,dsiz,doutlen;

  if ((id=GetPropertyId(gEqpName,devProperty)) < 0) return illegal_property;
  if ((i=GetDeviceNumberEx(gEqpName,devName,devProperty)) < 0) return illegal_equipment_number;
  if (!(bufferedProperties[id].arrayType & AT_SPECTRUM) &&
       dout->dArrayLength > bufferedProperties[id].dout.dArrayLength - i) 
     return dimension_error;
  if (din->dArrayLength > 0)
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    /* a fairly stringent condition : */
    if (din->dFormat != bufferedProperties[id].din.dFormat) return illegal_format;
    if (access & CA_FIRST) 
    {
      bufferedProperties[id].interimCompletionCode = not_ready;
    }
    else 
    {
      ccerr(bufferedProperties[id].interimCompletionCode);
    }
    iin = i;
    if (bufferedProperties[id].inputChanged) return operation_busy;
    if (bufferedProperties[id].din.dArrayLength == 1) iin = 0;
    if (din->dArrayLength > bufferedProperties[id].din.dArrayLength - iin) return dimension_error;
    p = fmtsizeof(bufferedProperties[id].din.dFormat) * iin;
    /* do range checking */
    if (areValuesOutOfBounds(id,din)) return out_of_range;
    /* dsiz = bufferedProperties[id].din.dArrayLength * fmtsizeof(bufferedProperties[id].din.dFormat); */
    dsiz = din->dArrayLength * fmtsizeof(bufferedProperties[id].din.dFormat);
    bufferedProperties[id].inputChanged = 1;
    if (memcmp(&bufferedProperties[id].din.data.bptr[p],din->data.bptr,dsiz)) bufferedProperties[id].inputChanged = -1;
    memcpy(&bufferedProperties[id].din.data.bptr[p],din->data.bptr,dsiz);
    currentPropertyId = id;
    currentDeviceNumber = i;
    if (bufferedProperties[id].notifier != NULL)
    {
      if (notifierCalled) return semaphore_busy;
      notifierCalled = -1;
      bufferedProperties[id].notifier(bufferedProperties[id].notifierId);
      return not_ready;
    }
    if (gBufferedNotifier != NULL)
    {
      if (notifierCalled) return semaphore_busy;
      notifierCalled = -1;
      strncpy(gPropertyNofified,devProperty,32);
      strncpy(gDeviceNotified,devName,16);
      gBufferedNotifier(gBufferedNotifierId);
      return not_ready;
    }
    if (bufferedProperties[id].callback != NULL) cc = bufferedProperties[id].callback();
  }
err:
  if (cc) return cc;
  if (dout->dArrayLength > 0)
  {
    if (dout->dFormat == CF_SPECTRUM) dout->dArrayLength += 24;
    /* if (dout->dFormat != bufferedProperties[id].dout.dFormat) return illegal_format; */
    doutlen = bufferedProperties[id].dout.dArrayLength;
    if (bufferedProperties[id].arrayType & AT_SPECTRUM) 
    {
      i *= bufferedProperties[id].dout.dArrayLength;
      doutlen *= gNumDevices;
    }
    switch (bufferedProperties[id].dout.dFormat)
    {
      case CF_SHORT:
        if ((cc=PutValuesFromShortEx(dout,bufferedProperties[id].dout.data.sptr,doutlen,i)) != 0) return cc;
        break;
      case CF_LONG:
        if ((cc=PutValuesFromLongEx(dout,bufferedProperties[id].dout.data.lptr,doutlen,i)) != 0) return cc;
        break;
      case CF_FLOAT:
        if ((cc=PutValuesFromFloatEx(dout,bufferedProperties[id].dout.data.fptr,doutlen,i)) != 0) return cc;
        break;
      case CF_DOUBLE:
        if ((cc=PutValuesFromDoubleEx(dout,bufferedProperties[id].dout.data.dptr,doutlen,i)) != 0) return cc;
        break;
      default:
        if (dout->dFormat != bufferedProperties[id].dout.dFormat) return illegal_format;
        i *= fmtsizeof(dout->dFormat);
        if (dout->dArrayLength < (UINT32)doutlen) doutlen = dout->dArrayLength;
        memcpy(dout->data.bptr,&bufferedProperties[id].dout.data.bptr[i],doutlen*fmtsizeof(dout->dFormat));
        break;
    }
    setDataTimeStamp(bufferedProperties[id].dataTimeStamp);
  }
  return 0;
}
EXPORT_B int SetBufferedCompletion(int compCode)
{
  int id = currentPropertyId;
  if (id < 0) return illegal_property;
  bufferedProperties[id].interimCompletionCode = compCode;
  notifierCalled = 0;
  return 0;
}
/** 
 * \brief Attaches the TINE server according to the input given
 *
 *
 * \param srvExportName is the desired export name.  If this is omitted (i.e. a zero-length 
 *        string is passed, then the export name must be obtained from an associated 'exports.csv'
 *        in the local FEC database repository.
 * \param srvEQPName is the desired local equipment module name.  If this is omitted (i.e. a zero-length
 *        string is passed, then the local equipment module name is extracted from an
 *        associated 'exports.csv' in the local FEC database repository.
 * \param ndevices is the number of devices (instances) associated with this device server.
 *        If this is omitted (i.e. 0 is passed), then the number of devices is extracted from an
 *        associated 'exports.csv' in the local FEC database repository.
 *
 * A call to AttachServer serves to initialize the device server.  It is 
 * analogous to setting the server ActiveX control's enabled property to 'true', where the
 * ActiveX interface is concerned.  Theoretically, you can call AttachServer() with all 'NULL'
 * parameters, in which case all registration information is obtained from the 'exports.csv' 
 * file (see <a href="csvFiles.html"><b><i>Server Startup csv-Files</i></b></a>.
 * Note however, that if you take this strategy, you should take care to avoid a lock-out
 * situation regarding the name resolution.  Namely, if all initialization names are to reside
 * in 'fecid.csv' (FEC name) and 'exports.csv' (device server name, local equipment module name,
 * and properties) then the fecid.csv should not make use of an 'Export_Name' column or a 
 * 'Local_Name' column.  That is, the fecid.csv file should only contain one entry, referring
 * to the server process in question.  In cases where more than one FEC process are running
 * on the same computer, each process should have its own fecid.csv in its working directory
 * when the export.csv file cannot make use of an 'Export_Name' column in order to latch
 * the appropriate device server.  Of course, if AttachServer() provides an 'Export Name' in
 * the first parameter, then there is no ambiguity.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \b Example:  Attach to server and set the server's export name (requires exports.csv file to
 *              set the local equipment module name and number of modules):
 *
 * \include eg_AttachServer0.txt
 *
 * \b Example:  Attach to server and set the server's export name, local equipment module name and
 *              number of modules (does not read exports.csv file!):
 *
 * \include eg_AttachServer1.txt
 */
EXPORT_B int AttachServerEx(char *srvExportName,char *srvEQPName,int ndevices,void (*tmr)(void),int tmrInterval) 
{
  int i, id, cc, dsiz, n, nelems;
  int nprps = MAX_BUFFERED_PROPERTIES, hasprplst = FALSE;
  char *lclName = NULL;
  ExportPropertyListStruct pls;

  feclog("Attach Server to %s",
    srvExportName != NULL ? srvExportName :
    srvEQPName != NULL ? srvEQPName : "(entry not given)");

  /* a LabView restart doesn't unload the library ! */
  FreeEQPModules();
  MinPollingRate = 20;
  if ((cc=SystemInit(TRUE)) != 0) 
  { 
#   ifndef TBUFSRVDLL_EXPORTS
    printf("init error: %s\n",GetRPCLastError((short)cc)); 
#   endif
    exit(1); 
  }
  resetLastMinuteInitialization();
  gSystemInitialized = TRUE;

  memset(gServerName,0,32); memset(gEqpName,0,16);
  if (srvExportName != NULL) strncpy(gServerName,srvExportName,16);
  if (srvEQPName != NULL) strncpy(gEqpName,srvEQPName,6);

  if (strlen(gServerName) > 0 && strlen(gEqpName) > 0)
  { /* all arguments passed */
    gNumDevices = ndevices;
    RegisterEquipmentModule(gServerName,gEqpName,ndevices,bufsrveqm,NULL,tmr,tmrInterval,NULL);
  }
  else
  { 
    if (strlen(gServerName) > 0)
    { /* just the export name was passed, find the local name */
      if ((lclName=getLocalName(gServerName)) == NULL) return non_existent_elem;
      strncpy(gEqpName, lclName, 6) ;
    }
    if (strlen(gEqpName) > 0)
    {
      RegisterEquipmentModule(NULL,gEqpName,0,bufsrveqm,NULL,tmr,tmrInterval,NULL);
    }
    else
    {
      return argument_list_error;
    }
  }
  hasprplst = hasDevicePropertyList(gEqpName);
  if ((cc=GetRegisteredPropertyList(gEqpName,registeredProperties,&nprps)) != 0) return cc;
  if ((n=GetNumberRegisteredDevices(gEqpName)) < 0) n = ndevices > 0 ? ndevices : 1;
  gNumDevices = n;
  for (i=0; i<nprps; i++)
  {
    if ((id=GetPropertyId(gEqpName,registeredProperties[i].name)) < 0) return -id;
    if ((cc=GetRegisteredPropertyListStruct(gEqpName,registeredProperties[i].name,&pls)) != 0) return cc;
    if (hasprplst) pls.prpArrayType |= AT_SPECTRUM; /* force separate storage for each device (can't be a channel array) */
    if (pls.prpFormat == CF_SPECTRUM) pls.prpSize += 24;
    if (pls.prpFormatIn == CF_SPECTRUM) pls.prpSizeIn += 24;
    strncpy(bufferedProperties[id].nam,pls.prpName,32);
    bufferedProperties[id].max = (float)pls.egu.max;
    bufferedProperties[id].min = (float)pls.egu.min;
    bufferedProperties[id].dout.dFormat = pls.prpFormat;
    bufferedProperties[id].dout.dArrayLength = pls.prpSize;
    bufferedProperties[id].arrayType = pls.prpArrayType;
    nelems = pls.prpArrayType & AT_SPECTRUM ? n : 1;
    if ((dsiz=pls.prpSize * fmtsizeof(pls.prpFormat)) > 0 &&
        (bufferedProperties[id].dout.data.bptr=calloc(nelems,dsiz)) == NULL)
      return -out_of_server_memory;
    bufferedProperties[id].din.dFormat = pls.prpFormatIn;
    bufferedProperties[id].din.dArrayLength = pls.prpSizeIn;
    if ((dsiz=pls.prpSizeIn * fmtsizeof(pls.prpFormatIn)) > 0 &&
        ((bufferedProperties[id].din.data.bptr=calloc(1,dsiz)) == NULL ||
        (bufferedProperties[id].blk.data.bptr=calloc(1,dsiz)) == NULL))
      return -out_of_server_memory;
    bufferedProperties[id].dataTimeStamp = 0;
    nBufferedProperties++;
  }

# ifdef _USRDLL
  TineStartCycler();
# endif
  return 0;
}
EXPORT_B int AttachServer(char *srvExportName,char *srvEQPName,int ndevices) 
{
  return AttachServerEx(srvExportName,srvEQPName,ndevices,NULL,0);
}
EXPORT_B void runServer(void)
{
  while (!gTerminateServer)
  {
    SystemCycle(TRUE);
  }
# ifndef TBUFSRVDLL_EXPORTS
  haltServer();
# endif
}
EXPORT_B void TerminateServer() 
{
  gTerminateServer = TRUE;
}

EXPORT_B int GetRegisteredProperties(NAME32 *prps, short *nprps)
{
  int i, n;
  if (prps == NULL || nprps == NULL) return -1;
  for (i=0, n=0; i<nBufferedProperties && n<*nprps; i++)
  {
    if (strlen(bufferedProperties[i].nam) == 0) continue;
    strncpy(prps[n++].name,bufferedProperties[i].nam,32);
  }
  *nprps = n;
  return nBufferedProperties;
}

/** 
 * \brief Refreshes the contents of the data buffer associated with the given property
 * 
 *
 * \param prpName is the property name into whose buffer the associated data are to be written
 * \param devName is the device name giving the starting address i.e. entry point into the
 *        property's buffer into which the associated data are to be written
 * \param prpData is a reference to the data to be written (pushed) into the property's data
 *        buffer.
 * \param prpSiz is the number of elements to be written into the property's data buffer.
 * \param prpSchedule is the schedule 'flag' determining whether the property is to be
 *        scheduled immediately ('true') or not.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \b Example:
 *
 * \include eg_pushBufferedData.txt
 */
EXPORT_B int pushBufferedData(char *prpName,char *devName,BYTE *prpData,long prpSiz,int prpSchedule)
{
  int id, i, p, fsz, isText = FALSE;

  if (prpData == NULL) return buffer_too_small;
  if ((id=GetPropertyId(gEqpName,prpName)) < 0) return -id;
  if ((i=GetDeviceNumberEx(gEqpName,devName,prpName)) < 0) return -i;
  if (bufferedProperties[id].dout.dFormat == CF_SPECTRUM) prpSiz += 24;
  if (bufferedProperties[id].dout.dFormat == CF_TEXT) isText = TRUE;
  if (prpSiz > i + (int)bufferedProperties[id].dout.dArrayLength) 
  {
    prpSiz = (int)bufferedProperties[id].dout.dArrayLength;
    if (bufferedProperties[id].arrayType != AT_SPECTRUM) prpSiz -= i;
  }

  fsz = fmtsizeof(bufferedProperties[id].dout.dFormat);
  p = fsz * i;
  if (bufferedProperties[id].arrayType & AT_SPECTRUM) 
  {
    p *= bufferedProperties[id].dout.dArrayLength;
    memset(&bufferedProperties[id].dout.data.bptr[p],0,bufferedProperties[id].dout.dArrayLength*fsz);
  }
  if (isText)
  {
    strncpy(&bufferedProperties[id].dout.data.bptr[p],prpData,prpSiz*fsz);
  }
  else
  {
    memcpy(&bufferedProperties[id].dout.data.bptr[p],prpData,prpSiz*fsz);
  }
  bufferedProperties[id].dataTimeStamp = makeDataTimeStamp();
  if (prpSchedule) SystemScheduleProperty(gEqpName,prpName);
  return 0;
}
/** 
 * \brief Retrieves the contents of the input data buffer associated with the given property
 * 
 *
 * \param prpName is the property name from whose input buffer the associated data are to be read.
 * \param devName is the device name giving the starting address i.e. entry point into the
 *        property's buffer from which the associated data are to be read.
 * \param prpData is a reference to the data buffer into which the incoming data should be read 
 *        (pulled).
 * \param prpSiz is the number of elements to be read into the property's data buffer.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \b Example:
 *
 * \include eg_pullBufferedData.txt
 */
EXPORT_B int pullBufferedData(char *prpName,char *devName,BYTE *prpData,long prpSiz)
{
  int id, i, p, fsz;

  if (prpData == NULL) return buffer_too_small;
  if ((id=GetPropertyId(gEqpName,prpName)) < 0) return -id;
  if ((i=GetDeviceNumberEx(gEqpName,devName,prpName)) < 0) return -i;
  if (bufferedProperties[id].din.dArrayLength == 1) i = 0;
  if (bufferedProperties[id].din.dFormat == CF_SPECTRUM) prpSiz += 24;
  if (prpSiz > (int)bufferedProperties[id].din.dArrayLength - i) return dimension_error;

  fsz = fmtsizeof(bufferedProperties[id].din.dFormat);
  p = fsz * i;
  memcpy(prpData,&bufferedProperties[id].din.data.bptr[p],prpSiz*fsz);
  bufferedProperties[id].inputChanged = 0;
  return 0;
}
/** 
 * \brief Retrieves the property which caused the notifier to be called
 * 
 *
 * \param prpName (output) is the property name which caused the notifer to be called
 *
 * \return 0 upon success otherwise a TINE return code
 *
 */
EXPORT_B int getNotifiedProperty(char *prpName)
{
  if (!notifierCalled) return not_signalled;
  strncpy(prpName,gPropertyNofified,32);
  return 0;
}
/** 
 * \brief Retrieves the property which caused the notifier to be called
 * 
 *
 * \param prpName (output) is the property name which caused the notifer to be called
 *
 * \return 0 upon success otherwise a TINE return code
 *
 */
EXPORT_B int getNotifiedPropertyAndDevice(char *prpName,char *devName)
{
  if (!notifierCalled) return not_signalled;
  strncpy(prpName,gPropertyNofified,32);
  strncpy(devName,gDeviceNotified,16);
  return 0;
}
/** 
 * \brief Checks whether there are new input data for the given property
 * 
 *
 * \param prpName is the property name whose input data buffer is to be checked for changes.
 *
 * \return 0 (False) if the input data have not changed or -1 (True) if there are new input
 * data to be acquired by calling PullBufferedData().
 *
 * \b Example:
 *
 * \include eg_hasInputChanged.txt
 */
EXPORT_B int hasInputChanged(char *prpName)
{
  int id;

  if ((id=GetPropertyId(gEqpName,prpName)) < 0) return 0;
  return bufferedProperties[id].inputChanged == -1 ? TRUE : FALSE;
}
/** 
 * \brief Returns the device number associated with the WRITE call.
 * 
 * This call can be made inside the server's callback routine along with
 * 'hasInputChanged' in order to ascertain which device was 'set'. 
 *
 * \return The device number (obtained from the registered device name) used
 * in the call.
 *
 * \b Example:
 *
 * \include eg_hasInputChanged.txt
 */
EXPORT_B int GetInputDeviceNumber(void)
{
  return currentDeviceNumber;
}
/** 
 * \brief Registers a callback routine to be called when a WRITE access property is
 * called.
 * 
 *
 *
 * \param prpName
 * \param cb
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \b Example:
 *
 * \include eg_RegisterServerCallback.txt
 */
EXPORT_B int RegisterServerCallback(char *prpName, int (*cb)(void))
{
  int id;

  if ((id=GetPropertyId(gEqpName,prpName)) < 0) return -id;
  bufferedProperties[id].callback = cb;
  return 0;
}
/** 
 * \brief Registers a Notifier routine to be called when a WRITE access property is
 * called.
 * 
 * The difference between a callback and a notifier is that it is assumed that the callback 
 * completes all activity associated with the transaction while inside the callback routine.
 * A notififier will call the notifier routine and expect the caller to call the 
 * SetBufferedCompletion() routine when the transaction is complete.  This scenario is useful 
 * for development platforms such as LabView.
 *
 * \param prpName the Property name for which the notifier is in effect
 * \param nf the notifier routine
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by a call to 
 * GetRPCLastError().
 *
 * \b Example:
 *
 * \include eg_RegisterServerNotifier.txt
 */
EXPORT_B int RegisterServerNotifier(char *prpName, void (*nf)(int))
{
  return RegisterServerNotifierEx(prpName,nf,0);
}
EXPORT_B int RegisterServerNotifierEx(char *prpName, void (*nf)(int), int nid)
{
  int id;

  if (prpName == NULL || strlen(prpName) == 0)
  {
    gBufferedNotifier = nf;
    gBufferedNotifierId = nid;
  }
  else
  {
    if ((id=GetPropertyId(gEqpName,prpName)) < 0) return -id;
    bufferedProperties[id].notifier = nf;
    bufferedProperties[id].notifierId = nid;
  }
  return 0;
}
// TODO: atyp, rowlength, egu ...
EXPORT_B int RegisterBufferedProperty
(
  char *prpName, long prpInSiz, short prpInFmt, long prpOutSiz, short prpOutFmt,
  float prpMax, float prpMin, char *prpEgu,	short access, char *prpDsc
) 
{
  int nelems=1,id,dsiz=0,atyp = AT_UNKNOWN, rowlength = 0;
  static char localEqpname[16],localProperty[36],localDescription[96];
  DTYPE dout, din;

  strncpy(localEqpname,gEqpName,6);
  strncpy(localProperty,prpName,32);
  sprintf(localDescription,"[%g:%g %s]%s",prpMin,prpMax,prpEgu,prpDsc);

  dout.dFormat = prpOutFmt;
  dout.dArrayLength = prpOutSiz;
  dout.dTag[0] = 0;
  din.dFormat = prpInFmt;
  din.dArrayLength = prpInSiz;
  din.dTag[0] = 0;

  if (atyp == AT_UNKNOWN && prpOutFmt == CF_SPECTRUM)
  {
    atyp = AT_SINGLE|AT_SPECTRUM;
    nelems = gNumDevices;
  }
  rowlength = prpOutSiz;

  id = RegisterPropertyInformation(localEqpname,localProperty,&dout,&din,
         access,(short)atyp,(UINT16)rowlength,prpDsc,-1,NULL); 

  if (id < 0 || id >= MAX_BUFFERED_PROPERTIES) return -illegal_property;

  if (prpOutFmt == CF_SPECTRUM) prpOutSiz += 24;
  if (prpInFmt == CF_SPECTRUM) prpInSiz += 24;

  strncpy(bufferedProperties[id].nam,prpName,32);
  bufferedProperties[id].max = prpMax;
  bufferedProperties[id].min = prpMin;
  bufferedProperties[id].dout.dFormat = prpOutFmt;
  bufferedProperties[id].dout.dArrayLength = prpOutSiz;
  bufferedProperties[id].arrayType = atyp;
  if ((dsiz=prpOutSiz * fmtsizeof(prpOutFmt)) > 0)
  {
    if ((bufferedProperties[id].dout.data.bptr=calloc(nelems,dsiz)) == NULL)
      return -out_of_server_memory;
  }
  bufferedProperties[id].din.dFormat= prpInFmt;
  bufferedProperties[id].din.dArrayLength = prpInSiz;
  if ((dsiz=prpInSiz * fmtsizeof(prpInFmt)) > 0)
  {
    if (((bufferedProperties[id].din.data.bptr=calloc(1,dsiz)) == NULL ||
        (bufferedProperties[id].blk.data.bptr=calloc(1,dsiz)) == NULL))
      return -out_of_server_memory;
  }
  nBufferedProperties++;
  return id;
}

