#include "prolog.h"
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include "servdef.h"
#include "rpctypes.h"
#include "errors.h"

extern int nofeclog;
extern int ListenForGlobals;
extern int ListenForMCasts;
extern int defaultClientTimeout;
int tget(char *dev,char *prp);
int thistory(char *context,char *keyword,char *devname,char *stoptime,char *depth);

char *usageTxt[] = {
  "\nRetrieves a stored archive data for the keyword and device specfied",
  "Data can be obtained from a central archive server or from a local history,",
  "depending on how the context and keyword are specified.\n",
  "A simple keyword along with a context will contact the central archive",
  "server for the context.",
  "\n        e.g. thistory HERA HPDCCur now 1hour\n",
  "The history depth can be specified in hours, days, weeks, or months and",
  "is parsed as",
  "\n        <number><time unit>\n",
  "The history target time can be 'now', a valid unix timestamp, or a date",
  "and time string in the form",
  "\n        <day>.<month>.<year>_<hour>.<minute>.<second>\n",
  "where all entries except <day> are optional from right to left.",
  "\n        e.g. thistory HERA HPDCCur 31.05.2006 1day",
  "          or thistory HERA HPDCCur 31.05.2006_12:00:00 1hour\n",
  "A local history can be obtained by specifying the target device server",
  "along with the context (with a leading \"/\")",
  "\n        e.g. thistory /HERA/HEEKOLLI SOLLWERT HESL66i 31.05.2006 1day\n",
  "A history snapshot of an array record at a specific time can be obtained",
  "by specifying 'snapshot' as the depth parameter.",
  "\n        e.g. thistory /HERA/HEPBPM ORBIT.X #1 7.06.2006_12:00:00 snapshot",
  "\nreturns the array data record stored at or after noon on June 7th 2006.",
  "The timestamp of the record found is always displayed along with the data.\n",
  "Usage : thistory <context> <keyword> <device name> <stop time> <depth>"
};

#define nUsageLines sizeof(usageTxt)/sizeof(char *)

int main(int argc, char *argv[])
{
  int cc,i;
  char *context=NULL, *keyword=NULL, *device=NULL, *stoptime="NOW", *depth="1 DAY";
  char tmpdev[32];
  switch (argc)
  {
    case 2:
      keyword = argv[1];
      break;
    case 3:
      keyword = argv[1];
      depth = argv[2];
      break;
    case 4:
      keyword = argv[1];
      stoptime = argv[2];
      depth = argv[3];
      break;
    case 5:
      context = argv[1];
      keyword = argv[2];
      stoptime = argv[3];
      depth = argv[4];
      break;
    case 6:
      context = argv[1];
      keyword = argv[2];
      device = argv[3];
      stoptime = argv[4];
      depth = argv[5];
      for (i=0; i<(int)strlen(device) && isdigit(device[i]); i++);
      if (i == (int)strlen(device)) 
      {
        sprintf(tmpdev,"#%s",device);
        device = tmpdev;
      }
      break;
    default:
      for (i=0; i<nUsageLines; i++)
      {
        printf("%s\n",usageTxt[i]);
      }
      printf("\n");
      exit(1);
  }
  MaxNumConnections = 3;
  ListenForGlobals = ListenForMCasts = FALSE;
  MaxPollingRate = MinPollingRate = 50;
  defaultClientTimeout = 50;
  nofeclog = TRUE;
//  printf("\n");
  if ((cc=thistory(context,keyword,device,stoptime,depth)) != 0)
    printf("error : %s\n",erlst[cc]);
  printf("\n");
  return cc;
}
