#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <sys/types.h>
#include <stddef.h>
#include "servdef.h"
#include "database.h"
#include "errors.h"
#include "srvip.h"
#include "unixcmd.h"

void closeClients(void);

struct CommandListStruct {
  char cmd[10];
  int (*fcn)(char *,char *);
};
struct CommandListStruct CommandList[] = {
  { "quit",quit},
  { "exit",quit},
  { "help",help},
  { "set",setVal},
  { "get",getVal}
};
int NumCommands = sizeof(CommandList)/sizeof(struct CommandListStruct);
int CommandLevel = 0;
char CommandString[CMDSTRSIZE];
int  IPCSocket = 0;
int  IPCfds[MAXIPCCLIENTS],nipcclients = 0;
int  ipcfd = -1;
extern int ALMdebug;
extern int foreground;

USRCMD *usrcmd = NULL;
int numUsrCmd = 0;

USRFCN *usrfcn = NULL;
int numUsrFcn = 0;

int RegisterUserCommand(
    char *cmd,
    int (*fcn)(int,int,int,int),
    int *iparam,
    float *fparam,
    int access)
{
  USRCMD *uc;
  
  if ((uc=(USRCMD *)calloc(1,sizeof(USRCMD))) == NULL) return out_of_local_memory;
  if (access == 0) access = CA_READ;
  strncpy(uc->cmd,cmd,32);
  uc->fcn = fcn;
  uc->iparam = iparam;
  uc->fparam = fparam;
  uc->access = access;
  uc->next = usrcmd;
  usrcmd = uc;
  numUsrCmd++;
  return 0;
}
int RegisterUserFunction(
    char *name,
    int (*fcn)(int argc,char *argv[]))
{
  USRFCN *uf;

  if ((uf=(USRFCN *)SystemCalloc(1,sizeof(USRFCN))) == NULL) return out_of_local_memory;
  strncpy(uf->name,name,32);
  uf->fcn = fcn;
  uf->next = usrfcn;
  usrfcn = uf;
  numUsrFcn++;
  return 0;
}
int InterpretCommand(char *cmd)
{
  char *c,*ce=NULL,*cp=NULL,*arg1=NULL,*arg2=NULL;
  int cc=0,i,len;
  int argc;
  char *argv[10],vstr[32],sstr[128],strbuf[CMDSIZE];
  USRFCN *uf;

  if (strlen(cmd) <= 1) ccerr(0);
  for (i=0; i<NumCommands; i++)
   if (!strncmp(cmd,CommandList[i].cmd,strlen(CommandList[i].cmd))) break;

  if (i < NumCommands) 
  {
    if ((c=strchr(cmd,'=')) != NULL)
    {
      arg2 = c+1;
      *c = 0;
    }
    arg1 = &cmd[strlen(CommandList[i].cmd)+1];
    ccerr(CommandList[i].fcn(arg1,arg2));
  }
  for (i=0; i<32 && cmd[i] == ' '; i++);
  strncpy(vstr,&cmd[i],32);
  if ((c=strchr(vstr,'(')) != NULL) *c++ = 0;
  for (uf=usrfcn; uf != NULL; uf=uf->next)
  {
    if (uf->name != NULL && !strncmp(vstr,uf->name,32))
    {
      argc = 0;
      len = strlen(uf->name);
      if (c != NULL && (ce=strchr(c,')')) != NULL)
      {
        *ce = 0; strncpy(sstr,c,62); sstr[63] = 0;
        for (i=0,c=strtok_r(sstr,",",&cp); c && i<10; i++,c=strtok_r(NULL,",",&cp))
        {
          argv[i] = c;
        }
        argc = i;
      }
      i = uf->fcn(argc,argv);
      sprintf(strbuf,"result is %d",i);
      ttyoutput(strbuf);
      ccerr(0);  
    }
  }
err: if (cc) ttyoutput("Syntax: set name = value or get name");
  ttyoutput("\n>");
  return cc;
}
int GetCommand(int fd,char *buf,int bufsiz)
{
  static int len=0;

  if ((len=read(fd,buf,bufsiz)) < 0) return -1;
  buf[len] = 0;
  if (len == 0) 
  {
    /* don't let it spin in the background ! */
    if (fd == STDIN_FILENO) foreground = 0;
    return 0;
  }
  if (fd != STDIN_FILENO) { ipcfd = fd; printf(buf); }
  else ipcfd = -1;
  CMDcount++;
  return len;
}
int help(char *s1,char *s2)
{
  USRCMD *uc;
  USRFCN *uf = NULL;
  char sstr[64],strbuf[CMDSIZE];
  
  ttyoutput("Currently available commands:\n>");
  ttyoutput("\tquit            - terminates program gracefully\n>");
  if (NrOfEqpTypes)
  {
    ttyoutput("\tget clients             - displays the current consumer list\n>");
    ttyoutput("\tget contracts           - displays the current contract list\n>");
    ttyoutput("\tget contract = <#>      - displays contract Nr. <#> information\n>");
  }
  if (nglobals)
  {
    ttyoutput("\tget globals     - displays the current globals list\n>");
  }
  if (nConnectionTableEntries)
  {
    ttyoutput("\tget connections          - displays the current connection list\n>");
    ttyoutput("\tget connection = <#>     - displays connection Nr. <#> information\n>");
  }
  ttyoutput("\tget BurstLimit           - displays Burst Limit in packets\n>");
  ttyoutput("\tset BurstLimit           - sets Burst Limit to num packets input\n>");
  ttyoutput("\tget CycleDelay           - displays Cycle Delay in msec\n>");
  ttyoutput("\tset CycleDelay           - sets Cycle Delay to msec input\n>");
  ttyoutput("\tget time                 - displays local time\n>");
  ttyoutput("\tget version              - displays TINE version number\n>");
  ttyoutput("\tget users                - displays WRITE access user list\n>");
  ttyoutput("\tget nets                 - displays WRITE access net list\n>");
  ttyoutput("\tget stats                - displays operational statistics\n>");
  ttyoutput("\tget filter               - displays current debug output filter string\n>");
  ttyoutput("\tset filter               - sets debug output filter string\n>");
  ttyoutput("\tset debug = 0            - turns debug printing off\n>");
  ttyoutput("\tset debug = 1            - sets debug level 1 (trace RPC commands)\n>");
  ttyoutput("\tset debug = 2            - sets debug level 2 (trace network activity)\n>");
  ttyoutput("\tset debug = 3            - sets debug level 2 (trace data exchange)\n>");
  ttyoutput("\tset debug = 4            - sets debug level 3 (full diagnostic trace )\n>");
  ttyoutput("\tset logdbg = 0           - turns debug logging off\n>");
  ttyoutput("\tset logdbg = 1           - turns debug logging on\n>");
  ttyoutput("\tset CommandLevel = <val> - sets global CommandLevel to <val>\n>");
  ttyoutput("\tset CommandString = <str> - sets global CommandString to <str>\n>");
  ttyoutput("\thelp            - display this list\n>");
  ttyoutput("\n>Extra commands:\n>");
  for (uc=usrcmd; uc != NULL; uc=uc->next)
  {
    sstr[0] = 0;
    if (uc->access & CA_READ) strcpy(sstr,"get ");
    if (uc->access & CA_WRITE)
    {
      if (strlen(sstr)) strcat(sstr," or set ");
      else strcpy(sstr,"set ");
    }
    if (uc->iparam != NULL) strcat(sstr,"integer value");
    else if (uc->fparam != NULL) strcat(sstr,"float value");
    else if (uc->fcn != NULL) strcat(sstr,"function value");
    sprintf(strbuf,"\t%s\t - %s\n>",uc->cmd,sstr);
    ttyoutput(strbuf);
  }
  for (uf=usrfcn; uf != NULL; uf=uf->next)
  {
    sstr[0] = 0;
    sprintf(strbuf,"\t%s()\n>",uf->name);
    ttyoutput(strbuf);
  }
  return 0;
}
int quit(char *s1,char *s2)
{
  ExitEQPModules();
  quitServer();                   /* non-servers must cover this routine */
  if (nConnectionTableEntries) closeClients();      /* non-clients must cover this routine */
  feclog("stopped");
  exit(0);
  return 0;
}
int getVal(char *var,char *s)
{
  USRCMD *uc;
  char *c,*cp=NULL,sstr[128],vstr[32],strbuf[CMDSIZE];
  int i,len,ival[4];

  if (var == NULL) return -1;
  for (i=0; i<32 && var[i] == ' '; i++);
  strncpy(vstr,&var[i],32);
  if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
  if ((c=strchr(vstr,' ')) != NULL) *c = 0;
  if ((c=strchr(vstr,'(')) != NULL) *c = 0;

  for (uc=usrcmd; uc != NULL; uc=uc->next)
  {
    if (!strncmp(vstr,uc->cmd,32) && uc->access & CA_READ)
    {
      if (uc->iparam != NULL)
      {
        sprintf(strbuf,"%s is %d",uc->cmd,*uc->iparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fparam != NULL)
      {
        sprintf(strbuf,"%s is %g",uc->cmd,*uc->fparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fcn != NULL)
      {
        len = strlen(uc->cmd);
        memset(ival,0,sizeof(ival));
        if (var[len] == '(' && (c=strchr(var,')')) != NULL)
        {
          *c = 0; strncpy(sstr,&var[len+1],62); sstr[63] = 0;
          for (i=0,c=strtok_r(sstr,",",&cp); i<4 && c; i++,c=strtok_r(NULL,",",&cp))
          {
            ival[i] = atoi(c);
          }
        }
        i = uc->fcn(ival[0],ival[1],ival[2],ival[3]);
        sprintf(strbuf,"result is %d",i);
        ttyoutput(strbuf);
        return 0;
      }
    }
  }
  if (strstr(var,"filter"))
  {
    sprintf(strbuf,"debug name filter : %s",tagNameFilter);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"version")) return dumpVersion();
  if (strstr(var,"debug"))
  {
    sprintf(strbuf,"Debug level %d",NGdebug);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"CommandLevel"))
  {
     sprintf(strbuf,"Command Level %d entered",CommandLevel); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"CommandString"))
  {
     sprintf(strbuf,"Command string %s entered",CommandString); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"users")) { dumpRegisteredUsers(); return 0; }
  if (strstr(var,"globals")) { dumpGlobals(); return 0; }
  if (strstr(var,"nets"))
  {
#   ifdef IPXBSD
    dumpIPXNets();
#   endif
    dumpIPNets();
    return 0;
  }
  if (strstr(var,"time")) return dumpServerTime();
  if (NrOfEqpTypes)
  {
    if (strstr(var,"clients")) return dumpConsumerList(); 
    if (strstr(var,"contract") && s != NULL)
    {
      i = atoi(s);
      dumpContractStats(i);
      return 0;
    }
    if (strstr(var,"contracts")) return dumpContractList();
  }
  if (nConnectionTableEntries)
  {
    if (strstr(var,"connection") && s != NULL)
    {
      i = atoi(s);
      dumpConnectionStats(i);      /* non-clients must cover this routine */
      return 0;
    }
    if (strstr(var,"connections"))
    {
      dumpConnectionList();      /* non-clients must cover this routine */
      ttyoutput("\n");
      sprintf(strbuf,"Currently: %d remote session(s)\n>",nipcclients);
      ttyoutput(strbuf);
      return 0;
    }
  }
  if (strstr(var,"BurstLimit"))
  {
    sprintf(strbuf,"Burst Limit set at %d",gBurstLimit);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"CycleDelay"))
  {
    sprintf(strbuf,"Cycle Delay set at %d",gCycleDelay);
    ttyoutput(strbuf);
    return 0;
  }
  if (strstr(var,"stats")) return dumpServerStats();
  sprintf(strbuf,"Variable unavailable: %s",var); ttyoutput(strbuf);
  return -1;
}
int setVal(char *var,char *val)
{
  int v;
  USRCMD *uc;
  char *c,*cp=NULL,sstr[64],vstr[32],strbuf[CMDSIZE];
  int i,len,ival[4];

  if (var == NULL || val == NULL) return -1;
  for (i=0; i<32 && var[i] == ' '; i++);
  strncpy(vstr,&var[i],32);
  if ((c=strchr(vstr,'\n')) != NULL) *c = 0;
  if ((c=strchr(vstr,' ')) != NULL) *c = 0;
  if ((c=strchr(vstr,'(')) != NULL) *c = 0;

  for (uc=usrcmd; uc != NULL; uc=uc->next)
  {
    if (!strncmp(vstr,uc->cmd,32) && uc->access & CA_WRITE)
    {
      if (uc->iparam != NULL)
      {
        *uc->iparam = atoi(val);
        sprintf(strbuf,"%s set to %d",uc->cmd,*uc->iparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fparam != NULL) 
      {
        *uc->fparam = (float)strtod(val,NULL);
        sprintf(strbuf,"%s is %g",uc->cmd,*uc->fparam);
        ttyoutput(strbuf);
        return 0;
      }
      if (uc->fcn != NULL) 
      {
        len = strlen(uc->cmd);
        memset(ival,0,sizeof(ival));
        if (var[len] == '(' && (c=strchr(var,')')) != NULL)
        {
          *c = 0; strncpy(sstr,&var[len+1],62); sstr[63] = 0;
          for (i=0,c=strtok_r(sstr,",",&cp); i<4 && c; i++,c=strtok_r(NULL,",",&cp))
          {
            ival[i] = atoi(c);
          }
          if (i<4) ival[i] = atoi(val);
        }
        i = uc->fcn(ival[0],ival[1],ival[2],ival[3]);
        sprintf(strbuf,"result is %d",i);
        ttyoutput(strbuf);
        return 0;  
      }
    }
  }
  if (strstr(var,"BurstLimit"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     v = SetBurstLimit(v);
     sprintf(strbuf,"Burst Limit set to %d",gBurstLimit); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"CycleDelay"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     v = SetCycleDelay(v);
     sprintf(strbuf,"Cycle Delay set to %d",gCycleDelay); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"debug"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     NGdebug = v;
     sprintf(strbuf,"Debug level %d",NGdebug); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"logdbg"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     if (v > 0)
     {
       if ((dbgfp=fopen("debug.log","w")) == NULL)
         sprintf(strbuf,"Couldn't open debug log file");
       else sprintf(strbuf,"debug logging ON");
       ttyoutput(strbuf);
     }
     else
     {
       if (dbgfp) fclose(dbgfp);
       dbgfp = NULL;
       sprintf(strbuf,"debug logging OFF");
       ttyoutput(strbuf);
     }
     return 0;
  }
  if (strstr(var,"foreground"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     foreground = v;
     sprintf(strbuf,"set foreground to %d",foreground); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"almdbg"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     ALMdebug = v;
     sprintf(strbuf,"Alarm Debug level %d",ALMdebug); ttyoutput(strbuf);
     return 0;
  }
  if (strstr(var,"CommandLevel"))
  {
     if ((v=atoi(val)) < 0) v = 0;
     CommandLevel = v;
     if (NGdebug)
     {
       sprintf(strbuf,"Command Level %d entered",CommandLevel);
       ttyoutput(strbuf);
     }
     return 0;
  }
  if (strstr(var,"filter"))
  {
     memset(tagNameFilter,0,64);
     if ((c=strchr(val,'\n')) != NULL) *c = 0;
     if (strlen(val) > 0) strncpy(tagNameFilter,val,63);
     if (NGdebug)
     {
       sprintf(strbuf,"debug name filter %s entered",tagNameFilter);
       ttyoutput(strbuf);
     }
     return 0;
  }
  if (strstr(var,"CommandString"))
  {
     memset(CommandString,0,CMDSTRSIZE);
     if ((c=strchr(val,'\n')) != NULL) *c = 0;
     strncpy(CommandString,val,CMDSTRSIZE-1);
     if (NGdebug)
     {
       sprintf(strbuf,"Command String %s entered",CommandString);
       ttyoutput(strbuf);
     }
     return 0;
  }
  sprintf(strbuf,"Variable %s unavailable",var); ttyoutput(strbuf);
  return -1;
}

int CMDlisten(const char *name)
{
  int fd,len;
  struct sockaddr_un unix_addr;
# ifdef SVR4
  int fdarr[2];
# endif

  unlink(name);

# ifndef SVR4
  memset(&unix_addr,0,sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  strcpy(unix_addr.sun_path,name);
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
  if ((fd=socket(AF_UNIX,SOCK_STREAM,0)) < 0) return -1;
  if (bind(fd, (struct sockaddr *)&unix_addr, len) < 0) return -2;
  if (listen(fd, 5) < 0) return -3;
# else
  if ((fd=creat(name,FIFO_MODE)) < 0) return -1;
  if (close(fd) < 0) return -2;
  if (pipe(fdarr) < 0) return -3;
  if (ioctl(fdarr[1],I_PUSH,"connld") < 0) return -4;
  if (fattach(fdarr[1],name) < 0) return -5;
  fd = fdarr[0];
# endif /* SVR4 */

  return fd;
}
int CMDconnect(const char *name)
{
  int cc = 0, fd = 0,len;
  struct sockaddr_un unix_addr;
# ifndef SVR4
  if ((fd=socket(AF_UNIX,SOCK_STREAM,0)) < 0) ccerr(-1);
  memset(&unix_addr,0,sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  sprintf(unix_addr.sun_path,
    strlen(CLN_PATH) == 9 ? "%s%05d" : "%s%09d", CLN_PATH, getpid());

# ifdef XSCM_RIGHTS
  len = sizeof(unix_addr.sun_len) + sizeof(unix_addr.sun_family) +
        strlen(unix_addr.sun_path) + 1;
  unix_addr.sun_len = len;
# else
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
  if (len != 16) ccerr(-5);
# endif
  unlink(unix_addr.sun_path);
  if (bind(fd, (struct sockaddr *)&unix_addr, len) < 0) ccerr(-2);
  if (chmod(unix_addr.sun_path, CLN_PERM) < 0) ccerr(-3);
  memset(&unix_addr, 0, sizeof(unix_addr));
  unix_addr.sun_family = AF_UNIX;
  strcpy(unix_addr.sun_path, name);
# ifdef XSCM_RIGHTS
  len = sizeof(unix_addr.sun_len) + sizeof(unix_addr.sun_family) +
  strlen(unix_addr.sun_path) + 1;
  unix_addr.sun_len = len;
# else
  len = strlen(unix_addr.sun_path) + sizeof(unix_addr.sun_family);
# ifdef FREEBSD
  len++;
# endif
# endif
  if (connect(fd, (struct sockaddr *)&unix_addr,len) < 0) ccerr(-4);
#else
  if ((fd=open(name,O_RDWR)) < 0) ccerr(-1);
  if (isastream(fd) == 0) 
  {
    close(fd);
    ccerr(-2);
  }
#endif /* SVR4 */
err:
  if (cc)
  {
    if (fd > 0) close(fd);
  }
  return fd;
}

#define STALE 30
int CMDaccept(int lsnfd, uid_t *uidptr)
{
  int cc=0,clnfd,len;
  time_t staletime;
  char pipename[15];
# ifndef SVR4
  struct sockaddr_un unix_addr;
  struct stat statbuf;
# else
  struct strrecvfd recvfd;
# endif

# ifndef SVR4
  len = sizeof(unix_addr);
  if ((clnfd=accept(lsnfd, (struct sockaddr *)&unix_addr,&len)) < 0) ccerr(-1);
  len -= sizeof(unix_addr.sun_family);

  strncpy(pipename,unix_addr.sun_path,14); pipename[14] = 0;
  if (stat(pipename,&statbuf) < 0) ccerr(-2);

# ifdef S_ISSOCK
  if (S_ISSOCK(statbuf.st_mode) == 0) ccerr(-3);
# endif

  if ((statbuf.st_mode & (S_IRWXG | S_IRWXO)) ||
      (statbuf.st_mode &  S_IRWXU) != S_IRWXU) ccerr(-4);

  staletime = time(NULL) - STALE;
  if (statbuf.st_atime < staletime ||
      statbuf.st_ctime < staletime ||
      statbuf.st_mtime < staletime) ccerr(-5);

  if (uidptr != NULL) *uidptr = statbuf.st_uid;

  unlink(pipename);
# else
  if (ioctl(lsnfd,I_RECVFD,&recvfd) < 0) ccerr(-1);
  if (uidptr != NULL) *uidptr = recvfd.uid;
  clnfd = recvfd.fd;
# endif /* SVR4 */
  return clnfd;
err:
  perror("CMDaccept");
  if (cc < -1) printf("name %s len %d\n",pipename,len);
  return cc;
}

#define IPCquit(txt) { perror(txt); printf("return %d\n",clnfd); return; }
void addIPCcln(void)
{
  int clnfd;
  uid_t uid;
  char strbuf[CMDSIZE];

  if ((clnfd = CMDaccept(IPCSocket,&uid)) < 0) IPCquit("accept error");
  if (nipcclients < MAXIPCCLIENTS)
  {
    IPCfds[nipcclients] = clnfd;
    nipcclients++;
    printf("Remote Session granted to uid %d\n",uid);
  }
  else
  {
    sprintf(strbuf,"Maximum %d clients exceeded\n",MAXIPCCLIENTS);
    write(clnfd,strbuf,strlen(strbuf));
  }
}
int dbglog(char *text,...)
{
# ifndef NO_REMOTE_PIPE
  int i;
# endif
  int cc=0;
  char str[256];
  va_list args;

  if (strlen(text) > 80) ccerr(string_too_long);
  va_start(args,text);
  vsprintf(str,text,args);
  if (str[strlen(str)-1] == '\n') strcat(str,">");
  else sprintf(&str[strlen(str)]," @%s\n>",getDataTimeString((double)time(NULL),FALSE));
  va_end(args);
  if (strlen(tagNameFilter) && !strstr(str,tagNameFilter)) return 0;
  printf(str);
  if (dbgfp != NULL) fprintf(dbgfp,str);
# ifndef NO_REMOTE_PIPE
  for (i=0; i<nipcclients; i++) write(IPCfds[i],str,strlen(str));
# endif
err:
  return cc;
}
int dbgoutput(char *str,time_t *ts)
{
  int i;
  char strbuf[CMDSIZE];

  if (strlen(tagNameFilter) && !strstr(str,tagNameFilter)) return 0;
  if (ts != NULL)
  {
    if (strlen(str) > CMDSIZE-29) str[CMDSIZE-29] = 0; /* just in case */
    sprintf(strbuf,"%s @%ld\n>",str,*ts);
  }
  else sprintf(strbuf,"%s\n>",str);

  printf(strbuf);
  if (dbgfp != NULL) fprintf(dbgfp,strbuf);
# ifndef NO_REMOTE_PIPE
  for (i=0; i<nipcclients; i++) write(IPCfds[i],strbuf,strlen(strbuf));
# endif
  return 0;
}
int ttyoutput(char *str)
{
  printf(str);
# ifndef NO_REMOTE_PIPE
  if (ipcfd > 0) write(ipcfd,str,strlen(str));
# endif
  return 0;
}
int rmtoutput(char *str)
{
  if (ipcfd > 0) write(ipcfd,str,strlen(str));
  return 0;
}
#ifndef BSDSIGNAL
Sigfunc *bsdsignal(int signo, Sigfunc *disp)
{
  struct sigaction act;

  act.sa_handler = (void(*)())disp;
  sigemptyset(&act.sa_mask);
  act.sa_flags = 0;

  if (signo != SIGALRM) act.sa_flags |= SA_RESTART;
  return sigaction(signo,&act,NULL) < 0 ? SIG_ERR : 0;
}
#endif

void sig_pipe(int signo)
{
  printf("Remote session terminated\n");
#ifdef INTEL
  if (bsdsignal(SIGPIPE, sig_pipe) == SIG_ERR) printf("signal error\n");
#endif
}
